/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.12
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */

#define SWIGRUBY

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#  ifndef GCC_HASCLASSVISIBILITY
#    define GCC_HASCLASSVISIBILITY
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif


/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#  ifndef GCC_HASCLASSVISIBILITY
#    define GCC_HASCLASSVISIBILITY
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif


/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      register size_t l = 0;
      register size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	register size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  register int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      register size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  register const unsigned char *u = (unsigned char *) ptr;
  register const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    register unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  register unsigned char *u = (unsigned char *) ptr;
  register const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    register char d = *(c++);
    register unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = ((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = ((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



#include <ruby.h>

/* Ruby 1.9.1 has a "memoisation optimisation" when compiling with GCC which
 * breaks using rb_intern as an lvalue, as SWIG does.  We work around this
 * issue for now by disabling this.
 * https://sourceforge.net/tracker/?func=detail&aid=2859614&group_id=1645&atid=101645
 */
#ifdef rb_intern
# undef rb_intern
#endif

/* Remove global macros defined in Ruby's win32.h */
#ifdef write
# undef write
#endif
#ifdef read
# undef read
#endif
#ifdef bind
# undef bind
#endif
#ifdef close
# undef close
#endif
#ifdef connect
# undef connect
#endif


/* Ruby 1.7 defines NUM2LL(), LL2NUM() and ULL2NUM() macros */
#ifndef NUM2LL
#define NUM2LL(x) NUM2LONG((x))
#endif
#ifndef LL2NUM
#define LL2NUM(x) INT2NUM((long) (x))
#endif
#ifndef ULL2NUM
#define ULL2NUM(x) UINT2NUM((unsigned long) (x))
#endif

/* Ruby 1.7 doesn't (yet) define NUM2ULL() */
#ifndef NUM2ULL
#ifdef HAVE_LONG_LONG
#define NUM2ULL(x) rb_num2ull((x))
#else
#define NUM2ULL(x) NUM2ULONG(x)
#endif
#endif

/* RSTRING_LEN, etc are new in Ruby 1.9, but ->ptr and ->len no longer work */
/* Define these for older versions so we can just write code the new way */
#ifndef RSTRING_LEN
# define RSTRING_LEN(x) RSTRING(x)->len
#endif
#ifndef RSTRING_PTR
# define RSTRING_PTR(x) RSTRING(x)->ptr
#endif
#ifndef RSTRING_END
# define RSTRING_END(x) (RSTRING_PTR(x) + RSTRING_LEN(x))
#endif
#ifndef RARRAY_LEN
# define RARRAY_LEN(x) RARRAY(x)->len
#endif
#ifndef RARRAY_PTR
# define RARRAY_PTR(x) RARRAY(x)->ptr
#endif
#ifndef RFLOAT_VALUE
# define RFLOAT_VALUE(x) RFLOAT(x)->value
#endif
#ifndef DOUBLE2NUM
# define DOUBLE2NUM(x) rb_float_new(x)
#endif
#ifndef RHASH_TBL
# define RHASH_TBL(x) (RHASH(x)->tbl)
#endif
#ifndef RHASH_ITER_LEV
# define RHASH_ITER_LEV(x) (RHASH(x)->iter_lev)
#endif
#ifndef RHASH_IFNONE
# define RHASH_IFNONE(x) (RHASH(x)->ifnone)
#endif
#ifndef RHASH_SIZE
# define RHASH_SIZE(x) (RHASH(x)->tbl->num_entries)
#endif
#ifndef RHASH_EMPTY_P
# define RHASH_EMPTY_P(x) (RHASH_SIZE(x) == 0)
#endif
#ifndef RSTRUCT_LEN
# define RSTRUCT_LEN(x) RSTRUCT(x)->len
#endif
#ifndef RSTRUCT_PTR
# define RSTRUCT_PTR(x) RSTRUCT(x)->ptr
#endif



/*
 * Need to be very careful about how these macros are defined, especially
 * when compiling C++ code or C code with an ANSI C compiler.
 *
 * VALUEFUNC(f) is a macro used to typecast a C function that implements
 * a Ruby method so that it can be passed as an argument to API functions
 * like rb_define_method() and rb_define_singleton_method().
 *
 * VOIDFUNC(f) is a macro used to typecast a C function that implements
 * either the "mark" or "free" stuff for a Ruby Data object, so that it
 * can be passed as an argument to API functions like Data_Wrap_Struct()
 * and Data_Make_Struct().
 */
 
#ifdef __cplusplus
#  ifndef RUBY_METHOD_FUNC /* These definitions should work for Ruby 1.4.6 */
#    define PROTECTFUNC(f) ((VALUE (*)()) f)
#    define VALUEFUNC(f) ((VALUE (*)()) f)
#    define VOIDFUNC(f)  ((void (*)()) f)
#  else
#    ifndef ANYARGS /* These definitions should work for Ruby 1.6 */
#      define PROTECTFUNC(f) ((VALUE (*)()) f)
#      define VALUEFUNC(f) ((VALUE (*)()) f)
#      define VOIDFUNC(f)  ((RUBY_DATA_FUNC) f)
#    else /* These definitions should work for Ruby 1.7+ */
#      define PROTECTFUNC(f) ((VALUE (*)(VALUE)) f)
#      define VALUEFUNC(f) ((VALUE (*)(ANYARGS)) f)
#      define VOIDFUNC(f)  ((RUBY_DATA_FUNC) f)
#    endif
#  endif
#else
#  define VALUEFUNC(f) (f)
#  define VOIDFUNC(f) (f)
#endif

/* Don't use for expressions have side effect */
#ifndef RB_STRING_VALUE
#define RB_STRING_VALUE(s) (TYPE(s) == T_STRING ? (s) : (*(volatile VALUE *)&(s) = rb_str_to_str(s)))
#endif
#ifndef StringValue
#define StringValue(s) RB_STRING_VALUE(s)
#endif
#ifndef StringValuePtr
#define StringValuePtr(s) RSTRING_PTR(RB_STRING_VALUE(s))
#endif
#ifndef StringValueLen
#define StringValueLen(s) RSTRING_LEN(RB_STRING_VALUE(s))
#endif
#ifndef SafeStringValue
#define SafeStringValue(v) do {\
    StringValue(v);\
    rb_check_safe_str(v);\
} while (0)
#endif

#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
#define rb_define_alloc_func(klass, func) rb_define_singleton_method((klass), "new", VALUEFUNC((func)), -1)
#define rb_undef_alloc_func(klass) rb_undef_method(CLASS_OF((klass)), "new")
#endif

static VALUE _mSWIG = Qnil;

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */


/* Define some additional error types */
#define SWIG_ObjectPreviouslyDeletedError  -100


/* Define custom exceptions for errors that do not map to existing Ruby
   exceptions.  Note this only works for C++ since a global cannot be
   initialized by a function in C.  For C, fallback to rb_eRuntimeError.*/

SWIGINTERN VALUE 
getNullReferenceError(void) {
  static int init = 0;
  static VALUE rb_eNullReferenceError ;
  if (!init) {
    init = 1;
    rb_eNullReferenceError = rb_define_class("NullReferenceError", rb_eRuntimeError);
  }
  return rb_eNullReferenceError;
} 

SWIGINTERN VALUE 
getObjectPreviouslyDeletedError(void) {
  static int init = 0;
  static VALUE rb_eObjectPreviouslyDeleted ;
  if (!init) {
    init = 1;
    rb_eObjectPreviouslyDeleted = rb_define_class("ObjectPreviouslyDeleted", rb_eRuntimeError);
  }
  return rb_eObjectPreviouslyDeleted;
} 


SWIGINTERN VALUE
SWIG_Ruby_ErrorType(int SWIG_code) {
  VALUE type;
  switch (SWIG_code) {
  case SWIG_MemoryError:
    type = rb_eNoMemError;
    break;
  case SWIG_IOError:
    type = rb_eIOError;
    break;
  case SWIG_RuntimeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_IndexError:
    type = rb_eIndexError;
    break;
  case SWIG_TypeError:
    type = rb_eTypeError;
    break;
  case SWIG_DivisionByZero:
    type = rb_eZeroDivError;
    break;
  case SWIG_OverflowError:
    type = rb_eRangeError;
    break;
  case SWIG_SyntaxError:
    type = rb_eSyntaxError;
    break;
  case SWIG_ValueError:
    type = rb_eArgError;
    break;
  case SWIG_SystemError:
    type = rb_eFatal;
    break;
  case SWIG_AttributeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_NullReferenceError:
    type = getNullReferenceError();
    break;
  case SWIG_ObjectPreviouslyDeletedError:
    type = getObjectPreviouslyDeletedError();
    break;
  case SWIG_UnknownError:
    type = rb_eRuntimeError;
    break;
  default:
    type = rb_eRuntimeError;
  }
  return type;
}


/* This function is called when a user inputs a wrong argument to
   a method.
 */
SWIGINTERN 
const char* Ruby_Format_TypeError( const char* msg,
				   const char* type, 
				   const char* name, 
				   const int argn,
				   VALUE input )
{
  char buf[128];
  VALUE str;
  VALUE asStr;
  if ( msg && *msg )
    {
      str = rb_str_new2(msg);
    }
  else
    {
      str = rb_str_new(NULL, 0);
    }

  str = rb_str_cat2( str, "Expected argument " );
  sprintf( buf, "%d of type ", argn-1 );
  str = rb_str_cat2( str, buf );
  str = rb_str_cat2( str, type );
  str = rb_str_cat2( str, ", but got " );
  str = rb_str_cat2( str, rb_obj_classname(input) );
  str = rb_str_cat2( str, " " );
  asStr = rb_inspect(input);
  if ( RSTRING_LEN(asStr) > 30 )
    {
      str = rb_str_cat( str, StringValuePtr(asStr), 30 );
      str = rb_str_cat2( str, "..." );
    }
  else
    {
      str = rb_str_append( str, asStr );
    }

  if ( name )
    {
      str = rb_str_cat2( str, "\n\tin SWIG method '" );
      str = rb_str_cat2( str, name );
      str = rb_str_cat2( str, "'" );
    }

  return StringValuePtr( str );
}

/* This function is called when an overloaded method fails */
SWIGINTERN 
void Ruby_Format_OverloadedError(
				 const int argc,
				 const int maxargs,
				 const char* method, 
				 const char* prototypes 
				 )
{
  const char* msg = "Wrong # of arguments";
  if ( argc <= maxargs ) msg = "Wrong arguments";
  rb_raise(rb_eArgError,"%s for overloaded method '%s'.\n"  
	   "Possible C/C++ prototypes are:\n%s",
	   msg, method, prototypes);
}

/* -----------------------------------------------------------------------------
 * rubytracking.swg
 *
 * This file contains support for tracking mappings from 
 * Ruby objects to C++ objects.  This functionality is needed
 * to implement mark functions for Ruby's mark and sweep
 * garbage collector.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* Ruby 1.8 actually assumes the first case. */
#if SIZEOF_VOIDP == SIZEOF_LONG
#  define SWIG2NUM(v) LONG2NUM((unsigned long)v)
#  define NUM2SWIG(x) (unsigned long)NUM2LONG(x)
#elif SIZEOF_VOIDP == SIZEOF_LONG_LONG
#  define SWIG2NUM(v) LL2NUM((unsigned long long)v)
#  define NUM2SWIG(x) (unsigned long long)NUM2LL(x)
#else
#  error sizeof(void*) is not the same as long or long long
#endif


/* Global Ruby hash table to store Trackings from C/C++
   structs to Ruby Objects. 
*/
static VALUE swig_ruby_trackings = Qnil;

/* Global variable that stores a reference to the ruby
   hash table delete function. */
static ID swig_ruby_hash_delete;

/* Setup a Ruby hash table to store Trackings */
SWIGRUNTIME void SWIG_RubyInitializeTrackings(void) {
  /* Create a ruby hash table to store Trackings from C++ 
     objects to Ruby objects. */

  /* Try to see if some other .so has already created a 
     tracking hash table, which we keep hidden in an instance var
     in the SWIG module.
     This is done to allow multiple DSOs to share the same
     tracking table.
  */
  ID trackings_id = rb_intern( "@__trackings__" );
  VALUE verbose = rb_gv_get("VERBOSE");
  rb_gv_set("VERBOSE", Qfalse);
  swig_ruby_trackings = rb_ivar_get( _mSWIG, trackings_id );
  rb_gv_set("VERBOSE", verbose);

  /* No, it hasn't.  Create one ourselves */ 
  if ( swig_ruby_trackings == Qnil )
    {
      swig_ruby_trackings = rb_hash_new();
      rb_ivar_set( _mSWIG, trackings_id, swig_ruby_trackings );
    }

  /* Now store a reference to the hash table delete function
     so that we only have to look it up once.*/
  swig_ruby_hash_delete = rb_intern("delete");
}

/* Get a Ruby number to reference a pointer */
SWIGRUNTIME VALUE SWIG_RubyPtrToReference(void* ptr) {
  /* We cast the pointer to an unsigned long
     and then store a reference to it using
     a Ruby number object. */

  /* Convert the pointer to a Ruby number */
  return SWIG2NUM(ptr);
}

/* Get a Ruby number to reference an object */
SWIGRUNTIME VALUE SWIG_RubyObjectToReference(VALUE object) {
  /* We cast the object to an unsigned long
     and then store a reference to it using
     a Ruby number object. */

  /* Convert the Object to a Ruby number */
  return SWIG2NUM(object);
}

/* Get a Ruby object from a previously stored reference */
SWIGRUNTIME VALUE SWIG_RubyReferenceToObject(VALUE reference) {
  /* The provided Ruby number object is a reference
     to the Ruby object we want.*/

  /* Convert the Ruby number to a Ruby object */
  return NUM2SWIG(reference);
}

/* Add a Tracking from a C/C++ struct to a Ruby object */
SWIGRUNTIME void SWIG_RubyAddTracking(void* ptr, VALUE object) {
  /* In a Ruby hash table we store the pointer and
     the associated Ruby object.  The trick here is
     that we cannot store the Ruby object directly - if
     we do then it cannot be garbage collected.  So
     instead we typecast it as a unsigned long and
     convert it to a Ruby number object.*/

  /* Get a reference to the pointer as a Ruby number */
  VALUE key = SWIG_RubyPtrToReference(ptr);

  /* Get a reference to the Ruby object as a Ruby number */
  VALUE value = SWIG_RubyObjectToReference(object);

  /* Store the mapping to the global hash table. */
  rb_hash_aset(swig_ruby_trackings, key, value);
}

/* Get the Ruby object that owns the specified C/C++ struct */
SWIGRUNTIME VALUE SWIG_RubyInstanceFor(void* ptr) {
  /* Get a reference to the pointer as a Ruby number */
  VALUE key = SWIG_RubyPtrToReference(ptr);

  /* Now lookup the value stored in the global hash table */
  VALUE value = rb_hash_aref(swig_ruby_trackings, key);
	
  if (value == Qnil) {
    /* No object exists - return nil. */
    return Qnil;
  }
  else {
    /* Convert this value to Ruby object */
    return SWIG_RubyReferenceToObject(value);
  }
}

/* Remove a Tracking from a C/C++ struct to a Ruby object.  It
   is very important to remove objects once they are destroyed
   since the same memory address may be reused later to create
   a new object. */
SWIGRUNTIME void SWIG_RubyRemoveTracking(void* ptr) {
  /* Get a reference to the pointer as a Ruby number */
  VALUE key = SWIG_RubyPtrToReference(ptr);

  /* Delete the object from the hash table by calling Ruby's
     do this we need to call the Hash.delete method.*/
  rb_funcall(swig_ruby_trackings, swig_ruby_hash_delete, 1, key);
}

/* This is a helper method that unlinks a Ruby object from its
   underlying C++ object.  This is needed if the lifetime of the
   Ruby object is longer than the C++ object */
SWIGRUNTIME void SWIG_RubyUnlinkObjects(void* ptr) {
  VALUE object = SWIG_RubyInstanceFor(ptr);

  if (object != Qnil) {
    DATA_PTR(object) = 0;
  }
}


#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Ruby API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

SWIGINTERN VALUE
SWIG_Ruby_AppendOutput(VALUE target, VALUE o) {
  if (NIL_P(target)) {
    target = o;
  } else {
    if (TYPE(target) != T_ARRAY) {
      VALUE o2 = target;
      target = rb_ary_new();
      rb_ary_push(target, o2);
    }
    rb_ary_push(target, o);
  }
  return target;
}

/* For ruby1.8.4 and earlier. */
#ifndef RUBY_INIT_STACK
   RUBY_EXTERN void Init_stack(VALUE* addr);
#  define RUBY_INIT_STACK \
   VALUE variable_in_this_stack_frame; \
   Init_stack(&variable_in_this_stack_frame);
#endif


#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * rubyrun.swg
 *
 * This file contains the runtime support for Ruby modules
 * and includes code for managing global variables and pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* For backward compatibility only */
#define SWIG_POINTER_EXCEPTION  0

/* for raw pointers */
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, own)
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Ruby_NewPointerObj(ptr, type, flags)
#define SWIG_AcquirePtr(ptr, own)                       SWIG_Ruby_AcquirePtr(ptr, own)
#define swig_owntype                                    ruby_owntype

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty, flags)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_ConvertPtr(obj, pptr, type, 0)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Ruby_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer) 		SWIG_Ruby_SetModule(pointer)


/* Error manipulation */

#define SWIG_ErrorType(code)                            SWIG_Ruby_ErrorType(code)               
#define SWIG_Error(code, msg)            		rb_raise(SWIG_Ruby_ErrorType(code), "%s", msg)
#define SWIG_fail                        		goto fail				 


/* Ruby-specific SWIG API */

#define SWIG_InitRuntime()                              SWIG_Ruby_InitRuntime()              
#define SWIG_define_class(ty)                        	SWIG_Ruby_define_class(ty)
#define SWIG_NewClassInstance(value, ty)             	SWIG_Ruby_NewClassInstance(value, ty)
#define SWIG_MangleStr(value)                        	SWIG_Ruby_MangleStr(value)		  
#define SWIG_CheckConvert(value, ty)                 	SWIG_Ruby_CheckConvert(value, ty)	  

#include "assert.h"

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  VALUE klass;
  VALUE mImpl;
  void  (*mark)(void *);
  void  (*destroy)(void *);
  int trackObjects;
} swig_class;


/* Global pointer used to keep some internal SWIG stuff */
static VALUE _cSWIG_Pointer = Qnil;
static VALUE swig_runtime_data_type_pointer = Qnil;

/* Global IDs used to keep some internal SWIG stuff */
static ID swig_arity_id = 0;
static ID swig_call_id  = 0;

/*
  If your swig extension is to be run within an embedded ruby and has
  director callbacks, you should set -DRUBY_EMBEDDED during compilation.  
  This will reset ruby's stack frame on each entry point from the main 
  program the first time a virtual director function is invoked (in a 
  non-recursive way).
  If this is not done, you run the risk of Ruby trashing the stack.
*/

#ifdef RUBY_EMBEDDED

#  define SWIG_INIT_STACK                            \
      if ( !swig_virtual_calls ) { RUBY_INIT_STACK } \
      ++swig_virtual_calls;
#  define SWIG_RELEASE_STACK --swig_virtual_calls;
#  define Ruby_DirectorTypeMismatchException(x) \
          rb_raise( rb_eTypeError, "%s", x ); return c_result;

      static unsigned int swig_virtual_calls = 0;

#else  /* normal non-embedded extension */

#  define SWIG_INIT_STACK
#  define SWIG_RELEASE_STACK
#  define Ruby_DirectorTypeMismatchException(x) \
          throw Swig::DirectorTypeMismatchException( x );

#endif  /* RUBY_EMBEDDED */


SWIGRUNTIME VALUE 
getExceptionClass(void) {
  static int init = 0;
  static VALUE rubyExceptionClass ;
  if (!init) {
    init = 1;
    rubyExceptionClass = rb_const_get(_mSWIG, rb_intern("Exception"));
  }
  return rubyExceptionClass;
} 

/* This code checks to see if the Ruby object being raised as part
   of an exception inherits from the Ruby class Exception.  If so,
   the object is simply returned.  If not, then a new Ruby exception
   object is created and that will be returned to Ruby.*/
SWIGRUNTIME VALUE
SWIG_Ruby_ExceptionType(swig_type_info *desc, VALUE obj) {
  VALUE exceptionClass = getExceptionClass();
  if (rb_obj_is_kind_of(obj, exceptionClass)) {
    return obj;
  }  else {
    return rb_exc_new3(rb_eRuntimeError, rb_obj_as_string(obj));
  }
}

/* Initialize Ruby runtime support */
SWIGRUNTIME void
SWIG_Ruby_InitRuntime(void)
{
  if (_mSWIG == Qnil) {
    _mSWIG = rb_define_module("SWIG");
    swig_call_id  = rb_intern("call");
    swig_arity_id = rb_intern("arity");
  }
}

/* Define Ruby class for C type */
SWIGRUNTIME void
SWIG_Ruby_define_class(swig_type_info *type)
{
  VALUE klass;
  char *klass_name = (char *) malloc(4 + strlen(type->name) + 1);
  sprintf(klass_name, "TYPE%s", type->name);
  if (NIL_P(_cSWIG_Pointer)) {
    _cSWIG_Pointer = rb_define_class_under(_mSWIG, "Pointer", rb_cObject);
    rb_undef_method(CLASS_OF(_cSWIG_Pointer), "new");
  }
  klass = rb_define_class_under(_mSWIG, klass_name, _cSWIG_Pointer);
  free((void *) klass_name);
}

/* Create a new pointer object */
SWIGRUNTIME VALUE
SWIG_Ruby_NewPointerObj(void *ptr, swig_type_info *type, int flags)
{
  int own =  flags & SWIG_POINTER_OWN; 
  int track;
  char *klass_name;
  swig_class *sklass;
  VALUE klass;
  VALUE obj;
  
  if (!ptr)
    return Qnil;
  
  if (type->clientdata) {
    sklass = (swig_class *) type->clientdata;
		
    /* Are we tracking this class and have we already returned this Ruby object? */
    track = sklass->trackObjects;
    if (track) {
      obj = SWIG_RubyInstanceFor(ptr);
      
      /* Check the object's type and make sure it has the correct type.
        It might not in cases where methods do things like 
        downcast methods. */
      if (obj != Qnil) {
        VALUE value = rb_iv_get(obj, "@__swigtype__");
        const char* type_name = RSTRING_PTR(value);
				
        if (strcmp(type->name, type_name) == 0) {
          return obj;
        }
      }
    }

    /* Create a new Ruby object */
    obj = Data_Wrap_Struct(sklass->klass, VOIDFUNC(sklass->mark), 
			   ( own ? VOIDFUNC(sklass->destroy) : 
			     (track ? VOIDFUNC(SWIG_RubyRemoveTracking) : 0 )
			     ), ptr);

    /* If tracking is on for this class then track this object. */
    if (track) {
      SWIG_RubyAddTracking(ptr, obj);
    }
  } else {
    klass_name = (char *) malloc(4 + strlen(type->name) + 1);
    sprintf(klass_name, "TYPE%s", type->name);
    klass = rb_const_get(_mSWIG, rb_intern(klass_name));
    free((void *) klass_name);
    obj = Data_Wrap_Struct(klass, 0, 0, ptr);
  }
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));
  
  return obj;
}

/* Create a new class instance (always owned) */
SWIGRUNTIME VALUE
SWIG_Ruby_NewClassInstance(VALUE klass, swig_type_info *type)
{
  VALUE obj;
  swig_class *sklass = (swig_class *) type->clientdata;
  obj = Data_Wrap_Struct(klass, VOIDFUNC(sklass->mark), VOIDFUNC(sklass->destroy), 0);
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));
  return obj;
}

/* Get type mangle from class name */
SWIGRUNTIMEINLINE char *
SWIG_Ruby_MangleStr(VALUE obj)
{
  VALUE stype = rb_iv_get(obj, "@__swigtype__");
  return StringValuePtr(stype);
}

/* Acquire a pointer value */
typedef void (*ruby_owntype)(void*);

SWIGRUNTIME ruby_owntype
SWIG_Ruby_AcquirePtr(VALUE obj, ruby_owntype own) {
  if (obj) {
    ruby_owntype oldown = RDATA(obj)->dfree;
    RDATA(obj)->dfree = own;
    return oldown;
  } else {
    return 0;
  }
}

/* Convert a pointer value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPtrAndOwn(VALUE obj, void **ptr, swig_type_info *ty, int flags, ruby_owntype *own)
{
  char *c;
  swig_cast_info *tc;
  void *vptr = 0;

  /* Grab the pointer */
  if (NIL_P(obj)) {
    *ptr = 0;
    return SWIG_OK;
  } else {
    if (TYPE(obj) != T_DATA) {
      return SWIG_ERROR;
    }
    Data_Get_Struct(obj, void, vptr);
  }
  
  if (own) *own = RDATA(obj)->dfree;
    
  /* Check to see if the input object is giving up ownership
     of the underlying C struct or C++ object.  If so then we
     need to reset the destructor since the Ruby object no 
     longer owns the underlying C++ object.*/ 
  if (flags & SWIG_POINTER_DISOWN) {
    /* Is tracking on for this class? */
    int track = 0;
    if (ty && ty->clientdata) {
      swig_class *sklass = (swig_class *) ty->clientdata;
      track = sklass->trackObjects;
    }
		
    if (track) {
      /* We are tracking objects for this class.  Thus we change the destructor
       * to SWIG_RubyRemoveTracking.  This allows us to
       * remove the mapping from the C++ to Ruby object
       * when the Ruby object is garbage collected.  If we don't
       * do this, then it is possible we will return a reference 
       * to a Ruby object that no longer exists thereby crashing Ruby. */
      RDATA(obj)->dfree = SWIG_RubyRemoveTracking;
    } else {    
      RDATA(obj)->dfree = 0;
    }
  }

  /* Do type-checking if type info was provided */
  if (ty) {
    if (ty->clientdata) {
      if (rb_obj_is_kind_of(obj, ((swig_class *) (ty->clientdata))->klass)) {
        if (vptr == 0) {
          /* The object has already been deleted */
          return SWIG_ObjectPreviouslyDeletedError;
        }
        *ptr = vptr;
        return SWIG_OK;
      }
    }
    if ((c = SWIG_MangleStr(obj)) == NULL) {
      return SWIG_ERROR;
    }
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) {
      return SWIG_ERROR;
    } else {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc, vptr, &newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    }
  } else {
    *ptr = vptr;
  }
  
  return SWIG_OK;
}

/* Check convert */
SWIGRUNTIMEINLINE int
SWIG_Ruby_CheckConvert(VALUE obj, swig_type_info *ty)
{
  char *c = SWIG_MangleStr(obj);
  if (!c) return 0;
  return SWIG_TypeCheck(c,ty) != 0;
}

SWIGRUNTIME VALUE
SWIG_Ruby_NewPackedObj(void *ptr, int sz, swig_type_info *type) {
  char result[1024];
  char *r = result;
  if ((2*sz + 1 + strlen(type->name)) > 1000) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r, ptr, sz);
  strcpy(r, type->name);
  return rb_str_new2(result);
}

/* Convert a packed value value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPacked(VALUE obj, void *ptr, int sz, swig_type_info *ty) {
  swig_cast_info *tc;
  const char  *c;

  if (TYPE(obj) != T_STRING) goto type_error;
  c = StringValuePtr(obj);
  /* Pointer values must start with leading underscore */
  if (*c != '_') goto type_error;
  c++;
  c = SWIG_UnpackData(c, ptr, sz);
  if (ty) {
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) goto type_error;
  }
  return SWIG_OK;

 type_error:
  return SWIG_ERROR;
}

SWIGRUNTIME swig_module_info *
SWIG_Ruby_GetModule(void *SWIGUNUSEDPARM(clientdata))
{
  VALUE pointer;
  swig_module_info *ret = 0;
  VALUE verbose = rb_gv_get("VERBOSE");

 /* temporarily disable warnings, since the pointer check causes warnings with 'ruby -w' */
  rb_gv_set("VERBOSE", Qfalse);
  
  /* first check if pointer already created */
  pointer = rb_gv_get("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME);
  if (pointer != Qnil) {
    Data_Get_Struct(pointer, swig_module_info, ret);
  }

  /* reinstate warnings */
  rb_gv_set("VERBOSE", verbose);
  return ret;
}

SWIGRUNTIME void 
SWIG_Ruby_SetModule(swig_module_info *pointer)
{
  /* register a new class */
  VALUE cl = rb_define_class("swig_runtime_data", rb_cObject);
  /* create and store the structure pointer to a global variable */
  swig_runtime_data_type_pointer = Data_Wrap_Struct(cl, 0, 0, pointer);
  rb_define_readonly_variable("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, &swig_runtime_data_type_pointer);
}

/* This function can be used to check whether a proc or method or similarly
   callable function has been passed.  Usually used in a %typecheck, like:

   %typecheck(c_callback_t, precedence=SWIG_TYPECHECK_POINTER) {
        $result = SWIG_Ruby_isCallable( $input );
   }
 */
SWIGINTERN
int SWIG_Ruby_isCallable( VALUE proc )
{
  if ( rb_respond_to( proc, swig_call_id ) )
    return 1;
  return 0;
}

/* This function can be used to check the arity (number of arguments)
   a proc or method can take.  Usually used in a %typecheck.
   Valid arities will be that equal to minimal or those < 0
   which indicate a variable number of parameters at the end.
 */
SWIGINTERN
int SWIG_Ruby_arity( VALUE proc, int minimal )
{
  if ( rb_respond_to( proc, swig_arity_id ) )
    {
      VALUE num = rb_funcall( proc, swig_arity_id, 0 );
      int arity = NUM2INT(num);
      if ( arity < 0 && (arity+1) < -minimal ) return 1;
      if ( arity == minimal ) return 1;
      return 1;
    }
  return 0;
}


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg);; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_apr_array_header_t swig_types[0]
#define SWIGTYPE_p_apr_file_t swig_types[1]
#define SWIGTYPE_p_apr_getopt_t swig_types[2]
#define SWIGTYPE_p_apr_hash_t swig_types[3]
#define SWIGTYPE_p_apr_int32_t swig_types[4]
#define SWIGTYPE_p_apr_int64_t swig_types[5]
#define SWIGTYPE_p_apr_pool_t swig_types[6]
#define SWIGTYPE_p_char swig_types[7]
#define SWIGTYPE_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void swig_types[8]
#define SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[9]
#define SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[10]
#define SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[11]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[12]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[13]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[14]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[15]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[16]
#define SWIGTYPE_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[17]
#define SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[18]
#define SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[19]
#define SWIGTYPE_p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[20]
#define SWIGTYPE_p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[21]
#define SWIGTYPE_p_f_p_void__p_svn_error_t swig_types[22]
#define SWIGTYPE_p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[23]
#define SWIGTYPE_p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[24]
#define SWIGTYPE_p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[25]
#define SWIGTYPE_p_f_p_void_p_q_const__char__int swig_types[26]
#define SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void swig_types[27]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[28]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[29]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[30]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t swig_types[31]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t swig_types[32]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[33]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[34]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t swig_types[35]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void swig_types[36]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void swig_types[37]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t swig_types[38]
#define SWIGTYPE_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void swig_types[39]
#define SWIGTYPE_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[40]
#define SWIGTYPE_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t swig_types[41]
#define SWIGTYPE_p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t swig_types[42]
#define SWIGTYPE_p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[43]
#define SWIGTYPE_p_int swig_types[44]
#define SWIGTYPE_p_long swig_types[45]
#define SWIGTYPE_p_p_apr_array_header_t swig_types[46]
#define SWIGTYPE_p_p_apr_hash_t swig_types[47]
#define SWIGTYPE_p_p_char swig_types[48]
#define SWIGTYPE_p_p_svn_auth_provider_object_t swig_types[49]
#define SWIGTYPE_p_p_svn_client_commit_info_t swig_types[50]
#define SWIGTYPE_p_p_svn_client_ctx_t swig_types[51]
#define SWIGTYPE_p_p_svn_commit_info_t swig_types[52]
#define SWIGTYPE_p_p_svn_ra_session_t swig_types[53]
#define SWIGTYPE_p_p_svn_string_t swig_types[54]
#define SWIGTYPE_p_svn_auth_baton_t swig_types[55]
#define SWIGTYPE_p_svn_auth_cred_simple_t swig_types[56]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t swig_types[57]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t swig_types[58]
#define SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t swig_types[59]
#define SWIGTYPE_p_svn_auth_cred_username_t swig_types[60]
#define SWIGTYPE_p_svn_auth_iterstate_t swig_types[61]
#define SWIGTYPE_p_svn_auth_provider_object_t swig_types[62]
#define SWIGTYPE_p_svn_auth_provider_t swig_types[63]
#define SWIGTYPE_p_svn_auth_ssl_server_cert_info_t swig_types[64]
#define SWIGTYPE_p_svn_checksum_ctx_t swig_types[65]
#define SWIGTYPE_p_svn_checksum_kind_t swig_types[66]
#define SWIGTYPE_p_svn_checksum_t swig_types[67]
#define SWIGTYPE_p_svn_client_commit_info_t swig_types[68]
#define SWIGTYPE_p_svn_client_commit_item2_t swig_types[69]
#define SWIGTYPE_p_svn_client_commit_item3_t swig_types[70]
#define SWIGTYPE_p_svn_client_commit_item_t swig_types[71]
#define SWIGTYPE_p_svn_client_copy_source_t swig_types[72]
#define SWIGTYPE_p_svn_client_ctx_t swig_types[73]
#define SWIGTYPE_p_svn_client_diff_summarize_kind_t swig_types[74]
#define SWIGTYPE_p_svn_client_diff_summarize_t swig_types[75]
#define SWIGTYPE_p_svn_client_info2_t swig_types[76]
#define SWIGTYPE_p_svn_client_proplist_item_t swig_types[77]
#define SWIGTYPE_p_svn_client_status_t swig_types[78]
#define SWIGTYPE_p_svn_commit_info_t swig_types[79]
#define SWIGTYPE_p_svn_config_t swig_types[80]
#define SWIGTYPE_p_svn_delta_editor_t swig_types[81]
#define SWIGTYPE_p_svn_depth_t swig_types[82]
#define SWIGTYPE_p_svn_diff_conflict_display_style_t swig_types[83]
#define SWIGTYPE_p_svn_diff_datasource_e swig_types[84]
#define SWIGTYPE_p_svn_diff_file_ignore_space_t swig_types[85]
#define SWIGTYPE_p_svn_diff_file_options_t swig_types[86]
#define SWIGTYPE_p_svn_diff_fns2_t swig_types[87]
#define SWIGTYPE_p_svn_diff_fns_t swig_types[88]
#define SWIGTYPE_p_svn_diff_hunk_t swig_types[89]
#define SWIGTYPE_p_svn_diff_operation_kind_e swig_types[90]
#define SWIGTYPE_p_svn_diff_output_fns_t swig_types[91]
#define SWIGTYPE_p_svn_diff_t swig_types[92]
#define SWIGTYPE_p_svn_dirent_t swig_types[93]
#define SWIGTYPE_p_svn_errno_t swig_types[94]
#define SWIGTYPE_p_svn_error_t swig_types[95]
#define SWIGTYPE_p_svn_info_t swig_types[96]
#define SWIGTYPE_p_svn_io_dirent2_t swig_types[97]
#define SWIGTYPE_p_svn_io_dirent_t swig_types[98]
#define SWIGTYPE_p_svn_io_file_del_t swig_types[99]
#define SWIGTYPE_p_svn_location_segment_t swig_types[100]
#define SWIGTYPE_p_svn_lock_t swig_types[101]
#define SWIGTYPE_p_svn_log_changed_path2_t swig_types[102]
#define SWIGTYPE_p_svn_log_changed_path_t swig_types[103]
#define SWIGTYPE_p_svn_log_entry_t swig_types[104]
#define SWIGTYPE_p_svn_merge_range_t swig_types[105]
#define SWIGTYPE_p_svn_mergeinfo_inheritance_t swig_types[106]
#define SWIGTYPE_p_svn_node_kind_t swig_types[107]
#define SWIGTYPE_p_svn_opt_revision_range_t swig_types[108]
#define SWIGTYPE_p_svn_opt_revision_t swig_types[109]
#define SWIGTYPE_p_svn_opt_revision_value_t swig_types[110]
#define SWIGTYPE_p_svn_opt_subcommand_desc2_t swig_types[111]
#define SWIGTYPE_p_svn_opt_subcommand_desc_t swig_types[112]
#define SWIGTYPE_p_svn_patch_file_t swig_types[113]
#define SWIGTYPE_p_svn_patch_t swig_types[114]
#define SWIGTYPE_p_svn_prop_inherited_item_t swig_types[115]
#define SWIGTYPE_p_svn_prop_kind swig_types[116]
#define SWIGTYPE_p_svn_prop_patch_t swig_types[117]
#define SWIGTYPE_p_svn_ra_callbacks2_t swig_types[118]
#define SWIGTYPE_p_svn_ra_callbacks_t swig_types[119]
#define SWIGTYPE_p_svn_ra_plugin_t swig_types[120]
#define SWIGTYPE_p_svn_ra_reporter2_t swig_types[121]
#define SWIGTYPE_p_svn_ra_reporter3_t swig_types[122]
#define SWIGTYPE_p_svn_ra_reporter_t swig_types[123]
#define SWIGTYPE_p_svn_ra_session_t swig_types[124]
#define SWIGTYPE_p_svn_stream_mark_t swig_types[125]
#define SWIGTYPE_p_svn_stream_t swig_types[126]
#define SWIGTYPE_p_svn_string_t swig_types[127]
#define SWIGTYPE_p_svn_stringbuf_t swig_types[128]
#define SWIGTYPE_p_svn_tristate_t swig_types[129]
#define SWIGTYPE_p_svn_txdelta_op_t swig_types[130]
#define SWIGTYPE_p_svn_txdelta_stream_t swig_types[131]
#define SWIGTYPE_p_svn_txdelta_window_t swig_types[132]
#define SWIGTYPE_p_svn_version_checklist_t swig_types[133]
#define SWIGTYPE_p_svn_version_ext_linked_lib_t swig_types[134]
#define SWIGTYPE_p_svn_version_ext_loaded_lib_t swig_types[135]
#define SWIGTYPE_p_svn_version_extended_t swig_types[136]
#define SWIGTYPE_p_svn_version_t swig_types[137]
#define SWIGTYPE_p_svn_wc_adm_access_t swig_types[138]
#define SWIGTYPE_p_svn_wc_committed_queue_t swig_types[139]
#define SWIGTYPE_p_svn_wc_conflict_action_t swig_types[140]
#define SWIGTYPE_p_svn_wc_conflict_choice_t swig_types[141]
#define SWIGTYPE_p_svn_wc_conflict_description_t swig_types[142]
#define SWIGTYPE_p_svn_wc_conflict_kind_t swig_types[143]
#define SWIGTYPE_p_svn_wc_conflict_reason_t swig_types[144]
#define SWIGTYPE_p_svn_wc_conflict_version_t swig_types[145]
#define SWIGTYPE_p_svn_wc_context_t swig_types[146]
#define SWIGTYPE_p_svn_wc_diff_callbacks2_t swig_types[147]
#define SWIGTYPE_p_svn_wc_diff_callbacks3_t swig_types[148]
#define SWIGTYPE_p_svn_wc_diff_callbacks4_t swig_types[149]
#define SWIGTYPE_p_svn_wc_diff_callbacks_t swig_types[150]
#define SWIGTYPE_p_svn_wc_entry_callbacks2_t swig_types[151]
#define SWIGTYPE_p_svn_wc_entry_callbacks_t swig_types[152]
#define SWIGTYPE_p_svn_wc_entry_t swig_types[153]
#define SWIGTYPE_p_svn_wc_external_item2_t swig_types[154]
#define SWIGTYPE_p_svn_wc_external_item_t swig_types[155]
#define SWIGTYPE_p_svn_wc_info_t swig_types[156]
#define SWIGTYPE_p_svn_wc_merge_outcome_t swig_types[157]
#define SWIGTYPE_p_svn_wc_notify_action_t swig_types[158]
#define SWIGTYPE_p_svn_wc_notify_lock_state_t swig_types[159]
#define SWIGTYPE_p_svn_wc_notify_state_t swig_types[160]
#define SWIGTYPE_p_svn_wc_notify_t swig_types[161]
#define SWIGTYPE_p_svn_wc_operation_t swig_types[162]
#define SWIGTYPE_p_svn_wc_revision_status_t swig_types[163]
#define SWIGTYPE_p_svn_wc_schedule_t swig_types[164]
#define SWIGTYPE_p_svn_wc_status2_t swig_types[165]
#define SWIGTYPE_p_svn_wc_status3_t swig_types[166]
#define SWIGTYPE_p_svn_wc_status_t swig_types[167]
#define SWIGTYPE_p_svn_wc_traversal_info_t swig_types[168]
#define SWIGTYPE_p_unsigned_long swig_types[169]
#define SWIGTYPE_p_void swig_types[170]
static swig_type_info *swig_types[172];
static swig_module_info swig_module = {swig_types, 171, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#define SWIG_init    Init_client
#define SWIG_name    "Svn::Ext::Client"

static VALUE mClient;

#define SWIG_RUBY_THREAD_BEGIN_BLOCK
#define SWIG_RUBY_THREAD_END_BLOCK


#define SWIGVERSION 0x020012 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#ifdef __cplusplus
extern "C" {
#endif

// Ruby 1.9 changed the file name of this header
#ifdef HAVE_RUBY_IO_H
#include "ruby/io.h"
#else
#include "rubyio.h"
#endif

#ifdef __cplusplus
}
#endif


#ifdef __cplusplus
extern "C" {
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
struct timeval rb_time_timeval(VALUE);
#endif
#ifdef __cplusplus
}
#endif


#include "svn_time.h"
#include "svn_pools.h"


#include "swigutil_rb.h"


static VALUE _global_svn_swig_rb_pool = Qnil;
static apr_pool_t *_global_pool = NULL;
static VALUE vresult = Qnil;
static VALUE *_global_vresult_address = &vresult;


#include <apu.h>
#include <apr_xlate.h>


#include "svn_client.h"


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


SWIGINTERN VALUE
SWIG_ruby_failed(void)
{
  return Qnil;
} 


/*@SWIG:/home/evgeny/release/svnrm/1.9.3/prefix/share/swig/2.0.12/ruby/rubyprimtypes.swg,19,%ruby_aux_method@*/
SWIGINTERN VALUE SWIG_AUX_NUM2LONG(VALUE *args)
{
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  long *res = (long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2LONG(obj) : rb_big2long(obj);
  return obj;
}
/*@SWIG@*/

SWIGINTERN int
SWIG_AsVal_long (VALUE obj, long* val)
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(RUBY_METHOD_FUNC(SWIG_AUX_NUM2LONG), (VALUE)a, RUBY_METHOD_FUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int (VALUE obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}


  #define SWIG_From_long   LONG2NUM 


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(VALUE obj, char** cptr, size_t* psize, int *alloc)
{
  if (TYPE(obj) == T_STRING) {
    char *cstr = StringValuePtr(obj); 
    size_t size = RSTRING_LEN(obj) + 1;
    if (cptr)  {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = (char *)memcpy((char *)malloc((size)*sizeof(char)), cstr, sizeof(char)*(size));
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      }
    }
    if (psize) *psize = size;
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *)vptr;
	if (psize) *psize = vptr ? (strlen((char*)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }  
  return SWIG_TypeError;
}





SWIGINTERNINLINE VALUE
SWIG_From_int  (int value)
{    
  return SWIG_From_long  (value);
}


/*@SWIG:/home/evgeny/release/svnrm/1.9.3/prefix/share/swig/2.0.12/ruby/rubyprimtypes.swg,19,%ruby_aux_method@*/
SWIGINTERN VALUE SWIG_AUX_NUM2ULONG(VALUE *args)
{
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  unsigned long *res = (unsigned long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2ULONG(obj) : rb_big2ulong(obj);
  return obj;
}
/*@SWIG@*/

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (VALUE obj, unsigned long *val) 
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    unsigned long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(RUBY_METHOD_FUNC(SWIG_AUX_NUM2ULONG), (VALUE)a, RUBY_METHOD_FUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERNINLINE VALUE
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return ULONG2NUM(value); 
}

SWIGINTERN struct svn_client_commit_item3_t *new_svn_client_commit_item3_t(apr_pool_t *pool){
    svn_error_t *err;
    const svn_client_commit_item3_t *self;
    err = svn_client_commit_item_create(&self, pool);
    if (err)
      svn_swig_rb_handle_svn_error(err);
    return (svn_client_commit_item3_t *)self;
  }
SWIGINTERN svn_client_commit_item3_t *svn_client_commit_item3_t_dup(struct svn_client_commit_item3_t *self,apr_pool_t *pool){
    return svn_client_commit_item3_dup(self, pool);
  }
SWIGINTERN struct svn_client_ctx_t *new_svn_client_ctx_t(apr_pool_t *pool){
    svn_error_t *err;
    svn_client_ctx_t *self;
    apr_hash_t *cfg_hash;

    err = svn_config_get_config(&cfg_hash, NULL, pool);
    if (err)
      svn_swig_rb_handle_svn_error(err);

    err = svn_client_create_context2(&self, cfg_hash, pool);
    if (err)
      svn_swig_rb_handle_svn_error(err);
    return self;
  }

SWIGINTERNINLINE VALUE 
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > LONG_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_NewPointerObj((char *)(carray), pchar_descriptor, 0) : Qnil;
    } else {
      return rb_str_new(carray, (long)(size));
    }
  } else {
    return Qnil;
  }
}


SWIGINTERNINLINE VALUE 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}

SWIGINTERN struct svn_client_copy_source_t *new_svn_client_copy_source_t(char const *path,svn_opt_revision_t const *rev,svn_opt_revision_t const *peg_rev,apr_pool_t *pool){
    svn_client_copy_source_t *self;
    svn_opt_revision_t *revision;
    svn_opt_revision_t *peg_revision;

    self = apr_palloc(pool, sizeof(*self));
    self->path = path ? apr_pstrdup(pool, path) : NULL;

    revision = apr_palloc(pool, sizeof(*revision));
    revision->kind = rev->kind;
    revision->value.number = rev->value.number;
    revision->value.date = rev->value.date;
    self->revision = revision;

    peg_revision = apr_palloc(pool, sizeof(*peg_revision));
    peg_revision->kind = peg_rev->kind;
    peg_revision->value.number = peg_rev->value.number;
    peg_revision->value.date = peg_rev->value.date;
    self->peg_revision = peg_revision;

    return self;
  }
SWIGINTERN char const *svn_client_copy_source_t__path(struct svn_client_copy_source_t *self){
    return self->path;
  }
SWIGINTERN svn_opt_revision_t const *svn_client_copy_source_t__revision(struct svn_client_copy_source_t *self){
    return self->revision;
  }
SWIGINTERN svn_opt_revision_t const *svn_client_copy_source_t__peg_revision(struct svn_client_copy_source_t *self){
    return self->peg_revision;
  }

static svn_error_t * svn_proplist_invoke_receiver2(
  svn_proplist_receiver2_t _obj, void *baton, const char *path, apr_hash_t *prop_hash, apr_array_header_t *inherited_props, apr_pool_t *scratch_pool) {
  return _obj(baton, path, prop_hash, inherited_props, scratch_pool);
}

static svn_error_t * svn_proplist_invoke_receiver(
  svn_proplist_receiver_t _obj, void *baton, const char *path, apr_hash_t *prop_hash, apr_pool_t *pool) {
  return _obj(baton, path, prop_hash, pool);
}

static svn_error_t * svn_client_invoke_get_commit_log3(
  svn_client_get_commit_log3_t _obj, const char **log_msg, const char **tmp_file, const apr_array_header_t *commit_items, void *baton, apr_pool_t *pool) {
  return _obj(log_msg, tmp_file, commit_items, baton, pool);
}

static svn_error_t * svn_client_invoke_get_commit_log2(
  svn_client_get_commit_log2_t _obj, const char **log_msg, const char **tmp_file, const apr_array_header_t *commit_items, void *baton, apr_pool_t *pool) {
  return _obj(log_msg, tmp_file, commit_items, baton, pool);
}

static svn_error_t * svn_client_invoke_get_commit_log(
  svn_client_get_commit_log_t _obj, const char **log_msg, const char **tmp_file, apr_array_header_t *commit_items, void *baton, apr_pool_t *pool) {
  return _obj(log_msg, tmp_file, commit_items, baton, pool);
}

static svn_error_t * svn_client_invoke_blame_receiver3(
  svn_client_blame_receiver3_t _obj, void *baton, svn_revnum_t start_revnum, svn_revnum_t end_revnum, apr_int64_t line_no, svn_revnum_t revision, apr_hash_t *rev_props, svn_revnum_t merged_revision, apr_hash_t *merged_rev_props, const char *merged_path, const char *line, svn_boolean_t local_change, apr_pool_t *pool) {
  return _obj(baton, start_revnum, end_revnum, line_no, revision, rev_props, merged_revision, merged_rev_props, merged_path, line, local_change, pool);
}

static svn_error_t * svn_client_invoke_blame_receiver2(
  svn_client_blame_receiver2_t _obj, void *baton, apr_int64_t line_no, svn_revnum_t revision, const char *author, const char *date, svn_revnum_t merged_revision, const char *merged_author, const char *merged_date, const char *merged_path, const char *line, apr_pool_t *pool) {
  return _obj(baton, line_no, revision, author, date, merged_revision, merged_author, merged_date, merged_path, line, pool);
}

static svn_error_t * svn_client_invoke_blame_receiver(
  svn_client_blame_receiver_t _obj, void *baton, apr_int64_t line_no, svn_revnum_t revision, const char *author, const char *date, const char *line, apr_pool_t *pool) {
  return _obj(baton, line_no, revision, author, date, line, pool);
}

static svn_error_t * svn_client_invoke_diff_summarize_func(
  svn_client_diff_summarize_func_t _obj, const svn_client_diff_summarize_t *diff, void *baton, apr_pool_t *pool) {
  return _obj(diff, baton, pool);
}

static svn_error_t * svn_client_invoke_import_filter_func(
  svn_client_import_filter_func_t _obj, void *baton, svn_boolean_t *filtered, const char *local_abspath, const svn_io_dirent2_t *dirent, apr_pool_t *scratch_pool) {
  return _obj(baton, filtered, local_abspath, dirent, scratch_pool);
}

static svn_error_t * svn_client_invoke_status_func(
  svn_client_status_func_t _obj, void *baton, const char *path, const svn_client_status_t *status, apr_pool_t *scratch_pool) {
  return _obj(baton, path, status, scratch_pool);
}

static svn_error_t * svn_client_invoke_list_func2(
  svn_client_list_func2_t _obj, void *baton, const char *path, const svn_dirent_t *dirent, const svn_lock_t *lock, const char *abs_path, const char *external_parent_url, const char *external_target, apr_pool_t *scratch_pool) {
  return _obj(baton, path, dirent, lock, abs_path, external_parent_url, external_target, scratch_pool);
}

static svn_error_t * svn_client_invoke_list_func(
  svn_client_list_func_t _obj, void *baton, const char *path, const svn_dirent_t *dirent, const svn_lock_t *lock, const char *abs_path, apr_pool_t *pool) {
  return _obj(baton, path, dirent, lock, abs_path, pool);
}

static svn_error_t * svn_info_invoke_receiver(
  svn_info_receiver_t _obj, void *baton, const char *path, const svn_info_t *info, apr_pool_t *pool) {
  return _obj(baton, path, info, pool);
}

static svn_error_t * svn_client_invoke_info_receiver2(
  svn_client_info_receiver2_t _obj, void *baton, const char *abspath_or_url, const svn_client_info2_t *info, apr_pool_t *scratch_pool) {
  return _obj(baton, abspath_or_url, info, scratch_pool);
}

static svn_error_t * svn_client_invoke_patch_func(
  svn_client_patch_func_t _obj, void *baton, svn_boolean_t *filtered, const char *canon_path_from_patchfile, const char *patch_abspath, const char *reject_abspath, apr_pool_t *scratch_pool) {
  return _obj(baton, filtered, canon_path_from_patchfile, patch_abspath, reject_abspath, scratch_pool);
}



static VALUE
svn_client_set_log_msg_func3(svn_client_ctx_t *ctx,
                             svn_client_get_commit_log3_t log_msg_func,
                             void *log_msg_baton,
                             apr_pool_t *pool)
{
  ctx->log_msg_func3 = log_msg_func;
  ctx->log_msg_baton3 = log_msg_baton;
  return (VALUE) log_msg_baton;
}

static VALUE
svn_client_set_notify_func2(svn_client_ctx_t *ctx,
                            svn_wc_notify_func2_t notify_func2,
                            void *notify_baton2,
                            apr_pool_t *pool)
{
  ctx->notify_func2 = notify_func2;
  ctx->notify_baton2 = notify_baton2;
  return (VALUE)notify_baton2;
}

static VALUE
svn_client_set_cancel_func(svn_client_ctx_t *ctx,
                           svn_cancel_func_t cancel_func,
                           void *cancel_baton,
                           apr_pool_t *pool)
{
  ctx->cancel_func = cancel_func;
  ctx->cancel_baton = cancel_baton;
  return (VALUE)cancel_baton;
}


static VALUE
svn_client_set_config(svn_client_ctx_t *ctx,
                      apr_hash_t *config,
                      apr_pool_t *pool)
{
  svn_error_t *err;

  apr_hash_clear(ctx->config);
  err = svn_config_copy_config(&ctx->config, config,
                               apr_hash_pool_get(ctx->config));
  if (err)
    svn_swig_rb_handle_svn_error(err);
  return Qnil;
}

static svn_error_t *
svn_client_get_config(svn_client_ctx_t *ctx,
                      apr_hash_t **cfg_hash,
                      apr_pool_t *pool)
{
  *cfg_hash = ctx->config;
  return SVN_NO_ERROR;
}

SWIGINTERN VALUE
_wrap_svn_client_version(int argc, VALUE *argv, VALUE self) {
  svn_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_version_t *)svn_client_version();
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_simple_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_simple_prompt_func_t arg2 = (svn_auth_simple_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_simple_prompt_func_t) svn_swig_rb_auth_simple_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_client_get_simple_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_client_get_simple_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_username_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_username_prompt_func_t arg2 = (svn_auth_username_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_username_prompt_func_t) svn_swig_rb_auth_username_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_client_get_username_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_client_get_username_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_simple_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_client_get_simple_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_username_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_client_get_username_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_ssl_server_trust_file_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_client_get_ssl_server_trust_file_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_ssl_client_cert_file_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_client_get_ssl_client_cert_file_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_ssl_client_cert_pw_file_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_client_get_ssl_client_cert_pw_file_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_ssl_server_trust_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_server_trust_prompt_func_t arg2 = (svn_auth_ssl_server_trust_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_server_trust_prompt_func_t) svn_swig_rb_auth_ssl_server_trust_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  if (argc > 1) {
    
  }
  {
    svn_client_get_ssl_server_trust_prompt_provider(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_ssl_client_cert_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_client_cert_prompt_func_t arg2 = (svn_auth_ssl_client_cert_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_client_cert_prompt_func_t) svn_swig_rb_auth_ssl_client_cert_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_client_get_ssl_client_cert_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_client_get_ssl_client_cert_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_ssl_client_cert_pw_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_client_cert_pw_prompt_func_t arg2 = (svn_auth_ssl_client_cert_pw_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_client_cert_pw_prompt_func_t) svn_swig_rb_auth_ssl_client_cert_pw_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_client_get_ssl_client_cert_pw_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_client_get_ssl_client_cert_pw_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_proplist_item_t;

SWIGINTERN VALUE
_wrap_svn_client_proplist_item_t_node_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_proplist_item_t *arg1 = (struct svn_client_proplist_item_t *) 0 ;
  svn_stringbuf_t *arg2 = (svn_stringbuf_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_proplist_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_proplist_item_t *","node_name", 1, self )); 
  }
  arg1 = (struct svn_client_proplist_item_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      
      arg2 = svn_stringbuf_ncreate(StringValuePtr(argv[0]),
        RSTRING_LEN(argv[0]),
        pool);
    }
  }
  if (arg1) (arg1)->node_name = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_proplist_item_t_node_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_proplist_item_t *arg1 = (struct svn_client_proplist_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_stringbuf_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_proplist_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_proplist_item_t *","node_name", 1, self )); 
  }
  arg1 = (struct svn_client_proplist_item_t *)(argp1);
  result = (svn_stringbuf_t *) ((arg1)->node_name);
  {
    vresult = rb_str_new(result->data, result->len);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_proplist_item_t_prop_hash_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_proplist_item_t *arg1 = (struct svn_client_proplist_item_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_proplist_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_proplist_item_t *","prop_hash", 1, self )); 
  }
  arg1 = (struct svn_client_proplist_item_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","prop_hash", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (arg1) (arg1)->prop_hash = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_proplist_item_t_prop_hash_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_proplist_item_t *arg1 = (struct svn_client_proplist_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_proplist_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_proplist_item_t *","prop_hash", 1, self )); 
  }
  arg1 = (struct svn_client_proplist_item_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->prop_hash);
  {
    vresult = svn_swig_rb_prop_hash_to_hash(result);
  }
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_proplist_item_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_proplist_item_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_proplist_item_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_proplist_item_t(int argc, VALUE *argv, VALUE self) {
  struct svn_client_proplist_item_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_client_proplist_item_t *)calloc(1, sizeof(struct svn_client_proplist_item_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_client_proplist_item_t(struct svn_client_proplist_item_t *arg1) {
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_client_proplist_item_dup(int argc, VALUE *argv, VALUE self) {
  svn_client_proplist_item_t *arg1 = (svn_client_proplist_item_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_proplist_item_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_proplist_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_proplist_item_t const *","svn_client_proplist_item_dup", 1, argv[0] )); 
  }
  arg1 = (svn_client_proplist_item_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_client_proplist_item_t *)svn_client_proplist_item_dup((struct svn_client_proplist_item_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_client_proplist_item_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_commit_info_t;

SWIGINTERN VALUE
_wrap_svn_client_commit_info_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_info_t *arg1 = (struct svn_client_commit_info_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_info_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_info_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_info_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_info_t *arg1 = (struct svn_client_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_info_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_info_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_info_t_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_info_t *arg1 = (struct svn_client_commit_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_info_t *","date", 1, self )); 
  }
  arg1 = (struct svn_client_commit_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","date", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->date) free((char *)arg1->date);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->date = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_info_t_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_info_t *arg1 = (struct svn_client_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_info_t *","date", 1, self )); 
  }
  arg1 = (struct svn_client_commit_info_t *)(argp1);
  result = (char *) ((arg1)->date);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_info_t_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_info_t *arg1 = (struct svn_client_commit_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_info_t *","author", 1, self )); 
  }
  arg1 = (struct svn_client_commit_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->author) free((char *)arg1->author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_info_t_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_info_t *arg1 = (struct svn_client_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_info_t *","author", 1, self )); 
  }
  arg1 = (struct svn_client_commit_info_t *)(argp1);
  result = (char *) ((arg1)->author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_commit_info_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_commit_info_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_commit_info_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_commit_info_t(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_info_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_client_commit_info_t *)calloc(1, sizeof(struct svn_client_commit_info_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_client_commit_info_t(struct svn_client_commit_info_t *arg1) {
    free((char *) arg1);
}

static swig_class SwigClassSvn_client_commit_item3_t;

SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_copyfrom_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_copyfrom_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_copyfrom_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","copyfrom_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->copyfrom_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_copyfrom_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_state_flags_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  apr_byte_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","state_flags", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_byte_t","state_flags", 2, argv[0] ));
  } 
  arg2 = (apr_byte_t)(val2);
  if (arg1) (arg1)->state_flags = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_state_flags_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_byte_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","state_flags", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result =  ((arg1)->state_flags);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_incoming_prop_changes_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","incoming_prop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg2 = svn_swig_rb_to_apr_array_prop(argv[0], pool);
    }
  }
  if (arg1) (arg1)->incoming_prop_changes = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_incoming_prop_changes_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","incoming_prop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (apr_array_header_t *) ((arg1)->incoming_prop_changes);
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, result ? svn_swig_rb_prop_apr_array_to_hash_prop(result) : Qnil);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_outgoing_prop_changes_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","outgoing_prop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg2 = svn_swig_rb_to_apr_array_prop(argv[0], pool);
    }
  }
  if (arg1) (arg1)->outgoing_prop_changes = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_outgoing_prop_changes_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","outgoing_prop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (apr_array_header_t *) ((arg1)->outgoing_prop_changes);
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, result ? svn_swig_rb_prop_apr_array_to_hash_prop(result) : Qnil);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_session_relpath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","session_relpath", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","session_relpath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->session_relpath) free((char *)arg1->session_relpath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->session_relpath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_session_relpath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","session_relpath", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (char *) ((arg1)->session_relpath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_moved_from_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","moved_from_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_from_abspath) free((char *)arg1->moved_from_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_from_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_moved_from_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  result = (char *) ((arg1)->moved_from_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_commit_item3_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_commit_item3_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_commit_item3_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_client_commit_item3_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (struct svn_client_commit_item3_t *)new_svn_client_commit_item3_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_client_commit_item3_t(struct svn_client_commit_item3_t *self){
  }
SWIGINTERN void
free_svn_client_commit_item3_t(struct svn_client_commit_item3_t *arg1) {
    delete_svn_client_commit_item3_t(arg1);
}


/*
  Document-method: Svn::Ext::Client::svn_client_commit_item3_t.dup

  call-seq:
    dup(pool) -> svn_client_commit_item3_t

Create a duplicate of the class and unfreeze it if needed.
*/
SWIGINTERN VALUE
_wrap_svn_client_commit_item3_t_dup(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item3_t *arg1 = (struct svn_client_commit_item3_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_commit_item3_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item3_t *","dup", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item3_t *)(argp1);
  if (argc > 0) {
    
  }
  {
    result = (svn_client_commit_item3_t *)svn_client_commit_item3_t_dup(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_commit_item2_t;

SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_copyfrom_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_copyfrom_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_copyfrom_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","copyfrom_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->copyfrom_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_copyfrom_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_state_flags_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  apr_byte_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","state_flags", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_byte_t","state_flags", 2, argv[0] ));
  } 
  arg2 = (apr_byte_t)(val2);
  if (arg1) (arg1)->state_flags = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_state_flags_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_byte_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","state_flags", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result =  ((arg1)->state_flags);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_wcprop_changes_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","wcprop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg2 = svn_swig_rb_to_apr_array_prop(argv[0], pool);
  }
  if (arg1) (arg1)->wcprop_changes = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_wcprop_changes_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *arg1 = (struct svn_client_commit_item2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item2_t *","wcprop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item2_t *)(argp1);
  result = (apr_array_header_t *) ((arg1)->wcprop_changes);
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(result));
  }
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_commit_item2_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_commit_item2_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_commit_item2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_commit_item2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_client_commit_item2_t *)calloc(1, sizeof(struct svn_client_commit_item2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_client_commit_item2_t(struct svn_client_commit_item2_t *arg1) {
    free((char *) arg1);
}

static swig_class SwigClassSvn_client_commit_item_t;

SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_copyfrom_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_copyfrom_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_state_flags_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  apr_byte_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","state_flags", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_byte_t","state_flags", 2, argv[0] ));
  } 
  arg2 = (apr_byte_t)(val2);
  if (arg1) (arg1)->state_flags = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_state_flags_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_byte_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","state_flags", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  result =  ((arg1)->state_flags);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_wcprop_changes_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","wcprop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg2 = svn_swig_rb_to_apr_array_prop(argv[0], pool);
  }
  if (arg1) (arg1)->wcprop_changes = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_wcprop_changes_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *arg1 = (struct svn_client_commit_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_commit_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_commit_item_t *","wcprop_changes", 1, self )); 
  }
  arg1 = (struct svn_client_commit_item_t *)(argp1);
  result = (apr_array_header_t *) ((arg1)->wcprop_changes);
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(result));
  }
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_commit_item_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_commit_item_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_commit_item_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_commit_item_t(int argc, VALUE *argv, VALUE self) {
  struct svn_client_commit_item_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_client_commit_item_t *)calloc(1, sizeof(struct svn_client_commit_item_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_client_commit_item_t(struct svn_client_commit_item_t *arg1) {
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_client_commit_item3_create(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_item3_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_client_commit_item3_t *)svn_client_commit_item3_create(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_client_commit_item3_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_diff_summarize_t;

SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","path", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_summarize_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  svn_client_diff_summarize_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","summarize_kind", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_client_diff_summarize_kind_t","summarize_kind", 2, argv[0] ));
  } 
  arg2 = (svn_client_diff_summarize_kind_t)(val2);
  if (arg1) (arg1)->summarize_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_summarize_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_diff_summarize_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","summarize_kind", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  result = (svn_client_diff_summarize_kind_t) ((arg1)->summarize_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_prop_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","prop_changed", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->prop_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_prop_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","prop_changed", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->prop_changed);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_node_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","node_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->node_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_node_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *arg1 = (struct svn_client_diff_summarize_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_diff_summarize_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_client_diff_summarize_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->node_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_diff_summarize_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_diff_summarize_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_diff_summarize_t(int argc, VALUE *argv, VALUE self) {
  struct svn_client_diff_summarize_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_client_diff_summarize_t *)calloc(1, sizeof(struct svn_client_diff_summarize_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_client_diff_summarize_t(struct svn_client_diff_summarize_t *arg1) {
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_dup(int argc, VALUE *argv, VALUE self) {
  svn_client_diff_summarize_t *arg1 = (svn_client_diff_summarize_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_diff_summarize_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_diff_summarize_t const *","svn_client_diff_summarize_dup", 1, argv[0] )); 
  }
  arg1 = (svn_client_diff_summarize_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_client_diff_summarize_t *)svn_client_diff_summarize_dup((struct svn_client_diff_summarize_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_ctx_t;

SWIGINTERN VALUE
_wrap_svn_client_ctx_t_auth_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_auth_baton_t *arg2 = (svn_auth_baton_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","auth_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_auth_baton_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_auth_baton_t *","auth_baton", 2, argv[0] )); 
  }
  arg2 = (svn_auth_baton_t *)(argp2);
  if (arg1) (arg1)->auth_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_auth_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_auth_baton_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","auth_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_auth_baton_t *) ((arg1)->auth_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_auth_baton_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_wc_notify_func_t arg2 = (svn_wc_notify_func_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","notify_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->notify_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_notify_func_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_wc_notify_func_t) ((arg1)->notify_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","notify_baton", 2, argv[0] )); 
  }
  if (arg1) (arg1)->notify_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->notify_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_client_get_commit_log_t arg2 = (svn_client_get_commit_log_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_get_commit_log_t","log_msg_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->log_msg_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_get_commit_log_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_client_get_commit_log_t) ((arg1)->log_msg_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","log_msg_baton", 2, argv[0] )); 
  }
  if (arg1) (arg1)->log_msg_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->log_msg_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_cancel_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_cancel_func_t arg2 = (svn_cancel_func_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","cancel_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_cancel_func_t","cancel_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->cancel_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_cancel_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_cancel_func_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","cancel_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_cancel_func_t) ((arg1)->cancel_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_cancel_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","cancel_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    arg2 = (void *)argv[0];
  }
  if (arg1) (arg1)->cancel_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_cancel_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","cancel_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->cancel_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_func2_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_wc_notify_func2_t arg2 = (svn_wc_notify_func2_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_func2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func2_t","notify_func2", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->notify_func2 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_func2_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_notify_func2_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_func2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_wc_notify_func2_t) ((arg1)->notify_func2);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_baton2_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_baton2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    arg2 = (void *)argv[0];
  }
  if (arg1) (arg1)->notify_baton2 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_notify_baton2_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","notify_baton2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->notify_baton2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_func2_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_client_get_commit_log2_t arg2 = (svn_client_get_commit_log2_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_func2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_get_commit_log2_t","log_msg_func2", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->log_msg_func2 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_func2_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_get_commit_log2_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_func2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_client_get_commit_log2_t) ((arg1)->log_msg_func2);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_baton2_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_baton2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","log_msg_baton2", 2, argv[0] )); 
  }
  if (arg1) (arg1)->log_msg_baton2 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_baton2_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_baton2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->log_msg_baton2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_progress_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_ra_progress_notify_func_t arg2 = (svn_ra_progress_notify_func_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","progress_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_ra_progress_notify_func_t","progress_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->progress_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_progress_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_ra_progress_notify_func_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","progress_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_ra_progress_notify_func_t) ((arg1)->progress_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_progress_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","progress_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","progress_baton", 2, argv[0] )); 
  }
  if (arg1) (arg1)->progress_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_progress_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","progress_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->progress_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_func3_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_client_get_commit_log3_t arg2 = (svn_client_get_commit_log3_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_func3", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_get_commit_log3_t","log_msg_func3", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->log_msg_func3 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_func3_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_get_commit_log3_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_func3", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_client_get_commit_log3_t) ((arg1)->log_msg_func3);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_baton3_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_baton3", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    arg2 = (void *)argv[0];
  }
  if (arg1) (arg1)->log_msg_baton3 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_log_msg_baton3_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","log_msg_baton3", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->log_msg_baton3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_mimetypes_map_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","mimetypes_map", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg2 = (NIL_P(argv[0])) ? NULL :
    svn_swig_rb_hash_to_apr_hash_string(argv[0], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg2)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[0], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (arg1) (arg1)->mimetypes_map = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_mimetypes_map_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","mimetypes_map", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->mimetypes_map);
  {
    vresult = result ? svn_swig_rb_apr_hash_to_hash_string(result) : Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_wc_conflict_resolver_func_t arg2 = (svn_wc_conflict_resolver_func_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func_t","conflict_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->conflict_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_resolver_func_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_wc_conflict_resolver_func_t) ((arg1)->conflict_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","conflict_baton", 2, argv[0] )); 
  }
  if (arg1) (arg1)->conflict_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->conflict_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_client_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","client_name", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","client_name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->client_name) free((char *)arg1->client_name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->client_name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_client_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","client_name", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (char *) ((arg1)->client_name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_func2_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_wc_conflict_resolver_func2_t arg2 = (svn_wc_conflict_resolver_func2_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_func2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func2_t","conflict_func2", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->conflict_func2 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_func2_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_resolver_func2_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_func2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_wc_conflict_resolver_func2_t) ((arg1)->conflict_func2);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_baton2_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_baton2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","conflict_baton2", 2, argv[0] )); 
  }
  if (arg1) (arg1)->conflict_baton2 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_conflict_baton2_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","conflict_baton2", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->conflict_baton2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_wc_ctx_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","wc_ctx", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","wc_ctx", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  if (arg1) (arg1)->wc_ctx = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_wc_ctx_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_context_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","wc_ctx", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_wc_context_t *) ((arg1)->wc_ctx);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_check_tunnel_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_ra_check_tunnel_func_t arg2 = (svn_ra_check_tunnel_func_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","check_tunnel_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_ra_check_tunnel_func_t","check_tunnel_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->check_tunnel_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_check_tunnel_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_ra_check_tunnel_func_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","check_tunnel_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_ra_check_tunnel_func_t) ((arg1)->check_tunnel_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char__int);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_open_tunnel_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  svn_ra_open_tunnel_func_t arg2 = (svn_ra_open_tunnel_func_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","open_tunnel_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_ra_open_tunnel_func_t","open_tunnel_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->open_tunnel_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_open_tunnel_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_ra_open_tunnel_func_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","open_tunnel_func", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (svn_ra_open_tunnel_func_t) ((arg1)->open_tunnel_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_tunnel_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","tunnel_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","tunnel_baton", 2, argv[0] )); 
  }
  if (arg1) (arg1)->tunnel_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ctx_t_tunnel_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_ctx_t *arg1 = (struct svn_client_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_ctx_t *","tunnel_baton", 1, self )); 
  }
  arg1 = (struct svn_client_ctx_t *)(argp1);
  result = (void *) ((arg1)->tunnel_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_ctx_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_ctx_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_ctx_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_ctx_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_client_ctx_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (struct svn_client_ctx_t *)new_svn_client_ctx_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_client_ctx_t(struct svn_client_ctx_t *self){
  }
SWIGINTERN void
free_svn_client_ctx_t(struct svn_client_ctx_t *arg1) {
    delete_svn_client_ctx_t(arg1);
}

SWIGINTERN VALUE
_wrap_svn_client_create_context2(int argc, VALUE *argv, VALUE self) {
  svn_client_ctx_t **arg1 = (svn_client_ctx_t **) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_ctx_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","svn_client_create_context2", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_client_create_context2(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_ctx_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_args_to_target_array2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_getopt_t *arg2 = (apr_getopt_t *) 0 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_getopt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_getopt_t *","svn_client_args_to_target_array2", 2, argv[0] )); 
  }
  arg2 = (apr_getopt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_client_args_to_target_array2", 3, argv[1] )); 
  }
  arg3 = (apr_array_header_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_args_to_target_array2", 4, argv[2] )); 
  }
  arg4 = (svn_client_ctx_t *)(argp4);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_args_to_target_array2(arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_string(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_args_to_target_array(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_getopt_t *arg2 = (apr_getopt_t *) 0 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_getopt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_getopt_t *","svn_client_args_to_target_array", 2, argv[0] )); 
  }
  arg2 = (apr_getopt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_client_args_to_target_array", 3, argv[1] )); 
  }
  arg3 = (apr_array_header_t *)(argp3);
  if (argc > 2) {
    res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_args_to_target_array", 4, argv[2] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_args_to_target_array(arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_string(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_checkout3(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_checkout3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_checkout3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_checkout3", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_checkout3(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_checkout2(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_checkout2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_checkout2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_checkout2", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_checkout2(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_checkout(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_checkout", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_checkout", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_checkout", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_checkout(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_update4(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  svn_opt_revision_t rev3 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[2]);
  }
  arg5 = RTEST(argv[3]);
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_update4", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_update4(arg1,(apr_array_header_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_svn_rev(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_update3(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  svn_opt_revision_t rev3 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[2]);
  }
  arg5 = RTEST(argv[3]);
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_update3", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_update3(arg1,(apr_array_header_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_svn_rev(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_update2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  svn_opt_revision_t rev3 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_update2", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_update2(arg1,(apr_array_header_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_svn_rev(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_update(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_update", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_update", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_update(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_switch3(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_switch3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_switch3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  arg10 = RTEST(argv[8]);
  if (argc > 9) {
    res11 = SWIG_ConvertPtr(argv[9], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_switch3", 11, argv[9] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_switch3(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_switch2(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_switch2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_switch2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_switch2", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_switch2(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_switch(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_switch", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_switch", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_switch", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_switch(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_add5(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_depth_t arg2 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_add5", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_add5", 7, argv[6] )); 
  }
  arg7 = (svn_client_ctx_t *)(argp7);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_add5((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_add4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_depth_t arg2 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_add4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_add4", 6, argv[5] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_add4((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_add3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_add3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  if (argc > 4) {
    res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_add3", 5, argv[4] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_add3((char const *)arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_add2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  svn_boolean_t arg3 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_add2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_add2", 4, argv[3] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_add2((char const *)arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_add(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_add", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  if (argc > 2) {
    res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_add", 3, argv[2] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_add((char const *)arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mkdir4(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg2 ;
  apr_hash_t *arg3 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg4 = (svn_commit_callback2_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg2 = RTEST(argv[1]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg3 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[2], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg3)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[2], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg4 = svn_swig_rb_commit_callback2;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mkdir4", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_mkdir4((apr_array_header_t const *)arg1,arg2,(apr_hash_t const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mkdir3(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg3 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg3 = RTEST(argv[1]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg4 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[2], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg4)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[2], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mkdir3", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_mkdir3(arg1,(apr_array_header_t const *)arg2,arg3,(apr_hash_t const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mkdir2(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  if (argc > 1) {
    res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mkdir2", 3, argv[1] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_client_mkdir2(arg1,(apr_array_header_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mkdir(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  if (argc > 1) {
    res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mkdir", 3, argv[1] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_client_mkdir(arg1,(apr_array_header_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_delete4(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg2 ;
  svn_boolean_t arg3 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg5 = (svn_commit_callback2_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg2 = RTEST(argv[1]);
  arg3 = RTEST(argv[2]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg4 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[3], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg4)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[3], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg5 = svn_swig_rb_commit_callback2;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_delete4", 7, argv[5] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_delete4((apr_array_header_t const *)arg1,arg2,arg3,(apr_hash_t const *)arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_delete3(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg5 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[3], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg5)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[3], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_delete3", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_delete3(arg1,(apr_array_header_t const *)arg2,arg3,arg4,(apr_hash_t const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_delete2(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_delete2", 4, argv[2] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_delete2(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_delete(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_delete", 4, argv[2] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_delete(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_import5(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_client_import_filter_func_t arg8 = (svn_client_import_filter_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_commit_callback2_t arg10 = (svn_commit_callback2_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res9 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_import5", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_import5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg7 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[6], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg7)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[6], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_import_filter_func_t","svn_client_import5", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_client_import5", 9, argv[8] )); 
  }
  {
    arg10 = svn_swig_rb_commit_callback2;
    arg11 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  res12 = SWIG_ConvertPtr(argv[10], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_import5", 12, argv[10] )); 
  }
  arg12 = (svn_client_ctx_t *)(argp12);
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_client_import5((char const *)arg1,(char const *)arg2,arg3,arg4,arg5,arg6,(apr_hash_t const *)arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_import4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg7 = (svn_commit_callback2_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_import4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_import4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg6)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg7 = svn_swig_rb_commit_callback2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_import4", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_client_import4((char const *)arg1,(char const *)arg2,arg3,arg4,arg5,(apr_hash_t const *)arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg8);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_import3(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_depth_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_import3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_import3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = svn_swig_rb_to_depth(argv[2]);
  }
  arg5 = RTEST(argv[3]);
  arg6 = RTEST(argv[4]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg7 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg7)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_import3", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_import3(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,(apr_hash_t const *)arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_import2(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_import2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_import2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_import2", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_import2(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_import(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_import", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_import", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_import", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_import(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit6(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_depth_t arg2 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg9 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg10 = (svn_commit_callback2_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (NIL_P(argv[7])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[7], _global_pool);
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg9 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[8], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg9)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[8], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg10 = svn_swig_rb_commit_callback2;
    arg11 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    res12 = SWIG_ConvertPtr(argv[10], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_commit6", 12, argv[10] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_client_commit6((apr_array_header_t const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,(apr_array_header_t const *)arg8,(apr_hash_t const *)arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit5(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_depth_t arg2 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_array_header_t *arg6 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg8 = (svn_commit_callback2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    arg6 = (NIL_P(argv[5])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[5], _global_pool);
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg7 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[6], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg7)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[6], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg8 = svn_swig_rb_commit_callback2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_commit5", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_client_commit5((apr_array_header_t const *)arg1,arg2,arg3,arg4,arg5,(apr_array_header_t const *)arg6,(apr_hash_t const *)arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit4(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_array_header_t *arg6 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg3 = svn_swig_rb_to_depth(argv[1]);
  }
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  {
    arg6 = (NIL_P(argv[4])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[4], _global_pool);
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg7 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg7)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_commit4", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_commit4(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5,(apr_array_header_t const *)arg6,(apr_hash_t const *)arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit3(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_commit3", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_commit3(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit2(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_commit2", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_commit2(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_commit(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_commit", 4, argv[2] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_commit(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_status_t;

SWIGINTERN VALUE
_wrap_svn_client_status_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_local_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","local_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","local_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->local_abspath) free((char *)arg1->local_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->local_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_local_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","local_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->local_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_filesize_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","filesize", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->filesize = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_filesize_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","filesize", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result =  ((arg1)->filesize);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_versioned_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","versioned", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->versioned = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_versioned_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","versioned", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->versioned);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_conflicted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","conflicted", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->conflicted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_conflicted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","conflicted", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->conflicted);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_node_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","node_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","node_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->node_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_node_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","node_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->node_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_wc_is_locked_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","wc_is_locked", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->wc_is_locked = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_wc_is_locked_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","wc_is_locked", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->wc_is_locked);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_copied_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->copied = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_copied_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->copied);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_root_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_root_url", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_root_url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_root_url) free((char *)arg1->repos_root_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_root_url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_root_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_root_url", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->repos_root_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_uuid_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_uuid", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_uuid", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_uuid) free((char *)arg1->repos_uuid);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_uuid = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_uuid_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_uuid", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->repos_uuid);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_relpath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_relpath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_relpath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_relpath) free((char *)arg1->repos_relpath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_relpath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_relpath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_relpath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->repos_relpath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changed_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changed_rev", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","changed_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->changed_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changed_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changed_rev", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->changed_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changed_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changed_date", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->changed_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changed_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changed_date", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result =  ((arg1)->changed_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changed_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changed_author", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","changed_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changed_author) free((char *)arg1->changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changed_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changed_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changed_author", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->changed_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_switched_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->switched = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_switched_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->switched);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_file_external_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","file_external", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->file_external = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_file_external_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","file_external", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->file_external);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t const *","lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->lock = (svn_lock_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changelist_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  {
    arg2 = StringValueCStr(argv[0]);
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_changelist_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->changelist);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_depth_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_depth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  {
    arg2 = svn_swig_rb_to_depth(argv[0]);
  }
  if (arg1) (arg1)->depth = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_depth_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_depth_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_depth_t) ((arg1)->depth);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_kind", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","ood_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->ood_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_kind", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->ood_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_node_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_node_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_node_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_node_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_node_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_node_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_node_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_lock", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t const *","repos_lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->repos_lock = (svn_lock_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_repos_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","repos_lock", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->repos_lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_changed_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","ood_changed_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->ood_changed_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_changed_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->ood_changed_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_changed_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_changed_date", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->ood_changed_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_changed_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_changed_date", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result =  ((arg1)->ood_changed_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_changed_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_changed_author", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","ood_changed_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->ood_changed_author) free((char *)arg1->ood_changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->ood_changed_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_ood_changed_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","ood_changed_author", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->ood_changed_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_backwards_compatibility_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","backwards_compatibility_baton", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void const *","backwards_compatibility_baton", 2, argv[0] )); 
  }
  if (arg1) (arg1)->backwards_compatibility_baton = (void const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_backwards_compatibility_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","backwards_compatibility_baton", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (void *) ((arg1)->backwards_compatibility_baton);
  {
    char *value = result;
    if (value) {
      vresult = rb_str_new2(value);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_moved_from_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","moved_from_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_from_abspath) free((char *)arg1->moved_from_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_from_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_moved_from_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->moved_from_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_moved_to_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","moved_to_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","moved_to_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_to_abspath) free((char *)arg1->moved_to_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_to_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status_t_moved_to_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *arg1 = (struct svn_client_status_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_status_t *","moved_to_abspath", 1, self )); 
  }
  arg1 = (struct svn_client_status_t *)(argp1);
  result = (char *) ((arg1)->moved_to_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_status_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_status_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_status_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_status_t(int argc, VALUE *argv, VALUE self) {
  struct svn_client_status_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_client_status_t *)calloc(1, sizeof(struct svn_client_status_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_client_status_t(struct svn_client_status_t *arg1) {
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_client_status_dup(int argc, VALUE *argv, VALUE self) {
  svn_client_status_t *arg1 = (svn_client_status_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_status_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_status_t const *","svn_client_status_dup", 1, argv[0] )); 
  }
  arg1 = (svn_client_status_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_client_status_t *)svn_client_status_dup((struct svn_client_status_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status6(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_client_ctx_t *arg2 = (svn_client_ctx_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  svn_client_status_func_t arg13 = (svn_client_status_func_t) 0 ;
  void *arg14 = (void *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  int res14 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 13) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 13)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_status6", 2, argv[0] )); 
  }
  arg2 = (svn_client_ctx_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_status6", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[3]);
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  arg10 = RTEST(argv[8]);
  arg11 = RTEST(argv[9]);
  {
    arg12 = (NIL_P(argv[10])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[10], _global_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[11], (void**)(&arg13), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_status_func_t","svn_client_status6", 13, argv[11] )); 
    }
  }
  res14 = SWIG_ConvertPtr(argv[12],SWIG_as_voidptrptr(&arg14), 0, 0);
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "void *","svn_client_status6", 14, argv[12] )); 
  }
  if (argc > 13) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_status6(arg1,arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,(apr_array_header_t const *)arg12,arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status5(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_client_ctx_t *arg2 = (svn_client_ctx_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  svn_client_status_func_t arg12 = (svn_client_status_func_t) 0 ;
  void *arg13 = (void *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  int res13 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 12) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_status5", 2, argv[0] )); 
  }
  arg2 = (svn_client_ctx_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_status5", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[3]);
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  arg10 = RTEST(argv[8]);
  {
    arg11 = (NIL_P(argv[9])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[9], _global_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[10], (void**)(&arg12), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_status_func_t","svn_client_status5", 12, argv[10] )); 
    }
  }
  res13 = SWIG_ConvertPtr(argv[11],SWIG_as_voidptrptr(&arg13), 0, 0);
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "void *","svn_client_status5", 13, argv[11] )); 
  }
  if (argc > 12) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_status5(arg1,arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status4(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_wc_status_func3_t arg4 = (svn_wc_status_func3_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res5 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_status4", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[2], (void**)(&arg4), SWIGTYPE_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func3_t","svn_client_status4", 4, argv[2] )); 
    }
  }
  res5 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_client_status4", 5, argv[3] )); 
  }
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  arg10 = RTEST(argv[8]);
  {
    arg11 = (NIL_P(argv[9])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[9], _global_pool);
  }
  if (argc > 10) {
    res12 = SWIG_ConvertPtr(argv[10], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_status4", 12, argv[10] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 11) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_status4(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status3(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_wc_status_func2_t arg4 = (svn_wc_status_func2_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_status3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    arg4 = (svn_wc_status_func2_t) svn_swig_rb_wc_status_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    arg6 = svn_swig_rb_to_depth(argv[3]);
  }
  arg7 = RTEST(argv[4]);
  arg8 = RTEST(argv[5]);
  arg9 = RTEST(argv[6]);
  arg10 = RTEST(argv[7]);
  {
    arg11 = (NIL_P(argv[8])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[8], _global_pool);
  }
  if (argc > 9) {
    res12 = SWIG_ConvertPtr(argv[9], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_status3", 12, argv[9] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_status3(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status2(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_wc_status_func2_t arg4 = (svn_wc_status_func2_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_status2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    arg4 = (svn_wc_status_func2_t) svn_swig_rb_wc_status_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  arg6 = RTEST(argv[3]);
  arg7 = RTEST(argv[4]);
  arg8 = RTEST(argv[5]);
  arg9 = RTEST(argv[6]);
  arg10 = RTEST(argv[7]);
  if (argc > 8) {
    res11 = SWIG_ConvertPtr(argv[8], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_status2", 11, argv[8] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 9) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_status2(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_status(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_wc_status_func_t arg4 = (svn_wc_status_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res5 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_status", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[2], (void**)(&arg4), SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func_t","svn_client_status", 4, argv[2] )); 
    }
  }
  res5 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_client_status", 5, argv[3] )); 
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_status", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_client_status(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_log5(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  int arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  svn_log_entry_receiver_t arg9 = (svn_log_entry_receiver_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev2 ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = svn_swig_rb_array_to_apr_array_revision_range(argv[2], _global_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_client_log5", 4, argv[3] ));
  } 
  arg4 = (int)(val4);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (NIL_P(argv[7])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[7], _global_pool);
  }
  {
    arg9 = (svn_log_entry_receiver_t) svn_swig_rb_log_entry_receiver;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  if (argc > 9) {
    res11 = SWIG_ConvertPtr(argv[9], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_log5", 11, argv[9] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_log5((apr_array_header_t const *)arg1,(struct svn_opt_revision_t const *)arg2,(apr_array_header_t const *)arg3,arg4,arg5,arg6,arg7,(apr_array_header_t const *)arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_log4(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  int arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  apr_array_header_t *arg9 = (apr_array_header_t *) 0 ;
  svn_log_entry_receiver_t arg10 = (svn_log_entry_receiver_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  ecode5 = SWIG_AsVal_int(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_client_log4", 5, argv[4] ));
  } 
  arg5 = (int)(val5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  {
    arg9 = (NIL_P(argv[8])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[8], _global_pool);
  }
  {
    arg10 = (svn_log_entry_receiver_t) svn_swig_rb_log_entry_receiver;
    arg11 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    res12 = SWIG_ConvertPtr(argv[10], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_log4", 12, argv[10] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 11) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_log4((apr_array_header_t const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8,(apr_array_header_t const *)arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_log3(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  int arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_log_message_receiver_t arg8 = (svn_log_message_receiver_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  ecode5 = SWIG_AsVal_int(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_client_log3", 5, argv[4] ));
  } 
  arg5 = (int)(val5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (svn_log_message_receiver_t) svn_swig_rb_log_receiver;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_log3", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_log3((apr_array_header_t const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_log2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  int arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_log_message_receiver_t arg7 = (svn_log_message_receiver_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  ecode4 = SWIG_AsVal_int(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_client_log2", 4, argv[3] ));
  } 
  arg4 = (int)(val4);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_log_message_receiver_t) svn_swig_rb_log_receiver;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_log2", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_client_log2((apr_array_header_t const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_log(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_log_message_receiver_t arg6 = (svn_log_message_receiver_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    arg6 = (svn_log_message_receiver_t) svn_swig_rb_log_receiver;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_log", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_log((apr_array_header_t const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_blame5(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_diff_file_options_t *arg5 = (svn_diff_file_options_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_client_blame_receiver3_t arg8 = (svn_client_blame_receiver3_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res9 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_blame5", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_client_blame5", 5, argv[4] )); 
  }
  arg5 = (svn_diff_file_options_t *)(argp5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_blame_receiver3_t","svn_client_blame5", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_client_blame5", 9, argv[8] )); 
  }
  if (argc > 9) {
    res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_blame5", 10, argv[9] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_blame5((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_diff_file_options_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_blame4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_diff_file_options_t *arg5 = (svn_diff_file_options_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_client_blame_receiver2_t arg8 = (svn_client_blame_receiver2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res9 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_blame4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_client_blame4", 5, argv[4] )); 
  }
  arg5 = (svn_diff_file_options_t *)(argp5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_blame_receiver2_t","svn_client_blame4", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_client_blame4", 9, argv[8] )); 
  }
  if (argc > 9) {
    res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_blame4", 10, argv[9] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_blame4((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_diff_file_options_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_blame3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_diff_file_options_t *arg5 = (svn_diff_file_options_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_client_blame_receiver_t arg7 = (svn_client_blame_receiver_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_blame3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_client_blame3", 5, argv[4] )); 
  }
  arg5 = (svn_diff_file_options_t *)(argp5);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_client_blame_receiver_t) svn_swig_rb_client_blame_receiver_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_blame3", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_blame3((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_diff_file_options_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_blame2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_client_blame_receiver_t arg5 = (svn_client_blame_receiver_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_blame2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = (svn_client_blame_receiver_t) svn_swig_rb_client_blame_receiver_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_blame2", 7, argv[5] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 6) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_blame2((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_blame(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_client_blame_receiver_t arg4 = (svn_client_blame_receiver_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_blame", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = (svn_client_blame_receiver_t) svn_swig_rb_client_blame_receiver_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_blame", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_blame((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff6(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  svn_boolean_t arg13 ;
  svn_boolean_t arg14 ;
  svn_boolean_t arg15 ;
  char *arg16 = (char *) 0 ;
  svn_stream_t *arg17 = (svn_stream_t *) 0 ;
  svn_stream_t *arg18 = (svn_stream_t *) 0 ;
  apr_array_header_t *arg19 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg20 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg21 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  void *argp20 = 0 ;
  int res20 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg21);
    _global_pool = arg21;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 19) || (argc > 21)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 19)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff6", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_diff6", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[6]);
  }
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  arg12 = RTEST(argv[11]);
  arg13 = RTEST(argv[12]);
  arg14 = RTEST(argv[13]);
  arg15 = RTEST(argv[14]);
  {
    arg16 = NULL;
    
    if (NIL_P(argv[15])) {
      
    } else if (TYPE(argv[15]) == T_FIXNUM) {
      arg16 = (char *)NUM2INT(argv[15]);
      if (!(arg16 == APR_LOCALE_CHARSET || arg16 == APR_DEFAULT_CHARSET)) {
        arg16 = NULL;
      }
    } else {
      arg16 = StringValuePtr(argv[15]);
    }
    
    if (!arg16) {
      arg16 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg17 = svn_swig_rb_make_stream(argv[16]);
  }
  {
    arg18 = svn_swig_rb_make_stream(argv[17]);
  }
  {
    arg19 = (NIL_P(argv[18])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[18], _global_pool);
  }
  if (argc > 19) {
    res20 = SWIG_ConvertPtr(argv[19], &argp20,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res20)) {
      SWIG_exception_fail(SWIG_ArgError(res20), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff6", 20, argv[19] )); 
    }
    arg20 = (svn_client_ctx_t *)(argp20);
  }
  if (argc > 20) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff6((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15,(char const *)arg16,arg17,arg18,(apr_array_header_t const *)arg19,arg20,arg21);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff5(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  char *arg13 = (char *) 0 ;
  apr_file_t *arg14 = (apr_file_t *) 0 ;
  apr_file_t *arg15 = (apr_file_t *) 0 ;
  apr_array_header_t *arg16 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg17 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg18 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  void *argp17 = 0 ;
  int res17 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg18);
    _global_pool = arg18;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 16) || (argc > 18)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 16)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_diff5", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[6]);
  }
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  arg12 = RTEST(argv[11]);
  {
    arg13 = NULL;
    
    if (NIL_P(argv[12])) {
      
    } else if (TYPE(argv[12]) == T_FIXNUM) {
      arg13 = (char *)NUM2INT(argv[12]);
      if (!(arg13 == APR_LOCALE_CHARSET || arg13 == APR_DEFAULT_CHARSET)) {
        arg13 = NULL;
      }
    } else {
      arg13 = StringValuePtr(argv[12]);
    }
    
    if (!arg13) {
      arg13 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg14 = svn_swig_rb_make_file(argv[13], _global_pool);
  }
  {
    arg15 = svn_swig_rb_make_file(argv[14], _global_pool);
  }
  {
    arg16 = (NIL_P(argv[15])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[15], _global_pool);
  }
  if (argc > 16) {
    res17 = SWIG_ConvertPtr(argv[16], &argp17,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res17)) {
      SWIG_exception_fail(SWIG_ArgError(res17), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff5", 17, argv[16] )); 
    }
    arg17 = (svn_client_ctx_t *)(argp17);
  }
  if (argc > 17) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff5((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,(char const *)arg13,arg14,arg15,(apr_array_header_t const *)arg16,arg17,arg18);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff4(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  char *arg11 = (char *) 0 ;
  apr_file_t *arg12 = (apr_file_t *) 0 ;
  apr_file_t *arg13 = (apr_file_t *) 0 ;
  apr_array_header_t *arg14 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg15 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg16 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg16);
    _global_pool = arg16;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 14) || (argc > 16)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_diff4", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[6]);
  }
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  {
    arg11 = NULL;
    
    if (NIL_P(argv[10])) {
      
    } else if (TYPE(argv[10]) == T_FIXNUM) {
      arg11 = (char *)NUM2INT(argv[10]);
      if (!(arg11 == APR_LOCALE_CHARSET || arg11 == APR_DEFAULT_CHARSET)) {
        arg11 = NULL;
      }
    } else {
      arg11 = StringValuePtr(argv[10]);
    }
    
    if (!arg11) {
      arg11 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg12 = svn_swig_rb_make_file(argv[11], _global_pool);
  }
  {
    arg13 = svn_swig_rb_make_file(argv[12], _global_pool);
  }
  {
    arg14 = (NIL_P(argv[13])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[13], _global_pool);
  }
  if (argc > 14) {
    res15 = SWIG_ConvertPtr(argv[14], &argp15,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res15)) {
      SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff4", 15, argv[14] )); 
    }
    arg15 = (svn_client_ctx_t *)(argp15);
  }
  if (argc > 15) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff4((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,(char const *)arg11,arg12,arg13,(apr_array_header_t const *)arg14,arg15,arg16);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff3(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  char *arg10 = (char *) 0 ;
  apr_file_t *arg11 = (apr_file_t *) 0 ;
  apr_file_t *arg12 = (apr_file_t *) 0 ;
  svn_client_ctx_t *arg13 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 12) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_diff3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  {
    arg10 = NULL;
    
    if (NIL_P(argv[9])) {
      
    } else if (TYPE(argv[9]) == T_FIXNUM) {
      arg10 = (char *)NUM2INT(argv[9]);
      if (!(arg10 == APR_LOCALE_CHARSET || arg10 == APR_DEFAULT_CHARSET)) {
        arg10 = NULL;
      }
    } else {
      arg10 = StringValuePtr(argv[9]);
    }
    
    if (!arg10) {
      arg10 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg11 = svn_swig_rb_make_file(argv[10], _global_pool);
  }
  {
    arg12 = svn_swig_rb_make_file(argv[11], _global_pool);
  }
  if (argc > 12) {
    res13 = SWIG_ConvertPtr(argv[12], &argp13,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res13)) {
      SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff3", 13, argv[12] )); 
    }
    arg13 = (svn_client_ctx_t *)(argp13);
  }
  if (argc > 13) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff3((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,(char const *)arg10,arg11,arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_file_t *arg10 = (apr_file_t *) 0 ;
  apr_file_t *arg11 = (apr_file_t *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_diff2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  {
    arg10 = svn_swig_rb_make_file(argv[9], _global_pool);
  }
  {
    arg11 = svn_swig_rb_make_file(argv[10], _global_pool);
  }
  if (argc > 11) {
    res12 = SWIG_ConvertPtr(argv[11], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff2", 12, argv[11] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff2((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  apr_file_t *arg9 = (apr_file_t *) 0 ;
  apr_file_t *arg10 = (apr_file_t *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_diff", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  {
    arg9 = svn_swig_rb_make_file(argv[8], _global_pool);
  }
  {
    arg10 = svn_swig_rb_make_file(argv[9], _global_pool);
  }
  if (argc > 10) {
    res11 = SWIG_ConvertPtr(argv[10], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff", 11, argv[10] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_peg6(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  svn_boolean_t arg13 ;
  svn_boolean_t arg14 ;
  svn_boolean_t arg15 ;
  char *arg16 = (char *) 0 ;
  svn_stream_t *arg17 = (svn_stream_t *) 0 ;
  svn_stream_t *arg18 = (svn_stream_t *) 0 ;
  apr_array_header_t *arg19 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg20 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg21 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp20 = 0 ;
  int res20 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg21);
    _global_pool = arg21;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 19) || (argc > 21)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 19)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff_peg6", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[6]);
  }
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  arg12 = RTEST(argv[11]);
  arg13 = RTEST(argv[12]);
  arg14 = RTEST(argv[13]);
  arg15 = RTEST(argv[14]);
  {
    arg16 = NULL;
    
    if (NIL_P(argv[15])) {
      
    } else if (TYPE(argv[15]) == T_FIXNUM) {
      arg16 = (char *)NUM2INT(argv[15]);
      if (!(arg16 == APR_LOCALE_CHARSET || arg16 == APR_DEFAULT_CHARSET)) {
        arg16 = NULL;
      }
    } else {
      arg16 = StringValuePtr(argv[15]);
    }
    
    if (!arg16) {
      arg16 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg17 = svn_swig_rb_make_stream(argv[16]);
  }
  {
    arg18 = svn_swig_rb_make_stream(argv[17]);
  }
  {
    arg19 = (NIL_P(argv[18])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[18], _global_pool);
  }
  if (argc > 19) {
    res20 = SWIG_ConvertPtr(argv[19], &argp20,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res20)) {
      SWIG_exception_fail(SWIG_ArgError(res20), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_peg6", 20, argv[19] )); 
    }
    arg20 = (svn_client_ctx_t *)(argp20);
  }
  if (argc > 20) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_peg6((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15,(char const *)arg16,arg17,arg18,(apr_array_header_t const *)arg19,arg20,arg21);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_peg5(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  char *arg13 = (char *) 0 ;
  apr_file_t *arg14 = (apr_file_t *) 0 ;
  apr_file_t *arg15 = (apr_file_t *) 0 ;
  apr_array_header_t *arg16 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg17 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg18 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp17 = 0 ;
  int res17 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg18);
    _global_pool = arg18;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 16) || (argc > 18)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 16)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff_peg5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[6]);
  }
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  arg12 = RTEST(argv[11]);
  {
    arg13 = NULL;
    
    if (NIL_P(argv[12])) {
      
    } else if (TYPE(argv[12]) == T_FIXNUM) {
      arg13 = (char *)NUM2INT(argv[12]);
      if (!(arg13 == APR_LOCALE_CHARSET || arg13 == APR_DEFAULT_CHARSET)) {
        arg13 = NULL;
      }
    } else {
      arg13 = StringValuePtr(argv[12]);
    }
    
    if (!arg13) {
      arg13 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg14 = svn_swig_rb_make_file(argv[13], _global_pool);
  }
  {
    arg15 = svn_swig_rb_make_file(argv[14], _global_pool);
  }
  {
    arg16 = (NIL_P(argv[15])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[15], _global_pool);
  }
  if (argc > 16) {
    res17 = SWIG_ConvertPtr(argv[16], &argp17,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res17)) {
      SWIG_exception_fail(SWIG_ArgError(res17), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_peg5", 17, argv[16] )); 
    }
    arg17 = (svn_client_ctx_t *)(argp17);
  }
  if (argc > 17) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_peg5((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,(char const *)arg13,arg14,arg15,(apr_array_header_t const *)arg16,arg17,arg18);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_peg4(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  char *arg11 = (char *) 0 ;
  apr_file_t *arg12 = (apr_file_t *) 0 ;
  apr_file_t *arg13 = (apr_file_t *) 0 ;
  apr_array_header_t *arg14 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg15 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg16 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg16);
    _global_pool = arg16;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 14) || (argc > 16)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff_peg4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[6]);
  }
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  {
    arg11 = NULL;
    
    if (NIL_P(argv[10])) {
      
    } else if (TYPE(argv[10]) == T_FIXNUM) {
      arg11 = (char *)NUM2INT(argv[10]);
      if (!(arg11 == APR_LOCALE_CHARSET || arg11 == APR_DEFAULT_CHARSET)) {
        arg11 = NULL;
      }
    } else {
      arg11 = StringValuePtr(argv[10]);
    }
    
    if (!arg11) {
      arg11 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg12 = svn_swig_rb_make_file(argv[11], _global_pool);
  }
  {
    arg13 = svn_swig_rb_make_file(argv[12], _global_pool);
  }
  {
    arg14 = (NIL_P(argv[13])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[13], _global_pool);
  }
  if (argc > 14) {
    res15 = SWIG_ConvertPtr(argv[14], &argp15,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res15)) {
      SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_peg4", 15, argv[14] )); 
    }
    arg15 = (svn_client_ctx_t *)(argp15);
  }
  if (argc > 15) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_peg4((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,(char const *)arg11,arg12,arg13,(apr_array_header_t const *)arg14,arg15,arg16);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_peg3(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  char *arg10 = (char *) 0 ;
  apr_file_t *arg11 = (apr_file_t *) 0 ;
  apr_file_t *arg12 = (apr_file_t *) 0 ;
  svn_client_ctx_t *arg13 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 12) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff_peg3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  {
    arg10 = NULL;
    
    if (NIL_P(argv[9])) {
      
    } else if (TYPE(argv[9]) == T_FIXNUM) {
      arg10 = (char *)NUM2INT(argv[9]);
      if (!(arg10 == APR_LOCALE_CHARSET || arg10 == APR_DEFAULT_CHARSET)) {
        arg10 = NULL;
      }
    } else {
      arg10 = StringValuePtr(argv[9]);
    }
    
    if (!arg10) {
      arg10 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    arg11 = svn_swig_rb_make_file(argv[10], _global_pool);
  }
  {
    arg12 = svn_swig_rb_make_file(argv[11], _global_pool);
  }
  if (argc > 12) {
    res13 = SWIG_ConvertPtr(argv[12], &argp13,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res13)) {
      SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_peg3", 13, argv[12] )); 
    }
    arg13 = (svn_client_ctx_t *)(argp13);
  }
  if (argc > 13) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_peg3((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,(char const *)arg10,arg11,arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_peg2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_file_t *arg10 = (apr_file_t *) 0 ;
  apr_file_t *arg11 = (apr_file_t *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff_peg2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  {
    arg10 = svn_swig_rb_make_file(argv[9], _global_pool);
  }
  {
    arg11 = svn_swig_rb_make_file(argv[10], _global_pool);
  }
  if (argc > 11) {
    res12 = SWIG_ConvertPtr(argv[11], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_peg2", 12, argv[11] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 12) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_peg2((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_peg(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  apr_file_t *arg9 = (apr_file_t *) 0 ;
  apr_file_t *arg10 = (apr_file_t *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_diff_peg", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  {
    arg9 = svn_swig_rb_make_file(argv[8], _global_pool);
  }
  {
    arg10 = svn_swig_rb_make_file(argv[9], _global_pool);
  }
  if (argc > 10) {
    res11 = SWIG_ConvertPtr(argv[10], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_peg", 11, argv[10] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 11) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_peg((apr_array_header_t const *)arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_client_diff_summarize_func_t arg8 = (svn_client_diff_summarize_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_diff_summarize2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_diff_summarize2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  {
    arg7 = (NIL_P(argv[6])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[6], _global_pool);
  }
  {
    arg8 = (svn_client_diff_summarize_func_t) svn_swig_rb_client_diff_summarize_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_summarize2", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff_summarize2((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_client_diff_summarize_func_t arg7 = (svn_client_diff_summarize_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_diff_summarize", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_diff_summarize", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_client_diff_summarize_func_t) svn_swig_rb_client_diff_summarize_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_summarize", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_client_diff_summarize((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_peg2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_client_diff_summarize_func_t arg8 = (svn_client_diff_summarize_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_diff_summarize_peg2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  {
    arg7 = (NIL_P(argv[6])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[6], _global_pool);
  }
  {
    arg8 = (svn_client_diff_summarize_func_t) svn_swig_rb_client_diff_summarize_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_summarize_peg2", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_summarize_peg2((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_diff_summarize_peg(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_client_diff_summarize_func_t arg7 = (svn_client_diff_summarize_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_diff_summarize_peg", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_client_diff_summarize_func_t) svn_swig_rb_client_diff_summarize_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_diff_summarize_peg", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_diff_summarize_peg((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_merging_summary(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  char **arg2 = (char **) 0 ;
  svn_revnum_t *arg3 = (svn_revnum_t *) 0 ;
  char **arg4 = (char **) 0 ;
  svn_revnum_t *arg5 = (svn_revnum_t *) 0 ;
  char **arg6 = (char **) 0 ;
  svn_revnum_t *arg7 = (svn_revnum_t *) 0 ;
  char **arg8 = (char **) 0 ;
  svn_revnum_t *arg9 = (svn_revnum_t *) 0 ;
  char **arg10 = (char **) 0 ;
  char *arg11 = (char *) 0 ;
  svn_opt_revision_t *arg12 = (svn_opt_revision_t *) 0 ;
  char *arg13 = (char *) 0 ;
  svn_opt_revision_t *arg14 = (svn_opt_revision_t *) 0 ;
  svn_client_ctx_t *arg15 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg16 = (apr_pool_t *) 0 ;
  apr_pool_t *arg17 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  char *temp2 ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  char *temp4 ;
  svn_revnum_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  char *temp6 ;
  svn_revnum_t temp7 ;
  int res7 = SWIG_TMPOBJ ;
  char *temp8 ;
  svn_revnum_t temp9 ;
  int res9 = SWIG_TMPOBJ ;
  char *temp10 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  svn_opt_revision_t rev12 ;
  int res13 ;
  char *buf13 = 0 ;
  int alloc13 = 0 ;
  svn_opt_revision_t rev14 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg16);
    _global_pool = arg16;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg17);
    _global_pool = arg17;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  arg6 = &temp6;
  arg7 = &temp7;
  arg8 = &temp8;
  arg9 = &temp9;
  arg10 = &temp10;
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res11 = SWIG_AsCharPtrAndSize(argv[0], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_client_get_merging_summary", 11, argv[0] ));
  }
  arg11 = (char *)(buf11);
  {
    arg12 = &rev12;
    svn_swig_rb_set_revision(&rev12, argv[1]);
  }
  res13 = SWIG_AsCharPtrAndSize(argv[2], &buf13, NULL, &alloc13);
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "char const *","svn_client_get_merging_summary", 13, argv[2] ));
  }
  arg13 = (char *)(buf13);
  {
    arg14 = &rev14;
    svn_swig_rb_set_revision(&rev14, argv[3]);
  }
  res15 = SWIG_ConvertPtr(argv[4], &argp15,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_get_merging_summary", 15, argv[4] )); 
  }
  arg15 = (svn_client_ctx_t *)(argp15);
  if (argc > 5) {
    
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_get_merging_summary(arg1,(char const **)arg2,arg3,(char const **)arg4,arg5,(char const **)arg6,arg7,(char const **)arg8,arg9,(char const **)arg10,(char const *)arg11,(struct svn_opt_revision_t const *)arg12,(char const *)arg13,(struct svn_opt_revision_t const *)arg14,arg15,arg16,arg17);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse);
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (SWIG_IsTmpObj(res3)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags));
  }
  {
    if (*arg4) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg4));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (SWIG_IsTmpObj(res5)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_long, new_flags));
  }
  {
    if (*arg6) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg6));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (SWIG_IsTmpObj(res7)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg7)));
  } else {
    int new_flags = SWIG_IsNewObj(res7) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg7), SWIGTYPE_p_long, new_flags));
  }
  {
    if (*arg8) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg8));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (SWIG_IsTmpObj(res9)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg9)));
  } else {
    int new_flags = SWIG_IsNewObj(res9) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg9), SWIGTYPE_p_long, new_flags));
  }
  {
    if (*arg10) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg10));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  if (alloc13 == SWIG_NEWOBJ) free((char*)buf13);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  if (alloc13 == SWIG_NEWOBJ) free((char*)buf13);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge5(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  apr_array_header_t *arg13 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg14 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 13) || (argc > 15)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 13)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge5", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_merge5", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_merge5", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  {
    arg6 = svn_swig_rb_to_depth(argv[5]);
  }
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  arg12 = RTEST(argv[11]);
  {
    if (NIL_P(argv[12])) {
      arg13 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg13 = svn_swig_rb_to_apr_array_prop(argv[12], pool);
    }
  }
  if (argc > 13) {
    res14 = SWIG_ConvertPtr(argv[13], &argp14,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res14)) {
      SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge5", 14, argv[13] )); 
    }
    arg14 = (svn_client_ctx_t *)(argp14);
  }
  if (argc > 14) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge5((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,(apr_array_header_t const *)arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg13 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 12) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_merge4", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_merge4", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  {
    arg6 = svn_swig_rb_to_depth(argv[5]);
  }
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  {
    if (NIL_P(argv[11])) {
      arg12 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg12 = svn_swig_rb_to_apr_array_prop(argv[11], pool);
    }
  }
  if (argc > 12) {
    res13 = SWIG_ConvertPtr(argv[12], &argp13,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res13)) {
      SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge4", 13, argv[12] )); 
    }
    arg13 = (svn_client_ctx_t *)(argp13);
  }
  if (argc > 13) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge4((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,(apr_array_header_t const *)arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_merge3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_merge3", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  {
    arg6 = svn_swig_rb_to_depth(argv[5]);
  }
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  {
    if (NIL_P(argv[10])) {
      arg11 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg11 = svn_swig_rb_to_apr_array_prop(argv[10], pool);
    }
  }
  if (argc > 11) {
    res12 = SWIG_ConvertPtr(argv[11], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge3", 12, argv[11] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge3((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_array_header_t *arg10 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_merge2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_merge2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  {
    if (NIL_P(argv[9])) {
      arg10 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg10 = svn_swig_rb_to_apr_array_prop(argv[9], pool);
    }
  }
  if (argc > 10) {
    res11 = SWIG_ConvertPtr(argv[10], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge2", 11, argv[10] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge2((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_merge", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_merge", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  if (argc > 9) {
    res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge", 10, argv[9] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge_reintegrate(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge_reintegrate", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_merge_reintegrate", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
    }
  }
  if (argc > 5) {
    res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge_reintegrate", 6, argv[5] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge_reintegrate((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge_peg5(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg13 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 12) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg5", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = svn_swig_rb_array_to_apr_array_revision_range(argv[1], _global_pool);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg5", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  {
    if (NIL_P(argv[11])) {
      arg12 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg12 = svn_swig_rb_to_apr_array_prop(argv[11], pool);
    }
  }
  if (argc > 12) {
    res13 = SWIG_ConvertPtr(argv[12], &argp13,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res13)) {
      SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge_peg5", 13, argv[12] )); 
    }
    arg13 = (svn_client_ctx_t *)(argp13);
  }
  if (argc > 13) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge_peg5((char const *)arg1,(apr_array_header_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,(apr_array_header_t const *)arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge_peg4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = svn_swig_rb_array_to_apr_array_revision_range(argv[1], _global_pool);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg4", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  arg10 = RTEST(argv[9]);
  {
    if (NIL_P(argv[10])) {
      arg11 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg11 = svn_swig_rb_to_apr_array_prop(argv[10], pool);
    }
  }
  if (argc > 11) {
    res12 = SWIG_ConvertPtr(argv[11], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge_peg4", 12, argv[11] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_client_merge_peg4((char const *)arg1,(apr_array_header_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge_peg3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_array_header_t *arg10 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = svn_swig_rb_array_to_apr_array_revision_range(argv[1], _global_pool);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  {
    if (NIL_P(argv[9])) {
      arg10 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg10 = svn_swig_rb_to_apr_array_prop(argv[9], pool);
    }
  }
  if (argc > 10) {
    res11 = SWIG_ConvertPtr(argv[10], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge_peg3", 11, argv[10] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 11) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_merge_peg3((char const *)arg1,(apr_array_header_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge_peg2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_array_header_t *arg10 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  {
    if (NIL_P(argv[9])) {
      arg10 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg10 = svn_swig_rb_to_apr_array_prop(argv[9], pool);
    }
  }
  if (argc > 10) {
    res11 = SWIG_ConvertPtr(argv[10], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge_peg2", 11, argv[10] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 11) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_merge_peg2((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_merge_peg(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_merge_peg", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  if (argc > 9) {
    res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_merge_peg", 10, argv[9] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_merge_peg((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_suggest_merge_sources(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_suggest_merge_sources", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  if (argc > 2) {
    res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_suggest_merge_sources", 4, argv[2] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 3) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_suggest_merge_sources(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_string(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mergeinfo_get_merged(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_get_merged", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  if (argc > 2) {
    res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mergeinfo_get_merged", 4, argv[2] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 3) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_mergeinfo_get_merged(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mergeinfo_log2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t arg1 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg6 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg7 = (svn_opt_revision_t *) 0 ;
  svn_log_entry_receiver_t arg8 = (svn_log_entry_receiver_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_boolean_t arg10 ;
  svn_depth_t arg11 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg13 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  svn_opt_revision_t rev6 ;
  svn_opt_revision_t rev7 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 12) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  arg1 = RTEST(argv[0]);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    arg6 = &rev6;
    svn_swig_rb_set_revision(&rev6, argv[5]);
  }
  {
    arg7 = &rev7;
    svn_swig_rb_set_revision(&rev7, argv[6]);
  }
  {
    arg8 = (svn_log_entry_receiver_t) svn_swig_rb_log_entry_receiver;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  arg10 = RTEST(argv[8]);
  {
    arg11 = svn_swig_rb_to_depth(argv[9]);
  }
  {
    arg12 = (NIL_P(argv[10])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[10], _global_pool);
  }
  res13 = SWIG_ConvertPtr(argv[11], &argp13,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mergeinfo_log2", 13, argv[11] )); 
  }
  arg13 = (svn_client_ctx_t *)(argp13);
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_client_mergeinfo_log2(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,(struct svn_opt_revision_t const *)arg6,(struct svn_opt_revision_t const *)arg7,arg8,arg9,arg10,arg11,(apr_array_header_t const *)arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mergeinfo_log(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t arg1 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_log_entry_receiver_t arg6 = (svn_log_entry_receiver_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_boolean_t arg8 ;
  svn_depth_t arg9 ;
  apr_array_header_t *arg10 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  arg1 = RTEST(argv[0]);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  {
    arg6 = (svn_log_entry_receiver_t) svn_swig_rb_log_entry_receiver;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  arg8 = RTEST(argv[6]);
  {
    arg9 = svn_swig_rb_to_depth(argv[7]);
  }
  {
    arg10 = (NIL_P(argv[8])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[8], _global_pool);
  }
  res11 = SWIG_ConvertPtr(argv[9], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mergeinfo_log", 11, argv[9] )); 
  }
  arg11 = (svn_client_ctx_t *)(argp11);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_client_mergeinfo_log(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mergeinfo_log_merged(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_log_entry_receiver_t arg5 = (svn_log_entry_receiver_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_boolean_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log_merged", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log_merged", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = (svn_log_entry_receiver_t) svn_swig_rb_log_entry_receiver;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  arg7 = RTEST(argv[5]);
  {
    arg8 = (NIL_P(argv[6])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[6], _global_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mergeinfo_log_merged", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_mergeinfo_log_merged((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,(apr_array_header_t const *)arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_mergeinfo_log_eligible(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_log_entry_receiver_t arg5 = (svn_log_entry_receiver_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_boolean_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log_eligible", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_mergeinfo_log_eligible", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  {
    arg5 = (svn_log_entry_receiver_t) svn_swig_rb_log_entry_receiver;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  arg7 = RTEST(argv[5]);
  {
    arg8 = (NIL_P(argv[6])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[6], _global_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_mergeinfo_log_eligible", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_mergeinfo_log_eligible((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,(apr_array_header_t const *)arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_vacuum(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_vacuum", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_vacuum", 7, argv[6] )); 
  }
  arg7 = (svn_client_ctx_t *)(argp7);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_vacuum((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_cleanup2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_cleanup2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_cleanup2", 7, argv[6] )); 
  }
  arg7 = (svn_client_ctx_t *)(argp7);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_cleanup2((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_cleanup(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_client_ctx_t *arg2 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_cleanup", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_cleanup", 2, argv[1] )); 
  }
  arg2 = (svn_client_ctx_t *)(argp2);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_client_cleanup((char const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_upgrade(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_client_ctx_t *arg2 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_upgrade", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_upgrade", 2, argv[1] )); 
  }
  arg2 = (svn_client_ctx_t *)(argp2);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_client_upgrade((char const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_relocate2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_relocate2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_relocate2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_relocate2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  if (argc > 4) {
    res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_relocate2", 5, argv[4] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_relocate2((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_relocate(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_relocate", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_relocate", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_relocate", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  if (argc > 4) {
    res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_relocate", 5, argv[4] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_relocate((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_revert3(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_depth_t arg2 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  {
    arg3 = (NIL_P(argv[2])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[2], _global_pool);
  }
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_revert3", 6, argv[5] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_revert3((apr_array_header_t const *)arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_revert2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_depth_t arg2 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  {
    arg3 = (NIL_P(argv[2])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[2], _global_pool);
  }
  if (argc > 3) {
    res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_revert2", 4, argv[3] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_revert2((apr_array_header_t const *)arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_revert(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg2 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg2 = RTEST(argv[1]);
  if (argc > 2) {
    res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_revert", 3, argv[2] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_revert((apr_array_header_t const *)arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_resolved(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_resolved", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  if (argc > 2) {
    res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_resolved", 3, argv[2] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_resolved((char const *)arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_resolve(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_depth_t arg2 ;
  svn_wc_conflict_choice_t arg3 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_resolve", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_wc_conflict_choice_t","svn_client_resolve", 3, argv[2] ));
  } 
  arg3 = (svn_wc_conflict_choice_t)(val3);
  if (argc > 3) {
    res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_resolve", 4, argv[3] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_resolve((char const *)arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_copy_source_t;

#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_copy_source_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_copy_source_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_copy_source_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_copy_source_t(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  struct svn_client_copy_source_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_copy_source_t", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  if (argc > 3) {
    
  }
  {
    result = (struct svn_client_copy_source_t *)new_svn_client_copy_source_t((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4);
    DATA_PTR(self) = result;
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_client_copy_source_t(struct svn_client_copy_source_t *self){
  }
SWIGINTERN void
free_svn_client_copy_source_t(struct svn_client_copy_source_t *arg1) {
    delete_svn_client_copy_source_t(arg1);
}

SWIGINTERN VALUE
_wrap_svn_client_copy_source_t_path(int argc, VALUE *argv, VALUE self) {
  struct svn_client_copy_source_t *arg1 = (struct svn_client_copy_source_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_copy_source_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_copy_source_t *","_path", 1, self )); 
  }
  arg1 = (struct svn_client_copy_source_t *)(argp1);
  {
    result = (char *)svn_client_copy_source_t__path(arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy_source_t_revision(int argc, VALUE *argv, VALUE self) {
  struct svn_client_copy_source_t *arg1 = (struct svn_client_copy_source_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_copy_source_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_copy_source_t *","_revision", 1, self )); 
  }
  arg1 = (struct svn_client_copy_source_t *)(argp1);
  {
    result = (svn_opt_revision_t *)svn_client_copy_source_t__revision(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy_source_t_peg_revision(int argc, VALUE *argv, VALUE self) {
  struct svn_client_copy_source_t *arg1 = (struct svn_client_copy_source_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_copy_source_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_copy_source_t *","_peg_revision", 1, self )); 
  }
  arg1 = (struct svn_client_copy_source_t *)(argp1);
  {
    result = (svn_opt_revision_t *)svn_client_copy_source_t__peg_revision(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy7(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  apr_hash_t *arg8 = (apr_hash_t *) 0 ;
  apr_hash_t *arg9 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg10 = (svn_commit_callback2_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_client_ctx_t *arg12 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_array_to_apr_array_copy_source(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_copy7", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "apr_hash_t const *","svn_client_copy7", 8, argv[7] )); 
  }
  arg8 = (apr_hash_t *)(argp8);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg9 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[8], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg9)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[8], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg10 = svn_swig_rb_commit_callback2;
    arg11 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    res12 = SWIG_ConvertPtr(argv[10], &argp12,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res12)) {
      SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_copy7", 12, argv[10] )); 
    }
    arg12 = (svn_client_ctx_t *)(argp12);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_client_copy7((apr_array_header_t const *)arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,(apr_hash_t const *)arg8,(apr_hash_t const *)arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy6(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg7 = (svn_commit_callback2_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_array_to_apr_array_copy_source(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_copy6", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg6)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg7 = svn_swig_rb_commit_callback2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_copy6", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_client_copy6((apr_array_header_t const *)arg1,(char const *)arg2,arg3,arg4,arg5,(apr_hash_t const *)arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg8);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy5(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_copy_source(argv[0], _global_pool);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_copy5", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  arg6 = RTEST(argv[4]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg7 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg7)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_copy5", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_copy5(arg1,(apr_array_header_t const *)arg2,(char const *)arg3,arg4,arg5,arg6,(apr_hash_t const *)arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy4(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_copy_source(argv[0], _global_pool);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_copy4", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[4], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg6)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[4], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 5) {
    res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_copy4", 7, argv[5] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_copy4(arg1,(apr_array_header_t const *)arg2,(char const *)arg3,arg4,arg5,(apr_hash_t const *)arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy3(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_copy3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_copy3", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_copy3", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_copy3(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy2(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_copy2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_copy2", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_copy2", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_copy2(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_copy(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_copy", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_copy", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_copy", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_copy(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_move7(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg8 = (svn_commit_callback2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_move7", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg7 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[6], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg7)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[6], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg8 = svn_swig_rb_commit_callback2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_move7", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_client_move7((apr_array_header_t const *)arg1,(char const *)arg2,arg3,arg4,arg5,arg6,(apr_hash_t const *)arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_move6(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg6 = (svn_commit_callback2_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_move6", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg5 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[4], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg5)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[4], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg6 = svn_swig_rb_commit_callback2;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_move6", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_move6((apr_array_header_t const *)arg1,(char const *)arg2,arg3,arg4,(apr_hash_t const *)arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_move5(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_move5", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  arg6 = RTEST(argv[4]);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg7 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg7)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_move5", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_move5(arg1,(apr_array_header_t const *)arg2,(char const *)arg3,arg4,arg5,arg6,(apr_hash_t const *)arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_move4(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_move4", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_move4", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_move4", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_move4(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_move3(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_move3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_move3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_move3", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_move3(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_move2(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_move2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_move2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_move2", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_move2(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_move(int argc, VALUE *argv, VALUE self) {
  svn_client_commit_info_t **arg1 = (svn_client_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_client_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_move", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_move", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_move", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_move(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(char const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_client_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propset_remote(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_revnum_t arg5 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  svn_commit_callback2_t arg7 = (svn_commit_callback2_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_propset_remote", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propset_remote", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  ecode5 = SWIG_AsVal_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_propset_remote", 5, argv[4] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg6)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg7 = svn_swig_rb_commit_callback2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propset_remote", 9, argv[7] )); 
  }
  arg9 = (svn_client_ctx_t *)(argp9);
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_client_propset_remote((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5,(apr_hash_t const *)arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg8);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propset_local(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_depth_t arg4 ;
  svn_boolean_t arg5 ;
  apr_array_header_t *arg6 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_propset_local", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  {
    arg3 = svn_swig_rb_strings_to_apr_array(argv[2], _global_pool);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  arg5 = RTEST(argv[4]);
  {
    arg6 = (NIL_P(argv[5])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[5], _global_pool);
  }
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propset_local", 7, argv[6] )); 
  }
  arg7 = (svn_client_ctx_t *)(argp7);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_propset_local((char const *)arg1,(struct svn_string_t const *)arg2,(apr_array_header_t const *)arg3,arg4,arg5,(apr_array_header_t const *)arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propset3(int argc, VALUE *argv, VALUE self) {
  svn_commit_info_t **arg1 = (svn_commit_info_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_revnum_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg9 = (apr_hash_t *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_commit_info_t *temp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_string_t value3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  long val7 ;
  int ecode7 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_propset3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    if (NIL_P(argv[1])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[1]);
      value3.len = RSTRING_LEN(argv[1]);
      arg3 = &value3;
    }
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_propset3", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = svn_swig_rb_to_depth(argv[3]);
  }
  arg6 = RTEST(argv[4]);
  ecode7 = SWIG_AsVal_long(argv[5], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_propset3", 7, argv[5] ));
  } 
  arg7 = (svn_revnum_t)(val7);
  {
    arg8 = (NIL_P(argv[6])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[6], _global_pool);
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg9 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[7], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg9)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[7], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propset3", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_client_propset3(arg1,(char const *)arg2,(struct svn_string_t const *)arg3,(char const *)arg4,arg5,arg6,arg7,(apr_array_header_t const *)arg8,(apr_hash_t const *)arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_commit_info_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propset2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_propset2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propset2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propset2", 6, argv[5] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_propset2((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propset(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_propset", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propset", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_propset((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_revprop_set2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_revnum_t *arg6 = (svn_revnum_t *) 0 ;
  svn_boolean_t arg7 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  svn_string_t value3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  svn_revnum_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg6 = &temp6;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_revprop_set2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[2]);
      value3.len = RSTRING_LEN(argv[2]);
      arg3 = &value3;
    }
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_revprop_set2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[4]);
  }
  arg7 = RTEST(argv[5]);
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_revprop_set2", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_revprop_set2((char const *)arg1,(struct svn_string_t const *)arg2,(struct svn_string_t const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res6)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg6)));
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_long, new_flags));
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_revprop_set(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_revnum_t *arg5 = (svn_revnum_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_revnum_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg5 = &temp5;
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_revprop_set", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_revprop_set", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  arg6 = RTEST(argv[4]);
  if (argc > 5) {
    res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_revprop_set", 7, argv[5] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 6) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_revprop_set((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res5)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_long, new_flags));
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propget5(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  apr_array_header_t **arg2 = (apr_array_header_t **) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg6 = (svn_opt_revision_t *) 0 ;
  svn_revnum_t *arg7 = (svn_revnum_t *) 0 ;
  svn_depth_t arg8 ;
  apr_array_header_t *arg9 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  apr_array_header_t *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_opt_revision_t rev5 ;
  svn_opt_revision_t rev6 ;
  svn_revnum_t temp7 ;
  int res7 = SWIG_TMPOBJ ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg7 = &temp7;
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propget5", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_propget5", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[2]);
  }
  {
    arg6 = &rev6;
    svn_swig_rb_set_revision(&rev6, argv[3]);
  }
  {
    arg8 = svn_swig_rb_to_depth(argv[4]);
  }
  {
    arg9 = (NIL_P(argv[5])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[5], _global_pool);
  }
  res10 = SWIG_ConvertPtr(argv[6], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propget5", 10, argv[6] )); 
  }
  arg10 = (svn_client_ctx_t *)(argp10);
  if (argc > 7) {
    
  }
  if (argc > 8) {
    
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg6) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_propget5(arg1,arg2,(char const *)arg3,(char const *)arg4,(struct svn_opt_revision_t const *)arg5,(struct svn_opt_revision_t const *)arg6,arg7,arg8,(apr_array_header_t const *)arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  {
    /* FIXME: Missing argout typemap: svn_client_propget5 arg 2 (apr_array_header_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_client_propget5 is not implemented yet");
    
  }
  if (SWIG_IsTmpObj(res7)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg7)));
  } else {
    int new_flags = SWIG_IsNewObj(res7) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg7), SWIGTYPE_p_long, new_flags));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propget4(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_revnum_t *arg6 = (svn_revnum_t *) 0 ;
  svn_depth_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  svn_revnum_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg6 = &temp6;
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_propget4", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propget4", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[4]);
  }
  {
    arg8 = (NIL_P(argv[5])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[5], _global_pool);
  }
  res9 = SWIG_ConvertPtr(argv[6], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propget4", 9, argv[6] )); 
  }
  arg9 = (svn_client_ctx_t *)(argp9);
  if (argc > 7) {
    
  }
  if (argc > 8) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_propget4(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,(apr_array_header_t const *)arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (SWIG_IsTmpObj(res6)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg6)));
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propget3(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_revnum_t *arg6 = (svn_revnum_t *) 0 ;
  svn_depth_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  svn_revnum_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg6 = &temp6;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_propget3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propget3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  {
    arg7 = svn_swig_rb_to_depth(argv[4]);
  }
  {
    arg8 = (NIL_P(argv[5])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[5], _global_pool);
  }
  if (argc > 6) {
    res9 = SWIG_ConvertPtr(argv[6], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propget3", 9, argv[6] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 7) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_propget3(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,(apr_array_header_t const *)arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (SWIG_IsTmpObj(res6)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg6)));
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propget2(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_propget2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propget2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  arg6 = RTEST(argv[4]);
  if (argc > 5) {
    res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propget2", 7, argv[5] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 6) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_propget2(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_propget(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_propget", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_propget", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_propget", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_propget(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_revprop_get(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_string_t **arg2 = (svn_string_t **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_revnum_t *arg5 = (svn_revnum_t *) 0 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_revnum_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg5 = &temp5;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_revprop_get", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_revprop_get", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  if (argc > 3) {
    res6 = SWIG_ConvertPtr(argv[3], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_revprop_get", 6, argv[3] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 4) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_revprop_get((char const *)arg1,arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg2)->data, (*arg2)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (SWIG_IsTmpObj(res5)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_long, new_flags));
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_proplist4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_proplist_receiver2_t arg7 = (svn_proplist_receiver2_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  int res8 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_proplist4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  {
    arg5 = (NIL_P(argv[4])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[4], _global_pool);
  }
  arg6 = RTEST(argv[5]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[6], (void**)(&arg7), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_proplist_receiver2_t","svn_client_proplist4", 7, argv[6] )); 
    }
  }
  res8 = SWIG_ConvertPtr(argv[7],SWIG_as_voidptrptr(&arg8), 0, 0);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "void *","svn_client_proplist4", 8, argv[7] )); 
  }
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_proplist4", 9, argv[8] )); 
  }
  arg9 = (svn_client_ctx_t *)(argp9);
  if (argc > 9) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_proplist4((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_proplist3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_proplist_receiver_t arg6 = (svn_proplist_receiver_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_proplist3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  {
    arg5 = (NIL_P(argv[4])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[4], _global_pool);
  }
  {
    arg6 = (svn_proplist_receiver_t) svn_swig_rb_proplist_receiver;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_proplist3", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_proplist3((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_proplist2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_proplist2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_proplist2", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_proplist2(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_proplist_item(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_proplist(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_proplist", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_proplist", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_proplist(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_proplist_item(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_revprop_list(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_revnum_t *arg4 = (svn_revnum_t *) 0 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg4 = &temp4;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_revprop_list", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  if (argc > 2) {
    res5 = SWIG_ConvertPtr(argv[2], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_revprop_list", 5, argv[2] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 3) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_revprop_list(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_export5(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_depth_t arg9 ;
  char *arg10 = (char *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_export5", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_export5", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  {
    arg9 = svn_swig_rb_to_depth(argv[7]);
  }
  {
    if (NIL_P(argv[8])) {
      arg10 = NULL;
    } else {
      arg10 = StringValuePtr(argv[8]);
    }
  }
  if (argc > 9) {
    res11 = SWIG_ConvertPtr(argv[9], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res11)) {
      SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_export5", 11, argv[9] )); 
    }
    arg11 = (svn_client_ctx_t *)(argp11);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_export5(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9,(char const *)arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_export4(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_depth_t arg8 ;
  char *arg9 = (char *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_export4", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_export4", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  {
    arg8 = svn_swig_rb_to_depth(argv[6]);
  }
  {
    if (NIL_P(argv[7])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[7]);
    }
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_export4", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_export4(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,(char const *)arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_export3(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  char *arg9 = (char *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_export3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_export3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  {
    if (NIL_P(argv[7])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[7]);
    }
  }
  if (argc > 8) {
    res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_export3", 10, argv[8] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 9) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_export3(arg1,(char const *)arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,(char const *)arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_export2(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  char *arg6 = (char *) 0 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_export2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_export2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  arg5 = RTEST(argv[3]);
  {
    if (NIL_P(argv[4])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[4]);
    }
  }
  if (argc > 5) {
    res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_export2", 7, argv[5] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_export2(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,(char const *)arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_export(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_export", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_export", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_export", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_export(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_list3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  apr_uint32_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_client_list_func2_t arg8 = (svn_client_list_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  int res9 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_list3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  ecode5 = SWIG_AsVal_unsigned_SS_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "apr_uint32_t","svn_client_list3", 5, argv[4] ));
  } 
  arg5 = (apr_uint32_t)(val5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_list_func2_t","svn_client_list3", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_client_list3", 9, argv[8] )); 
  }
  if (argc > 9) {
    res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_list3", 10, argv[9] )); 
    }
    arg10 = (svn_client_ctx_t *)(argp10);
  }
  if (argc > 10) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_list3((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_list2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  apr_uint32_t arg5 ;
  svn_boolean_t arg6 ;
  svn_client_list_func_t arg7 = (svn_client_list_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_list2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  ecode5 = SWIG_AsVal_unsigned_SS_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "apr_uint32_t","svn_client_list2", 5, argv[4] ));
  } 
  arg5 = (apr_uint32_t)(val5);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_client_list_func_t) svn_swig_rb_client_list_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_list2", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_list2((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_list(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_uint32_t arg5 ;
  svn_boolean_t arg6 ;
  svn_client_list_func_t arg7 = (svn_client_list_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_client_ctx_t *arg9 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_list", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  arg4 = RTEST(argv[3]);
  ecode5 = SWIG_AsVal_unsigned_SS_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "apr_uint32_t","svn_client_list", 5, argv[4] ));
  } 
  arg5 = (apr_uint32_t)(val5);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_client_list_func_t) svn_swig_rb_client_list_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_list", 9, argv[7] )); 
    }
    arg9 = (svn_client_ctx_t *)(argp9);
  }
  if (argc > 8) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_list((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ls3(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  apr_hash_t **arg2 = (apr_hash_t **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  apr_hash_t *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_ls3", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[1]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[2]);
  }
  arg6 = RTEST(argv[3]);
  if (argc > 4) {
    res7 = SWIG_ConvertPtr(argv[4], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_ls3", 7, argv[4] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_ls3(arg1,arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_dirent_t *"))
    
    ;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg2,
        "svn_lock_t *"))
    
    ;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ls2(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_ls2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_ls2", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_ls2(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_dirent_t *"))
    
    ;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_ls(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_ls", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[1]);
  }
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_ls", 5, argv[3] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_ls(arg1,(char const *)arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_dirent_t *"))
    
    ;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_cat3(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg5 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_opt_revision_t rev4 ;
  svn_opt_revision_t rev5 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_cat3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[2]);
  }
  {
    arg5 = &rev5;
    svn_swig_rb_set_revision(&rev5, argv[3]);
  }
  arg6 = RTEST(argv[4]);
  res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_cat3", 7, argv[5] )); 
  }
  arg7 = (svn_client_ctx_t *)(argp7);
  if (argc > 6) {
    
  }
  if (argc > 7) {
    
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_cat3(arg1,arg2,(char const *)arg3,(struct svn_opt_revision_t const *)arg4,(struct svn_opt_revision_t const *)arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_cat2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg4 = (svn_opt_revision_t *) 0 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  svn_opt_revision_t rev4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_cat2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = &rev4;
    svn_swig_rb_set_revision(&rev4, argv[3]);
  }
  if (argc > 4) {
    res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_cat2", 5, argv[4] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 5) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_cat2(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,(struct svn_opt_revision_t const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_cat(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_revision_t rev3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_cat", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  if (argc > 3) {
    res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_cat", 4, argv[3] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 4) {
    
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_cat(arg1,(char const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_add_to_changelist(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  apr_array_header_t *arg4 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = StringValueCStr(argv[1]);
  }
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  {
    arg4 = (NIL_P(argv[3])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[3], _global_pool);
  }
  if (argc > 4) {
    res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_add_to_changelist", 5, argv[4] )); 
    }
    arg5 = (svn_client_ctx_t *)(argp5);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_add_to_changelist((apr_array_header_t const *)arg1,(char const *)arg2,arg3,(apr_array_header_t const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_remove_from_changelists(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_depth_t arg2 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_to_depth(argv[1]);
  }
  {
    arg3 = (NIL_P(argv[2])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[2], _global_pool);
  }
  if (argc > 3) {
    res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_remove_from_changelists", 4, argv[3] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_remove_from_changelists((apr_array_header_t const *)arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_changelists(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  svn_depth_t arg3 ;
  svn_changelist_receiver_t arg4 = (svn_changelist_receiver_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_client_ctx_t *arg6 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_get_changelists", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = (NIL_P(argv[1])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[1], _global_pool);
  }
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  {
    arg4 = (svn_changelist_receiver_t) svn_swig_rb_changelist_receiver;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_get_changelists", 6, argv[4] )); 
    }
    arg6 = (svn_client_ctx_t *)(argp6);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_get_changelists((char const *)arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_lock(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[1]);
    }
  }
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_lock", 4, argv[3] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_lock((apr_array_header_t const *)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_unlock(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg2 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_strings_to_apr_array(argv[0], _global_pool);
  }
  arg2 = RTEST(argv[1]);
  if (argc > 2) {
    res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_unlock", 3, argv[2] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_unlock((apr_array_header_t const *)arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_info_t;

SWIGINTERN VALUE
_wrap_svn_info_t_URL_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","URL", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","URL", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->URL) free((char *)arg1->URL);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->URL = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_URL_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","URL", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->URL);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","rev", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","rev", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_repos_root_URL_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","repos_root_URL", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_root_URL", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_root_URL) free((char *)arg1->repos_root_URL);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_root_URL = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_repos_root_URL_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","repos_root_URL", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->repos_root_URL);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_repos_UUID_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","repos_UUID", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_UUID", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_UUID) free((char *)arg1->repos_UUID);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_UUID = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_repos_UUID_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","repos_UUID", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->repos_UUID);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_last_changed_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","last_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","last_changed_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->last_changed_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_last_changed_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","last_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->last_changed_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_last_changed_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","last_changed_date", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->last_changed_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_last_changed_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","last_changed_date", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result =  ((arg1)->last_changed_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_last_changed_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","last_changed_author", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","last_changed_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->last_changed_author) free((char *)arg1->last_changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->last_changed_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_last_changed_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","last_changed_author", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->last_changed_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t *","lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->lock = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_has_wc_info_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","has_wc_info", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->has_wc_info = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_has_wc_info_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","has_wc_info", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->has_wc_info);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_schedule_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_wc_schedule_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","schedule", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_schedule_t","schedule", 2, argv[0] ));
  } 
  arg2 = (svn_wc_schedule_t)(val2);
  if (arg1) (arg1)->schedule = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_schedule_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_schedule_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","schedule", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_wc_schedule_t) ((arg1)->schedule);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_copyfrom_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_copyfrom_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_copyfrom_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","copyfrom_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->copyfrom_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_copyfrom_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_text_time_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","text_time", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->text_time = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_text_time_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","text_time", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result =  ((arg1)->text_time);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_prop_time_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","prop_time", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->prop_time = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_prop_time_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","prop_time", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result =  ((arg1)->prop_time);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_checksum_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","checksum", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","checksum", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->checksum) free((char *)arg1->checksum);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->checksum = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_checksum_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","checksum", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->checksum);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_conflict_old_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","conflict_old", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","conflict_old", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_old) free((char *)arg1->conflict_old);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_old = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_conflict_old_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","conflict_old", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->conflict_old);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_conflict_new_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","conflict_new", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","conflict_new", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_new) free((char *)arg1->conflict_new);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_new = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_conflict_new_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","conflict_new", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->conflict_new);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_conflict_wrk_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","conflict_wrk", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","conflict_wrk", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_wrk) free((char *)arg1->conflict_wrk);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_wrk = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_conflict_wrk_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","conflict_wrk", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->conflict_wrk);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_prejfile_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","prejfile", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","prejfile", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->prejfile) free((char *)arg1->prejfile);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->prejfile = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_prejfile_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","prejfile", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->prejfile);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_changelist_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    arg2 = StringValueCStr(argv[0]);
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_changelist_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (char *) ((arg1)->changelist);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_depth_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_depth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    arg2 = svn_swig_rb_to_depth(argv[0]);
  }
  if (arg1) (arg1)->depth = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_depth_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_depth_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_depth_t) ((arg1)->depth);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_working_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  apr_size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","working_size", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","working_size", 2, argv[0] ));
  } 
  arg2 = (apr_size_t)(val2);
  if (arg1) (arg1)->working_size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_working_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_size_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","working_size", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result =  ((arg1)->working_size);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}



/*
  Document-method: Svn::Ext::Client::svn_info_t.size

  call-seq:
    size -> apr_size_t

Size or Length of the svn_info_t.
*/

/*
  Document-method: Svn::Ext::Client::svn_info_t.size=

  call-seq:
    size=(x) -> apr_size_t

Size or Length of the svn_info_t.
*/
SWIGINTERN VALUE
_wrap_svn_info_t_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  apr_size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","size", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","size", 2, argv[0] ));
  } 
  arg2 = (apr_size_t)(val2);
  if (arg1) (arg1)->size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_size_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","size", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result =  ((arg1)->size);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_size64_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","size64", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->size64 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_size64_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","size64", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result =  ((arg1)->size64);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_working_size64_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","working_size64", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->working_size64 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_working_size64_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","working_size64", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result =  ((arg1)->working_size64);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_tree_conflict_set(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  svn_wc_conflict_description_t *arg2 = (svn_wc_conflict_description_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","tree_conflict", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_conflict_description_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_conflict_description_t *","tree_conflict", 2, argv[0] )); 
  }
  arg2 = (svn_wc_conflict_description_t *)(argp2);
  if (arg1) (arg1)->tree_conflict = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_t_tree_conflict_get(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *arg1 = (struct svn_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_info_t *","tree_conflict", 1, self )); 
  }
  arg1 = (struct svn_info_t *)(argp1);
  result = (svn_wc_conflict_description_t *) ((arg1)->tree_conflict);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_info_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_info_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_info_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_info_t(int argc, VALUE *argv, VALUE self) {
  struct svn_info_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_info_t *)calloc(1, sizeof(struct svn_info_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_info_t(struct svn_info_t *arg1) {
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_info_dup(int argc, VALUE *argv, VALUE self) {
  svn_info_t *arg1 = (svn_info_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_info_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_info_t const *","svn_info_dup", 1, argv[0] )); 
  }
  arg1 = (svn_info_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_info_t *)svn_info_dup((struct svn_info_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_info_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_client_info2_t;

SWIGINTERN VALUE
_wrap_svn_client_info2_t_URL_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","URL", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","URL", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->URL) free((char *)arg1->URL);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->URL = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_URL_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","URL", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (char *) ((arg1)->URL);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","rev", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","rev", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_repos_root_URL_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","repos_root_URL", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_root_URL", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_root_URL) free((char *)arg1->repos_root_URL);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_root_URL = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_repos_root_URL_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","repos_root_URL", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (char *) ((arg1)->repos_root_URL);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_repos_UUID_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","repos_UUID", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_UUID", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_UUID) free((char *)arg1->repos_UUID);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_UUID = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_repos_UUID_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","repos_UUID", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (char *) ((arg1)->repos_UUID);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}



/*
  Document-method: Svn::Ext::Client::svn_client_info2_t.size

  call-seq:
    size -> svn_filesize_t

Size or Length of the svn_client_info2_t.
*/

/*
  Document-method: Svn::Ext::Client::svn_client_info2_t.size=

  call-seq:
    size=(x) -> svn_filesize_t

Size or Length of the svn_client_info2_t.
*/
SWIGINTERN VALUE
_wrap_svn_client_info2_t_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","size", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","size", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result =  ((arg1)->size);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_last_changed_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","last_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","last_changed_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->last_changed_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_last_changed_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","last_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->last_changed_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_last_changed_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","last_changed_date", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->last_changed_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_last_changed_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","last_changed_date", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result =  ((arg1)->last_changed_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_last_changed_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","last_changed_author", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","last_changed_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->last_changed_author) free((char *)arg1->last_changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->last_changed_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_last_changed_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","last_changed_author", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (char *) ((arg1)->last_changed_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t const *","lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->lock = (svn_lock_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_wc_info_set(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  svn_wc_info_t *arg2 = (svn_wc_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","wc_info", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_info_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_info_t const *","wc_info", 2, argv[0] )); 
  }
  arg2 = (svn_wc_info_t *)(argp2);
  if (arg1) (arg1)->wc_info = (svn_wc_info_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2_t_wc_info_get(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *arg1 = (struct svn_client_info2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_info_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_client_info2_t *","wc_info", 1, self )); 
  }
  arg1 = (struct svn_client_info2_t *)(argp1);
  result = (svn_wc_info_t *) ((arg1)->wc_info);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
SWIGINTERN VALUE
_wrap_svn_client_info2_t_allocate(VALUE self) {
#else
  SWIGINTERN VALUE
  _wrap_svn_client_info2_t_allocate(int argc, VALUE *argv, VALUE self) {
#endif
    
    
    VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_client_info2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
    rb_obj_call_init(vresult, argc, argv);
#endif
    return vresult;
  }
  

SWIGINTERN VALUE
_wrap_new_svn_client_info2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_client_info2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_client_info2_t *)calloc(1, sizeof(struct svn_client_info2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_client_info2_t(struct svn_client_info2_t *arg1) {
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_client_info2_dup(int argc, VALUE *argv, VALUE self) {
  svn_client_info2_t *arg1 = (svn_client_info2_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_client_info2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_info2_t const *","svn_client_info2_dup", 1, argv[0] )); 
  }
  arg1 = (svn_client_info2_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_client_info2_t *)svn_client_info2_dup((struct svn_client_info2_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  svn_client_info_receiver2_t arg9 = (svn_client_info_receiver2_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_client_ctx_t *arg11 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  int res10 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_info4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (NIL_P(argv[7])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[7], _global_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[8], (void**)(&arg9), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_info_receiver2_t","svn_client_info4", 9, argv[8] )); 
    }
  }
  res10 = SWIG_ConvertPtr(argv[9],SWIG_as_voidptrptr(&arg10), 0, 0);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "void *","svn_client_info4", 10, argv[9] )); 
  }
  res11 = SWIG_ConvertPtr(argv[10], &argp11,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_info4", 11, argv[10] )); 
  }
  arg11 = (svn_client_ctx_t *)(argp11);
  if (argc > 11) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_info4((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,(apr_array_header_t const *)arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_depth_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_client_info_receiver2_t arg8 = (svn_client_info_receiver2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  int res9 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_info3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (NIL_P(argv[6])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[6], _global_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_info_receiver2_t","svn_client_info3", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_client_info3", 9, argv[8] )); 
  }
  res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_info3", 10, argv[9] )); 
  }
  arg10 = (svn_client_ctx_t *)(argp10);
  if (argc > 10) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_info3((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_info_receiver_t arg4 = (svn_info_receiver_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_depth_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_client_ctx_t *arg8 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_info2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_info_receiver;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  {
    arg7 = (NIL_P(argv[5])) ? NULL :
    svn_swig_rb_strings_to_apr_array(argv[5], _global_pool);
  }
  if (argc > 6) {
    res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_info2", 8, argv[6] )); 
    }
    arg8 = (svn_client_ctx_t *)(argp8);
  }
  if (argc > 7) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_info2((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_info(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg3 = (svn_opt_revision_t *) 0 ;
  svn_info_receiver_t arg4 = (svn_info_receiver_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_boolean_t arg6 ;
  svn_client_ctx_t *arg7 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_opt_revision_t rev2 ;
  svn_opt_revision_t rev3 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_info", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  {
    arg3 = &rev3;
    svn_swig_rb_set_revision(&rev3, argv[2]);
  }
  {
    arg4 = svn_swig_rb_info_receiver;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  arg6 = RTEST(argv[4]);
  if (argc > 5) {
    res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_info", 7, argv[5] )); 
    }
    arg7 = (svn_client_ctx_t *)(argp7);
  }
  if (argc > 6) {
    
  }
  {
    if (!arg2) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_ValueError,"Received a NULL pointer.");
    }
  }
  {
    result = (svn_error_t *)svn_client_info((char const *)arg1,(struct svn_opt_revision_t const *)arg2,(struct svn_opt_revision_t const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_wc_root(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_get_wc_root", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_get_wc_root", 3, argv[1] )); 
  }
  arg3 = (svn_client_ctx_t *)(argp3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_get_wc_root((char const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_min_max_revisions(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_revnum_t *arg2 = (svn_revnum_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_client_ctx_t *arg5 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_revnum_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_min_max_revisions", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[1]);
  res5 = SWIG_ConvertPtr(argv[2], &argp5,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_min_max_revisions", 5, argv[2] )); 
  }
  arg5 = (svn_client_ctx_t *)(argp5);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_min_max_revisions(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  if (SWIG_IsTmpObj(res2)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_long, new_flags));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_patch(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  int arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_client_patch_func_t arg8 = (svn_client_patch_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_client_ctx_t *arg10 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res9 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_client_patch", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_patch", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  ecode4 = SWIG_AsVal_int(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_client_patch", 4, argv[3] ));
  } 
  arg4 = (int)(val4);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_patch_func_t","svn_client_patch", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_client_patch", 9, argv[8] )); 
  }
  res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_patch", 10, argv[9] )); 
  }
  arg10 = (svn_client_ctx_t *)(argp10);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_client_patch((char const *)arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_url_from_path2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_url_from_path2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_url_from_path2", 3, argv[1] )); 
  }
  arg3 = (svn_client_ctx_t *)(argp3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_url_from_path2((char const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_url_from_path(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_url_from_path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_client_url_from_path((char const **)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_repos_root(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_get_repos_root", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_get_repos_root", 4, argv[1] )); 
  }
  arg4 = (svn_client_ctx_t *)(argp4);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_get_repos_root((char const **)arg1,(char const **)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_root_url_from_path(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_root_url_from_path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_root_url_from_path", 3, argv[1] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_client_root_url_from_path((char const **)arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_uuid_from_url(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_uuid_from_url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_uuid_from_url", 3, argv[1] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_client_uuid_from_url((char const **)arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_uuid_from_path2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_uuid_from_path2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_uuid_from_path2", 3, argv[1] )); 
  }
  arg3 = (svn_client_ctx_t *)(argp3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_uuid_from_path2((char const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_uuid_from_path(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_uuid_from_path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_client_uuid_from_path", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_uuid_from_path", 4, argv[2] )); 
    }
    arg4 = (svn_client_ctx_t *)(argp4);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_uuid_from_path((char const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_open_ra_session2(int argc, VALUE *argv, VALUE self) {
  svn_ra_session_t **arg1 = (svn_ra_session_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_client_ctx_t *arg4 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_ra_session_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_open_ra_session2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_open_ra_session2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_open_ra_session2", 4, argv[2] )); 
  }
  arg4 = (svn_client_ctx_t *)(argp4);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_open_ra_session2(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_ra_session_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_open_ra_session(int argc, VALUE *argv, VALUE self) {
  svn_ra_session_t **arg1 = (svn_ra_session_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_client_ctx_t *arg3 = (svn_client_ctx_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_ra_session_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    int adjusted_argc = argc;
    VALUE *adjusted_argv = argv;
    
    svn_swig_rb_adjust_arg_for_client_ctx_and_pool(&adjusted_argc,
      &adjusted_argv);
    svn_swig_rb_get_pool(adjusted_argc, adjusted_argv, self,
      &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_client_open_ra_session", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_open_ra_session", 3, argv[1] )); 
    }
    arg3 = (svn_client_ctx_t *)(argp3);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_client_open_ra_session(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_ra_session_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_proplist_invoke_receiver2(int argc, VALUE *argv, VALUE self) {
  svn_proplist_receiver2_t arg1 = (svn_proplist_receiver2_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_proplist_receiver2_t","svn_proplist_invoke_receiver2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_proplist_invoke_receiver2", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_proplist_invoke_receiver2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_hash_t *","svn_proplist_invoke_receiver2", 4, argv[3] )); 
  }
  arg4 = (apr_hash_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_array_header_t *","svn_proplist_invoke_receiver2", 5, argv[4] )); 
  }
  arg5 = (apr_array_header_t *)(argp5);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_proplist_invoke_receiver2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_proplist_invoke_receiver(int argc, VALUE *argv, VALUE self) {
  svn_proplist_receiver_t arg1 = (svn_proplist_receiver_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_proplist_receiver_t","svn_proplist_invoke_receiver", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_proplist_invoke_receiver", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_proplist_invoke_receiver", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_hash_t *","svn_proplist_invoke_receiver", 4, argv[3] )); 
  }
  arg4 = (apr_hash_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_proplist_invoke_receiver(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_get_commit_log3(int argc, VALUE *argv, VALUE self) {
  svn_client_get_commit_log3_t arg1 = (svn_client_get_commit_log3_t) 0 ;
  char **arg2 = (char **) 0 ;
  char **arg3 = (char **) 0 ;
  apr_array_header_t *arg4 = (apr_array_header_t *) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp2 ;
  char *temp3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_get_commit_log3_t","svn_client_invoke_get_commit_log3", 1, argv[0] )); 
    }
  }
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_client_invoke_get_commit_log3", 4, argv[1] )); 
  }
  arg4 = (apr_array_header_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_client_invoke_get_commit_log3", 5, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_get_commit_log3(arg1,(char const **)arg2,(char const **)arg3,(apr_array_header_t const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_get_commit_log2(int argc, VALUE *argv, VALUE self) {
  svn_client_get_commit_log2_t arg1 = (svn_client_get_commit_log2_t) 0 ;
  char **arg2 = (char **) 0 ;
  char **arg3 = (char **) 0 ;
  apr_array_header_t *arg4 = (apr_array_header_t *) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp2 ;
  char *temp3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_get_commit_log2_t","svn_client_invoke_get_commit_log2", 1, argv[0] )); 
    }
  }
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_client_invoke_get_commit_log2", 4, argv[1] )); 
  }
  arg4 = (apr_array_header_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_client_invoke_get_commit_log2", 5, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_get_commit_log2(arg1,(char const **)arg2,(char const **)arg3,(apr_array_header_t const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_get_commit_log(int argc, VALUE *argv, VALUE self) {
  svn_client_get_commit_log_t arg1 = (svn_client_get_commit_log_t) 0 ;
  char **arg2 = (char **) 0 ;
  char **arg3 = (char **) 0 ;
  apr_array_header_t *arg4 = (apr_array_header_t *) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp2 ;
  char *temp3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_get_commit_log_t","svn_client_invoke_get_commit_log", 1, argv[0] )); 
    }
  }
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_array_header_t *","svn_client_invoke_get_commit_log", 4, argv[1] )); 
  }
  arg4 = (apr_array_header_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_client_invoke_get_commit_log", 5, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_get_commit_log(arg1,(char const **)arg2,(char const **)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_blame_receiver3(int argc, VALUE *argv, VALUE self) {
  svn_client_blame_receiver3_t arg1 = (svn_client_blame_receiver3_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  apr_int64_t arg5 ;
  svn_revnum_t arg6 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_revnum_t arg8 ;
  apr_hash_t *arg9 = (apr_hash_t *) 0 ;
  char *arg10 = (char *) 0 ;
  char *arg11 = (char *) 0 ;
  svn_boolean_t arg12 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  long val3 ;
  int ecode3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  long val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 12) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_blame_receiver3_t","svn_client_invoke_blame_receiver3", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_blame_receiver3", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_invoke_blame_receiver3", 3, argv[2] ));
  } 
  arg3 = (svn_revnum_t)(val3);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_invoke_blame_receiver3", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    arg5 = (apr_int64_t)NUM2LL(argv[4]);
  }
  ecode6 = SWIG_AsVal_long(argv[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_invoke_blame_receiver3", 6, argv[5] ));
  } 
  arg6 = (svn_revnum_t)(val6);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "apr_hash_t *","svn_client_invoke_blame_receiver3", 7, argv[6] )); 
  }
  arg7 = (apr_hash_t *)(argp7);
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_invoke_blame_receiver3", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "apr_hash_t *","svn_client_invoke_blame_receiver3", 9, argv[8] )); 
  }
  arg9 = (apr_hash_t *)(argp9);
  res10 = SWIG_AsCharPtrAndSize(argv[9], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver3", 10, argv[9] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_AsCharPtrAndSize(argv[10], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver3", 11, argv[10] ));
  }
  arg11 = (char *)(buf11);
  arg12 = RTEST(argv[11]);
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_blame_receiver3(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_blame_receiver2(int argc, VALUE *argv, VALUE self) {
  svn_client_blame_receiver2_t arg1 = (svn_client_blame_receiver2_t) 0 ;
  void *arg2 = (void *) 0 ;
  apr_int64_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_revnum_t arg7 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  char *arg10 = (char *) 0 ;
  char *arg11 = (char *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  long val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  long val7 ;
  int ecode7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  int res9 ;
  char *buf9 = 0 ;
  int alloc9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_blame_receiver2_t","svn_client_invoke_blame_receiver2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_blame_receiver2", 2, argv[1] )); 
  }
  {
    arg3 = (apr_int64_t)NUM2LL(argv[2]);
  }
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_invoke_blame_receiver2", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver2", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  ecode7 = SWIG_AsVal_long(argv[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_invoke_blame_receiver2", 7, argv[6] ));
  } 
  arg7 = (svn_revnum_t)(val7);
  res8 = SWIG_AsCharPtrAndSize(argv[7], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver2", 8, argv[7] ));
  }
  arg8 = (char *)(buf8);
  res9 = SWIG_AsCharPtrAndSize(argv[8], &buf9, NULL, &alloc9);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver2", 9, argv[8] ));
  }
  arg9 = (char *)(buf9);
  res10 = SWIG_AsCharPtrAndSize(argv[9], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver2", 10, argv[9] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_AsCharPtrAndSize(argv[10], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver2", 11, argv[10] ));
  }
  arg11 = (char *)(buf11);
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_blame_receiver2(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,arg7,(char const *)arg8,(char const *)arg9,(char const *)arg10,(char const *)arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_blame_receiver(int argc, VALUE *argv, VALUE self) {
  svn_client_blame_receiver_t arg1 = (svn_client_blame_receiver_t) 0 ;
  void *arg2 = (void *) 0 ;
  apr_int64_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  long val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_blame_receiver_t","svn_client_invoke_blame_receiver", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_blame_receiver", 2, argv[1] )); 
  }
  {
    arg3 = (apr_int64_t)NUM2LL(argv[2]);
  }
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_client_invoke_blame_receiver", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_blame_receiver", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_blame_receiver(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_diff_summarize_func(int argc, VALUE *argv, VALUE self) {
  svn_client_diff_summarize_func_t arg1 = (svn_client_diff_summarize_func_t) 0 ;
  svn_client_diff_summarize_t *arg2 = (svn_client_diff_summarize_t *) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_diff_summarize_func_t","svn_client_invoke_diff_summarize_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_client_diff_summarize_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_client_diff_summarize_t const *","svn_client_invoke_diff_summarize_func", 2, argv[1] )); 
  }
  arg2 = (svn_client_diff_summarize_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_client_invoke_diff_summarize_func", 3, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_diff_summarize_func(arg1,(struct svn_client_diff_summarize_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_import_filter_func(int argc, VALUE *argv, VALUE self) {
  svn_client_import_filter_func_t arg1 = (svn_client_import_filter_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_io_dirent2_t *arg5 = (svn_io_dirent2_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_import_filter_func_t","svn_client_invoke_import_filter_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_import_filter_func", 2, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_import_filter_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_io_dirent2_t const *","svn_client_invoke_import_filter_func", 5, argv[3] )); 
  }
  arg5 = (svn_io_dirent2_t *)(argp5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_import_filter_func(arg1,arg2,arg3,(char const *)arg4,(struct svn_io_dirent2_t const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_status_func(int argc, VALUE *argv, VALUE self) {
  svn_client_status_func_t arg1 = (svn_client_status_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_client_status_t *arg4 = (svn_client_status_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_status_func_t","svn_client_invoke_status_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_status_func", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_status_func", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_status_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_status_t const *","svn_client_invoke_status_func", 4, argv[3] )); 
  }
  arg4 = (svn_client_status_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_status_func(arg1,arg2,(char const *)arg3,(struct svn_client_status_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_list_func2(int argc, VALUE *argv, VALUE self) {
  svn_client_list_func2_t arg1 = (svn_client_list_func2_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_dirent_t *arg4 = (svn_dirent_t *) 0 ;
  svn_lock_t *arg5 = (svn_lock_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_list_func2_t","svn_client_invoke_list_func2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_list_func2", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_list_func2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_dirent_t const *","svn_client_invoke_list_func2", 4, argv[3] )); 
  }
  arg4 = (svn_dirent_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_lock_t const *","svn_client_invoke_list_func2", 5, argv[4] )); 
  }
  arg5 = (svn_lock_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_list_func2", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_list_func2", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[7], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_list_func2", 8, argv[7] ));
  }
  arg8 = (char *)(buf8);
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_list_func2(arg1,arg2,(char const *)arg3,(struct svn_dirent_t const *)arg4,(struct svn_lock_t const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_list_func(int argc, VALUE *argv, VALUE self) {
  svn_client_list_func_t arg1 = (svn_client_list_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_dirent_t *arg4 = (svn_dirent_t *) 0 ;
  svn_lock_t *arg5 = (svn_lock_t *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_list_func_t","svn_client_invoke_list_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_list_func", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_list_func", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_dirent_t const *","svn_client_invoke_list_func", 4, argv[3] )); 
  }
  arg4 = (svn_dirent_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_lock_t const *","svn_client_invoke_list_func", 5, argv[4] )); 
  }
  arg5 = (svn_lock_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_list_func", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_list_func(arg1,arg2,(char const *)arg3,(struct svn_dirent_t const *)arg4,(struct svn_lock_t const *)arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_info_invoke_receiver(int argc, VALUE *argv, VALUE self) {
  svn_info_receiver_t arg1 = (svn_info_receiver_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_info_t *arg4 = (svn_info_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_info_receiver_t","svn_info_invoke_receiver", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_info_invoke_receiver", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_info_invoke_receiver", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_info_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_info_t const *","svn_info_invoke_receiver", 4, argv[3] )); 
  }
  arg4 = (svn_info_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_info_invoke_receiver(arg1,arg2,(char const *)arg3,(struct svn_info_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_info_receiver2(int argc, VALUE *argv, VALUE self) {
  svn_client_info_receiver2_t arg1 = (svn_client_info_receiver2_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_client_info2_t *arg4 = (svn_client_info2_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_info_receiver2_t","svn_client_invoke_info_receiver2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_info_receiver2", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_info_receiver2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_client_info2_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_client_info2_t const *","svn_client_invoke_info_receiver2", 4, argv[3] )); 
  }
  arg4 = (svn_client_info2_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_info_receiver2(arg1,arg2,(char const *)arg3,(struct svn_client_info2_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_invoke_patch_func(int argc, VALUE *argv, VALUE self) {
  svn_client_patch_func_t arg1 = (svn_client_patch_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_client_patch_func_t","svn_client_invoke_patch_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_client_invoke_patch_func", 2, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_patch_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_patch_func", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_client_invoke_patch_func", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_client_invoke_patch_func(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_set_log_msg_func3(int argc, VALUE *argv, VALUE self) {
  svn_client_ctx_t *arg1 = (svn_client_ctx_t *) 0 ;
  svn_client_get_commit_log3_t arg2 = (svn_client_get_commit_log3_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_set_log_msg_func3", 1, argv[0] )); 
  }
  arg1 = (svn_client_ctx_t *)(argp1);
  {
    arg2 = (svn_client_get_commit_log3_t) svn_swig_rb_get_commit_log_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[1], _global_svn_swig_rb_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_client_set_log_msg_func3(arg1,arg2,arg3,arg4);
    
    
    
  }
  vresult = result;
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_set_notify_func2(int argc, VALUE *argv, VALUE self) {
  svn_client_ctx_t *arg1 = (svn_client_ctx_t *) 0 ;
  svn_wc_notify_func2_t arg2 = (svn_wc_notify_func2_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_set_notify_func2", 1, argv[0] )); 
  }
  arg1 = (svn_client_ctx_t *)(argp1);
  {
    arg2 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg3 = (void *)svn_swig_rb_make_baton(argv[1], _global_svn_swig_rb_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_client_set_notify_func2(arg1,arg2,arg3,arg4);
    
    
    
  }
  vresult = result;
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_set_cancel_func(int argc, VALUE *argv, VALUE self) {
  svn_client_ctx_t *arg1 = (svn_client_ctx_t *) 0 ;
  svn_cancel_func_t arg2 = (svn_cancel_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_set_cancel_func", 1, argv[0] )); 
  }
  arg1 = (svn_client_ctx_t *)(argp1);
  {
    arg2 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[1], _global_svn_swig_rb_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_client_set_cancel_func(arg1,arg2,arg3,arg4);
    
    
    
  }
  vresult = result;
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg3);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_set_config(int argc, VALUE *argv, VALUE self) {
  svn_client_ctx_t *arg1 = (svn_client_ctx_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_set_config", 1, argv[0] )); 
  }
  arg1 = (svn_client_ctx_t *)(argp1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      arg2 = svn_swig_rb_hash_to_apr_hash_swig_type(argv[1], "svn_config_t *", _global_pool);
    }
  }
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_client_set_config(arg1,arg2,arg3);
    
    
    
  }
  vresult = result;
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_client_get_config(int argc, VALUE *argv, VALUE self) {
  svn_client_ctx_t *arg1 = (svn_client_ctx_t *) 0 ;
  apr_hash_t **arg2 = (apr_hash_t **) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_client_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_client_ctx_t *","svn_client_get_config", 1, argv[0] )); 
  }
  arg1 = (svn_client_ctx_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_client_get_config(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg2,
        "svn_config_t *"))
    
    ;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}



/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_apr_array_header_t = {"_p_apr_array_header_t", "apr_array_header_t *|svn_rangelist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_file_t = {"_p_apr_file_t", "apr_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_getopt_t = {"_p_apr_getopt_t", "apr_getopt_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_hash_t = {"_p_apr_hash_t", "apr_hash_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int32_t = {"_p_apr_int32_t", "apr_int32_t *|time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int64_t = {"_p_apr_int64_t", "apr_int64_t *|svn_filesize_t *|apr_time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_pool_t = {"_p_apr_pool_t", "apr_pool_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void = {"_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void", "void (*)(apr_off_t,apr_off_t,void *,apr_pool_t *)|svn_ra_progress_notify_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_opt_subcommand_t *|struct svn_error_t *(*)(apr_getopt_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(char const **,char const **,apr_array_header_t *,void *,apr_pool_t *)|svn_client_get_commit_log_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(char const **,char const **,apr_array_header_t const *,void *,apr_pool_t *)|svn_client_get_commit_log3_t|svn_client_get_commit_log2_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_auth_simple_prompt_func_t|struct svn_error_t *(*)(svn_auth_cred_simple_t **,void *,char const *,char const *,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_auth_ssl_client_cert_pw_prompt_func_t|struct svn_error_t *(*)(svn_auth_cred_ssl_client_cert_pw_t **,void *,char const *,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_auth_ssl_client_cert_prompt_func_t|struct svn_error_t *(*)(svn_auth_cred_ssl_client_cert_t **,void *,char const *,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_auth_cred_ssl_server_trust_t **,void *,char const *,apr_uint32_t,svn_auth_ssl_server_cert_info_t const *,svn_boolean_t,apr_pool_t *)|svn_auth_ssl_server_trust_prompt_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_auth_cred_username_t **,void *,char const *,svn_boolean_t,apr_pool_t *)|svn_auth_username_prompt_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_ra_open_tunnel_func_t|struct svn_error_t *(*)(svn_stream_t **,svn_stream_t **,svn_ra_close_tunnel_func_t *,void **,void *,char const *,char const *,char const *,int,svn_cancel_func_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_wc_conflict_resolver_func2_t|struct svn_error_t *(*)(svn_wc_conflict_result_t **,svn_wc_conflict_description2_t const *,void *,apr_pool_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_wc_conflict_resolver_func_t|struct svn_error_t *(*)(svn_wc_conflict_result_t **,svn_wc_conflict_description_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_client_diff_summarize_func_t|struct svn_error_t *(*)(svn_client_diff_summarize_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_commit_callback2_t|struct svn_error_t *(*)(svn_commit_info_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__p_svn_error_t = {"_p_f_p_void__p_svn_error_t", "svn_cancel_func_t|struct svn_error_t *(*)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,apr_int64_t,svn_revnum_t,char const *,char const *,char const *,apr_pool_t *)|svn_client_blame_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_client_blame_receiver2_t|struct svn_error_t *(*)(void *,apr_int64_t,svn_revnum_t,char const *,char const *,svn_revnum_t,char const *,char const *,char const *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,apr_hash_t *,svn_revnum_t,char const *,char const *,char const *,apr_pool_t *)|svn_log_message_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char__int = {"_p_f_p_void_p_q_const__char__int", "int (*)(void *,char const *)|svn_ra_check_tunnel_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void = {"_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void", "void (*)(void *,char const *,enum svn_wc_notify_action_t,enum svn_node_kind_t,char const *,enum svn_wc_notify_state_t,enum svn_wc_notify_state_t,long)|svn_wc_notify_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,apr_hash_t *,apr_array_header_t *,apr_pool_t *)|svn_proplist_receiver2_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,apr_hash_t *,apr_pool_t *)|svn_proplist_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,char const *,apr_pool_t *)|svn_changelist_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t", "svn_client_info_receiver2_t|struct svn_error_t *(*)(void *,char const *,svn_client_info2_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t", "svn_client_status_func_t|struct svn_error_t *(*)(void *,char const *,svn_client_status_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,svn_dirent_t const *,svn_lock_t const *,char const *,apr_pool_t *)|svn_client_list_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,svn_dirent_t const *,svn_lock_t const *,char const *,char const *,char const *,apr_pool_t *)|svn_client_list_func2_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,svn_info_t const *,apr_pool_t *)|svn_info_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void = {"_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void", "void (*)(void *,char const *,struct svn_wc_status2_t *)|svn_wc_status_func2_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void = {"_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void", "void (*)(void *,char const *,struct svn_wc_status_t *)|svn_wc_status_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,svn_wc_status2_t *,apr_pool_t *)|svn_wc_status_func3_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void = {"_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void", "svn_wc_notify_func2_t|void (*)(void *,struct svn_wc_notify_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_boolean_t *,char const *,char const *,char const *,apr_pool_t *)|svn_client_patch_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_boolean_t *,char const *,svn_io_dirent2_t const *,apr_pool_t *)|svn_client_import_filter_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_log_entry_t *,apr_pool_t *)|svn_log_entry_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_revnum_t,svn_revnum_t,apr_int64_t,svn_revnum_t,apr_hash_t *,svn_revnum_t,apr_hash_t *,char const *,char const *,svn_boolean_t,apr_pool_t *)|svn_client_blame_receiver3_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "int *|svn_boolean_t *|apr_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "long *|svn_revnum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_array_header_t = {"_p_p_apr_array_header_t", "apr_array_header_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_hash_t = {"_p_p_apr_hash_t", "apr_hash_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_provider_object_t = {"_p_p_svn_auth_provider_object_t", "struct svn_auth_provider_object_t **|svn_auth_provider_object_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_client_commit_info_t = {"_p_p_svn_client_commit_info_t", "struct svn_client_commit_info_t **|svn_client_commit_info_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_client_ctx_t = {"_p_p_svn_client_ctx_t", "struct svn_client_ctx_t **|svn_client_ctx_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_commit_info_t = {"_p_p_svn_commit_info_t", "struct svn_commit_info_t **|svn_commit_info_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_ra_session_t = {"_p_p_svn_ra_session_t", "struct svn_ra_session_t **|svn_ra_session_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_string_t = {"_p_p_svn_string_t", "struct svn_string_t **|svn_string_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_baton_t = {"_p_svn_auth_baton_t", "struct svn_auth_baton_t *|svn_auth_baton_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_simple_t = {"_p_svn_auth_cred_simple_t", "struct svn_auth_cred_simple_t *|svn_auth_cred_simple_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_pw_t = {"_p_svn_auth_cred_ssl_client_cert_pw_t", "struct svn_auth_cred_ssl_client_cert_pw_t *|svn_auth_cred_ssl_client_cert_pw_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_t = {"_p_svn_auth_cred_ssl_client_cert_t", "struct svn_auth_cred_ssl_client_cert_t *|svn_auth_cred_ssl_client_cert_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_server_trust_t = {"_p_svn_auth_cred_ssl_server_trust_t", "struct svn_auth_cred_ssl_server_trust_t *|svn_auth_cred_ssl_server_trust_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_username_t = {"_p_svn_auth_cred_username_t", "struct svn_auth_cred_username_t *|svn_auth_cred_username_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_iterstate_t = {"_p_svn_auth_iterstate_t", "struct svn_auth_iterstate_t *|svn_auth_iterstate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_object_t = {"_p_svn_auth_provider_object_t", "struct svn_auth_provider_object_t *|svn_auth_provider_object_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_t = {"_p_svn_auth_provider_t", "struct svn_auth_provider_t *|svn_auth_provider_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_ssl_server_cert_info_t = {"_p_svn_auth_ssl_server_cert_info_t", "struct svn_auth_ssl_server_cert_info_t *|svn_auth_ssl_server_cert_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_ctx_t = {"_p_svn_checksum_ctx_t", "struct svn_checksum_ctx_t *|svn_checksum_ctx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_kind_t = {"_p_svn_checksum_kind_t", "enum svn_checksum_kind_t *|svn_checksum_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_t = {"_p_svn_checksum_t", "struct svn_checksum_t *|svn_checksum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_commit_info_t = {"_p_svn_client_commit_info_t", "struct svn_client_commit_info_t *|svn_client_commit_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_commit_item2_t = {"_p_svn_client_commit_item2_t", "struct svn_client_commit_item2_t *|svn_client_commit_item2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_commit_item3_t = {"_p_svn_client_commit_item3_t", "struct svn_client_commit_item3_t *|svn_client_commit_item3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_commit_item_t = {"_p_svn_client_commit_item_t", "struct svn_client_commit_item_t *|svn_client_commit_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_copy_source_t = {"_p_svn_client_copy_source_t", "struct svn_client_copy_source_t *|svn_client_copy_source_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_ctx_t = {"_p_svn_client_ctx_t", "struct svn_client_ctx_t *|svn_client_ctx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_diff_summarize_kind_t = {"_p_svn_client_diff_summarize_kind_t", "enum svn_client_diff_summarize_kind_t *|svn_client_diff_summarize_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_diff_summarize_t = {"_p_svn_client_diff_summarize_t", "struct svn_client_diff_summarize_t *|svn_client_diff_summarize_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_info2_t = {"_p_svn_client_info2_t", "struct svn_client_info2_t *|svn_client_info2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_proplist_item_t = {"_p_svn_client_proplist_item_t", "struct svn_client_proplist_item_t *|svn_client_proplist_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_status_t = {"_p_svn_client_status_t", "struct svn_client_status_t *|svn_client_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_commit_info_t = {"_p_svn_commit_info_t", "struct svn_commit_info_t *|svn_commit_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_config_t = {"_p_svn_config_t", "struct svn_config_t *|svn_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_editor_t = {"_p_svn_delta_editor_t", "struct svn_delta_editor_t *|svn_delta_editor_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_depth_t = {"_p_svn_depth_t", "enum svn_depth_t *|svn_depth_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_conflict_display_style_t = {"_p_svn_diff_conflict_display_style_t", "enum svn_diff_conflict_display_style_t *|svn_diff_conflict_display_style_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_datasource_e = {"_p_svn_diff_datasource_e", "enum svn_diff_datasource_e *|svn_diff_datasource_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_file_ignore_space_t = {"_p_svn_diff_file_ignore_space_t", "enum svn_diff_file_ignore_space_t *|svn_diff_file_ignore_space_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_file_options_t = {"_p_svn_diff_file_options_t", "struct svn_diff_file_options_t *|svn_diff_file_options_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_fns2_t = {"_p_svn_diff_fns2_t", "struct svn_diff_fns2_t *|svn_diff_fns2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_fns_t = {"_p_svn_diff_fns_t", "struct svn_diff_fns_t *|svn_diff_fns_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_hunk_t = {"_p_svn_diff_hunk_t", "struct svn_diff_hunk_t *|svn_diff_hunk_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_operation_kind_e = {"_p_svn_diff_operation_kind_e", "svn_diff_operation_kind_t *|enum svn_diff_operation_kind_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_output_fns_t = {"_p_svn_diff_output_fns_t", "struct svn_diff_output_fns_t *|svn_diff_output_fns_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_t = {"_p_svn_diff_t", "struct svn_diff_t *|svn_diff_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_dirent_t = {"_p_svn_dirent_t", "struct svn_dirent_t *|svn_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_errno_t = {"_p_svn_errno_t", "enum svn_errno_t *|svn_errno_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_error_t = {"_p_svn_error_t", "struct svn_error_t *|svn_error_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_info_t = {"_p_svn_info_t", "struct svn_info_t *|svn_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent2_t = {"_p_svn_io_dirent2_t", "struct svn_io_dirent2_t *|svn_io_dirent2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent_t = {"_p_svn_io_dirent_t", "struct svn_io_dirent_t *|svn_io_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_file_del_t = {"_p_svn_io_file_del_t", "enum svn_io_file_del_t *|svn_io_file_del_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_location_segment_t = {"_p_svn_location_segment_t", "struct svn_location_segment_t *|svn_location_segment_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_lock_t = {"_p_svn_lock_t", "struct svn_lock_t *|svn_lock_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path2_t = {"_p_svn_log_changed_path2_t", "struct svn_log_changed_path2_t *|svn_log_changed_path2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path_t = {"_p_svn_log_changed_path_t", "struct svn_log_changed_path_t *|svn_log_changed_path_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_entry_t = {"_p_svn_log_entry_t", "struct svn_log_entry_t *|svn_log_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_merge_range_t = {"_p_svn_merge_range_t", "struct svn_merge_range_t *|svn_merge_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_mergeinfo_inheritance_t = {"_p_svn_mergeinfo_inheritance_t", "enum svn_mergeinfo_inheritance_t *|svn_mergeinfo_inheritance_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_node_kind_t = {"_p_svn_node_kind_t", "enum svn_node_kind_t *|svn_node_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_range_t = {"_p_svn_opt_revision_range_t", "struct svn_opt_revision_range_t *|svn_opt_revision_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_t = {"_p_svn_opt_revision_t", "struct svn_opt_revision_t *|svn_opt_revision_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_value_t = {"_p_svn_opt_revision_value_t", "union svn_opt_revision_value_t *|svn_opt_revision_value_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc2_t = {"_p_svn_opt_subcommand_desc2_t", "struct svn_opt_subcommand_desc2_t *|svn_opt_subcommand_desc2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc_t = {"_p_svn_opt_subcommand_desc_t", "struct svn_opt_subcommand_desc_t *|svn_opt_subcommand_desc_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_patch_file_t = {"_p_svn_patch_file_t", "struct svn_patch_file_t *|svn_patch_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_patch_t = {"_p_svn_patch_t", "struct svn_patch_t *|svn_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_inherited_item_t = {"_p_svn_prop_inherited_item_t", "struct svn_prop_inherited_item_t *|svn_prop_inherited_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_kind = {"_p_svn_prop_kind", "svn_prop_kind_t *|enum svn_prop_kind *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_patch_t = {"_p_svn_prop_patch_t", "struct svn_prop_patch_t *|svn_prop_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks2_t = {"_p_svn_ra_callbacks2_t", "struct svn_ra_callbacks2_t *|svn_ra_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks_t = {"_p_svn_ra_callbacks_t", "struct svn_ra_callbacks_t *|svn_ra_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_plugin_t = {"_p_svn_ra_plugin_t", "struct svn_ra_plugin_t *|svn_ra_plugin_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter2_t = {"_p_svn_ra_reporter2_t", "struct svn_ra_reporter2_t *|svn_ra_reporter2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter3_t = {"_p_svn_ra_reporter3_t", "struct svn_ra_reporter3_t *|svn_ra_reporter3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter_t = {"_p_svn_ra_reporter_t", "struct svn_ra_reporter_t *|svn_ra_reporter_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_session_t = {"_p_svn_ra_session_t", "struct svn_ra_session_t *|svn_ra_session_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_mark_t = {"_p_svn_stream_mark_t", "struct svn_stream_mark_t *|svn_stream_mark_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_t = {"_p_svn_stream_t", "struct svn_stream_t *|svn_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_string_t = {"_p_svn_string_t", "struct svn_string_t *|svn_string_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stringbuf_t = {"_p_svn_stringbuf_t", "struct svn_stringbuf_t *|svn_stringbuf_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_tristate_t = {"_p_svn_tristate_t", "enum svn_tristate_t *|svn_tristate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_op_t = {"_p_svn_txdelta_op_t", "struct svn_txdelta_op_t *|svn_txdelta_op_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_stream_t = {"_p_svn_txdelta_stream_t", "struct svn_txdelta_stream_t *|svn_txdelta_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_window_t = {"_p_svn_txdelta_window_t", "struct svn_txdelta_window_t *|svn_txdelta_window_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_checklist_t = {"_p_svn_version_checklist_t", "struct svn_version_checklist_t *|svn_version_checklist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_linked_lib_t = {"_p_svn_version_ext_linked_lib_t", "struct svn_version_ext_linked_lib_t *|svn_version_ext_linked_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_loaded_lib_t = {"_p_svn_version_ext_loaded_lib_t", "struct svn_version_ext_loaded_lib_t *|svn_version_ext_loaded_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_extended_t = {"_p_svn_version_extended_t", "struct svn_version_extended_t *|svn_version_extended_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_t = {"_p_svn_version_t", "struct svn_version_t *|svn_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_adm_access_t = {"_p_svn_wc_adm_access_t", "struct svn_wc_adm_access_t *|svn_wc_adm_access_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_committed_queue_t = {"_p_svn_wc_committed_queue_t", "struct svn_wc_committed_queue_t *|svn_wc_committed_queue_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_action_t = {"_p_svn_wc_conflict_action_t", "enum svn_wc_conflict_action_t *|svn_wc_conflict_action_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_choice_t = {"_p_svn_wc_conflict_choice_t", "enum svn_wc_conflict_choice_t *|svn_wc_conflict_choice_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_description_t = {"_p_svn_wc_conflict_description_t", "struct svn_wc_conflict_description_t *|svn_wc_conflict_description_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_kind_t = {"_p_svn_wc_conflict_kind_t", "enum svn_wc_conflict_kind_t *|svn_wc_conflict_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_reason_t = {"_p_svn_wc_conflict_reason_t", "enum svn_wc_conflict_reason_t *|svn_wc_conflict_reason_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_version_t = {"_p_svn_wc_conflict_version_t", "struct svn_wc_conflict_version_t *|svn_wc_conflict_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_context_t = {"_p_svn_wc_context_t", "struct svn_wc_context_t *|svn_wc_context_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks2_t = {"_p_svn_wc_diff_callbacks2_t", "struct svn_wc_diff_callbacks2_t *|svn_wc_diff_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks3_t = {"_p_svn_wc_diff_callbacks3_t", "struct svn_wc_diff_callbacks3_t *|svn_wc_diff_callbacks3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks4_t = {"_p_svn_wc_diff_callbacks4_t", "struct svn_wc_diff_callbacks4_t *|svn_wc_diff_callbacks4_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks_t = {"_p_svn_wc_diff_callbacks_t", "struct svn_wc_diff_callbacks_t *|svn_wc_diff_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_callbacks2_t = {"_p_svn_wc_entry_callbacks2_t", "struct svn_wc_entry_callbacks2_t *|svn_wc_entry_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_callbacks_t = {"_p_svn_wc_entry_callbacks_t", "struct svn_wc_entry_callbacks_t *|svn_wc_entry_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_t = {"_p_svn_wc_entry_t", "struct svn_wc_entry_t *|svn_wc_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item2_t = {"_p_svn_wc_external_item2_t", "struct svn_wc_external_item2_t *|svn_wc_external_item2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item_t = {"_p_svn_wc_external_item_t", "struct svn_wc_external_item_t *|svn_wc_external_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_info_t = {"_p_svn_wc_info_t", "struct svn_wc_info_t *|svn_wc_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_merge_outcome_t = {"_p_svn_wc_merge_outcome_t", "enum svn_wc_merge_outcome_t *|svn_wc_merge_outcome_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_action_t = {"_p_svn_wc_notify_action_t", "enum svn_wc_notify_action_t *|svn_wc_notify_action_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_lock_state_t = {"_p_svn_wc_notify_lock_state_t", "enum svn_wc_notify_lock_state_t *|svn_wc_notify_lock_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_state_t = {"_p_svn_wc_notify_state_t", "enum svn_wc_notify_state_t *|svn_wc_notify_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_t = {"_p_svn_wc_notify_t", "struct svn_wc_notify_t *|svn_wc_notify_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_operation_t = {"_p_svn_wc_operation_t", "enum svn_wc_operation_t *|svn_wc_operation_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_revision_status_t = {"_p_svn_wc_revision_status_t", "struct svn_wc_revision_status_t *|svn_wc_revision_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_schedule_t = {"_p_svn_wc_schedule_t", "enum svn_wc_schedule_t *|svn_wc_schedule_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status2_t = {"_p_svn_wc_status2_t", "struct svn_wc_status2_t *|svn_wc_status2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status3_t = {"_p_svn_wc_status3_t", "struct svn_wc_status3_t *|svn_wc_status3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status_t = {"_p_svn_wc_status_t", "struct svn_wc_status_t *|svn_wc_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_traversal_info_t = {"_p_svn_wc_traversal_info_t", "struct svn_wc_traversal_info_t *|svn_wc_traversal_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "unsigned long *|svn_linenum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_apr_array_header_t,
  &_swigt__p_apr_file_t,
  &_swigt__p_apr_getopt_t,
  &_swigt__p_apr_hash_t,
  &_swigt__p_apr_int32_t,
  &_swigt__p_apr_int64_t,
  &_swigt__p_apr_pool_t,
  &_swigt__p_char,
  &_swigt__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void,
  &_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void__p_svn_error_t,
  &_swigt__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char__int,
  &_swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  &_swigt__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  &_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  &_swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  &_swigt__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_p_apr_array_header_t,
  &_swigt__p_p_apr_hash_t,
  &_swigt__p_p_char,
  &_swigt__p_p_svn_auth_provider_object_t,
  &_swigt__p_p_svn_client_commit_info_t,
  &_swigt__p_p_svn_client_ctx_t,
  &_swigt__p_p_svn_commit_info_t,
  &_swigt__p_p_svn_ra_session_t,
  &_swigt__p_p_svn_string_t,
  &_swigt__p_svn_auth_baton_t,
  &_swigt__p_svn_auth_cred_simple_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_pw_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_t,
  &_swigt__p_svn_auth_cred_ssl_server_trust_t,
  &_swigt__p_svn_auth_cred_username_t,
  &_swigt__p_svn_auth_iterstate_t,
  &_swigt__p_svn_auth_provider_object_t,
  &_swigt__p_svn_auth_provider_t,
  &_swigt__p_svn_auth_ssl_server_cert_info_t,
  &_swigt__p_svn_checksum_ctx_t,
  &_swigt__p_svn_checksum_kind_t,
  &_swigt__p_svn_checksum_t,
  &_swigt__p_svn_client_commit_info_t,
  &_swigt__p_svn_client_commit_item2_t,
  &_swigt__p_svn_client_commit_item3_t,
  &_swigt__p_svn_client_commit_item_t,
  &_swigt__p_svn_client_copy_source_t,
  &_swigt__p_svn_client_ctx_t,
  &_swigt__p_svn_client_diff_summarize_kind_t,
  &_swigt__p_svn_client_diff_summarize_t,
  &_swigt__p_svn_client_info2_t,
  &_swigt__p_svn_client_proplist_item_t,
  &_swigt__p_svn_client_status_t,
  &_swigt__p_svn_commit_info_t,
  &_swigt__p_svn_config_t,
  &_swigt__p_svn_delta_editor_t,
  &_swigt__p_svn_depth_t,
  &_swigt__p_svn_diff_conflict_display_style_t,
  &_swigt__p_svn_diff_datasource_e,
  &_swigt__p_svn_diff_file_ignore_space_t,
  &_swigt__p_svn_diff_file_options_t,
  &_swigt__p_svn_diff_fns2_t,
  &_swigt__p_svn_diff_fns_t,
  &_swigt__p_svn_diff_hunk_t,
  &_swigt__p_svn_diff_operation_kind_e,
  &_swigt__p_svn_diff_output_fns_t,
  &_swigt__p_svn_diff_t,
  &_swigt__p_svn_dirent_t,
  &_swigt__p_svn_errno_t,
  &_swigt__p_svn_error_t,
  &_swigt__p_svn_info_t,
  &_swigt__p_svn_io_dirent2_t,
  &_swigt__p_svn_io_dirent_t,
  &_swigt__p_svn_io_file_del_t,
  &_swigt__p_svn_location_segment_t,
  &_swigt__p_svn_lock_t,
  &_swigt__p_svn_log_changed_path2_t,
  &_swigt__p_svn_log_changed_path_t,
  &_swigt__p_svn_log_entry_t,
  &_swigt__p_svn_merge_range_t,
  &_swigt__p_svn_mergeinfo_inheritance_t,
  &_swigt__p_svn_node_kind_t,
  &_swigt__p_svn_opt_revision_range_t,
  &_swigt__p_svn_opt_revision_t,
  &_swigt__p_svn_opt_revision_value_t,
  &_swigt__p_svn_opt_subcommand_desc2_t,
  &_swigt__p_svn_opt_subcommand_desc_t,
  &_swigt__p_svn_patch_file_t,
  &_swigt__p_svn_patch_t,
  &_swigt__p_svn_prop_inherited_item_t,
  &_swigt__p_svn_prop_kind,
  &_swigt__p_svn_prop_patch_t,
  &_swigt__p_svn_ra_callbacks2_t,
  &_swigt__p_svn_ra_callbacks_t,
  &_swigt__p_svn_ra_plugin_t,
  &_swigt__p_svn_ra_reporter2_t,
  &_swigt__p_svn_ra_reporter3_t,
  &_swigt__p_svn_ra_reporter_t,
  &_swigt__p_svn_ra_session_t,
  &_swigt__p_svn_stream_mark_t,
  &_swigt__p_svn_stream_t,
  &_swigt__p_svn_string_t,
  &_swigt__p_svn_stringbuf_t,
  &_swigt__p_svn_tristate_t,
  &_swigt__p_svn_txdelta_op_t,
  &_swigt__p_svn_txdelta_stream_t,
  &_swigt__p_svn_txdelta_window_t,
  &_swigt__p_svn_version_checklist_t,
  &_swigt__p_svn_version_ext_linked_lib_t,
  &_swigt__p_svn_version_ext_loaded_lib_t,
  &_swigt__p_svn_version_extended_t,
  &_swigt__p_svn_version_t,
  &_swigt__p_svn_wc_adm_access_t,
  &_swigt__p_svn_wc_committed_queue_t,
  &_swigt__p_svn_wc_conflict_action_t,
  &_swigt__p_svn_wc_conflict_choice_t,
  &_swigt__p_svn_wc_conflict_description_t,
  &_swigt__p_svn_wc_conflict_kind_t,
  &_swigt__p_svn_wc_conflict_reason_t,
  &_swigt__p_svn_wc_conflict_version_t,
  &_swigt__p_svn_wc_context_t,
  &_swigt__p_svn_wc_diff_callbacks2_t,
  &_swigt__p_svn_wc_diff_callbacks3_t,
  &_swigt__p_svn_wc_diff_callbacks4_t,
  &_swigt__p_svn_wc_diff_callbacks_t,
  &_swigt__p_svn_wc_entry_callbacks2_t,
  &_swigt__p_svn_wc_entry_callbacks_t,
  &_swigt__p_svn_wc_entry_t,
  &_swigt__p_svn_wc_external_item2_t,
  &_swigt__p_svn_wc_external_item_t,
  &_swigt__p_svn_wc_info_t,
  &_swigt__p_svn_wc_merge_outcome_t,
  &_swigt__p_svn_wc_notify_action_t,
  &_swigt__p_svn_wc_notify_lock_state_t,
  &_swigt__p_svn_wc_notify_state_t,
  &_swigt__p_svn_wc_notify_t,
  &_swigt__p_svn_wc_operation_t,
  &_swigt__p_svn_wc_revision_status_t,
  &_swigt__p_svn_wc_schedule_t,
  &_swigt__p_svn_wc_status2_t,
  &_swigt__p_svn_wc_status3_t,
  &_swigt__p_svn_wc_status_t,
  &_swigt__p_svn_wc_traversal_info_t,
  &_swigt__p_unsigned_long,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_apr_array_header_t[] = {  {&_swigt__p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_file_t[] = {  {&_swigt__p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_getopt_t[] = {  {&_swigt__p_apr_getopt_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_hash_t[] = {  {&_swigt__p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int32_t[] = {  {&_swigt__p_apr_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int64_t[] = {  {&_swigt__p_apr_int64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_pool_t[] = {  {&_swigt__p_apr_pool_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void[] = {  {&_swigt__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char__int[] = {  {&_swigt__p_f_p_void_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_array_header_t[] = {  {&_swigt__p_p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_hash_t[] = {  {&_swigt__p_p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_provider_object_t[] = {  {&_swigt__p_p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_client_commit_info_t[] = {  {&_swigt__p_p_svn_client_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_client_ctx_t[] = {  {&_swigt__p_p_svn_client_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_commit_info_t[] = {  {&_swigt__p_p_svn_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_ra_session_t[] = {  {&_swigt__p_p_svn_ra_session_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_string_t[] = {  {&_swigt__p_p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_baton_t[] = {  {&_swigt__p_svn_auth_baton_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_simple_t[] = {  {&_swigt__p_svn_auth_cred_simple_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_pw_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_pw_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_server_trust_t[] = {  {&_swigt__p_svn_auth_cred_ssl_server_trust_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_username_t[] = {  {&_swigt__p_svn_auth_cred_username_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_iterstate_t[] = {  {&_swigt__p_svn_auth_iterstate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_object_t[] = {  {&_swigt__p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_t[] = {  {&_swigt__p_svn_auth_provider_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_ssl_server_cert_info_t[] = {  {&_swigt__p_svn_auth_ssl_server_cert_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_ctx_t[] = {  {&_swigt__p_svn_checksum_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_kind_t[] = {  {&_swigt__p_svn_checksum_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_t[] = {  {&_swigt__p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_commit_info_t[] = {  {&_swigt__p_svn_client_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_commit_item2_t[] = {  {&_swigt__p_svn_client_commit_item2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_commit_item3_t[] = {  {&_swigt__p_svn_client_commit_item3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_commit_item_t[] = {  {&_swigt__p_svn_client_commit_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_copy_source_t[] = {  {&_swigt__p_svn_client_copy_source_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_ctx_t[] = {  {&_swigt__p_svn_client_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_diff_summarize_kind_t[] = {  {&_swigt__p_svn_client_diff_summarize_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_diff_summarize_t[] = {  {&_swigt__p_svn_client_diff_summarize_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_info2_t[] = {  {&_swigt__p_svn_client_info2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_proplist_item_t[] = {  {&_swigt__p_svn_client_proplist_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_status_t[] = {  {&_swigt__p_svn_client_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_commit_info_t[] = {  {&_swigt__p_svn_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_config_t[] = {  {&_swigt__p_svn_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_editor_t[] = {  {&_swigt__p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_depth_t[] = {  {&_swigt__p_svn_depth_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_conflict_display_style_t[] = {  {&_swigt__p_svn_diff_conflict_display_style_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_datasource_e[] = {  {&_swigt__p_svn_diff_datasource_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_file_ignore_space_t[] = {  {&_swigt__p_svn_diff_file_ignore_space_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_file_options_t[] = {  {&_swigt__p_svn_diff_file_options_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_fns2_t[] = {  {&_swigt__p_svn_diff_fns2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_fns_t[] = {  {&_swigt__p_svn_diff_fns_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_hunk_t[] = {  {&_swigt__p_svn_diff_hunk_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_operation_kind_e[] = {  {&_swigt__p_svn_diff_operation_kind_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_output_fns_t[] = {  {&_swigt__p_svn_diff_output_fns_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_t[] = {  {&_swigt__p_svn_diff_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_dirent_t[] = {  {&_swigt__p_svn_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_errno_t[] = {  {&_swigt__p_svn_errno_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_error_t[] = {  {&_swigt__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_info_t[] = {  {&_swigt__p_svn_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent2_t[] = {  {&_swigt__p_svn_io_dirent2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent_t[] = {  {&_swigt__p_svn_io_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_file_del_t[] = {  {&_swigt__p_svn_io_file_del_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_location_segment_t[] = {  {&_swigt__p_svn_location_segment_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_lock_t[] = {  {&_swigt__p_svn_lock_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path2_t[] = {  {&_swigt__p_svn_log_changed_path2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path_t[] = {  {&_swigt__p_svn_log_changed_path_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_entry_t[] = {  {&_swigt__p_svn_log_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_merge_range_t[] = {  {&_swigt__p_svn_merge_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_mergeinfo_inheritance_t[] = {  {&_swigt__p_svn_mergeinfo_inheritance_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_node_kind_t[] = {  {&_swigt__p_svn_node_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_range_t[] = {  {&_swigt__p_svn_opt_revision_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_t[] = {  {&_swigt__p_svn_opt_revision_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_value_t[] = {  {&_swigt__p_svn_opt_revision_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc2_t[] = {  {&_swigt__p_svn_opt_subcommand_desc2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc_t[] = {  {&_swigt__p_svn_opt_subcommand_desc_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_patch_file_t[] = {  {&_swigt__p_svn_patch_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_patch_t[] = {  {&_swigt__p_svn_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_inherited_item_t[] = {  {&_swigt__p_svn_prop_inherited_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_kind[] = {  {&_swigt__p_svn_prop_kind, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_patch_t[] = {  {&_swigt__p_svn_prop_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks2_t[] = {  {&_swigt__p_svn_ra_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks_t[] = {  {&_swigt__p_svn_ra_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_plugin_t[] = {  {&_swigt__p_svn_ra_plugin_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter2_t[] = {  {&_swigt__p_svn_ra_reporter2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter3_t[] = {  {&_swigt__p_svn_ra_reporter3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter_t[] = {  {&_swigt__p_svn_ra_reporter_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_session_t[] = {  {&_swigt__p_svn_ra_session_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_mark_t[] = {  {&_swigt__p_svn_stream_mark_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_t[] = {  {&_swigt__p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_string_t[] = {  {&_swigt__p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stringbuf_t[] = {  {&_swigt__p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_tristate_t[] = {  {&_swigt__p_svn_tristate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_op_t[] = {  {&_swigt__p_svn_txdelta_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_stream_t[] = {  {&_swigt__p_svn_txdelta_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_window_t[] = {  {&_swigt__p_svn_txdelta_window_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_checklist_t[] = {  {&_swigt__p_svn_version_checklist_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_linked_lib_t[] = {  {&_swigt__p_svn_version_ext_linked_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_loaded_lib_t[] = {  {&_swigt__p_svn_version_ext_loaded_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_extended_t[] = {  {&_swigt__p_svn_version_extended_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_t[] = {  {&_swigt__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_adm_access_t[] = {  {&_swigt__p_svn_wc_adm_access_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_committed_queue_t[] = {  {&_swigt__p_svn_wc_committed_queue_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_action_t[] = {  {&_swigt__p_svn_wc_conflict_action_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_choice_t[] = {  {&_swigt__p_svn_wc_conflict_choice_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_description_t[] = {  {&_swigt__p_svn_wc_conflict_description_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_kind_t[] = {  {&_swigt__p_svn_wc_conflict_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_reason_t[] = {  {&_swigt__p_svn_wc_conflict_reason_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_version_t[] = {  {&_swigt__p_svn_wc_conflict_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_context_t[] = {  {&_swigt__p_svn_wc_context_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks2_t[] = {  {&_swigt__p_svn_wc_diff_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks3_t[] = {  {&_swigt__p_svn_wc_diff_callbacks3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks4_t[] = {  {&_swigt__p_svn_wc_diff_callbacks4_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks_t[] = {  {&_swigt__p_svn_wc_diff_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_callbacks2_t[] = {  {&_swigt__p_svn_wc_entry_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_callbacks_t[] = {  {&_swigt__p_svn_wc_entry_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_t[] = {  {&_swigt__p_svn_wc_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item2_t[] = {  {&_swigt__p_svn_wc_external_item2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item_t[] = {  {&_swigt__p_svn_wc_external_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_info_t[] = {  {&_swigt__p_svn_wc_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_merge_outcome_t[] = {  {&_swigt__p_svn_wc_merge_outcome_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_action_t[] = {  {&_swigt__p_svn_wc_notify_action_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_lock_state_t[] = {  {&_swigt__p_svn_wc_notify_lock_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_state_t[] = {  {&_swigt__p_svn_wc_notify_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_t[] = {  {&_swigt__p_svn_wc_notify_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_operation_t[] = {  {&_swigt__p_svn_wc_operation_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_revision_status_t[] = {  {&_swigt__p_svn_wc_revision_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_schedule_t[] = {  {&_swigt__p_svn_wc_schedule_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status2_t[] = {  {&_swigt__p_svn_wc_status2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status3_t[] = {  {&_swigt__p_svn_wc_status3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status_t[] = {  {&_swigt__p_svn_wc_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_traversal_info_t[] = {  {&_swigt__p_svn_wc_traversal_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_apr_array_header_t,
  _swigc__p_apr_file_t,
  _swigc__p_apr_getopt_t,
  _swigc__p_apr_hash_t,
  _swigc__p_apr_int32_t,
  _swigc__p_apr_int64_t,
  _swigc__p_apr_pool_t,
  _swigc__p_char,
  _swigc__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void,
  _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_q_const__apr_array_header_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__svn_client_diff_summarize_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void__p_svn_error_t,
  _swigc__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_apr_int64_t_svn_revnum_t_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char__int,
  _swigc__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  _swigc__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_client_info2_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_client_status_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_dirent_t_p_q_const__svn_lock_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_info_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  _swigc__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  _swigc__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_boolean_t_p_q_const__char_p_q_const__svn_io_dirent2_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_svn_revnum_t_svn_revnum_t_apr_int64_t_svn_revnum_t_p_apr_hash_t_svn_revnum_t_p_apr_hash_t_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_p_apr_array_header_t,
  _swigc__p_p_apr_hash_t,
  _swigc__p_p_char,
  _swigc__p_p_svn_auth_provider_object_t,
  _swigc__p_p_svn_client_commit_info_t,
  _swigc__p_p_svn_client_ctx_t,
  _swigc__p_p_svn_commit_info_t,
  _swigc__p_p_svn_ra_session_t,
  _swigc__p_p_svn_string_t,
  _swigc__p_svn_auth_baton_t,
  _swigc__p_svn_auth_cred_simple_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_pw_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_t,
  _swigc__p_svn_auth_cred_ssl_server_trust_t,
  _swigc__p_svn_auth_cred_username_t,
  _swigc__p_svn_auth_iterstate_t,
  _swigc__p_svn_auth_provider_object_t,
  _swigc__p_svn_auth_provider_t,
  _swigc__p_svn_auth_ssl_server_cert_info_t,
  _swigc__p_svn_checksum_ctx_t,
  _swigc__p_svn_checksum_kind_t,
  _swigc__p_svn_checksum_t,
  _swigc__p_svn_client_commit_info_t,
  _swigc__p_svn_client_commit_item2_t,
  _swigc__p_svn_client_commit_item3_t,
  _swigc__p_svn_client_commit_item_t,
  _swigc__p_svn_client_copy_source_t,
  _swigc__p_svn_client_ctx_t,
  _swigc__p_svn_client_diff_summarize_kind_t,
  _swigc__p_svn_client_diff_summarize_t,
  _swigc__p_svn_client_info2_t,
  _swigc__p_svn_client_proplist_item_t,
  _swigc__p_svn_client_status_t,
  _swigc__p_svn_commit_info_t,
  _swigc__p_svn_config_t,
  _swigc__p_svn_delta_editor_t,
  _swigc__p_svn_depth_t,
  _swigc__p_svn_diff_conflict_display_style_t,
  _swigc__p_svn_diff_datasource_e,
  _swigc__p_svn_diff_file_ignore_space_t,
  _swigc__p_svn_diff_file_options_t,
  _swigc__p_svn_diff_fns2_t,
  _swigc__p_svn_diff_fns_t,
  _swigc__p_svn_diff_hunk_t,
  _swigc__p_svn_diff_operation_kind_e,
  _swigc__p_svn_diff_output_fns_t,
  _swigc__p_svn_diff_t,
  _swigc__p_svn_dirent_t,
  _swigc__p_svn_errno_t,
  _swigc__p_svn_error_t,
  _swigc__p_svn_info_t,
  _swigc__p_svn_io_dirent2_t,
  _swigc__p_svn_io_dirent_t,
  _swigc__p_svn_io_file_del_t,
  _swigc__p_svn_location_segment_t,
  _swigc__p_svn_lock_t,
  _swigc__p_svn_log_changed_path2_t,
  _swigc__p_svn_log_changed_path_t,
  _swigc__p_svn_log_entry_t,
  _swigc__p_svn_merge_range_t,
  _swigc__p_svn_mergeinfo_inheritance_t,
  _swigc__p_svn_node_kind_t,
  _swigc__p_svn_opt_revision_range_t,
  _swigc__p_svn_opt_revision_t,
  _swigc__p_svn_opt_revision_value_t,
  _swigc__p_svn_opt_subcommand_desc2_t,
  _swigc__p_svn_opt_subcommand_desc_t,
  _swigc__p_svn_patch_file_t,
  _swigc__p_svn_patch_t,
  _swigc__p_svn_prop_inherited_item_t,
  _swigc__p_svn_prop_kind,
  _swigc__p_svn_prop_patch_t,
  _swigc__p_svn_ra_callbacks2_t,
  _swigc__p_svn_ra_callbacks_t,
  _swigc__p_svn_ra_plugin_t,
  _swigc__p_svn_ra_reporter2_t,
  _swigc__p_svn_ra_reporter3_t,
  _swigc__p_svn_ra_reporter_t,
  _swigc__p_svn_ra_session_t,
  _swigc__p_svn_stream_mark_t,
  _swigc__p_svn_stream_t,
  _swigc__p_svn_string_t,
  _swigc__p_svn_stringbuf_t,
  _swigc__p_svn_tristate_t,
  _swigc__p_svn_txdelta_op_t,
  _swigc__p_svn_txdelta_stream_t,
  _swigc__p_svn_txdelta_window_t,
  _swigc__p_svn_version_checklist_t,
  _swigc__p_svn_version_ext_linked_lib_t,
  _swigc__p_svn_version_ext_loaded_lib_t,
  _swigc__p_svn_version_extended_t,
  _swigc__p_svn_version_t,
  _swigc__p_svn_wc_adm_access_t,
  _swigc__p_svn_wc_committed_queue_t,
  _swigc__p_svn_wc_conflict_action_t,
  _swigc__p_svn_wc_conflict_choice_t,
  _swigc__p_svn_wc_conflict_description_t,
  _swigc__p_svn_wc_conflict_kind_t,
  _swigc__p_svn_wc_conflict_reason_t,
  _swigc__p_svn_wc_conflict_version_t,
  _swigc__p_svn_wc_context_t,
  _swigc__p_svn_wc_diff_callbacks2_t,
  _swigc__p_svn_wc_diff_callbacks3_t,
  _swigc__p_svn_wc_diff_callbacks4_t,
  _swigc__p_svn_wc_diff_callbacks_t,
  _swigc__p_svn_wc_entry_callbacks2_t,
  _swigc__p_svn_wc_entry_callbacks_t,
  _swigc__p_svn_wc_entry_t,
  _swigc__p_svn_wc_external_item2_t,
  _swigc__p_svn_wc_external_item_t,
  _swigc__p_svn_wc_info_t,
  _swigc__p_svn_wc_merge_outcome_t,
  _swigc__p_svn_wc_notify_action_t,
  _swigc__p_svn_wc_notify_lock_state_t,
  _swigc__p_svn_wc_notify_state_t,
  _swigc__p_svn_wc_notify_t,
  _swigc__p_svn_wc_operation_t,
  _swigc__p_svn_wc_revision_status_t,
  _swigc__p_svn_wc_schedule_t,
  _swigc__p_svn_wc_status2_t,
  _swigc__p_svn_wc_status3_t,
  _swigc__p_svn_wc_status_t,
  _swigc__p_svn_wc_traversal_info_t,
  _swigc__p_unsigned_long,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned staticly to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int found, init;

  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }

  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
    module_head = &swig_module;
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    found=0;
    iter=module_head;
    do {
      if (iter==&swig_module) {
        found=1;
        break;
      }
      iter=iter->next;
    } while (iter!= module_head);

    /* if the is found in the list, then all is done and we may leave */
    if (found) return;
    /* otherwise we must add out module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }

  /* When multiple interpreters are used, a module could have already been initialized in
     a different interpreter, but not yet have a pointer in this interpreter.
     In this case, we do not want to continue adding types... everything should be
     set up already */
  if (init == 0) return;

  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %d\n", swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;

#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
#endif

    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
	type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }

    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {

      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
	if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
	if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
	  printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
	  cast->type = ret;
	  ret = 0;
	} else {
	  /* Check for casting already in the list */
	  swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
	  if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
	  if (!ocast) ret = 0;
	}
      }

      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
	printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;

#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
  printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;

  if (init_run) return;
  init_run = 1;

  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
            SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{ /* c-mode */
#endif
}
#endif

/*

*/
#ifdef __cplusplus
extern "C"
#endif
SWIGEXPORT void Init_client(void) {
  size_t i;
  
  SWIG_InitRuntime();
  mClient = rb_define_module("Svn");
  mClient = rb_define_module_under(mClient, "Ext");
  mClient = rb_define_module_under(mClient, "Client");
  
  SWIG_InitializeModule(0);
  for (i = 0; i < swig_module.size; i++) {
    SWIG_define_class(swig_module.types[i]);
  }
  
  SWIG_RubyInitializeTrackings();
  rb_require("svn/ext/core");
  rb_require("svn/ext/delta");
  rb_require("svn/ext/wc");
  rb_require("svn/ext/ra");
  rb_define_module_function(mClient, "svn_client_version", _wrap_svn_client_version, -1);
  rb_define_module_function(mClient, "svn_client_get_simple_prompt_provider", _wrap_svn_client_get_simple_prompt_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_username_prompt_provider", _wrap_svn_client_get_username_prompt_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_simple_provider", _wrap_svn_client_get_simple_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_username_provider", _wrap_svn_client_get_username_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_ssl_server_trust_file_provider", _wrap_svn_client_get_ssl_server_trust_file_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_ssl_client_cert_file_provider", _wrap_svn_client_get_ssl_client_cert_file_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_ssl_client_cert_pw_file_provider", _wrap_svn_client_get_ssl_client_cert_pw_file_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_ssl_server_trust_prompt_provider", _wrap_svn_client_get_ssl_server_trust_prompt_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_ssl_client_cert_prompt_provider", _wrap_svn_client_get_ssl_client_cert_prompt_provider, -1);
  rb_define_module_function(mClient, "svn_client_get_ssl_client_cert_pw_prompt_provider", _wrap_svn_client_get_ssl_client_cert_pw_prompt_provider, -1);
  
  SwigClassSvn_client_proplist_item_t.klass = rb_define_class_under(mClient, "Svn_client_proplist_item_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_proplist_item_t, (void *) &SwigClassSvn_client_proplist_item_t);
  rb_define_alloc_func(SwigClassSvn_client_proplist_item_t.klass, _wrap_svn_client_proplist_item_t_allocate);
  rb_define_method(SwigClassSvn_client_proplist_item_t.klass, "initialize", _wrap_new_svn_client_proplist_item_t, -1);
  rb_define_method(SwigClassSvn_client_proplist_item_t.klass, "node_name=", _wrap_svn_client_proplist_item_t_node_name_set, -1);
  rb_define_method(SwigClassSvn_client_proplist_item_t.klass, "node_name", _wrap_svn_client_proplist_item_t_node_name_get, -1);
  rb_define_method(SwigClassSvn_client_proplist_item_t.klass, "prop_hash=", _wrap_svn_client_proplist_item_t_prop_hash_set, -1);
  rb_define_method(SwigClassSvn_client_proplist_item_t.klass, "prop_hash", _wrap_svn_client_proplist_item_t_prop_hash_get, -1);
  SwigClassSvn_client_proplist_item_t.mark = 0;
  SwigClassSvn_client_proplist_item_t.destroy = (void (*)(void *)) free_svn_client_proplist_item_t;
  SwigClassSvn_client_proplist_item_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_client_proplist_item_dup", _wrap_svn_client_proplist_item_dup, -1);
  
  SwigClassSvn_client_commit_info_t.klass = rb_define_class_under(mClient, "Svn_client_commit_info_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_commit_info_t, (void *) &SwigClassSvn_client_commit_info_t);
  rb_define_alloc_func(SwigClassSvn_client_commit_info_t.klass, _wrap_svn_client_commit_info_t_allocate);
  rb_define_method(SwigClassSvn_client_commit_info_t.klass, "initialize", _wrap_new_svn_client_commit_info_t, -1);
  rb_define_method(SwigClassSvn_client_commit_info_t.klass, "revision=", _wrap_svn_client_commit_info_t_revision_set, -1);
  rb_define_method(SwigClassSvn_client_commit_info_t.klass, "revision", _wrap_svn_client_commit_info_t_revision_get, -1);
  rb_define_method(SwigClassSvn_client_commit_info_t.klass, "date=", _wrap_svn_client_commit_info_t_date_set, -1);
  rb_define_method(SwigClassSvn_client_commit_info_t.klass, "date", _wrap_svn_client_commit_info_t_date_get, -1);
  rb_define_method(SwigClassSvn_client_commit_info_t.klass, "author=", _wrap_svn_client_commit_info_t_author_set, -1);
  rb_define_method(SwigClassSvn_client_commit_info_t.klass, "author", _wrap_svn_client_commit_info_t_author_get, -1);
  SwigClassSvn_client_commit_info_t.mark = 0;
  SwigClassSvn_client_commit_info_t.destroy = (void (*)(void *)) free_svn_client_commit_info_t;
  SwigClassSvn_client_commit_info_t.trackObjects = 0;
  rb_define_const(mClient, "SVN_CLIENT_COMMIT_ITEM_ADD", SWIG_From_int((int)(0x01)));
  rb_define_const(mClient, "SVN_CLIENT_COMMIT_ITEM_DELETE", SWIG_From_int((int)(0x02)));
  rb_define_const(mClient, "SVN_CLIENT_COMMIT_ITEM_TEXT_MODS", SWIG_From_int((int)(0x04)));
  rb_define_const(mClient, "SVN_CLIENT_COMMIT_ITEM_PROP_MODS", SWIG_From_int((int)(0x08)));
  rb_define_const(mClient, "SVN_CLIENT_COMMIT_ITEM_IS_COPY", SWIG_From_int((int)(0x10)));
  rb_define_const(mClient, "SVN_CLIENT_COMMIT_ITEM_LOCK_TOKEN", SWIG_From_int((int)(0x20)));
  rb_define_const(mClient, "SVN_CLIENT_COMMIT_ITEM_MOVED_HERE", SWIG_From_int((int)(0x40)));
  
  SwigClassSvn_client_commit_item3_t.klass = rb_define_class_under(mClient, "Svn_client_commit_item3_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_commit_item3_t, (void *) &SwigClassSvn_client_commit_item3_t);
  rb_define_alloc_func(SwigClassSvn_client_commit_item3_t.klass, _wrap_svn_client_commit_item3_t_allocate);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "initialize", _wrap_new_svn_client_commit_item3_t, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "path=", _wrap_svn_client_commit_item3_t_path_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "path", _wrap_svn_client_commit_item3_t_path_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "kind=", _wrap_svn_client_commit_item3_t_kind_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "kind", _wrap_svn_client_commit_item3_t_kind_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "url=", _wrap_svn_client_commit_item3_t_url_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "url", _wrap_svn_client_commit_item3_t_url_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "revision=", _wrap_svn_client_commit_item3_t_revision_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "revision", _wrap_svn_client_commit_item3_t_revision_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "copyfrom_url=", _wrap_svn_client_commit_item3_t_copyfrom_url_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "copyfrom_url", _wrap_svn_client_commit_item3_t_copyfrom_url_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "copyfrom_rev=", _wrap_svn_client_commit_item3_t_copyfrom_rev_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "copyfrom_rev", _wrap_svn_client_commit_item3_t_copyfrom_rev_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "state_flags=", _wrap_svn_client_commit_item3_t_state_flags_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "state_flags", _wrap_svn_client_commit_item3_t_state_flags_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "incoming_prop_changes=", _wrap_svn_client_commit_item3_t_incoming_prop_changes_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "incoming_prop_changes", _wrap_svn_client_commit_item3_t_incoming_prop_changes_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "outgoing_prop_changes=", _wrap_svn_client_commit_item3_t_outgoing_prop_changes_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "outgoing_prop_changes", _wrap_svn_client_commit_item3_t_outgoing_prop_changes_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "session_relpath=", _wrap_svn_client_commit_item3_t_session_relpath_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "session_relpath", _wrap_svn_client_commit_item3_t_session_relpath_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "moved_from_abspath=", _wrap_svn_client_commit_item3_t_moved_from_abspath_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "moved_from_abspath", _wrap_svn_client_commit_item3_t_moved_from_abspath_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item3_t.klass, "dup", _wrap_svn_client_commit_item3_t_dup, -1);
  SwigClassSvn_client_commit_item3_t.mark = 0;
  SwigClassSvn_client_commit_item3_t.destroy = (void (*)(void *)) free_svn_client_commit_item3_t;
  SwigClassSvn_client_commit_item3_t.trackObjects = 0;
  
  SwigClassSvn_client_commit_item2_t.klass = rb_define_class_under(mClient, "Svn_client_commit_item2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_commit_item2_t, (void *) &SwigClassSvn_client_commit_item2_t);
  rb_define_alloc_func(SwigClassSvn_client_commit_item2_t.klass, _wrap_svn_client_commit_item2_t_allocate);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "initialize", _wrap_new_svn_client_commit_item2_t, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "path=", _wrap_svn_client_commit_item2_t_path_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "path", _wrap_svn_client_commit_item2_t_path_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "kind=", _wrap_svn_client_commit_item2_t_kind_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "kind", _wrap_svn_client_commit_item2_t_kind_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "url=", _wrap_svn_client_commit_item2_t_url_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "url", _wrap_svn_client_commit_item2_t_url_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "revision=", _wrap_svn_client_commit_item2_t_revision_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "revision", _wrap_svn_client_commit_item2_t_revision_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "copyfrom_url=", _wrap_svn_client_commit_item2_t_copyfrom_url_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "copyfrom_url", _wrap_svn_client_commit_item2_t_copyfrom_url_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "copyfrom_rev=", _wrap_svn_client_commit_item2_t_copyfrom_rev_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "copyfrom_rev", _wrap_svn_client_commit_item2_t_copyfrom_rev_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "state_flags=", _wrap_svn_client_commit_item2_t_state_flags_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "state_flags", _wrap_svn_client_commit_item2_t_state_flags_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "wcprop_changes=", _wrap_svn_client_commit_item2_t_wcprop_changes_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item2_t.klass, "wcprop_changes", _wrap_svn_client_commit_item2_t_wcprop_changes_get, -1);
  SwigClassSvn_client_commit_item2_t.mark = 0;
  SwigClassSvn_client_commit_item2_t.destroy = (void (*)(void *)) free_svn_client_commit_item2_t;
  SwigClassSvn_client_commit_item2_t.trackObjects = 0;
  
  SwigClassSvn_client_commit_item_t.klass = rb_define_class_under(mClient, "Svn_client_commit_item_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_commit_item_t, (void *) &SwigClassSvn_client_commit_item_t);
  rb_define_alloc_func(SwigClassSvn_client_commit_item_t.klass, _wrap_svn_client_commit_item_t_allocate);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "initialize", _wrap_new_svn_client_commit_item_t, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "path=", _wrap_svn_client_commit_item_t_path_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "path", _wrap_svn_client_commit_item_t_path_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "kind=", _wrap_svn_client_commit_item_t_kind_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "kind", _wrap_svn_client_commit_item_t_kind_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "url=", _wrap_svn_client_commit_item_t_url_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "url", _wrap_svn_client_commit_item_t_url_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "revision=", _wrap_svn_client_commit_item_t_revision_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "revision", _wrap_svn_client_commit_item_t_revision_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "copyfrom_url=", _wrap_svn_client_commit_item_t_copyfrom_url_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "copyfrom_url", _wrap_svn_client_commit_item_t_copyfrom_url_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "state_flags=", _wrap_svn_client_commit_item_t_state_flags_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "state_flags", _wrap_svn_client_commit_item_t_state_flags_get, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "wcprop_changes=", _wrap_svn_client_commit_item_t_wcprop_changes_set, -1);
  rb_define_method(SwigClassSvn_client_commit_item_t.klass, "wcprop_changes", _wrap_svn_client_commit_item_t_wcprop_changes_get, -1);
  SwigClassSvn_client_commit_item_t.mark = 0;
  SwigClassSvn_client_commit_item_t.destroy = (void (*)(void *)) free_svn_client_commit_item_t;
  SwigClassSvn_client_commit_item_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_client_commit_item3_create", _wrap_svn_client_commit_item3_create, -1);
  rb_define_const(mClient, "Svn_client_diff_summarize_kind_normal", SWIG_From_int((int)(svn_client_diff_summarize_kind_normal)));
  rb_define_const(mClient, "Svn_client_diff_summarize_kind_added", SWIG_From_int((int)(svn_client_diff_summarize_kind_added)));
  rb_define_const(mClient, "Svn_client_diff_summarize_kind_modified", SWIG_From_int((int)(svn_client_diff_summarize_kind_modified)));
  rb_define_const(mClient, "Svn_client_diff_summarize_kind_deleted", SWIG_From_int((int)(svn_client_diff_summarize_kind_deleted)));
  
  SwigClassSvn_client_diff_summarize_t.klass = rb_define_class_under(mClient, "Svn_client_diff_summarize_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_diff_summarize_t, (void *) &SwigClassSvn_client_diff_summarize_t);
  rb_define_alloc_func(SwigClassSvn_client_diff_summarize_t.klass, _wrap_svn_client_diff_summarize_t_allocate);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "initialize", _wrap_new_svn_client_diff_summarize_t, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "path=", _wrap_svn_client_diff_summarize_t_path_set, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "path", _wrap_svn_client_diff_summarize_t_path_get, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "summarize_kind=", _wrap_svn_client_diff_summarize_t_summarize_kind_set, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "summarize_kind", _wrap_svn_client_diff_summarize_t_summarize_kind_get, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "prop_changed=", _wrap_svn_client_diff_summarize_t_prop_changed_set, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "prop_changed", _wrap_svn_client_diff_summarize_t_prop_changed_get, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "node_kind=", _wrap_svn_client_diff_summarize_t_node_kind_set, -1);
  rb_define_method(SwigClassSvn_client_diff_summarize_t.klass, "node_kind", _wrap_svn_client_diff_summarize_t_node_kind_get, -1);
  SwigClassSvn_client_diff_summarize_t.mark = 0;
  SwigClassSvn_client_diff_summarize_t.destroy = (void (*)(void *)) free_svn_client_diff_summarize_t;
  SwigClassSvn_client_diff_summarize_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_client_diff_summarize_dup", _wrap_svn_client_diff_summarize_dup, -1);
  
  SwigClassSvn_client_ctx_t.klass = rb_define_class_under(mClient, "Svn_client_ctx_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_ctx_t, (void *) &SwigClassSvn_client_ctx_t);
  rb_define_alloc_func(SwigClassSvn_client_ctx_t.klass, _wrap_svn_client_ctx_t_allocate);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "initialize", _wrap_new_svn_client_ctx_t, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "auth_baton=", _wrap_svn_client_ctx_t_auth_baton_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "auth_baton", _wrap_svn_client_ctx_t_auth_baton_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_func=", _wrap_svn_client_ctx_t_notify_func_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_func", _wrap_svn_client_ctx_t_notify_func_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_baton=", _wrap_svn_client_ctx_t_notify_baton_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_baton", _wrap_svn_client_ctx_t_notify_baton_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_func=", _wrap_svn_client_ctx_t_log_msg_func_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_func", _wrap_svn_client_ctx_t_log_msg_func_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_baton=", _wrap_svn_client_ctx_t_log_msg_baton_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_baton", _wrap_svn_client_ctx_t_log_msg_baton_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "cancel_func=", _wrap_svn_client_ctx_t_cancel_func_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "cancel_func", _wrap_svn_client_ctx_t_cancel_func_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "cancel_baton=", _wrap_svn_client_ctx_t_cancel_baton_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "cancel_baton", _wrap_svn_client_ctx_t_cancel_baton_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_func2=", _wrap_svn_client_ctx_t_notify_func2_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_func2", _wrap_svn_client_ctx_t_notify_func2_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_baton2=", _wrap_svn_client_ctx_t_notify_baton2_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "notify_baton2", _wrap_svn_client_ctx_t_notify_baton2_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_func2=", _wrap_svn_client_ctx_t_log_msg_func2_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_func2", _wrap_svn_client_ctx_t_log_msg_func2_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_baton2=", _wrap_svn_client_ctx_t_log_msg_baton2_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_baton2", _wrap_svn_client_ctx_t_log_msg_baton2_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "progress_func=", _wrap_svn_client_ctx_t_progress_func_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "progress_func", _wrap_svn_client_ctx_t_progress_func_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "progress_baton=", _wrap_svn_client_ctx_t_progress_baton_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "progress_baton", _wrap_svn_client_ctx_t_progress_baton_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_func3=", _wrap_svn_client_ctx_t_log_msg_func3_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_func3", _wrap_svn_client_ctx_t_log_msg_func3_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_baton3=", _wrap_svn_client_ctx_t_log_msg_baton3_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "log_msg_baton3", _wrap_svn_client_ctx_t_log_msg_baton3_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "mimetypes_map=", _wrap_svn_client_ctx_t_mimetypes_map_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "mimetypes_map", _wrap_svn_client_ctx_t_mimetypes_map_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_func=", _wrap_svn_client_ctx_t_conflict_func_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_func", _wrap_svn_client_ctx_t_conflict_func_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_baton=", _wrap_svn_client_ctx_t_conflict_baton_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_baton", _wrap_svn_client_ctx_t_conflict_baton_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "client_name=", _wrap_svn_client_ctx_t_client_name_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "client_name", _wrap_svn_client_ctx_t_client_name_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_func2=", _wrap_svn_client_ctx_t_conflict_func2_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_func2", _wrap_svn_client_ctx_t_conflict_func2_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_baton2=", _wrap_svn_client_ctx_t_conflict_baton2_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "conflict_baton2", _wrap_svn_client_ctx_t_conflict_baton2_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "wc_ctx=", _wrap_svn_client_ctx_t_wc_ctx_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "wc_ctx", _wrap_svn_client_ctx_t_wc_ctx_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "check_tunnel_func=", _wrap_svn_client_ctx_t_check_tunnel_func_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "check_tunnel_func", _wrap_svn_client_ctx_t_check_tunnel_func_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "open_tunnel_func=", _wrap_svn_client_ctx_t_open_tunnel_func_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "open_tunnel_func", _wrap_svn_client_ctx_t_open_tunnel_func_get, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "tunnel_baton=", _wrap_svn_client_ctx_t_tunnel_baton_set, -1);
  rb_define_method(SwigClassSvn_client_ctx_t.klass, "tunnel_baton", _wrap_svn_client_ctx_t_tunnel_baton_get, -1);
  SwigClassSvn_client_ctx_t.mark = 0;
  SwigClassSvn_client_ctx_t.destroy = (void (*)(void *)) free_svn_client_ctx_t;
  SwigClassSvn_client_ctx_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_client_create_context2", _wrap_svn_client_create_context2, -1);
  rb_define_const(mClient, "SVN_CLIENT_AUTH_USERNAME", SWIG_FromCharPtr("username"));
  rb_define_const(mClient, "SVN_CLIENT_AUTH_PASSWORD", SWIG_FromCharPtr("password"));
  rb_define_module_function(mClient, "svn_client_args_to_target_array2", _wrap_svn_client_args_to_target_array2, -1);
  rb_define_module_function(mClient, "svn_client_args_to_target_array", _wrap_svn_client_args_to_target_array, -1);
  rb_define_module_function(mClient, "svn_client_checkout3", _wrap_svn_client_checkout3, -1);
  rb_define_module_function(mClient, "svn_client_checkout2", _wrap_svn_client_checkout2, -1);
  rb_define_module_function(mClient, "svn_client_checkout", _wrap_svn_client_checkout, -1);
  rb_define_module_function(mClient, "svn_client_update4", _wrap_svn_client_update4, -1);
  rb_define_module_function(mClient, "svn_client_update3", _wrap_svn_client_update3, -1);
  rb_define_module_function(mClient, "svn_client_update2", _wrap_svn_client_update2, -1);
  rb_define_module_function(mClient, "svn_client_update", _wrap_svn_client_update, -1);
  rb_define_module_function(mClient, "svn_client_switch3", _wrap_svn_client_switch3, -1);
  rb_define_module_function(mClient, "svn_client_switch2", _wrap_svn_client_switch2, -1);
  rb_define_module_function(mClient, "svn_client_switch", _wrap_svn_client_switch, -1);
  rb_define_module_function(mClient, "svn_client_add5", _wrap_svn_client_add5, -1);
  rb_define_module_function(mClient, "svn_client_add4", _wrap_svn_client_add4, -1);
  rb_define_module_function(mClient, "svn_client_add3", _wrap_svn_client_add3, -1);
  rb_define_module_function(mClient, "svn_client_add2", _wrap_svn_client_add2, -1);
  rb_define_module_function(mClient, "svn_client_add", _wrap_svn_client_add, -1);
  rb_define_module_function(mClient, "svn_client_mkdir4", _wrap_svn_client_mkdir4, -1);
  rb_define_module_function(mClient, "svn_client_mkdir3", _wrap_svn_client_mkdir3, -1);
  rb_define_module_function(mClient, "svn_client_mkdir2", _wrap_svn_client_mkdir2, -1);
  rb_define_module_function(mClient, "svn_client_mkdir", _wrap_svn_client_mkdir, -1);
  rb_define_module_function(mClient, "svn_client_delete4", _wrap_svn_client_delete4, -1);
  rb_define_module_function(mClient, "svn_client_delete3", _wrap_svn_client_delete3, -1);
  rb_define_module_function(mClient, "svn_client_delete2", _wrap_svn_client_delete2, -1);
  rb_define_module_function(mClient, "svn_client_delete", _wrap_svn_client_delete, -1);
  rb_define_module_function(mClient, "svn_client_import5", _wrap_svn_client_import5, -1);
  rb_define_module_function(mClient, "svn_client_import4", _wrap_svn_client_import4, -1);
  rb_define_module_function(mClient, "svn_client_import3", _wrap_svn_client_import3, -1);
  rb_define_module_function(mClient, "svn_client_import2", _wrap_svn_client_import2, -1);
  rb_define_module_function(mClient, "svn_client_import", _wrap_svn_client_import, -1);
  rb_define_module_function(mClient, "svn_client_commit6", _wrap_svn_client_commit6, -1);
  rb_define_module_function(mClient, "svn_client_commit5", _wrap_svn_client_commit5, -1);
  rb_define_module_function(mClient, "svn_client_commit4", _wrap_svn_client_commit4, -1);
  rb_define_module_function(mClient, "svn_client_commit3", _wrap_svn_client_commit3, -1);
  rb_define_module_function(mClient, "svn_client_commit2", _wrap_svn_client_commit2, -1);
  rb_define_module_function(mClient, "svn_client_commit", _wrap_svn_client_commit, -1);
  
  SwigClassSvn_client_status_t.klass = rb_define_class_under(mClient, "Svn_client_status_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_status_t, (void *) &SwigClassSvn_client_status_t);
  rb_define_alloc_func(SwigClassSvn_client_status_t.klass, _wrap_svn_client_status_t_allocate);
  rb_define_method(SwigClassSvn_client_status_t.klass, "initialize", _wrap_new_svn_client_status_t, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "kind=", _wrap_svn_client_status_t_kind_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "kind", _wrap_svn_client_status_t_kind_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "local_abspath=", _wrap_svn_client_status_t_local_abspath_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "local_abspath", _wrap_svn_client_status_t_local_abspath_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "filesize=", _wrap_svn_client_status_t_filesize_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "filesize", _wrap_svn_client_status_t_filesize_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "versioned=", _wrap_svn_client_status_t_versioned_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "versioned", _wrap_svn_client_status_t_versioned_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "conflicted=", _wrap_svn_client_status_t_conflicted_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "conflicted", _wrap_svn_client_status_t_conflicted_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "node_status=", _wrap_svn_client_status_t_node_status_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "node_status", _wrap_svn_client_status_t_node_status_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "text_status=", _wrap_svn_client_status_t_text_status_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "text_status", _wrap_svn_client_status_t_text_status_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "prop_status=", _wrap_svn_client_status_t_prop_status_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "prop_status", _wrap_svn_client_status_t_prop_status_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "wc_is_locked=", _wrap_svn_client_status_t_wc_is_locked_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "wc_is_locked", _wrap_svn_client_status_t_wc_is_locked_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "copied=", _wrap_svn_client_status_t_copied_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "copied", _wrap_svn_client_status_t_copied_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_root_url=", _wrap_svn_client_status_t_repos_root_url_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_root_url", _wrap_svn_client_status_t_repos_root_url_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_uuid=", _wrap_svn_client_status_t_repos_uuid_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_uuid", _wrap_svn_client_status_t_repos_uuid_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_relpath=", _wrap_svn_client_status_t_repos_relpath_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_relpath", _wrap_svn_client_status_t_repos_relpath_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "revision=", _wrap_svn_client_status_t_revision_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "revision", _wrap_svn_client_status_t_revision_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changed_rev=", _wrap_svn_client_status_t_changed_rev_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changed_rev", _wrap_svn_client_status_t_changed_rev_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changed_date=", _wrap_svn_client_status_t_changed_date_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changed_date", _wrap_svn_client_status_t_changed_date_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changed_author=", _wrap_svn_client_status_t_changed_author_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changed_author", _wrap_svn_client_status_t_changed_author_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "switched=", _wrap_svn_client_status_t_switched_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "switched", _wrap_svn_client_status_t_switched_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "file_external=", _wrap_svn_client_status_t_file_external_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "file_external", _wrap_svn_client_status_t_file_external_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "lock=", _wrap_svn_client_status_t_lock_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "lock", _wrap_svn_client_status_t_lock_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changelist=", _wrap_svn_client_status_t_changelist_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "changelist", _wrap_svn_client_status_t_changelist_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "depth=", _wrap_svn_client_status_t_depth_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "depth", _wrap_svn_client_status_t_depth_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_kind=", _wrap_svn_client_status_t_ood_kind_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_kind", _wrap_svn_client_status_t_ood_kind_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_node_status=", _wrap_svn_client_status_t_repos_node_status_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_node_status", _wrap_svn_client_status_t_repos_node_status_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_text_status=", _wrap_svn_client_status_t_repos_text_status_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_text_status", _wrap_svn_client_status_t_repos_text_status_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_prop_status=", _wrap_svn_client_status_t_repos_prop_status_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_prop_status", _wrap_svn_client_status_t_repos_prop_status_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_lock=", _wrap_svn_client_status_t_repos_lock_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "repos_lock", _wrap_svn_client_status_t_repos_lock_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_changed_rev=", _wrap_svn_client_status_t_ood_changed_rev_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_changed_rev", _wrap_svn_client_status_t_ood_changed_rev_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_changed_date=", _wrap_svn_client_status_t_ood_changed_date_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_changed_date", _wrap_svn_client_status_t_ood_changed_date_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_changed_author=", _wrap_svn_client_status_t_ood_changed_author_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "ood_changed_author", _wrap_svn_client_status_t_ood_changed_author_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "backwards_compatibility_baton=", _wrap_svn_client_status_t_backwards_compatibility_baton_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "backwards_compatibility_baton", _wrap_svn_client_status_t_backwards_compatibility_baton_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "moved_from_abspath=", _wrap_svn_client_status_t_moved_from_abspath_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "moved_from_abspath", _wrap_svn_client_status_t_moved_from_abspath_get, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "moved_to_abspath=", _wrap_svn_client_status_t_moved_to_abspath_set, -1);
  rb_define_method(SwigClassSvn_client_status_t.klass, "moved_to_abspath", _wrap_svn_client_status_t_moved_to_abspath_get, -1);
  SwigClassSvn_client_status_t.mark = 0;
  SwigClassSvn_client_status_t.destroy = (void (*)(void *)) free_svn_client_status_t;
  SwigClassSvn_client_status_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_client_status_dup", _wrap_svn_client_status_dup, -1);
  rb_define_module_function(mClient, "svn_client_status6", _wrap_svn_client_status6, -1);
  rb_define_module_function(mClient, "svn_client_status5", _wrap_svn_client_status5, -1);
  rb_define_module_function(mClient, "svn_client_status4", _wrap_svn_client_status4, -1);
  rb_define_module_function(mClient, "svn_client_status3", _wrap_svn_client_status3, -1);
  rb_define_module_function(mClient, "svn_client_status2", _wrap_svn_client_status2, -1);
  rb_define_module_function(mClient, "svn_client_status", _wrap_svn_client_status, -1);
  rb_define_module_function(mClient, "svn_client_log5", _wrap_svn_client_log5, -1);
  rb_define_module_function(mClient, "svn_client_log4", _wrap_svn_client_log4, -1);
  rb_define_module_function(mClient, "svn_client_log3", _wrap_svn_client_log3, -1);
  rb_define_module_function(mClient, "svn_client_log2", _wrap_svn_client_log2, -1);
  rb_define_module_function(mClient, "svn_client_log", _wrap_svn_client_log, -1);
  rb_define_module_function(mClient, "svn_client_blame5", _wrap_svn_client_blame5, -1);
  rb_define_module_function(mClient, "svn_client_blame4", _wrap_svn_client_blame4, -1);
  rb_define_module_function(mClient, "svn_client_blame3", _wrap_svn_client_blame3, -1);
  rb_define_module_function(mClient, "svn_client_blame2", _wrap_svn_client_blame2, -1);
  rb_define_module_function(mClient, "svn_client_blame", _wrap_svn_client_blame, -1);
  rb_define_module_function(mClient, "svn_client_diff6", _wrap_svn_client_diff6, -1);
  rb_define_module_function(mClient, "svn_client_diff5", _wrap_svn_client_diff5, -1);
  rb_define_module_function(mClient, "svn_client_diff4", _wrap_svn_client_diff4, -1);
  rb_define_module_function(mClient, "svn_client_diff3", _wrap_svn_client_diff3, -1);
  rb_define_module_function(mClient, "svn_client_diff2", _wrap_svn_client_diff2, -1);
  rb_define_module_function(mClient, "svn_client_diff", _wrap_svn_client_diff, -1);
  rb_define_module_function(mClient, "svn_client_diff_peg6", _wrap_svn_client_diff_peg6, -1);
  rb_define_module_function(mClient, "svn_client_diff_peg5", _wrap_svn_client_diff_peg5, -1);
  rb_define_module_function(mClient, "svn_client_diff_peg4", _wrap_svn_client_diff_peg4, -1);
  rb_define_module_function(mClient, "svn_client_diff_peg3", _wrap_svn_client_diff_peg3, -1);
  rb_define_module_function(mClient, "svn_client_diff_peg2", _wrap_svn_client_diff_peg2, -1);
  rb_define_module_function(mClient, "svn_client_diff_peg", _wrap_svn_client_diff_peg, -1);
  rb_define_module_function(mClient, "svn_client_diff_summarize2", _wrap_svn_client_diff_summarize2, -1);
  rb_define_module_function(mClient, "svn_client_diff_summarize", _wrap_svn_client_diff_summarize, -1);
  rb_define_module_function(mClient, "svn_client_diff_summarize_peg2", _wrap_svn_client_diff_summarize_peg2, -1);
  rb_define_module_function(mClient, "svn_client_diff_summarize_peg", _wrap_svn_client_diff_summarize_peg, -1);
  rb_define_module_function(mClient, "svn_client_get_merging_summary", _wrap_svn_client_get_merging_summary, -1);
  rb_define_module_function(mClient, "svn_client_merge5", _wrap_svn_client_merge5, -1);
  rb_define_module_function(mClient, "svn_client_merge4", _wrap_svn_client_merge4, -1);
  rb_define_module_function(mClient, "svn_client_merge3", _wrap_svn_client_merge3, -1);
  rb_define_module_function(mClient, "svn_client_merge2", _wrap_svn_client_merge2, -1);
  rb_define_module_function(mClient, "svn_client_merge", _wrap_svn_client_merge, -1);
  rb_define_module_function(mClient, "svn_client_merge_reintegrate", _wrap_svn_client_merge_reintegrate, -1);
  rb_define_module_function(mClient, "svn_client_merge_peg5", _wrap_svn_client_merge_peg5, -1);
  rb_define_module_function(mClient, "svn_client_merge_peg4", _wrap_svn_client_merge_peg4, -1);
  rb_define_module_function(mClient, "svn_client_merge_peg3", _wrap_svn_client_merge_peg3, -1);
  rb_define_module_function(mClient, "svn_client_merge_peg2", _wrap_svn_client_merge_peg2, -1);
  rb_define_module_function(mClient, "svn_client_merge_peg", _wrap_svn_client_merge_peg, -1);
  rb_define_module_function(mClient, "svn_client_suggest_merge_sources", _wrap_svn_client_suggest_merge_sources, -1);
  rb_define_module_function(mClient, "svn_client_mergeinfo_get_merged", _wrap_svn_client_mergeinfo_get_merged, -1);
  rb_define_module_function(mClient, "svn_client_mergeinfo_log2", _wrap_svn_client_mergeinfo_log2, -1);
  rb_define_module_function(mClient, "svn_client_mergeinfo_log", _wrap_svn_client_mergeinfo_log, -1);
  rb_define_module_function(mClient, "svn_client_mergeinfo_log_merged", _wrap_svn_client_mergeinfo_log_merged, -1);
  rb_define_module_function(mClient, "svn_client_mergeinfo_log_eligible", _wrap_svn_client_mergeinfo_log_eligible, -1);
  rb_define_module_function(mClient, "svn_client_vacuum", _wrap_svn_client_vacuum, -1);
  rb_define_module_function(mClient, "svn_client_cleanup2", _wrap_svn_client_cleanup2, -1);
  rb_define_module_function(mClient, "svn_client_cleanup", _wrap_svn_client_cleanup, -1);
  rb_define_module_function(mClient, "svn_client_upgrade", _wrap_svn_client_upgrade, -1);
  rb_define_module_function(mClient, "svn_client_relocate2", _wrap_svn_client_relocate2, -1);
  rb_define_module_function(mClient, "svn_client_relocate", _wrap_svn_client_relocate, -1);
  rb_define_module_function(mClient, "svn_client_revert3", _wrap_svn_client_revert3, -1);
  rb_define_module_function(mClient, "svn_client_revert2", _wrap_svn_client_revert2, -1);
  rb_define_module_function(mClient, "svn_client_revert", _wrap_svn_client_revert, -1);
  rb_define_module_function(mClient, "svn_client_resolved", _wrap_svn_client_resolved, -1);
  rb_define_module_function(mClient, "svn_client_resolve", _wrap_svn_client_resolve, -1);
  
  SwigClassSvn_client_copy_source_t.klass = rb_define_class_under(mClient, "Svn_client_copy_source_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_copy_source_t, (void *) &SwigClassSvn_client_copy_source_t);
  rb_define_alloc_func(SwigClassSvn_client_copy_source_t.klass, _wrap_svn_client_copy_source_t_allocate);
  rb_define_method(SwigClassSvn_client_copy_source_t.klass, "initialize", _wrap_new_svn_client_copy_source_t, -1);
  rb_define_method(SwigClassSvn_client_copy_source_t.klass, "path", _wrap_svn_client_copy_source_t_path, -1);
  rb_define_method(SwigClassSvn_client_copy_source_t.klass, "revision", _wrap_svn_client_copy_source_t_revision, -1);
  rb_define_method(SwigClassSvn_client_copy_source_t.klass, "peg_revision", _wrap_svn_client_copy_source_t_peg_revision, -1);
  SwigClassSvn_client_copy_source_t.mark = 0;
  SwigClassSvn_client_copy_source_t.destroy = (void (*)(void *)) free_svn_client_copy_source_t;
  SwigClassSvn_client_copy_source_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_client_copy7", _wrap_svn_client_copy7, -1);
  rb_define_module_function(mClient, "svn_client_copy6", _wrap_svn_client_copy6, -1);
  rb_define_module_function(mClient, "svn_client_copy5", _wrap_svn_client_copy5, -1);
  rb_define_module_function(mClient, "svn_client_copy4", _wrap_svn_client_copy4, -1);
  rb_define_module_function(mClient, "svn_client_copy3", _wrap_svn_client_copy3, -1);
  rb_define_module_function(mClient, "svn_client_copy2", _wrap_svn_client_copy2, -1);
  rb_define_module_function(mClient, "svn_client_copy", _wrap_svn_client_copy, -1);
  rb_define_module_function(mClient, "svn_client_move7", _wrap_svn_client_move7, -1);
  rb_define_module_function(mClient, "svn_client_move6", _wrap_svn_client_move6, -1);
  rb_define_module_function(mClient, "svn_client_move5", _wrap_svn_client_move5, -1);
  rb_define_module_function(mClient, "svn_client_move4", _wrap_svn_client_move4, -1);
  rb_define_module_function(mClient, "svn_client_move3", _wrap_svn_client_move3, -1);
  rb_define_module_function(mClient, "svn_client_move2", _wrap_svn_client_move2, -1);
  rb_define_module_function(mClient, "svn_client_move", _wrap_svn_client_move, -1);
  rb_define_module_function(mClient, "svn_client_propset_remote", _wrap_svn_client_propset_remote, -1);
  rb_define_module_function(mClient, "svn_client_propset_local", _wrap_svn_client_propset_local, -1);
  rb_define_module_function(mClient, "svn_client_propset3", _wrap_svn_client_propset3, -1);
  rb_define_module_function(mClient, "svn_client_propset2", _wrap_svn_client_propset2, -1);
  rb_define_module_function(mClient, "svn_client_propset", _wrap_svn_client_propset, -1);
  rb_define_module_function(mClient, "svn_client_revprop_set2", _wrap_svn_client_revprop_set2, -1);
  rb_define_module_function(mClient, "svn_client_revprop_set", _wrap_svn_client_revprop_set, -1);
  rb_define_module_function(mClient, "svn_client_propget5", _wrap_svn_client_propget5, -1);
  rb_define_module_function(mClient, "svn_client_propget4", _wrap_svn_client_propget4, -1);
  rb_define_module_function(mClient, "svn_client_propget3", _wrap_svn_client_propget3, -1);
  rb_define_module_function(mClient, "svn_client_propget2", _wrap_svn_client_propget2, -1);
  rb_define_module_function(mClient, "svn_client_propget", _wrap_svn_client_propget, -1);
  rb_define_module_function(mClient, "svn_client_revprop_get", _wrap_svn_client_revprop_get, -1);
  rb_define_module_function(mClient, "svn_client_proplist4", _wrap_svn_client_proplist4, -1);
  rb_define_module_function(mClient, "svn_client_proplist3", _wrap_svn_client_proplist3, -1);
  rb_define_module_function(mClient, "svn_client_proplist2", _wrap_svn_client_proplist2, -1);
  rb_define_module_function(mClient, "svn_client_proplist", _wrap_svn_client_proplist, -1);
  rb_define_module_function(mClient, "svn_client_revprop_list", _wrap_svn_client_revprop_list, -1);
  rb_define_module_function(mClient, "svn_client_export5", _wrap_svn_client_export5, -1);
  rb_define_module_function(mClient, "svn_client_export4", _wrap_svn_client_export4, -1);
  rb_define_module_function(mClient, "svn_client_export3", _wrap_svn_client_export3, -1);
  rb_define_module_function(mClient, "svn_client_export2", _wrap_svn_client_export2, -1);
  rb_define_module_function(mClient, "svn_client_export", _wrap_svn_client_export, -1);
  rb_define_module_function(mClient, "svn_client_list3", _wrap_svn_client_list3, -1);
  rb_define_module_function(mClient, "svn_client_list2", _wrap_svn_client_list2, -1);
  rb_define_module_function(mClient, "svn_client_list", _wrap_svn_client_list, -1);
  rb_define_module_function(mClient, "svn_client_ls3", _wrap_svn_client_ls3, -1);
  rb_define_module_function(mClient, "svn_client_ls2", _wrap_svn_client_ls2, -1);
  rb_define_module_function(mClient, "svn_client_ls", _wrap_svn_client_ls, -1);
  rb_define_module_function(mClient, "svn_client_cat3", _wrap_svn_client_cat3, -1);
  rb_define_module_function(mClient, "svn_client_cat2", _wrap_svn_client_cat2, -1);
  rb_define_module_function(mClient, "svn_client_cat", _wrap_svn_client_cat, -1);
  rb_define_module_function(mClient, "svn_client_add_to_changelist", _wrap_svn_client_add_to_changelist, -1);
  rb_define_module_function(mClient, "svn_client_remove_from_changelists", _wrap_svn_client_remove_from_changelists, -1);
  rb_define_module_function(mClient, "svn_client_get_changelists", _wrap_svn_client_get_changelists, -1);
  rb_define_module_function(mClient, "svn_client_lock", _wrap_svn_client_lock, -1);
  rb_define_module_function(mClient, "svn_client_unlock", _wrap_svn_client_unlock, -1);
  
  SwigClassSvn_info_t.klass = rb_define_class_under(mClient, "Svn_info_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_info_t, (void *) &SwigClassSvn_info_t);
  rb_define_alloc_func(SwigClassSvn_info_t.klass, _wrap_svn_info_t_allocate);
  rb_define_method(SwigClassSvn_info_t.klass, "initialize", _wrap_new_svn_info_t, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "URL=", _wrap_svn_info_t_URL_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "URL", _wrap_svn_info_t_URL_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "rev=", _wrap_svn_info_t_rev_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "rev", _wrap_svn_info_t_rev_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "kind=", _wrap_svn_info_t_kind_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "kind", _wrap_svn_info_t_kind_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "repos_root_URL=", _wrap_svn_info_t_repos_root_URL_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "repos_root_URL", _wrap_svn_info_t_repos_root_URL_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "repos_UUID=", _wrap_svn_info_t_repos_UUID_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "repos_UUID", _wrap_svn_info_t_repos_UUID_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "last_changed_rev=", _wrap_svn_info_t_last_changed_rev_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "last_changed_rev", _wrap_svn_info_t_last_changed_rev_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "last_changed_date=", _wrap_svn_info_t_last_changed_date_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "last_changed_date", _wrap_svn_info_t_last_changed_date_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "last_changed_author=", _wrap_svn_info_t_last_changed_author_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "last_changed_author", _wrap_svn_info_t_last_changed_author_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "lock=", _wrap_svn_info_t_lock_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "lock", _wrap_svn_info_t_lock_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "has_wc_info=", _wrap_svn_info_t_has_wc_info_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "has_wc_info", _wrap_svn_info_t_has_wc_info_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "schedule=", _wrap_svn_info_t_schedule_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "schedule", _wrap_svn_info_t_schedule_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "copyfrom_url=", _wrap_svn_info_t_copyfrom_url_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "copyfrom_url", _wrap_svn_info_t_copyfrom_url_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "copyfrom_rev=", _wrap_svn_info_t_copyfrom_rev_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "copyfrom_rev", _wrap_svn_info_t_copyfrom_rev_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "text_time=", _wrap_svn_info_t_text_time_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "text_time", _wrap_svn_info_t_text_time_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "prop_time=", _wrap_svn_info_t_prop_time_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "prop_time", _wrap_svn_info_t_prop_time_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "checksum=", _wrap_svn_info_t_checksum_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "checksum", _wrap_svn_info_t_checksum_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "conflict_old=", _wrap_svn_info_t_conflict_old_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "conflict_old", _wrap_svn_info_t_conflict_old_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "conflict_new=", _wrap_svn_info_t_conflict_new_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "conflict_new", _wrap_svn_info_t_conflict_new_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "conflict_wrk=", _wrap_svn_info_t_conflict_wrk_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "conflict_wrk", _wrap_svn_info_t_conflict_wrk_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "prejfile=", _wrap_svn_info_t_prejfile_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "prejfile", _wrap_svn_info_t_prejfile_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "changelist=", _wrap_svn_info_t_changelist_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "changelist", _wrap_svn_info_t_changelist_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "depth=", _wrap_svn_info_t_depth_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "depth", _wrap_svn_info_t_depth_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "working_size=", _wrap_svn_info_t_working_size_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "working_size", _wrap_svn_info_t_working_size_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "size=", _wrap_svn_info_t_size_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "size", _wrap_svn_info_t_size_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "size64=", _wrap_svn_info_t_size64_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "size64", _wrap_svn_info_t_size64_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "working_size64=", _wrap_svn_info_t_working_size64_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "working_size64", _wrap_svn_info_t_working_size64_get, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "tree_conflict=", _wrap_svn_info_t_tree_conflict_set, -1);
  rb_define_method(SwigClassSvn_info_t.klass, "tree_conflict", _wrap_svn_info_t_tree_conflict_get, -1);
  SwigClassSvn_info_t.mark = 0;
  SwigClassSvn_info_t.destroy = (void (*)(void *)) free_svn_info_t;
  SwigClassSvn_info_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_info_dup", _wrap_svn_info_dup, -1);
  
  SwigClassSvn_client_info2_t.klass = rb_define_class_under(mClient, "Svn_client_info2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_client_info2_t, (void *) &SwigClassSvn_client_info2_t);
  rb_define_alloc_func(SwigClassSvn_client_info2_t.klass, _wrap_svn_client_info2_t_allocate);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "initialize", _wrap_new_svn_client_info2_t, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "URL=", _wrap_svn_client_info2_t_URL_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "URL", _wrap_svn_client_info2_t_URL_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "rev=", _wrap_svn_client_info2_t_rev_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "rev", _wrap_svn_client_info2_t_rev_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "repos_root_URL=", _wrap_svn_client_info2_t_repos_root_URL_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "repos_root_URL", _wrap_svn_client_info2_t_repos_root_URL_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "repos_UUID=", _wrap_svn_client_info2_t_repos_UUID_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "repos_UUID", _wrap_svn_client_info2_t_repos_UUID_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "kind=", _wrap_svn_client_info2_t_kind_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "kind", _wrap_svn_client_info2_t_kind_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "size=", _wrap_svn_client_info2_t_size_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "size", _wrap_svn_client_info2_t_size_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "last_changed_rev=", _wrap_svn_client_info2_t_last_changed_rev_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "last_changed_rev", _wrap_svn_client_info2_t_last_changed_rev_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "last_changed_date=", _wrap_svn_client_info2_t_last_changed_date_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "last_changed_date", _wrap_svn_client_info2_t_last_changed_date_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "last_changed_author=", _wrap_svn_client_info2_t_last_changed_author_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "last_changed_author", _wrap_svn_client_info2_t_last_changed_author_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "lock=", _wrap_svn_client_info2_t_lock_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "lock", _wrap_svn_client_info2_t_lock_get, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "wc_info=", _wrap_svn_client_info2_t_wc_info_set, -1);
  rb_define_method(SwigClassSvn_client_info2_t.klass, "wc_info", _wrap_svn_client_info2_t_wc_info_get, -1);
  SwigClassSvn_client_info2_t.mark = 0;
  SwigClassSvn_client_info2_t.destroy = (void (*)(void *)) free_svn_client_info2_t;
  SwigClassSvn_client_info2_t.trackObjects = 0;
  rb_define_module_function(mClient, "svn_client_info2_dup", _wrap_svn_client_info2_dup, -1);
  rb_define_module_function(mClient, "svn_client_info4", _wrap_svn_client_info4, -1);
  rb_define_module_function(mClient, "svn_client_info3", _wrap_svn_client_info3, -1);
  rb_define_module_function(mClient, "svn_client_info2", _wrap_svn_client_info2, -1);
  rb_define_module_function(mClient, "svn_client_info", _wrap_svn_client_info, -1);
  rb_define_module_function(mClient, "svn_client_get_wc_root", _wrap_svn_client_get_wc_root, -1);
  rb_define_module_function(mClient, "svn_client_min_max_revisions", _wrap_svn_client_min_max_revisions, -1);
  rb_define_module_function(mClient, "svn_client_patch", _wrap_svn_client_patch, -1);
  rb_define_module_function(mClient, "svn_client_url_from_path2", _wrap_svn_client_url_from_path2, -1);
  rb_define_module_function(mClient, "svn_client_url_from_path", _wrap_svn_client_url_from_path, -1);
  rb_define_module_function(mClient, "svn_client_get_repos_root", _wrap_svn_client_get_repos_root, -1);
  rb_define_module_function(mClient, "svn_client_root_url_from_path", _wrap_svn_client_root_url_from_path, -1);
  rb_define_module_function(mClient, "svn_client_uuid_from_url", _wrap_svn_client_uuid_from_url, -1);
  rb_define_module_function(mClient, "svn_client_uuid_from_path2", _wrap_svn_client_uuid_from_path2, -1);
  rb_define_module_function(mClient, "svn_client_uuid_from_path", _wrap_svn_client_uuid_from_path, -1);
  rb_define_module_function(mClient, "svn_client_open_ra_session2", _wrap_svn_client_open_ra_session2, -1);
  rb_define_module_function(mClient, "svn_client_open_ra_session", _wrap_svn_client_open_ra_session, -1);
  rb_define_module_function(mClient, "svn_proplist_invoke_receiver2", _wrap_svn_proplist_invoke_receiver2, -1);
  rb_define_module_function(mClient, "svn_proplist_invoke_receiver", _wrap_svn_proplist_invoke_receiver, -1);
  rb_define_module_function(mClient, "svn_client_invoke_get_commit_log3", _wrap_svn_client_invoke_get_commit_log3, -1);
  rb_define_module_function(mClient, "svn_client_invoke_get_commit_log2", _wrap_svn_client_invoke_get_commit_log2, -1);
  rb_define_module_function(mClient, "svn_client_invoke_get_commit_log", _wrap_svn_client_invoke_get_commit_log, -1);
  rb_define_module_function(mClient, "svn_client_invoke_blame_receiver3", _wrap_svn_client_invoke_blame_receiver3, -1);
  rb_define_module_function(mClient, "svn_client_invoke_blame_receiver2", _wrap_svn_client_invoke_blame_receiver2, -1);
  rb_define_module_function(mClient, "svn_client_invoke_blame_receiver", _wrap_svn_client_invoke_blame_receiver, -1);
  rb_define_module_function(mClient, "svn_client_invoke_diff_summarize_func", _wrap_svn_client_invoke_diff_summarize_func, -1);
  rb_define_module_function(mClient, "svn_client_invoke_import_filter_func", _wrap_svn_client_invoke_import_filter_func, -1);
  rb_define_module_function(mClient, "svn_client_invoke_status_func", _wrap_svn_client_invoke_status_func, -1);
  rb_define_module_function(mClient, "svn_client_invoke_list_func2", _wrap_svn_client_invoke_list_func2, -1);
  rb_define_module_function(mClient, "svn_client_invoke_list_func", _wrap_svn_client_invoke_list_func, -1);
  rb_define_module_function(mClient, "svn_info_invoke_receiver", _wrap_svn_info_invoke_receiver, -1);
  rb_define_module_function(mClient, "svn_client_invoke_info_receiver2", _wrap_svn_client_invoke_info_receiver2, -1);
  rb_define_module_function(mClient, "svn_client_invoke_patch_func", _wrap_svn_client_invoke_patch_func, -1);
  rb_define_const(mClient, "SWIG_SVN_INFO_SIZE_UNKNOWN", SWIG_From_unsigned_SS_long((unsigned long)(-1)));
  rb_define_module_function(mClient, "svn_client_set_log_msg_func3", _wrap_svn_client_set_log_msg_func3, -1);
  rb_define_module_function(mClient, "svn_client_set_notify_func2", _wrap_svn_client_set_notify_func2, -1);
  rb_define_module_function(mClient, "svn_client_set_cancel_func", _wrap_svn_client_set_cancel_func, -1);
  rb_define_module_function(mClient, "svn_client_set_config", _wrap_svn_client_set_config, -1);
  rb_define_module_function(mClient, "svn_client_get_config", _wrap_svn_client_get_config, -1);
}

