/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.12
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */

#define SWIGPYTHON
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE
#define SWIG_PYTHON_CLASSIC

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#  ifndef GCC_HASCLASSVISIBILITY
#    define GCC_HASCLASSVISIBILITY
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif



#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */
# undef _DEBUG
# include <Python.h>
# define _DEBUG
#else
# include <Python.h>
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      register size_t l = 0;
      register size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	register size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  register int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      register size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  register const unsigned char *u = (unsigned char *) ptr;
  register const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    register unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  register unsigned char *u = (unsigned char *) ptr;
  register const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    register char d = *(c++);
    register unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = ((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = ((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define PyString_AS_STRING(x) PyUnicode_AS_STRING(x)
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

#ifndef Py_TYPE
#  define Py_TYPE(op) ((op)->ob_type)
#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif


/* Warning: This function will allocate a new string in Python 3,
 * so please call SWIG_Python_str_DelForPy3(x) to free the space.
 */
SWIGINTERN char*
SWIG_Python_str_AsChar(PyObject *str)
{
#if PY_VERSION_HEX >= 0x03000000
  char *cstr;
  char *newstr;
  Py_ssize_t len;
  str = PyUnicode_AsUTF8String(str);
  PyBytes_AsStringAndSize(str, &cstr, &len);
  newstr = (char *) malloc(len+1);
  memcpy(newstr, cstr, len+1);
  Py_XDECREF(str);
  return newstr;
#else
  return PyString_AsString(str);
#endif
}

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_DelForPy3(x) free( (void*) (x) )
#else
#  define SWIG_Python_str_DelForPy3(x) 
#endif


SWIGINTERN PyObject*
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

/* Add PyOS_snprintf for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# if defined(_MSC_VER) || defined(__BORLANDC__) || defined(_WATCOM)
#  define PyOS_snprintf _snprintf
# else
#  define PyOS_snprintf snprintf
# endif
#endif

/* A crude PyString_FromFormat implementation for old Pythons */
#if PY_VERSION_HEX < 0x02020000

#ifndef SWIG_PYBUFFER_SIZE
# define SWIG_PYBUFFER_SIZE 1024
#endif

static PyObject *
PyString_FromFormat(const char *fmt, ...) {
  va_list ap;
  char buf[SWIG_PYBUFFER_SIZE * 2];
  int res;
  va_start(ap, fmt);
  res = vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap);
  return (res < 0 || res >= (int)sizeof(buf)) ? 0 : PyString_FromString(buf);
}
#endif

/* Add PyObject_Del for old Pythons */
#if PY_VERSION_HEX < 0x01060000
# define PyObject_Del(op) PyMem_DEL((op))
#endif
#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

/* A crude PyExc_StopIteration exception for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# ifndef PyExc_StopIteration
#  define PyExc_StopIteration PyExc_RuntimeError
# endif
# ifndef PyObject_GenericGetAttr
#  define PyObject_GenericGetAttr 0
# endif
#endif

/* Py_NotImplemented is defined in 2.1 and up. */
#if PY_VERSION_HEX < 0x02010000
# ifndef Py_NotImplemented
#  define Py_NotImplemented PyExc_RuntimeError
# endif
#endif

/* A crude PyString_AsStringAndSize implementation for old Pythons */
#if PY_VERSION_HEX < 0x02010000
# ifndef PyString_AsStringAndSize
#  define PyString_AsStringAndSize(obj, s, len) {*s = PyString_AsString(obj); *len = *s ? strlen(*s) : 0;}
# endif
#endif

/* PySequence_Size for old Pythons */
#if PY_VERSION_HEX < 0x02000000
# ifndef PySequence_Size
#  define PySequence_Size PySequence_Length
# endif
#endif

/* PyBool_FromLong for old Pythons */
#if PY_VERSION_HEX < 0x02030000
static
PyObject *PyBool_FromLong(long ok)
{
  PyObject *result = ok ? Py_True : Py_False;
  Py_INCREF(result);
  return result;
}
#endif

/* Py_ssize_t for old Pythons */
/* This code is as recommended by: */
/* http://www.python.org/dev/peps/pep-0353/#conversion-guidelines */
#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
typedef int Py_ssize_t;
# define PY_SSIZE_T_MAX INT_MAX
# define PY_SSIZE_T_MIN INT_MIN
typedef inquiry lenfunc;
typedef intargfunc ssizeargfunc;
typedef intintargfunc ssizessizeargfunc;
typedef intobjargproc ssizeobjargproc;
typedef intintobjargproc ssizessizeobjargproc;
typedef getreadbufferproc readbufferproc;
typedef getwritebufferproc writebufferproc;
typedef getsegcountproc segcountproc;
typedef getcharbufferproc charbufferproc;
static long PyNumber_AsSsize_t (PyObject *x, void *SWIGUNUSEDPARM(exc))
{
  long result = 0;
  PyObject *i = PyNumber_Int(x);
  if (i) {
    result = PyInt_AsLong(i);
    Py_DECREF(i);
  }
  return result;
}
#endif

#if PY_VERSION_HEX < 0x02050000
#define PyInt_FromSize_t(x) PyInt_FromLong((long)x)
#endif

#if PY_VERSION_HEX < 0x02040000
#define Py_VISIT(op)				\
  do { 						\
    if (op) {					\
      int vret = visit((op), arg);		\
      if (vret)					\
        return vret;				\
    }						\
  } while (0)
#endif

#if PY_VERSION_HEX < 0x02030000
typedef struct {
  PyTypeObject type;
  PyNumberMethods as_number;
  PyMappingMethods as_mapping;
  PySequenceMethods as_sequence;
  PyBufferProcs as_buffer;
  PyObject *name, *slots;
} PyHeapTypeObject;
#endif

#if PY_VERSION_HEX < 0x02030000
typedef destructor freefunc;
#endif

#if ((PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION > 6) || \
     (PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION > 0) || \
     (PY_MAJOR_VERSION > 3))
# define SWIGPY_USE_CAPSULE
# define SWIGPY_CAPSULE_NAME ((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION ".type_pointer_capsule" SWIG_TYPE_TABLE_NAME)
#endif

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred()) PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    char *tmp;
    PyObject *old_str = PyObject_Str(value);
    PyErr_Clear();
    Py_XINCREF(type);

    PyErr_Format(type, "%s %s", tmp = SWIG_Python_str_AsChar(old_str), mesg);
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    if (PY_VERSION_HEX >= 0x02030000) /* For 2.3 or later, use the PyGILState calls */
#      define SWIG_PYTHON_USE_GIL
#    endif
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    ifndef SWIG_PYTHON_INITIALIZE_THREADS
#     define SWIG_PYTHON_INITIALIZE_THREADS  PyEval_InitThreads() 
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { PyEval_RestoreThread(save); status = false; }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;


/* -----------------------------------------------------------------------------
 * Wrapper of PyInstanceMethod_New() used in Python 3
 * It is exported to the generated module, used for -fastproxy
 * ----------------------------------------------------------------------------- */
#if PY_VERSION_HEX >= 0x03000000
SWIGRUNTIME PyObject* SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func)
{
  return PyInstanceMethod_New(func);
}
#else
SWIGRUNTIME PyObject* SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *SWIGUNUSEDPARM(func))
{
  return NULL;
}
#endif

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
#if PY_VERSION_HEX < 0x02030000
  PyDict_SetItemString(d, (char *)name, obj);
#else
  PyDict_SetItemString(d, name, obj);
#endif
  Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
#if PY_VERSION_HEX < 0x02030000
  PyDict_SetItemString(d, (char *)name, obj);
#else
  PyDict_SetItemString(d, name, obj);
#endif
  Py_DECREF(obj);                            
}

#endif

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
#if !defined(SWIG_PYTHON_OUTPUT_TUPLE)
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      PyList_SetItem(result, 0, o2);
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
#else
  PyObject*   o2;
  PyObject*   o3;
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyTuple_Check(result)) {
      o2 = result;
      result = PyTuple_New(1);
      PyTuple_SET_ITEM(result, 0, o2);
    }
    o3 = PyTuple_New(1);
    PyTuple_SET_ITEM(o3, 0, obj);
    o2 = result;
    result = PySequence_Concat(o2, o3);
    Py_DECREF(o2);
    Py_DECREF(o3);
  }
  return result;
#endif
}

/* Unpack the argument tuple */

SWIGINTERN int
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      register int i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    register Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      register int i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

/* A functor is a function object with one single object argument */
#if PY_VERSION_HEX >= 0x02020000
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);
#else
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunction(functor, "O", obj);
#endif

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/*  How to access Py_None */
#if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#  ifndef SWIG_PYTHON_NO_BUILD_NONE
#    ifndef SWIG_PYTHON_BUILD_NONE
#      define SWIG_PYTHON_BUILD_NONE
#    endif
#  endif
#endif

#ifdef SWIG_PYTHON_BUILD_NONE
#  ifdef Py_None
#   undef Py_None
#   define Py_None SWIG_Py_None()
#  endif
SWIGRUNTIMEINLINE PyObject * 
_SWIG_Py_None(void)
{
  PyObject *none = Py_BuildValue((char*)"");
  Py_DECREF(none);
  return none;
}
SWIGRUNTIME PyObject * 
SWIG_Py_None(void)
{
  static PyObject *SWIG_STATIC_POINTER(none) = _SWIG_Py_None();
  return none;
}
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  return data ? data->implicitconv : 0;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      data->newargs = obj;
      Py_INCREF(obj);
    } else {
#if (PY_VERSION_HEX < 0x02020000)
      data->newraw = 0;
#else
      data->newraw = PyObject_GetAttrString(data->klass, (char *)"__new__");
#endif
      if (data->newraw) {
	Py_INCREF(data->newraw);
	data->newargs = PyTuple_New(1);
	PyTuple_SetItem(data->newargs, 0, obj);
      } else {
	data->newargs = obj;
      }
      Py_INCREF(data->newargs);
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, (char *)"__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      int flags;
      Py_INCREF(data->destroy);
      flags = PyCFunction_GET_FLAGS(data->destroy);
#ifdef METH_O
      data->delargs = !(flags & (METH_O));
#else
      data->delargs = 0;
#endif
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data) {
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
#ifdef SWIGPYTHON_BUILTIN
  PyObject *dict;
#endif
} SwigPyObject;

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    if (PyTuple_SetItem(args, 0, SwigPyObject_long(v)) == 0) {
      PyObject *ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
	res = PyUnicode_Format(ofmt,args);
#else
	res = PyString_Format(ofmt,args);
#endif
	Py_DECREF(ofmt);
      }
      Py_DECREF(args);
    }
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
#ifdef METH_NOARGS
SwigPyObject_repr(SwigPyObject *v)
#else
SwigPyObject_repr(SwigPyObject *v, PyObject *args)
#endif
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (v->next) {
# ifdef METH_NOARGS
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
# else
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next, args);
# endif
# if PY_VERSION_HEX >= 0x03000000
    PyObject *joined = PyUnicode_Concat(repr, nrep);
    Py_DecRef(repr);
    Py_DecRef(nrep);
    repr = joined;
# else
    PyString_ConcatAndDel(&repr,nrep);
# endif
  }
  return repr;  
}

SWIGRUNTIME int
SwigPyObject_compare(SwigPyObject *v, SwigPyObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(SwigPyObject *v, SwigPyObject *w, int op)
{
  PyObject* res;
  if( op != Py_EQ && op != Py_NE ) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  return res;  
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
#ifdef SWIGPYTHON_BUILTIN
  PyTypeObject *target_tp = SwigPyObject_type();
  if (PyType_IsSubtype(op->ob_type, target_tp))
    return 1;
  return (strcmp(op->ob_type->tp_name, "SwigPyObject") == 0);
#else
  return (Py_TYPE(op) == SwigPyObject_type())
    || (strcmp(Py_TYPE(op)->tp_name,"SwigPyObject") == 0);
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;
      if (data->delargs) {
	/* we need to create a temporary object to carry the destroy operation */
	PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
	res = SWIG_Python_CallFunctor(destroy, tmp);
	Py_DECREF(tmp);
      } else {
	PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
	PyObject *mself = PyCFunction_GET_SELF(destroy);
	res = ((*meth)(mself, v));
      }
      Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
  } 
  Py_XDECREF(next);
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
#ifndef METH_O
  PyObject *tmp = 0;
  if (!PyArg_ParseTuple(next,(char *)"O:append", &tmp)) return NULL;
  next = tmp;
#endif
  if (!SwigPyObject_Check(next)) {
    return NULL;
  }
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
#ifdef METH_NOARGS
SwigPyObject_next(PyObject* v)
#else
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
SwigPyObject_disown(PyObject *v)
#else
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
SwigPyObject_acquire(PyObject *v)
#else
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
#if (PY_VERSION_HEX < 0x02020000)
  if (!PyArg_ParseTuple(args,(char *)"|O:own",&val))
#elif (PY_VERSION_HEX < 0x02050000)
  if (!PyArg_UnpackTuple(args, (char *)"own", 0, 1, &val)) 
#else
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) 
#endif
    {
      return NULL;
    } 
  else
    {
      SwigPyObject *sobj = (SwigPyObject *)v;
      PyObject *obj = PyBool_FromLong(sobj->own);
      if (val) {
#ifdef METH_NOARGS
	if (PyObject_IsTrue(val)) {
	  SwigPyObject_acquire(v);
	} else {
	  SwigPyObject_disown(v);
	}
#else
	if (PyObject_IsTrue(val)) {
	  SwigPyObject_acquire(v,args);
	} else {
	  SwigPyObject_disown(v,args);
	}
#endif
      } 
      return obj;
    }
}

#ifdef METH_O
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)SwigPyObject_disown,  METH_NOARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)SwigPyObject_acquire, METH_NOARGS,  (char *)"acquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)SwigPyObject_own,     METH_VARARGS, (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)SwigPyObject_append,  METH_O,       (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)SwigPyObject_next,    METH_NOARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)SwigPyObject_repr,    METH_NOARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#else
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)SwigPyObject_disown,  METH_VARARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)SwigPyObject_acquire, METH_VARARGS,  (char *)"aquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)SwigPyObject_own,     METH_VARARGS,  (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)SwigPyObject_append,  METH_VARARGS,  (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)SwigPyObject_next,    METH_VARARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)SwigPyObject_repr,   METH_VARARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#endif

#if PY_VERSION_HEX < 0x02020000
SWIGINTERN PyObject *
SwigPyObject_getattr(SwigPyObject *sobj,char *name)
{
  return Py_FindMethod(swigobject_methods, (PyObject *)sobj, name);
}
#endif

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";

  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#elif PY_VERSION_HEX >= 0x02050000 /* 2.5.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#elif PY_VERSION_HEX >= 0x02020000 /* 2.2.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_true_divide */
#elif PY_VERSION_HEX >= 0x02000000 /* 2.0.0 */
    0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_or */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
      /* PyObject header changed in Python 3 */
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      (char *)"SwigPyObject",               /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
      0,				    /* tp_print */
#if PY_VERSION_HEX < 0x02020000
      (getattrfunc)SwigPyObject_getattr,    /* tp_getattr */
#else
      (getattrfunc)0,                       /* tp_getattr */
#endif
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,				    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigobject_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      swigobject_methods,                   /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
      0,                                    /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
      0,                                    /* tp_version */
#endif
#ifdef COUNT_ALLOCS
      0,0,0,0                               /* tp_alloc -> tp_next */
#endif
    };
    swigpyobject_type = tmp;
    type_init = 1;
#if PY_VERSION_HEX < 0x02020000
    swigpyobject_type.ob_type = &PyType_Type;
#else
    if (PyType_Ready(&swigpyobject_type) < 0)
      return NULL;
#endif
  }
  return &swigpyobject_type;
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_NEW(SwigPyObject, SwigPyObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME int
SwigPyPacked_print(SwigPyPacked *v, FILE *fp, int SWIGUNUSEDPARM(flags))
{
  char result[SWIG_BUFFER_SIZE];
  fputs("<Swig Packed ", fp); 
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    fputs("at ", fp); 
    fputs(result, fp); 
  }
  fputs(v->ty->name,fp); 
  fputs(">", fp);
  return 0; 
}
  
SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((char *)v->pack, (char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  return ((op)->ob_type == SwigPyPacked_TypeOnce()) 
    || (strcmp((op)->ob_type->tp_name,"SwigPyPacked") == 0);
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
      /* PyObject header changed in Python 3 */
#if PY_VERSION_HEX>=0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      (char *)"SwigPyPacked",               /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
      (printfunc)SwigPyPacked_print,        /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX>=0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigpacked_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
      0,                                    /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
      0,                                    /* tp_version */
#endif
#ifdef COUNT_ALLOCS
      0,0,0,0                               /* tp_alloc -> tp_next */
#endif
    };
    swigpypacked_type = tmp;
    type_init = 1;
#if PY_VERSION_HEX < 0x02020000
    swigpypacked_type.ob_type = &PyType_Type;
#else
    if (PyType_Ready(&swigpypacked_type) < 0)
      return NULL;
#endif
  }
  return &swigpypacked_type;
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_NEW(SwigPyPacked, SwigPyPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIMEINLINE PyObject *
_SWIG_This(void)
{
    return SWIG_Python_str_FromChar("this");
}

static PyObject *swig_this = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (swig_this == NULL)
    swig_this = _SWIG_This();
  return swig_this;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX>=0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
# ifdef PyWeakref_CheckProxy
  if (PyWeakref_CheckProxy(pyobj)) {
    pyobj = PyWeakref_GET_OBJECT(pyobj);
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
# endif
  return NULL;
#else

  obj = 0;

#if (!defined(SWIG_PYTHON_SLOW_GETSET_THIS) && (PY_VERSION_HEX >= 0x02030000))
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
#ifdef PyWeakref_CheckProxy
      if (PyWeakref_CheckProxy(pyobj)) {
	PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
#endif
      obj = PyObject_GetAttr(pyobj,SWIG_This());
      if (obj) {
	Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj,SWIG_This());
  if (obj) {
    Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (own)
      *own = *own | sobj->own;
    if (flags & SWIG_POINTER_DISOWN) {
      sobj->own = 0;
    }
    res = SWIG_OK;
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            Py_DECREF(impconv);
          }
        }
      }
    }
    if (!SWIG_IsOK(res) && obj == Py_None) {
      if (ptr)
        *ptr = 0;
      if (PyErr_Occurred())
        PyErr_Clear();
      res = SWIG_OK;
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    
    /* here we get the method pointer for callbacks */
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
    if (!desc) 
      return SWIG_ERROR;
    if (ty) {
      swig_cast_info *tc = SWIG_TypeCheck(desc,ty);
      if (tc) {
        int newmemory = 0;
        *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
        assert(!newmemory); /* newmemory handling not yet implemented */
      } else {
        return SWIG_ERROR;
      }
    } else {
      *ptr = vptr;
    }
    return SWIG_OK;
  }
}

/* Convert a packed value value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
#if (PY_VERSION_HEX >= 0x02020000)
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	if (dict == NULL) {
	  dict = PyDict_New();
	  *dictptr = dict;
	  PyDict_SetItem(dict, SWIG_This(), swig_this);
	}
      }
#else
      PyObject *key = SWIG_This();
      PyObject_SetAttr(inst, key, swig_this);
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    inst = PyBaseObject_Type.tp_new((PyTypeObject*) data->newargs, Py_None, Py_None);
    if (inst) {
      PyObject_SetAttr(inst, SWIG_This(), swig_this);
      Py_TYPE(inst)->tp_flags &= ~Py_TPFLAGS_VALID_VERSION_TAG;
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      Py_DECREF(dict);
    }
#endif
  }
  return inst;
#else
#if (PY_VERSION_HEX >= 0x02010000)
  PyObject *inst = 0;
  PyObject *dict = PyDict_New();
  if (dict) {
    PyDict_SetItem(dict, SWIG_This(), swig_this);
    inst = PyInstance_NewRaw(data->newargs, dict);
    Py_DECREF(dict);
  }
  return (PyObject *) inst;
#else
  PyInstanceObject *inst = PyObject_NEW(PyInstanceObject, &PyInstance_Type);
  if (inst == NULL) {
    return NULL;
  }
  inst->in_class = (PyClassObject *)data->newargs;
  Py_INCREF(inst->in_class);
  inst->in_dict = PyDict_New();
  if (inst->in_dict == NULL) {
    Py_DECREF(inst);
    return NULL;
  }
#ifdef Py_TPFLAGS_HAVE_WEAKREFS
  inst->in_weakreflist = NULL;
#endif
#ifdef Py_TPFLAGS_GC
  PyObject_GC_Init(inst);
#endif
  PyDict_SetItem(inst->in_dict, SWIG_This(), swig_this);
  return (PyObject *) inst;
#endif
#endif
}

SWIGRUNTIME void
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
 PyObject *dict;
#if (PY_VERSION_HEX >= 0x02020000) && !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
 PyObject **dictptr = _PyObject_GetDictPtr(inst);
 if (dictptr != NULL) {
   dict = *dictptr;
   if (dict == NULL) {
     dict = PyDict_New();
     *dictptr = dict;
   }
   PyDict_SetItem(dict, SWIG_This(), swig_this);
   return;
 }
#endif
 dict = PyObject_GetAttrString(inst, (char*)"__dict__");
 PyDict_SetItem(dict, SWIG_This(), swig_this);
 Py_DECREF(dict);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SwigPyObject_append((PyObject*) sthis, obj[1]);
    } else {
      SWIG_Python_SetSwigThis(obj[0], obj[1]);
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
        PyObject *next_self = clientdata->pytype->tp_alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
#ifdef SWIGPYTHON_BUILTIN
      newobj->dict = 0;
#endif
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
#ifdef SWIG_LINK_RUNTIME
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
#else
# ifdef SWIGPY_USE_CAPSULE
    type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
# else
    type_pointer = PyCObject_Import((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION,
				    (char*)"type_pointer" SWIG_TYPE_TABLE_NAME);
# endif
    if (PyErr_Occurred()) {
      PyErr_Clear();
      type_pointer = (void *)0;
    }
#endif
  }
  return (swig_module_info *) type_pointer;
}

#if PY_MAJOR_VERSION < 2
/* PyModule_AddObject function was introduced in Python 2.0.  The following function
   is copied out of Python/modsupport.c in python version 2.3.4 */
SWIGINTERN int
PyModule_AddObject(PyObject *m, char *name, PyObject *o)
{
  PyObject *dict;
  if (!PyModule_Check(m)) {
    PyErr_SetString(PyExc_TypeError,
		    "PyModule_AddObject() needs module as first arg");
    return SWIG_ERROR;
  }
  if (!o) {
    PyErr_SetString(PyExc_TypeError,
		    "PyModule_AddObject() needs non-NULL value");
    return SWIG_ERROR;
  }
  
  dict = PyModule_GetDict(m);
  if (dict == NULL) {
    /* Internal error -- modules must have a dict! */
    PyErr_Format(PyExc_SystemError, "module '%s' has no __dict__",
		 PyModule_GetName(m));
    return SWIG_ERROR;
  }
  if (PyDict_SetItemString(dict, name, o))
    return SWIG_ERROR;
  Py_DECREF(o);
  return SWIG_OK;
}
#endif

SWIGRUNTIME void
#ifdef SWIGPY_USE_CAPSULE
SWIG_Python_DestroyModule(PyObject *obj)
#else
SWIG_Python_DestroyModule(void *vptr)
#endif
{
#ifdef SWIGPY_USE_CAPSULE
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
#else
  swig_module_info *swig_module = (swig_module_info *) vptr;
#endif
  swig_type_info **types = swig_module->types;
  size_t i;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      if (data) SwigPyClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
  swig_this = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
#if PY_VERSION_HEX >= 0x03000000
 /* Add a dummy module object into sys.modules */
  PyObject *module = PyImport_AddModule((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION);
#else
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
  PyObject *module = Py_InitModule((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION, swig_empty_runtime_method_table);
#endif
#ifdef SWIGPY_USE_CAPSULE
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, (char*)"type_pointer_capsule" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
#else
  PyObject *pointer = PyCObject_FromVoidPtr((void *) swig_module, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, (char*)"type_pointer" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
#endif
}

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  static PyObject *SWIG_STATIC_POINTER(cache) = PyDict_New();
  return cache;
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = SWIG_Python_str_FromChar(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
#ifdef SWIGPY_USE_CAPSULE
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
#else
    descriptor = (swig_type_info *) PyCObject_AsVoidPtr(obj);
#endif
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
#ifdef SWIGPY_USE_CAPSULE
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
#else
      obj = PyCObject_FromVoidPtr(descriptor, NULL);
#endif
      PyDict_SetItem(cache, key, obj);
      Py_DECREF(obj);
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      char *tmp;
      PyObject *old_str = PyObject_Str(value);
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, tmp = SWIG_Python_str_AsChar(old_str));
      } else {
	PyErr_Format(type, "%s %s", tmp = SWIG_Python_str_AsChar(old_str), mesg);
      }
      SWIG_Python_str_DelForPy3(tmp);
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && SwigPyObject_Check(obj)) {
      const char *otype = (const char *) SwigPyObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'SwigPyObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	const char *cstr = str ? SWIG_Python_str_AsChar(str) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
          SWIG_Python_str_DelForPy3(cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(str);
	return;
      }
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
#if SWIG_POINTER_EXCEPTION
    if (flags) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(ty), obj);
      SWIG_Python_ArgFail(argnum);
    }
#endif
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = obj->ob_type;
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%.200s'", name->ob_type->tp_name);
    return -1;
  } else {
    Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) < 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = descr->ob_type->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
    }
    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '%.200s'", tp->tp_name, PyString_AsString(encoded_name));
    Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_apr_array_header_t swig_types[0]
#define SWIGTYPE_p_apr_file_t swig_types[1]
#define SWIGTYPE_p_apr_hash_t swig_types[2]
#define SWIGTYPE_p_apr_int32_t swig_types[3]
#define SWIGTYPE_p_apr_int64_t swig_types[4]
#define SWIGTYPE_p_apr_off_t swig_types[5]
#define SWIGTYPE_p_apr_pool_t swig_types[6]
#define SWIGTYPE_p_char swig_types[7]
#define SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[8]
#define SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[9]
#define SWIGTYPE_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[10]
#define SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[11]
#define SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[12]
#define SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[13]
#define SWIGTYPE_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[14]
#define SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[15]
#define SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[16]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t swig_types[17]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[18]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t swig_types[19]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t swig_types[20]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t swig_types[21]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t swig_types[22]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t swig_types[23]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[24]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t swig_types[25]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t swig_types[26]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t swig_types[27]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[28]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[29]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t swig_types[30]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[31]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t swig_types[32]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[33]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[34]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[35]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[36]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[37]
#define SWIGTYPE_p_f_p_void__p_svn_error_t swig_types[38]
#define SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[39]
#define SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void swig_types[40]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t swig_types[41]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[42]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[43]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[44]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t swig_types[45]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t swig_types[46]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void swig_types[47]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void swig_types[48]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t swig_types[49]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[50]
#define SWIGTYPE_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void swig_types[51]
#define SWIGTYPE_p_int swig_types[52]
#define SWIGTYPE_p_long swig_types[53]
#define SWIGTYPE_p_p_apr_array_header_t swig_types[54]
#define SWIGTYPE_p_p_apr_file_t swig_types[55]
#define SWIGTYPE_p_p_apr_hash_t swig_types[56]
#define SWIGTYPE_p_p_char swig_types[57]
#define SWIGTYPE_p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[58]
#define SWIGTYPE_p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[59]
#define SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[60]
#define SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[61]
#define SWIGTYPE_p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[62]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void swig_types[63]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t swig_types[64]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[65]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[66]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[67]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t swig_types[68]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t swig_types[69]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void swig_types[70]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void swig_types[71]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t swig_types[72]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[73]
#define SWIGTYPE_p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void swig_types[74]
#define SWIGTYPE_p_p_svn_checksum_t swig_types[75]
#define SWIGTYPE_p_p_svn_delta_editor_t swig_types[76]
#define SWIGTYPE_p_p_svn_stream_t swig_types[77]
#define SWIGTYPE_p_p_svn_string_t swig_types[78]
#define SWIGTYPE_p_p_svn_wc_adm_access_t swig_types[79]
#define SWIGTYPE_p_p_svn_wc_committed_queue_t swig_types[80]
#define SWIGTYPE_p_p_svn_wc_conflict_result_t swig_types[81]
#define SWIGTYPE_p_p_svn_wc_context_t swig_types[82]
#define SWIGTYPE_p_p_svn_wc_entry_t swig_types[83]
#define SWIGTYPE_p_p_svn_wc_external_item2_t swig_types[84]
#define SWIGTYPE_p_p_svn_wc_revision_status_t swig_types[85]
#define SWIGTYPE_p_p_svn_wc_status2_t swig_types[86]
#define SWIGTYPE_p_p_svn_wc_status3_t swig_types[87]
#define SWIGTYPE_p_p_svn_wc_status_t swig_types[88]
#define SWIGTYPE_p_p_void swig_types[89]
#define SWIGTYPE_p_svn_auth_baton_t swig_types[90]
#define SWIGTYPE_p_svn_auth_cred_simple_t swig_types[91]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t swig_types[92]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t swig_types[93]
#define SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t swig_types[94]
#define SWIGTYPE_p_svn_auth_cred_username_t swig_types[95]
#define SWIGTYPE_p_svn_auth_iterstate_t swig_types[96]
#define SWIGTYPE_p_svn_auth_provider_object_t swig_types[97]
#define SWIGTYPE_p_svn_auth_provider_t swig_types[98]
#define SWIGTYPE_p_svn_auth_ssl_server_cert_info_t swig_types[99]
#define SWIGTYPE_p_svn_checksum_ctx_t swig_types[100]
#define SWIGTYPE_p_svn_checksum_kind_t swig_types[101]
#define SWIGTYPE_p_svn_checksum_t swig_types[102]
#define SWIGTYPE_p_svn_commit_info_t swig_types[103]
#define SWIGTYPE_p_svn_config_t swig_types[104]
#define SWIGTYPE_p_svn_delta_editor_t swig_types[105]
#define SWIGTYPE_p_svn_depth_t swig_types[106]
#define SWIGTYPE_p_svn_diff_hunk_t swig_types[107]
#define SWIGTYPE_p_svn_dirent_t swig_types[108]
#define SWIGTYPE_p_svn_errno_t swig_types[109]
#define SWIGTYPE_p_svn_error_t swig_types[110]
#define SWIGTYPE_p_svn_fs_lock_target_t swig_types[111]
#define SWIGTYPE_p_svn_io_dirent2_t swig_types[112]
#define SWIGTYPE_p_svn_io_dirent_t swig_types[113]
#define SWIGTYPE_p_svn_io_file_del_t swig_types[114]
#define SWIGTYPE_p_svn_location_segment_t swig_types[115]
#define SWIGTYPE_p_svn_lock_t swig_types[116]
#define SWIGTYPE_p_svn_log_changed_path2_t swig_types[117]
#define SWIGTYPE_p_svn_log_changed_path_t swig_types[118]
#define SWIGTYPE_p_svn_log_entry_t swig_types[119]
#define SWIGTYPE_p_svn_merge_range_t swig_types[120]
#define SWIGTYPE_p_svn_mergeinfo_inheritance_t swig_types[121]
#define SWIGTYPE_p_svn_node_kind_t swig_types[122]
#define SWIGTYPE_p_svn_opt_revision_range_t swig_types[123]
#define SWIGTYPE_p_svn_opt_revision_t swig_types[124]
#define SWIGTYPE_p_svn_opt_revision_value_t swig_types[125]
#define SWIGTYPE_p_svn_opt_subcommand_desc2_t swig_types[126]
#define SWIGTYPE_p_svn_opt_subcommand_desc_t swig_types[127]
#define SWIGTYPE_p_svn_prop_inherited_item_t swig_types[128]
#define SWIGTYPE_p_svn_prop_kind swig_types[129]
#define SWIGTYPE_p_svn_ra_callbacks2_t swig_types[130]
#define SWIGTYPE_p_svn_ra_callbacks_t swig_types[131]
#define SWIGTYPE_p_svn_ra_plugin_t swig_types[132]
#define SWIGTYPE_p_svn_ra_reporter2_t swig_types[133]
#define SWIGTYPE_p_svn_ra_reporter3_t swig_types[134]
#define SWIGTYPE_p_svn_ra_reporter_t swig_types[135]
#define SWIGTYPE_p_svn_ra_session_t swig_types[136]
#define SWIGTYPE_p_svn_stream_mark_t swig_types[137]
#define SWIGTYPE_p_svn_stream_t swig_types[138]
#define SWIGTYPE_p_svn_string_t swig_types[139]
#define SWIGTYPE_p_svn_stringbuf_t swig_types[140]
#define SWIGTYPE_p_svn_tristate_t swig_types[141]
#define SWIGTYPE_p_svn_txdelta_op_t swig_types[142]
#define SWIGTYPE_p_svn_txdelta_stream_t swig_types[143]
#define SWIGTYPE_p_svn_txdelta_window_t swig_types[144]
#define SWIGTYPE_p_svn_version_checklist_t swig_types[145]
#define SWIGTYPE_p_svn_version_ext_linked_lib_t swig_types[146]
#define SWIGTYPE_p_svn_version_ext_loaded_lib_t swig_types[147]
#define SWIGTYPE_p_svn_version_extended_t swig_types[148]
#define SWIGTYPE_p_svn_version_t swig_types[149]
#define SWIGTYPE_p_svn_wc_adm_access_t swig_types[150]
#define SWIGTYPE_p_svn_wc_committed_queue_t swig_types[151]
#define SWIGTYPE_p_svn_wc_conflict_action_t swig_types[152]
#define SWIGTYPE_p_svn_wc_conflict_choice_t swig_types[153]
#define SWIGTYPE_p_svn_wc_conflict_description2_t swig_types[154]
#define SWIGTYPE_p_svn_wc_conflict_description_t swig_types[155]
#define SWIGTYPE_p_svn_wc_conflict_kind_t swig_types[156]
#define SWIGTYPE_p_svn_wc_conflict_reason_t swig_types[157]
#define SWIGTYPE_p_svn_wc_conflict_result_t swig_types[158]
#define SWIGTYPE_p_svn_wc_conflict_version_t swig_types[159]
#define SWIGTYPE_p_svn_wc_context_t swig_types[160]
#define SWIGTYPE_p_svn_wc_diff_callbacks2_t swig_types[161]
#define SWIGTYPE_p_svn_wc_diff_callbacks3_t swig_types[162]
#define SWIGTYPE_p_svn_wc_diff_callbacks4_t swig_types[163]
#define SWIGTYPE_p_svn_wc_diff_callbacks_t swig_types[164]
#define SWIGTYPE_p_svn_wc_entry_callbacks2_t swig_types[165]
#define SWIGTYPE_p_svn_wc_entry_callbacks_t swig_types[166]
#define SWIGTYPE_p_svn_wc_entry_t swig_types[167]
#define SWIGTYPE_p_svn_wc_external_item2_t swig_types[168]
#define SWIGTYPE_p_svn_wc_external_item_t swig_types[169]
#define SWIGTYPE_p_svn_wc_info_t swig_types[170]
#define SWIGTYPE_p_svn_wc_merge_outcome_t swig_types[171]
#define SWIGTYPE_p_svn_wc_notify_action_t swig_types[172]
#define SWIGTYPE_p_svn_wc_notify_lock_state_t swig_types[173]
#define SWIGTYPE_p_svn_wc_notify_state_t swig_types[174]
#define SWIGTYPE_p_svn_wc_notify_t swig_types[175]
#define SWIGTYPE_p_svn_wc_operation_t swig_types[176]
#define SWIGTYPE_p_svn_wc_revision_status_t swig_types[177]
#define SWIGTYPE_p_svn_wc_schedule_t swig_types[178]
#define SWIGTYPE_p_svn_wc_status2_t swig_types[179]
#define SWIGTYPE_p_svn_wc_status3_t swig_types[180]
#define SWIGTYPE_p_svn_wc_status_t swig_types[181]
#define SWIGTYPE_p_svn_wc_traversal_info_t swig_types[182]
#define SWIGTYPE_p_unsigned_char swig_types[183]
#define SWIGTYPE_p_unsigned_long swig_types[184]
#define SWIGTYPE_p_void swig_types[185]
static swig_type_info *swig_types[187];
static swig_module_info swig_module = {swig_types, 186, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#if (PY_VERSION_HEX <= 0x02000000)
# if !defined(SWIG_PYTHON_CLASSIC)
#  error "This python version requires swig to be run with the '-classic' option"
# endif
#endif

/*-----------------------------------------------
              @(target):= _wc.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__wc

#else
#  define SWIG_init    init_wc

#endif
#define SWIG_name    "_wc"

#define SWIGVERSION 0x020012 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#include "svn_time.h"
#include "svn_pools.h"


#include "swigutil_py.h"


static PyObject * _global_py_pool = NULL;


#define svn_argnum_obj0 1
#define svn_argnum_obj1 2
#define svn_argnum_obj2 3
#define svn_argnum_obj3 4
#define svn_argnum_obj4 5
#define svn_argnum_obj5 6
#define svn_argnum_obj6 7
#define svn_argnum_obj7 8
#define svn_argnum_obj8 9
#define svn_argnum_obj9 10
#define svn_argnum_obj10 11
#define svn_argnum_obj11 12
#define svn_argnum_obj12 13
#define svn_argnum_obj13 14
#define svn_argnum_obj14 15
#define svn_argnum_obj15 16
#define svn_argnum_obj16 17
#define svn_argnum_obj17 18
#define svn_argnum_obj18 19
#define svn_argnum_obj19 20
#define svn_argnum_obj20 21
#define svn_argnum_obj21 22
#define svn_argnum_obj22 23
#define svn_argnum_obj23 24
#define svn_argnum_obj24 25
#define svn_argnum_obj25 26
#define svn_argnum_obj26 27
#define svn_argnum_obj27 28
#define svn_argnum_obj28 29
#define svn_argnum_obj29 30
#define svn_argnum_obj30 31
#define svn_argnum_obj31 32
#define svn_argnum_obj32 33
#define svn_argnum_obj33 34
#define svn_argnum_obj34 35
#define svn_argnum_obj35 36
#define svn_argnum_obj36 37
#define svn_argnum_obj37 38
#define svn_argnum_obj38 39
#define svn_argnum_obj39 40


#include <apr_md5.h>
#include "svn_md5.h"


#include "svn_wc.h"


  #define SWIG_From_long   PyLong_FromLong 


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
  } else if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx = floor(x);
   double cx = ceil(x);
   double rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE long
SWIG_As_long (PyObject* obj)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (!SWIG_IsOK(res)) {
    /*
      this is needed to make valgrind/purify happier. 
     */
    memset((void*)&v, 0, sizeof(long));
    SWIG_Error(res, "");
  }
  return v;
}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char** cptr, size_t* psize, int *alloc)
{
#if PY_VERSION_HEX>=0x03000000
  if (PyUnicode_Check(obj))
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
#if PY_VERSION_HEX>=0x03000000
    if (!alloc && cptr) {
        /* We can't allow converting without allocation, since the internal
           representation of string in Python 3 is UCS-2/UCS-4 but we require
           a UTF-8 representation.
           TODO(bhy) More detailed explanation */
        return SWIG_RuntimeError;
    }
    obj = PyUnicode_AsUTF8String(obj);
    PyBytes_AsStringAndSize(obj, &cstr, &len);
    if(alloc) *alloc = SWIG_NEWOBJ;
#else
    PyString_AsStringAndSize(obj, &cstr, &len);
#endif
    if (cptr) {
      if (alloc) {
	/* 
	   In python the user should not be able to modify the inner
	   string representation. To warranty that, if you define
	   SWIG_PYTHON_SAFE_CSTRINGS, a new/copy of the python string
	   buffer is always returned.

	   The default behavior is just to return the pointer value,
	   so, be careful.
	*/ 
#if defined(SWIG_PYTHON_SAFE_CSTRINGS)
	if (*alloc != SWIG_OLDOBJ) 
#else
	if (*alloc == SWIG_NEWOBJ) 
#endif
	  {
	    *cptr = (char *)memcpy((char *)malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
	    *alloc = SWIG_NEWOBJ;
	  }
	else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      } else {
        #if PY_VERSION_HEX>=0x03000000
        assert(0); /* Should never reach here in Python 3 */
        #endif
	*cptr = SWIG_Python_str_AsChar(obj);
      }
    }
    if (psize) *psize = len + 1;
#if PY_VERSION_HEX>=0x03000000
    Py_XDECREF(obj);
#endif
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}





SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > INT_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj((char *)(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
      return PyUnicode_FromStringAndSize(carray, (int)(size));
#else
      return PyString_FromStringAndSize(carray, (int)(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
#if PY_VERSION_HEX >= 0x03000000
      {
        long v = PyLong_AsLong(obj);
        if (!PyErr_Occurred()) {
          if (v < 0) {
            return SWIG_OverflowError;
          }
        } else {
          PyErr_Clear();
        }
      }
#endif
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE unsigned long
SWIG_As_unsigned_SS_long (PyObject* obj)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (!SWIG_IsOK(res)) {
    /*
      this is needed to make valgrind/purify happier. 
     */
    memset((void*)&v, 0, sizeof(unsigned long));
    SWIG_Error(res, "");
  }
  return v;
}


SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyLong_FromLong((long)(value)); 
}


static svn_error_t * svn_wc_diff_callbacks4_invoke_file_opened(
  svn_wc_diff_callbacks4_t * _obj, svn_boolean_t *tree_conflicted, svn_boolean_t *skip, const char *path, svn_revnum_t rev, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_opened)(tree_conflicted, skip, path, rev, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_file_changed(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_changed)(contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_file_added(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const char *copyfrom_path, svn_revnum_t copyfrom_revision, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_added)(contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, copyfrom_path, copyfrom_revision, propchanges, originalprops, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_file_deleted(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, apr_hash_t *originalprops, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_deleted)(state, tree_conflicted, path, tmpfile1, tmpfile2, mimetype1, mimetype2, originalprops, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_deleted(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_deleted)(state, tree_conflicted, path, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_opened(
  svn_wc_diff_callbacks4_t * _obj, svn_boolean_t *tree_conflicted, svn_boolean_t *skip, svn_boolean_t *skip_children, const char *path, svn_revnum_t rev, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_opened)(tree_conflicted, skip, skip_children, path, rev, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_added(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, svn_boolean_t *skip, svn_boolean_t *skip_children, const char *path, svn_revnum_t rev, const char *copyfrom_path, svn_revnum_t copyfrom_revision, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_added)(state, tree_conflicted, skip, skip_children, path, rev, copyfrom_path, copyfrom_revision, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_props_changed(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, svn_boolean_t dir_was_added, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_props_changed)(propstate, tree_conflicted, path, dir_was_added, propchanges, original_props, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_closed(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, svn_boolean_t dir_was_added, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_closed)(contentstate, propstate, tree_conflicted, path, dir_was_added, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_file_changed(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_changed)(adm_access, contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_file_added(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_added)(adm_access, contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_file_deleted(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_deleted)(adm_access, state, tree_conflicted, path, tmpfile1, tmpfile2, mimetype1, mimetype2, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_added(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_added)(adm_access, state, tree_conflicted, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_deleted(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, void *diff_baton) {
  return (_obj->dir_deleted)(adm_access, state, tree_conflicted, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_props_changed(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton) {
  return (_obj->dir_props_changed)(adm_access, propstate, tree_conflicted, path, propchanges, original_props, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_opened(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_boolean_t *tree_conflicted, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_opened)(adm_access, tree_conflicted, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_closed(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, void *diff_baton) {
  return (_obj->dir_closed)(adm_access, contentstate, propstate, tree_conflicted, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_file_changed(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_changed)(adm_access, contentstate, propstate, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_file_added(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_added)(adm_access, contentstate, propstate, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_file_deleted(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_deleted)(adm_access, state, path, tmpfile1, tmpfile2, mimetype1, mimetype2, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_dir_added(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_added)(adm_access, state, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_dir_deleted(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, void *diff_baton) {
  return (_obj->dir_deleted)(adm_access, state, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_dir_props_changed(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton) {
  return (_obj->dir_props_changed)(adm_access, state, path, propchanges, original_props, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_file_changed(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, void *diff_baton) {
  return (_obj->file_changed)(adm_access, state, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_file_added(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, void *diff_baton) {
  return (_obj->file_added)(adm_access, state, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_file_deleted(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, void *diff_baton) {
  return (_obj->file_deleted)(adm_access, state, path, tmpfile1, tmpfile2, mimetype1, mimetype2, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_dir_added(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_added)(adm_access, state, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_dir_deleted(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, void *diff_baton) {
  return (_obj->dir_deleted)(adm_access, state, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_props_changed(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton) {
  return (_obj->props_changed)(adm_access, state, path, propchanges, original_props, diff_baton);
}

static svn_error_t * svn_wc_entry_callbacks2_invoke_found_entry(
  svn_wc_entry_callbacks2_t * _obj, const char *path, const svn_wc_entry_t *entry, void *walk_baton, apr_pool_t *pool) {
  return (_obj->found_entry)(path, entry, walk_baton, pool);
}

static svn_error_t * svn_wc_entry_callbacks2_invoke_handle_error(
  svn_wc_entry_callbacks2_t * _obj, const char *path, svn_error_t *err, void *walk_baton, apr_pool_t *pool) {
  return (_obj->handle_error)(path, err, walk_baton, pool);
}

static svn_error_t * svn_wc_entry_callbacks_invoke_found_entry(
  svn_wc_entry_callbacks_t * _obj, const char *path, const svn_wc_entry_t *entry, void *walk_baton, apr_pool_t *pool) {
  return (_obj->found_entry)(path, entry, walk_baton, pool);
}

static svn_error_t * svn_wc_invoke_external_update(
  svn_wc_external_update_t _obj, void *baton, const char *local_abspath, const svn_string_t *old_val, const svn_string_t *new_val, svn_depth_t depth, apr_pool_t *scratch_pool) {
  return _obj(baton, local_abspath, old_val, new_val, depth, scratch_pool);
}

static void svn_wc_invoke_notify_func2(
  svn_wc_notify_func2_t _obj, void *baton, const svn_wc_notify_t *notify, apr_pool_t *pool) {
  _obj(baton, notify, pool);
}

static void svn_wc_invoke_notify_func(
  svn_wc_notify_func_t _obj, void *baton, const char *path, svn_wc_notify_action_t action, svn_node_kind_t kind, const char *mime_type, svn_wc_notify_state_t content_state, svn_wc_notify_state_t prop_state, svn_revnum_t revision) {
  _obj(baton, path, action, kind, mime_type, content_state, prop_state, revision);
}

static svn_error_t * svn_wc_invoke_conflict_resolver_func2(
  svn_wc_conflict_resolver_func2_t _obj, svn_wc_conflict_result_t **result, const svn_wc_conflict_description2_t *description, void *baton, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(result, description, baton, result_pool, scratch_pool);
}

static svn_error_t * svn_wc_invoke_conflict_resolver_func(
  svn_wc_conflict_resolver_func_t _obj, svn_wc_conflict_result_t **result, const svn_wc_conflict_description_t *description, void *baton, apr_pool_t *pool) {
  return _obj(result, description, baton, pool);
}

static svn_error_t * svn_wc_invoke_status_func4(
  svn_wc_status_func4_t _obj, void *baton, const char *local_abspath, const svn_wc_status3_t *status, apr_pool_t *scratch_pool) {
  return _obj(baton, local_abspath, status, scratch_pool);
}

static svn_error_t * svn_wc_invoke_status_func3(
  svn_wc_status_func3_t _obj, void *baton, const char *path, svn_wc_status2_t *status, apr_pool_t *pool) {
  return _obj(baton, path, status, pool);
}

static void svn_wc_invoke_status_func2(
  svn_wc_status_func2_t _obj, void *baton, const char *path, svn_wc_status2_t *status) {
  _obj(baton, path, status);
}

static void svn_wc_invoke_status_func(
  svn_wc_status_func_t _obj, void *baton, const char *path, svn_wc_status_t *status) {
  _obj(baton, path, status);
}

static svn_error_t * svn_wc_invoke_get_file(
  svn_wc_get_file_t _obj, void *baton, const char *path, svn_revnum_t revision, svn_stream_t *stream, svn_revnum_t *fetched_rev, apr_hash_t **props, apr_pool_t *pool) {
  return _obj(baton, path, revision, stream, fetched_rev, props, pool);
}

static svn_error_t * svn_wc_invoke_dirents_func(
  svn_wc_dirents_func_t _obj, void *baton, apr_hash_t **dirents, const char *repos_root_url, const char *repos_relpath, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(baton, dirents, repos_root_url, repos_relpath, result_pool, scratch_pool);
}

static svn_error_t * svn_wc_invoke_canonicalize_svn_prop_get_file(
  svn_wc_canonicalize_svn_prop_get_file_t _obj, const svn_string_t **mime_type, svn_stream_t *stream, void *baton, apr_pool_t *pool) {
  return _obj(mime_type, stream, baton, pool);
}

static svn_error_t * svn_wc_invoke_upgrade_get_repos_info(
  svn_wc_upgrade_get_repos_info_t _obj, const char **repos_root, const char **repos_uuid, void *baton, const char *url, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(repos_root, repos_uuid, baton, url, result_pool, scratch_pool);
}

static svn_error_t * svn_wc_invoke_relocation_validator3(
  svn_wc_relocation_validator3_t _obj, void *baton, const char *uuid, const char *url, const char *root_url, apr_pool_t *pool) {
  return _obj(baton, uuid, url, root_url, pool);
}

static svn_error_t * svn_wc_invoke_relocation_validator2(
  svn_wc_relocation_validator2_t _obj, void *baton, const char *uuid, const char *url, svn_boolean_t root, apr_pool_t *pool) {
  return _obj(baton, uuid, url, root, pool);
}

static svn_error_t * svn_wc_invoke_relocation_validator(
  svn_wc_relocation_validator_t _obj, void *baton, const char *uuid, const char *url) {
  return _obj(baton, uuid, url);
}

static svn_error_t * svn_changelist_invoke_receiver(
  svn_changelist_receiver_t _obj, void *baton, const char *path, const char *changelist, apr_pool_t *pool) {
  return _obj(baton, path, changelist, pool);
}



static svn_error_t *
svn_wc_swig_init_asp_dot_net_hack (apr_pool_t *pool)
{
#if defined(WIN32) || defined(__CYGWIN__)
  if (getenv ("SVN_ASP_DOT_NET_HACK"))
    SVN_ERR (svn_wc_set_adm_dir("_svn", pool));
#endif /* WIN32 */
  return SVN_NO_ERROR;
}

#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_svn_wc_version(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_version_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":svn_wc_version")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_version_t *)svn_wc_version();
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_version_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_context_create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t **arg1 = (svn_wc_context_t **) 0 ;
  svn_config_t *arg2 = (svn_config_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_context_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"O|OO:svn_wc_context_create",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (svn_config_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_config_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_context_create(arg1,(struct svn_config_t const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_context_create arg 1 (svn_wc_context_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_context_create is not implemented yet");
    
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_context_destroy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_context_destroy",&obj0)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_context_destroy(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_open3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_adm_open3",&obj0,&arg3,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_open3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_open2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_adm_open2",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_open2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_open(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_adm_open",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_open(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_probe_open3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_adm_probe_open3",&obj0,&arg3,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_probe_open3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_probe_open2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_adm_probe_open2",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_probe_open2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_probe_open(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_adm_probe_open",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_probe_open(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_open_anchor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t **arg2 = (svn_wc_adm_access_t **) 0 ;
  char **arg3 = (char **) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  int arg6 ;
  svn_cancel_func_t arg7 = (svn_cancel_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  svn_wc_adm_access_t *temp2 ;
  char *temp3 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"sOOO|O:svn_wc_adm_open_anchor",&arg4,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (int)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg8 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_open_anchor(arg1,arg2,(char const **)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    PyObject *s;
    if (*arg3 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg3);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_retrieve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_adm_retrieve",&obj0,&arg3,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_retrieve(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_probe_retrieve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_adm_probe_retrieve",&obj0,&arg3,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_probe_retrieve(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_probe_try3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_adm_probe_try3",&obj0,&arg3,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_probe_try3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_probe_try2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_adm_probe_try2",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_probe_try2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_probe_try(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t **arg1 = (svn_wc_adm_access_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_adm_access_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_adm_probe_try",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_probe_try(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_close2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_adm_close2",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_close2(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_close(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_adm_close",&obj0)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_adm_close(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_access_path(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_adm_access_path",&obj0)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (char *)svn_wc_adm_access_path((struct svn_wc_adm_access_t const *)arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_access_pool(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_pool_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_adm_access_pool",&obj0)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (apr_pool_t *)svn_wc_adm_access_pool((struct svn_wc_adm_access_t const *)arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_apr_pool_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_adm_locked(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_adm_locked",&obj0)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_boolean_t)svn_wc_adm_locked((struct svn_wc_adm_access_t const *)arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_locked2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_wc_context_t *arg3 = (svn_wc_context_t *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_locked2",&obj0,&arg4,&obj2)) SWIG_fail;
  {
    arg3 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_locked2(arg1,arg2,arg3,(char const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_locked(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"s|O:svn_wc_locked",&arg2,&obj1)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_locked(arg1,(char const *)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_is_adm_dir(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  svn_boolean_t result;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"s|O:svn_wc_is_adm_dir",&arg1,&obj1)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_boolean_t)svn_wc_is_adm_dir((char const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_From_long((long)(result));
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_adm_dir(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg1 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"|O:svn_wc_get_adm_dir",&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (char *)svn_wc_get_adm_dir(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_set_adm_dir(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"s|O:svn_wc_set_adm_dir",&arg1,&obj1)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_set_adm_dir((char const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_init_traversal_info(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  svn_wc_traversal_info_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg1 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"|O:svn_wc_init_traversal_info",&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_traversal_info_t *)svn_wc_init_traversal_info(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_traversal_info_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_edited_externals(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  apr_hash_t **arg2 = (apr_hash_t **) 0 ;
  svn_wc_traversal_info_t *arg3 = (svn_wc_traversal_info_t *) 0 ;
  apr_hash_t *temp1 ;
  apr_hash_t *temp2 ;
  PyObject * obj0 = 0 ;
  
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_edited_externals",&obj0)) SWIG_fail;
  {
    arg3 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_wc_edited_externals(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_stringhash_to_dict(*arg1));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_stringhash_to_dict(*arg2));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_traversed_depths(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  svn_wc_traversal_info_t *arg2 = (svn_wc_traversal_info_t *) 0 ;
  apr_hash_t *temp1 ;
  PyObject * obj0 = 0 ;
  
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_traversed_depths",&obj0)) SWIG_fail;
  {
    arg2 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_wc_traversed_depths(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_stringhash_to_dict(*arg1));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_target_dir_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_external_item2_t_target_dir_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->target_dir) free((char *)arg1->target_dir);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->target_dir = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_target_dir_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_external_item2_t_target_dir_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->target_dir);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_external_item2_t_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_external_item2_t_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_external_item2_t_revision_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_opt_revision_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_opt_revision_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->revision = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_opt_revision_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_external_item2_t_revision_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_opt_revision_t *)& ((arg1)->revision);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_opt_revision_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_peg_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_external_item2_t_peg_revision_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_opt_revision_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_opt_revision_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->peg_revision = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_t_peg_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_opt_revision_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_external_item2_t_peg_revision_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_opt_revision_t *)& ((arg1)->peg_revision);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_opt_revision_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_external_item2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_external_item2_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_external_item2_t *)calloc(1, sizeof(struct svn_wc_external_item2_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_external_item2_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_external_item2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_external_item2_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_external_item2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_external_item2_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_external_item2_create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_external_item2_t **arg1 = (svn_wc_external_item2_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_external_item2_t *temp1 ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"|O:svn_wc_external_item2_create",&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_external_item2_create(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_external_item2_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item_create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_external_item2_t **arg1 = (svn_wc_external_item2_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_external_item2_t *temp1 ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"|O:svn_wc_external_item_create",&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_external_item_create((struct svn_wc_external_item2_t const **)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_external_item2_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item2_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_external_item2_t *arg1 = (svn_wc_external_item2_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_external_item2_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_external_item2_dup",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_external_item2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_external_item2_t *)svn_wc_external_item2_dup((struct svn_wc_external_item2_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_external_item2_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item_t_target_dir_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_external_item_t_target_dir_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->target_dir) free((char *)arg1->target_dir);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->target_dir = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item_t_target_dir_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_external_item_t_target_dir_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->target_dir);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item_t_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_external_item_t_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item_t_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_external_item_t_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item_t_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_external_item_t_revision_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_opt_revision_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_opt_revision_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->revision = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_external_item_t_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_opt_revision_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_external_item_t_revision_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_opt_revision_t *)& ((arg1)->revision);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_opt_revision_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_external_item_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_external_item_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_external_item_t *)calloc(1, sizeof(struct svn_wc_external_item_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_external_item_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_external_item_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_external_item_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_external_item_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_external_item_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_external_item_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_external_item_t *arg1 = (svn_wc_external_item_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_external_item_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_external_item_dup",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_external_item_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_external_item_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_external_item_t *)svn_wc_external_item_dup((struct svn_wc_external_item_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_external_item_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_parse_externals_description3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp1 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"ssO|O:svn_wc_parse_externals_description3",&arg2,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_parse_externals_description3(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_pointerlist_to_list(*arg1, SWIGTYPE_p_svn_wc_external_item2_t,
        _global_py_pool))
    
    
    ;
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_parse_externals_description2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp1 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"ss|O:svn_wc_parse_externals_description2",&arg2,&arg3,&obj2)) SWIG_fail;
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_parse_externals_description2(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_pointerlist_to_list(*arg1, SWIGTYPE_p_svn_wc_external_item2_t,
        _global_py_pool))
    
    
    ;
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_parse_externals_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp1 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"ss|O:svn_wc_parse_externals_description",&arg2,&arg3,&obj2)) SWIG_fail;
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_parse_externals_description(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg1,
        SWIGTYPE_p_svn_wc_external_item_t,
        _global_py_pool))
    
    
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_path_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_notify_t_path_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_path_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_path_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->path);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_action_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_wc_notify_action_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_action_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_action_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->action = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_action_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_notify_action_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_action_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_notify_action_t) ((arg1)->action);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_mime_type_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_notify_t_mime_type_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->mime_type) free((char *)arg1->mime_type);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->mime_type = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_mime_type_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_mime_type_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->mime_type);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_lock_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_lock_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_lock_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_lock_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->lock = (svn_lock_t const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_lock_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_lock_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_lock_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_lock_t *) ((arg1)->lock);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_lock_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_err_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_error_t *arg2 = (svn_error_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_err_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_error_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_error_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->err = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_err_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_err_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *) ((arg1)->err);
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_content_state_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_wc_notify_state_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_content_state_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->content_state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_content_state_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_notify_state_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_content_state_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_notify_state_t) ((arg1)->content_state);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_prop_state_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_wc_notify_state_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_prop_state_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->prop_state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_prop_state_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_notify_state_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_prop_state_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_notify_state_t) ((arg1)->prop_state);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_lock_state_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_wc_notify_lock_state_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_lock_state_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_lock_state_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->lock_state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_lock_state_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_notify_lock_state_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_lock_state_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_notify_lock_state_t) ((arg1)->lock_state);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_revision_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->revision = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_revision_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->revision);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_changelist_name_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_notify_t_changelist_name_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist_name) free((char *)arg1->changelist_name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist_name = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_changelist_name_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_changelist_name_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->changelist_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_merge_range_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_merge_range_t *arg2 = (svn_merge_range_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_merge_range_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_merge_range_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_merge_range_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->merge_range = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_merge_range_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_merge_range_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_merge_range_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_merge_range_t *) ((arg1)->merge_range);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_merge_range_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_notify_t_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_path_prefix_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_notify_t_path_prefix_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path_prefix) free((char *)arg1->path_prefix);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path_prefix = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_path_prefix_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_path_prefix_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->path_prefix);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_prop_name_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_notify_t_prop_name_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->prop_name) free((char *)arg1->prop_name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->prop_name = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_prop_name_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_prop_name_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->prop_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_rev_props_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_rev_props_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_apr_hash_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->rev_props = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_rev_props_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_hash_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_rev_props_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (apr_hash_t *) ((arg1)->rev_props);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_apr_hash_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_old_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_old_revision_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->old_revision = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_old_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_old_revision_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->old_revision);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_original_start_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_linenum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_hunk_original_start_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_linenum_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->hunk_original_start = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_original_start_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_linenum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_hunk_original_start_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_linenum_t) ((arg1)->hunk_original_start);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_original_length_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_linenum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_hunk_original_length_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_linenum_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->hunk_original_length = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_original_length_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_linenum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_hunk_original_length_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_linenum_t) ((arg1)->hunk_original_length);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_modified_start_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_linenum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_hunk_modified_start_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_linenum_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->hunk_modified_start = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_modified_start_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_linenum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_hunk_modified_start_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_linenum_t) ((arg1)->hunk_modified_start);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_modified_length_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_linenum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_hunk_modified_length_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_linenum_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->hunk_modified_length = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_modified_length_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_linenum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_hunk_modified_length_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_linenum_t) ((arg1)->hunk_modified_length);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_matched_line_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_linenum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_hunk_matched_line_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_linenum_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->hunk_matched_line = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_matched_line_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_linenum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_hunk_matched_line_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_linenum_t) ((arg1)->hunk_matched_line);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_fuzz_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  svn_linenum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_notify_t_hunk_fuzz_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_linenum_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->hunk_fuzz = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_notify_t_hunk_fuzz_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_linenum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_notify_t_hunk_fuzz_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_linenum_t) ((arg1)->hunk_fuzz);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_notify_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_notify_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_notify_t *)calloc(1, sizeof(struct svn_wc_notify_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_notify_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_notify_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_notify_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_notify_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_notify_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_create_notify(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_notify_action_t arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_wc_notify_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_create_notify",&arg1,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_notify_action_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_notify_t *)svn_wc_create_notify((char const *)arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_notify_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_create_notify_url(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_notify_action_t arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_wc_notify_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_create_notify_url",&arg1,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_notify_action_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_notify_t *)svn_wc_create_notify_url((char const *)arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_notify_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_dup_notify(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_notify_t *arg1 = (svn_wc_notify_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_notify_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_dup_notify",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_notify_t *)svn_wc_dup_notify((struct svn_wc_notify_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_notify_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_repos_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_version_t_repos_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_url) free((char *)arg1->repos_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_repos_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_version_t_repos_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->repos_url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_peg_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_version_t_peg_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->peg_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_peg_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_version_t_peg_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->peg_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_path_in_repos_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_version_t_path_in_repos_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path_in_repos) free((char *)arg1->path_in_repos);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path_in_repos = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_path_in_repos_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_version_t_path_in_repos_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->path_in_repos);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_node_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_version_t_node_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->node_kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_node_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_version_t_node_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->node_kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_repos_uuid_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_version_t_repos_uuid_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_uuid) free((char *)arg1->repos_uuid);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_uuid = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_t_repos_uuid_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_version_t_repos_uuid_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->repos_uuid);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_conflict_version_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_conflict_version_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_conflict_version_t *)calloc(1, sizeof(struct svn_wc_conflict_version_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_version_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_conflict_version_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_conflict_version_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_conflict_version_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_conflict_version_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_create2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  svn_node_kind_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sssOO|O:svn_wc_conflict_version_create2",&arg1,&arg2,&arg3,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_node_kind_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_version_t *)svn_wc_conflict_version_create2((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_version_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_revnum_t arg3 ;
  svn_node_kind_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"ssOO|O:svn_wc_conflict_version_create",&arg1,&arg2,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_node_kind_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_version_t *)svn_wc_conflict_version_create((char const *)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_version_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_version_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_conflict_version_t *arg1 = (svn_wc_conflict_version_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_conflict_version_dup",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_version_t *)svn_wc_conflict_version_dup((struct svn_wc_conflict_version_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_version_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_path_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_description_t_path_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_path_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_path_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->path);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_node_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_node_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->node_kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_node_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_node_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->node_kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_wc_conflict_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_conflict_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_conflict_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_conflict_kind_t) ((arg1)->kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_property_name_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_description_t_property_name_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->property_name) free((char *)arg1->property_name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->property_name = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_property_name_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_property_name_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->property_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_is_binary_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_is_binary_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->is_binary = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_is_binary_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_is_binary_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->is_binary);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_mime_type_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_description_t_mime_type_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->mime_type) free((char *)arg1->mime_type);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->mime_type = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_mime_type_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_mime_type_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->mime_type);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_access_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_access_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->access = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_access_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_adm_access_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_access_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_adm_access_t *) ((arg1)->access);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_adm_access_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_action_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_wc_conflict_action_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_action_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_conflict_action_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->action = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_action_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_conflict_action_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_action_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_conflict_action_t) ((arg1)->action);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_reason_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_wc_conflict_reason_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_reason_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_conflict_reason_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->reason = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_reason_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_conflict_reason_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_reason_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_conflict_reason_t) ((arg1)->reason);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_base_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_description_t_base_file_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->base_file) free((char *)arg1->base_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->base_file = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_base_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_base_file_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->base_file);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_their_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_description_t_their_file_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->their_file) free((char *)arg1->their_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->their_file = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_their_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_their_file_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->their_file);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_my_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_description_t_my_file_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->my_file) free((char *)arg1->my_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->my_file = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_my_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_my_file_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->my_file);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_merged_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_conflict_description_t_merged_file_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->merged_file) free((char *)arg1->merged_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->merged_file = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_merged_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_merged_file_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->merged_file);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_operation_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_wc_operation_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_operation_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_operation_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->operation = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_operation_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_operation_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_operation_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_operation_t) ((arg1)->operation);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_src_left_version_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_wc_conflict_version_t *arg2 = (svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_src_left_version_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->src_left_version = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_src_left_version_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_src_left_version_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_conflict_version_t *) ((arg1)->src_left_version);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_version_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_src_right_version_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  svn_wc_conflict_version_t *arg2 = (svn_wc_conflict_version_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_conflict_description_t_src_right_version_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->src_right_version = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_t_src_right_version_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_conflict_description_t_src_right_version_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_conflict_version_t *) ((arg1)->src_right_version);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_version_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_conflict_description_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_conflict_description_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_conflict_description_t *)calloc(1, sizeof(struct svn_wc_conflict_description_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_conflict_description_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_conflict_description_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_conflict_description_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_conflict_description_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_create_text2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"s|O:svn_wc_conflict_description_create_text2",&arg1,&obj1)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description_create_text2((char const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description2_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_create_text(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_conflict_description_create_text",&arg1,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description_t *)svn_wc_conflict_description_create_text((char const *)arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_create_prop2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_node_kind_t arg2 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOs|O:svn_wc_conflict_description_create_prop2",&arg1,&obj1,&arg3,&obj3)) SWIG_fail;
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description_create_prop2((char const *)arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description2_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_create_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_node_kind_t arg3 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOs|O:svn_wc_conflict_description_create_prop",&arg1,&obj1,&obj2,&arg4,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_node_kind_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description_t *)svn_wc_conflict_description_create_prop((char const *)arg1,arg2,arg3,(char const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_create_tree2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_node_kind_t arg2 ;
  svn_wc_operation_t arg3 ;
  svn_wc_conflict_version_t *arg4 = (svn_wc_conflict_version_t *) 0 ;
  svn_wc_conflict_version_t *arg5 = (svn_wc_conflict_version_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOO|O:svn_wc_conflict_description_create_tree2",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_operation_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description_create_tree2((char const *)arg1,arg2,arg3,(struct svn_wc_conflict_version_t const *)arg4,(struct svn_wc_conflict_version_t const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description2_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description_create_tree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_node_kind_t arg3 ;
  svn_wc_operation_t arg4 ;
  svn_wc_conflict_version_t *arg5 = (svn_wc_conflict_version_t *) 0 ;
  svn_wc_conflict_version_t *arg6 = (svn_wc_conflict_version_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOO|O:svn_wc_conflict_description_create_tree",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_node_kind_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_operation_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description_t *)svn_wc_conflict_description_create_tree((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflict_description2_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_conflict_description2_t *arg1 = (svn_wc_conflict_description2_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_conflict_description2_dup",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_conflict_description2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description2_dup((struct svn_wc_conflict_description2_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description2_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc__conflict_description2_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_conflict_description2_t *arg1 = (svn_wc_conflict_description2_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc__conflict_description2_dup",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_conflict_description2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_conflict_description2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_description2_t *)svn_wc__conflict_description2_dup((struct svn_wc_conflict_description2_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description2_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_create_conflict_result(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_conflict_choice_t arg1 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_wc_conflict_result_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_create_conflict_result",&obj0,&arg2,&obj2)) SWIG_fail;
  {
    arg1 = (svn_wc_conflict_choice_t)SWIG_As_long (obj0);
    if (SWIG_arg_fail(svn_argnum_obj0)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_conflict_result_t *)svn_wc_create_conflict_result(arg1,(char const *)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_result_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_opened_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = (svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_file_opened_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_file_opened_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->file_opened = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_opened_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_file_opened_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)) ((arg1)->file_opened);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_file_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_file_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_file_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->file_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_file_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_file_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_file_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->file_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *) = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_file_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_file_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_file_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_file_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->file_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *) = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_dir_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_dir_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_dir_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)) ((arg1)->dir_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_opened_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = (svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_dir_opened_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_dir_opened_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_opened = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_opened_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_dir_opened_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)) ((arg1)->dir_opened);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *) = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_dir_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_dir_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_dir_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)) ((arg1)->dir_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_props_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_dir_props_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_dir_props_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_props_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_props_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_dir_props_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->dir_props_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_closed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *) = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks4_t_dir_closed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks4_t_dir_closed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_closed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_t_dir_closed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks4_t_dir_closed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)) ((arg1)->dir_closed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_diff_callbacks4_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_diff_callbacks4_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_diff_callbacks4_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks4_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_diff_callbacks4_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_diff_callbacks4_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_diff_callbacks4_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_diff_callbacks4_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_diff_callbacks4_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_file_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_file_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_file_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_file_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_file_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_file_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_file_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_file_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_file_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_file_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_file_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_file_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_file_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_file_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_file_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)) ((arg1)->file_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_dir_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_dir_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_dir_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_dir_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_dir_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_dir_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)) ((arg1)->dir_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_props_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_dir_props_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_dir_props_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_props_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_props_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_dir_props_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->dir_props_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_opened_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_dir_opened_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_dir_opened_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_opened = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_opened_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_dir_opened_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_opened);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_closed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks3_t_dir_closed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks3_t_dir_closed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_closed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_t_dir_closed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks3_t_dir_closed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)) ((arg1)->dir_closed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_diff_callbacks3_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_diff_callbacks3_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_diff_callbacks3_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks3_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_diff_callbacks3_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_diff_callbacks3_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_diff_callbacks3_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_diff_callbacks3_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_diff_callbacks3_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_file_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks2_t_file_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks2_t_file_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_file_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks2_t_file_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_file_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks2_t_file_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks2_t_file_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_file_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks2_t_file_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_file_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks2_t_file_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks2_t_file_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_file_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks2_t_file_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)) ((arg1)->file_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_dir_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks2_t_dir_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks2_t_dir_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_dir_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks2_t_dir_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_dir_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks2_t_dir_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks2_t_dir_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_dir_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks2_t_dir_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)) ((arg1)->dir_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_dir_props_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks2_t_dir_props_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks2_t_dir_props_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_props_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_t_dir_props_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks2_t_dir_props_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->dir_props_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_diff_callbacks2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_diff_callbacks2_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_diff_callbacks2_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks2_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_diff_callbacks2_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_diff_callbacks2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_diff_callbacks2_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_diff_callbacks2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_diff_callbacks2_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_file_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks_t_file_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks_t_file_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_file_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks_t_file_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)) ((arg1)->file_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_file_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks_t_file_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks_t_file_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_file_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks_t_file_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)) ((arg1)->file_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_file_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks_t_file_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks_t_file_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_file_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks_t_file_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)) ((arg1)->file_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_dir_added_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks_t_dir_added_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks_t_dir_added_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_dir_added_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks_t_dir_added_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_added);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_dir_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks_t_dir_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks_t_dir_deleted_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_dir_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks_t_dir_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)) ((arg1)->dir_deleted);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_props_changed_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_diff_callbacks_t_props_changed_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_diff_callbacks_t_props_changed_set" "', argument " "2"" of type '" "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)""'"); 
    }
  }
  if (arg1) (arg1)->props_changed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_t_props_changed_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_diff_callbacks_t_props_changed_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->props_changed);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_diff_callbacks_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_diff_callbacks_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_diff_callbacks_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_diff_callbacks_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_diff_callbacks_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_diff_callbacks_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_diff_callbacks_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_diff_callbacks_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_check_wc2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  int *arg1 = (int *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  int temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_check_wc2",&obj0,&arg3,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_check_wc2(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_check_wc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int *arg2 = (int *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"s|O:svn_wc_check_wc",&arg1,&obj1)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_check_wc((char const *)arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_has_binary_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_has_binary_prop",&arg2,&obj1,&obj2)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_has_binary_prop(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_text_modified_p2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsO|O:svn_wc_text_modified_p2",&obj0,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_text_modified_p2(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_text_modified_p(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sOO|O:svn_wc_text_modified_p",&arg2,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_text_modified_p(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_props_modified_p2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_props_modified_p2",&obj0,&arg3,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_props_modified_p2(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_props_modified_p(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_props_modified_p",&arg2,&obj1,&obj2)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_props_modified_p(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_name_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_name_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_name_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_name_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_revision_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->revision = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_revision_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->revision);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_repos_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Oz:svn_wc_entry_t_repos_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos) free((char *)arg1->repos);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_repos_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_repos_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->repos);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_uuid_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Oz:svn_wc_entry_t_uuid_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->uuid) free((char *)arg1->uuid);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->uuid = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_uuid_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_uuid_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->uuid);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_schedule_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_wc_schedule_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_schedule_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_schedule_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->schedule = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_schedule_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_schedule_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_schedule_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_schedule_t) ((arg1)->schedule);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_copied_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_copied_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->copied = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_copied_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_copied_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->copied);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_deleted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_deleted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->deleted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_deleted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_deleted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->deleted);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_absent_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_absent_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->absent = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_absent_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_absent_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->absent);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_incomplete_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_incomplete_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->incomplete = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_incomplete_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_incomplete_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->incomplete);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_copyfrom_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Oz:svn_wc_entry_t_copyfrom_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_copyfrom_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_copyfrom_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->copyfrom_url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_copyfrom_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_copyfrom_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->copyfrom_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_copyfrom_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_copyfrom_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_conflict_old_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_conflict_old_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_old) free((char *)arg1->conflict_old);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_old = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_conflict_old_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_conflict_old_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->conflict_old);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_conflict_new_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_conflict_new_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_new) free((char *)arg1->conflict_new);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_new = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_conflict_new_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_conflict_new_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->conflict_new);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_conflict_wrk_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_conflict_wrk_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_wrk) free((char *)arg1->conflict_wrk);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_wrk = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_conflict_wrk_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_conflict_wrk_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->conflict_wrk);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_prejfile_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_prejfile_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->prejfile) free((char *)arg1->prejfile);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->prejfile = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_prejfile_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_prejfile_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->prejfile);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_text_time_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_text_time_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->text_time = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_text_time_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_text_time_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->text_time);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_prop_time_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_prop_time_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->prop_time = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_prop_time_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_prop_time_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->prop_time);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_checksum_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_checksum_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->checksum) free((char *)arg1->checksum);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->checksum = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_checksum_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_checksum_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->checksum);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cmt_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_cmt_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->cmt_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cmt_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_cmt_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->cmt_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cmt_date_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_cmt_date_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->cmt_date = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cmt_date_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_cmt_date_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->cmt_date);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cmt_author_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_cmt_author_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->cmt_author) free((char *)arg1->cmt_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->cmt_author = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cmt_author_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_cmt_author_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->cmt_author);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_token_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Oz:svn_wc_entry_t_lock_token_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->lock_token) free((char *)arg1->lock_token);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->lock_token = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_token_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_lock_token_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->lock_token);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_owner_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_lock_owner_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->lock_owner) free((char *)arg1->lock_owner);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->lock_owner = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_owner_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_lock_owner_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->lock_owner);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_comment_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_lock_comment_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->lock_comment) free((char *)arg1->lock_comment);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->lock_comment = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_comment_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_lock_comment_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->lock_comment);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_creation_date_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_lock_creation_date_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->lock_creation_date = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_lock_creation_date_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_lock_creation_date_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->lock_creation_date);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_has_props_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_has_props_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->has_props = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_has_props_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_has_props_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->has_props);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_has_prop_mods_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_has_prop_mods_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->has_prop_mods = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_has_prop_mods_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_has_prop_mods_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->has_prop_mods);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cachable_props_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_cachable_props_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->cachable_props) free((char *)arg1->cachable_props);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->cachable_props = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_cachable_props_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_cachable_props_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->cachable_props);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_present_props_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_present_props_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->present_props) free((char *)arg1->present_props);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->present_props = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_present_props_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_present_props_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->present_props);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_changelist_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_changelist_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_changelist_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_changelist_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->changelist);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_working_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  apr_off_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_working_size_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_off_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->working_size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_working_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_off_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_working_size_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->working_size);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_keep_local_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_keep_local_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->keep_local = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_keep_local_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_keep_local_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->keep_local);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_depth_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_depth_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_depth_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_depth_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->depth = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_depth_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_depth_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_depth_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_depth_t) ((arg1)->depth);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_tree_conflict_data_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_tree_conflict_data_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->tree_conflict_data) free((char *)arg1->tree_conflict_data);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->tree_conflict_data = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_tree_conflict_data_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_tree_conflict_data_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->tree_conflict_data);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_file_external_path_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_entry_t_file_external_path_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->file_external_path) free((char *)arg1->file_external_path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->file_external_path = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_file_external_path_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_file_external_path_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->file_external_path);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_file_external_peg_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_file_external_peg_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_opt_revision_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_opt_revision_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->file_external_peg_rev = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_file_external_peg_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_opt_revision_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_file_external_peg_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_opt_revision_t *)& ((arg1)->file_external_peg_rev);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_opt_revision_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_file_external_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_t_file_external_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_opt_revision_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_opt_revision_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->file_external_rev = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_t_file_external_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_opt_revision_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_t_file_external_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_opt_revision_t *)& ((arg1)->file_external_rev);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_opt_revision_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_entry_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_entry_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_entry_t *)calloc(1, sizeof(struct svn_wc_entry_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_entry_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_entry_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_entry_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_entry_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_entry_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_entry(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_entry_t **arg1 = (svn_wc_entry_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_entry_t *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sOO|O:svn_wc_entry",&arg2,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_entry((struct svn_wc_entry_t const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_entry_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entries_read(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OO|O:svn_wc_entries_read",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_entries_read(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg1,
        SWIGTYPE_p_svn_wc_entry_t,
        _global_py_pool))
    
    
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_entry_t *arg1 = (svn_wc_entry_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_entry_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_entry_dup",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_entry_t *)svn_wc_entry_dup((struct svn_wc_entry_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_entry_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_schedule_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  svn_wc_schedule_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_info_t_schedule_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_schedule_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->schedule = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_schedule_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_schedule_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_schedule_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_schedule_t) ((arg1)->schedule);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_copyfrom_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Oz:svn_wc_info_t_copyfrom_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_copyfrom_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_copyfrom_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->copyfrom_url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_copyfrom_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_info_t_copyfrom_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->copyfrom_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_copyfrom_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_copyfrom_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_checksum_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  svn_checksum_t *arg2 = (svn_checksum_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_info_t_checksum_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_checksum_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_checksum_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->checksum = (svn_checksum_t const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_checksum_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_checksum_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_checksum_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_checksum_t *) ((arg1)->checksum);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_checksum_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_changelist_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_info_t_changelist_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_changelist_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_changelist_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->changelist);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_depth_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  svn_depth_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_info_t_depth_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_depth_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->depth = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_depth_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_depth_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_depth_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_depth_t) ((arg1)->depth);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_recorded_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  svn_filesize_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_info_t_recorded_size_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (svn_filesize_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->recorded_size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_recorded_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_filesize_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_recorded_size_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->recorded_size);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_recorded_time_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_info_t_recorded_time_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->recorded_time = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_recorded_time_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_recorded_time_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->recorded_time);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_conflicts_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_info_t_conflicts_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->conflicts = (apr_array_header_t const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_conflicts_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_array_header_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_conflicts_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (apr_array_header_t *) ((arg1)->conflicts);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_apr_array_header_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_wcroot_abspath_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_info_t_wcroot_abspath_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->wcroot_abspath) free((char *)arg1->wcroot_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->wcroot_abspath = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_wcroot_abspath_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_wcroot_abspath_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->wcroot_abspath);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_moved_from_abspath_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_info_t_moved_from_abspath_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_from_abspath) free((char *)arg1->moved_from_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_from_abspath = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_moved_from_abspath_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_moved_from_abspath_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->moved_from_abspath);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_moved_to_abspath_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_info_t_moved_to_abspath_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_to_abspath) free((char *)arg1->moved_to_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_to_abspath = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_info_t_moved_to_abspath_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_info_t_moved_to_abspath_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->moved_to_abspath);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_info_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_info_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_info_t *)calloc(1, sizeof(struct svn_wc_info_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_info_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_info_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_info_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_info_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_info_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_info_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_info_t *arg1 = (svn_wc_info_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_info_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_info_dup",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_info_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_info_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_info_t *)svn_wc_info_dup((struct svn_wc_info_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_info_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflicted_p3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  svn_wc_context_t *arg4 = (svn_wc_context_t *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_conflicted_p3",&obj0,&arg5,&obj2)) SWIG_fail;
  {
    arg4 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_conflicted_p3(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflicted_p2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_wc_adm_access_t *arg5 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_conflicted_p2",&arg4,&obj1,&obj2)) SWIG_fail;
  {
    arg5 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_conflicted_p2(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_conflicted_p(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_entry_t *arg4 = (svn_wc_entry_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_conflicted_p",&arg3,&obj1,&obj2)) SWIG_fail;
  {
    arg4 = (svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_conflicted_p(arg1,arg2,(char const *)arg3,(struct svn_wc_entry_t const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_ancestry(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  svn_revnum_t *arg2 = (svn_revnum_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp1 ;
  svn_revnum_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_get_ancestry",&arg3,&obj1,&obj2)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_ancestry(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg1);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks2_t_found_entry_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks2_t *arg1 = (struct svn_wc_entry_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = (svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_callbacks2_t_found_entry_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_entry_callbacks2_t_found_entry_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->found_entry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks2_t_found_entry_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks2_t *arg1 = (struct svn_wc_entry_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_callbacks2_t_found_entry_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)) ((arg1)->found_entry);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks2_t_handle_error_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks2_t *arg1 = (struct svn_wc_entry_callbacks2_t *) 0 ;
  svn_error_t *(*arg2)(char const *,svn_error_t *,void *,apr_pool_t *) = (svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_callbacks2_t_handle_error_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_entry_callbacks2_t_handle_error_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->handle_error = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks2_t_handle_error_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks2_t *arg1 = (struct svn_wc_entry_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,svn_error_t *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_callbacks2_t_handle_error_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)) ((arg1)->handle_error);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_entry_callbacks2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks2_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_entry_callbacks2_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_entry_callbacks2_t *)calloc(1, sizeof(struct svn_wc_entry_callbacks2_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_entry_callbacks2_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_entry_callbacks2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks2_t *arg1 = (struct svn_wc_entry_callbacks2_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_entry_callbacks2_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_entry_callbacks2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_entry_callbacks2_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks_t_found_entry_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks_t *arg1 = (struct svn_wc_entry_callbacks_t *) 0 ;
  svn_error_t *(*arg2)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = (svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_entry_callbacks_t_found_entry_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_wc_entry_callbacks_t_found_entry_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->found_entry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks_t_found_entry_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks_t *arg1 = (struct svn_wc_entry_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_entry_callbacks_t_found_entry_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)) ((arg1)->found_entry);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_entry_callbacks_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_entry_callbacks_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_entry_callbacks_t *)calloc(1, sizeof(struct svn_wc_entry_callbacks_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_entry_callbacks_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_entry_callbacks_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_entry_callbacks_t *arg1 = (struct svn_wc_entry_callbacks_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_entry_callbacks_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_entry_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_entry_callbacks_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_entry_callbacks_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_walk_entries3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_entry_callbacks2_t *arg3 = (svn_wc_entry_callbacks2_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_cancel_func_t arg7 = (svn_cancel_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOO|O:svn_wc_walk_entries3",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg8 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_walk_entries3((char const *)arg1,arg2,(struct svn_wc_entry_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_walk_entries2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_entry_callbacks_t *arg3 = (svn_wc_entry_callbacks_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOO|O:svn_wc_walk_entries2",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_entry_callbacks_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_entry_callbacks_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_walk_entries2((char const *)arg1,arg2,(struct svn_wc_entry_callbacks_t const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_walk_entries(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_entry_callbacks_t *arg3 = (svn_wc_entry_callbacks_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOO|O:svn_wc_walk_entries",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_entry_callbacks_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_entry_callbacks_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_walk_entries((char const *)arg1,arg2,(struct svn_wc_entry_callbacks_t const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_mark_missing_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_mark_missing_deleted",&arg1,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_mark_missing_deleted((char const *)arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_ensure_adm4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_revnum_t arg6 ;
  svn_depth_t arg7 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OssssOO|O:svn_wc_ensure_adm4",&obj0,&arg2,&arg3,&arg4,&arg5,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_revnum_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_depth_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_ensure_adm4(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_ensure_adm3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  svn_depth_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"szszOO|O:svn_wc_ensure_adm3",&arg1,&arg2,&arg3,&arg4,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_ensure_adm3((char const *)arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_ensure_adm2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"szszO|O:svn_wc_ensure_adm2",&arg1,&arg2,&arg3,&arg4,&obj4,&obj5)) SWIG_fail;
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_ensure_adm2((char const *)arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_ensure_adm(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"szsO|O:svn_wc_ensure_adm",&arg1,&arg2,&arg3,&obj3,&obj4)) SWIG_fail;
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_ensure_adm((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_maybe_set_repos_root(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"Osz|O:svn_wc_maybe_set_repos_root",&obj0,&arg2,&arg3,&obj3)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_maybe_set_repos_root(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_depth_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_depth_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_depth_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_depth_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->depth = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_depth_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_depth_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_depth_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_depth_t) ((arg1)->depth);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_filesize_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_filesize_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_filesize_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (svn_filesize_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->filesize = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_filesize_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_filesize_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_filesize_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->filesize);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_versioned_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_versioned_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->versioned = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_versioned_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_versioned_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->versioned);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_conflicted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_conflicted_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->conflicted = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_conflicted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_conflicted_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->conflicted);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_node_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_node_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->node_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_node_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_node_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->node_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_text_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_text_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->text_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_text_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_text_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->text_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_prop_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_prop_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->prop_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_prop_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_prop_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->prop_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_copied_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_copied_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->copied = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_copied_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_copied_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->copied);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_revision_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->revision = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_revision_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->revision);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changed_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_changed_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->changed_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changed_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_changed_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->changed_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changed_date_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_changed_date_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->changed_date = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changed_date_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_changed_date_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->changed_date);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changed_author_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_changed_author_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changed_author) free((char *)arg1->changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changed_author = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changed_author_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_changed_author_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->changed_author);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_root_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_repos_root_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_root_url) free((char *)arg1->repos_root_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_root_url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_root_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_repos_root_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->repos_root_url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_uuid_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_repos_uuid_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_uuid) free((char *)arg1->repos_uuid);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_uuid = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_uuid_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_repos_uuid_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->repos_uuid);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_relpath_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_repos_relpath_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_relpath) free((char *)arg1->repos_relpath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_relpath = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_relpath_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_repos_relpath_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->repos_relpath);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_switched_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_switched_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->switched = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_switched_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_switched_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->switched);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_locked_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_locked_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->locked = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_locked_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_locked_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->locked);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_lock_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_lock_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_lock_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_lock_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->lock = (svn_lock_t const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_lock_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_lock_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_lock_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_lock_t *) ((arg1)->lock);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_lock_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changelist_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_changelist_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_changelist_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_changelist_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->changelist);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_ood_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->ood_kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_ood_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->ood_kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_node_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_repos_node_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_node_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_node_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_repos_node_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->repos_node_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_text_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_repos_text_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_text_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_text_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_repos_text_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->repos_text_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_prop_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_repos_prop_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_prop_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_prop_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_repos_prop_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->repos_prop_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_lock_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_repos_lock_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_lock_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_lock_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_lock = (svn_lock_t const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_repos_lock_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_lock_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_repos_lock_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_lock_t *) ((arg1)->repos_lock);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_lock_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_changed_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_ood_changed_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->ood_changed_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_changed_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_ood_changed_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->ood_changed_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_changed_date_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_ood_changed_date_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->ood_changed_date = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_changed_date_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_ood_changed_date_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->ood_changed_date);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_changed_author_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_ood_changed_author_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->ood_changed_author) free((char *)arg1->ood_changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->ood_changed_author = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_ood_changed_author_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_ood_changed_author_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->ood_changed_author);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_moved_from_abspath_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_moved_from_abspath_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_from_abspath) free((char *)arg1->moved_from_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_from_abspath = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_moved_from_abspath_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_moved_from_abspath_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->moved_from_abspath);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_moved_to_abspath_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status3_t_moved_to_abspath_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_to_abspath) free((char *)arg1->moved_to_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_to_abspath = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_moved_to_abspath_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_moved_to_abspath_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->moved_to_abspath);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_file_external_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_file_external_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->file_external = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_file_external_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_file_external_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->file_external);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_actual_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status3_t_actual_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->actual_kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3_t_actual_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status3_t_actual_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->actual_kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_status3_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_status3_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_status3_t *)calloc(1, sizeof(struct svn_wc_status3_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_status3_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_status3_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_status3_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_status3_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_status3_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_status2_t_entry_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_wc_entry_t *arg2 = (svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_entry_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->entry = (svn_wc_entry_t const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_entry_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_entry_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_entry_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_entry_t *) ((arg1)->entry);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_entry_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_text_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_text_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->text_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_text_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_text_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->text_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_prop_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_prop_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->prop_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_prop_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_prop_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->prop_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_locked_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_locked_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->locked = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_locked_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_locked_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->locked);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_copied_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_copied_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->copied = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_copied_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_copied_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->copied);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_switched_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_switched_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->switched = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_switched_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_switched_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->switched);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_repos_text_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_repos_text_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_text_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_repos_text_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_repos_text_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->repos_text_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_repos_prop_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_repos_prop_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_prop_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_repos_prop_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_repos_prop_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->repos_prop_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_repos_lock_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_repos_lock_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_lock_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_lock_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_lock = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_repos_lock_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_lock_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_repos_lock_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_lock_t *) ((arg1)->repos_lock);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_lock_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_url_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status2_t_url_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_url_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_url_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->url);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_last_cmt_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_ood_last_cmt_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->ood_last_cmt_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_last_cmt_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_ood_last_cmt_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->ood_last_cmt_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_last_cmt_date_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  apr_time_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_ood_last_cmt_date_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->ood_last_cmt_date = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_last_cmt_date_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_time_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_ood_last_cmt_date_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->ood_last_cmt_date);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_kind_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_node_kind_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_ood_kind_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_node_kind_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->ood_kind = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_kind_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_node_kind_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_ood_kind_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_node_kind_t) ((arg1)->ood_kind);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_last_cmt_author_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"Os:svn_wc_status2_t_ood_last_cmt_author_set",&obj0,&arg2)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->ood_last_cmt_author) free((char *)arg1->ood_last_cmt_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->ood_last_cmt_author = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_ood_last_cmt_author_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_ood_last_cmt_author_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->ood_last_cmt_author);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_tree_conflict_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_wc_conflict_description_t *arg2 = (svn_wc_conflict_description_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_tree_conflict_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->tree_conflict = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_tree_conflict_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_tree_conflict_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_conflict_description_t *) ((arg1)->tree_conflict);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_conflict_description_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_file_external_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_file_external_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->file_external = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_file_external_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_file_external_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->file_external);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_pristine_text_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_pristine_text_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->pristine_text_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_pristine_text_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_pristine_text_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->pristine_text_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_pristine_prop_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status2_t_pristine_prop_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->pristine_prop_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2_t_pristine_prop_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status2_t_pristine_prop_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->pristine_prop_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_status2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_status2_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_status2_t *)calloc(1, sizeof(struct svn_wc_status2_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_status2_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_status2_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_status2_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_status2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_status2_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_status_t_entry_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  svn_wc_entry_t *arg2 = (svn_wc_entry_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_entry_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->entry = (svn_wc_entry_t const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_entry_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_wc_entry_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_entry_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_wc_entry_t *) ((arg1)->entry);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_entry_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_text_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_text_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->text_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_text_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_text_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->text_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_prop_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_prop_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->prop_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_prop_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_prop_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->prop_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_locked_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_locked_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->locked = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_locked_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_locked_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->locked);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_copied_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_copied_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->copied = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_copied_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_copied_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->copied);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_switched_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_switched_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->switched = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_switched_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_switched_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->switched);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_repos_text_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_repos_text_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_text_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_repos_text_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_repos_text_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->repos_text_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_repos_prop_status_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  enum svn_wc_status_kind arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_status_t_repos_prop_status_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_wc_status_kind)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->repos_prop_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_t_repos_prop_status_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_wc_status_kind result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_status_t_repos_prop_status_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_wc_status_kind) ((arg1)->repos_prop_status);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_status_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_status_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_status_t *)calloc(1, sizeof(struct svn_wc_status_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_status_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_status_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_status_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_status_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_status_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_dup_status3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status3_t *arg1 = (svn_wc_status3_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_status3_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_dup_status3",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_status3_t *)svn_wc_dup_status3((struct svn_wc_status3_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_status3_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_dup_status2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status2_t *arg1 = (svn_wc_status2_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_status2_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_dup_status2",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_status2_t *)svn_wc_dup_status2((struct svn_wc_status2_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_status2_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_dup_status(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status_t *arg1 = (svn_wc_status_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_wc_status_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_dup_status",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_wc_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_status_t *)svn_wc_dup_status((struct svn_wc_status_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_status_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status3_t **arg1 = (svn_wc_status3_t **) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_status3_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|OO:svn_wc_status3",&obj0,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_status3(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_status3 arg 1 (svn_wc_status3_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_status3 is not implemented yet");
    
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status2_t **arg1 = (svn_wc_status2_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_status2_t *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_status2",&arg2,&obj1,&obj2)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_status2(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_status2_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status_t **arg1 = (svn_wc_status_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_status_t *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_status",&arg2,&obj1,&obj2)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_status(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_status_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_walk_status(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_wc_status_func4_t arg8 = (svn_wc_status_func4_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_cancel_func_t arg10 = (svn_cancel_func_t) 0 ;
  void *arg11 = (void *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOOO|O:svn_wc_walk_status",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_status_func4_t * tmp =
    svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj7);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg8 = *tmp;
  }
  {
    if (obj8 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj8;
      PyErr_Clear();
    }
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg11 = obj9;
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_walk_status(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_status_editor5(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  void **arg3 = (void **) 0 ;
  svn_revnum_t *arg4 = (svn_revnum_t *) 0 ;
  svn_wc_context_t *arg5 = (svn_wc_context_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_depth_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  apr_array_header_t *arg13 = (apr_array_header_t *) 0 ;
  svn_wc_status_func4_t arg14 = (svn_wc_status_func4_t) 0 ;
  void *arg15 = (void *) 0 ;
  svn_cancel_func_t arg16 = (svn_cancel_func_t) 0 ;
  void *arg17 = (void *) 0 ;
  apr_pool_t *arg18 = (apr_pool_t *) 0 ;
  apr_pool_t *arg19 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg18 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg19 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOOOOOOO|OO:svn_wc_get_status_editor5",&obj0,&arg6,&arg7,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11,&obj12,&obj13)) SWIG_fail;
  {
    arg5 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg12 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_status_func4_t * tmp =
    svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj9);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg14 = *tmp;
  }
  {
    if (obj10 == Py_None) {
      arg15 = NULL;
    } else if (SWIG_ConvertPtr(obj10, (void **) &arg15, 0, 0) == -1) {
      arg15 = (void *) obj10;
      PyErr_Clear();
    }
  }
  {
    arg16 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg17 = obj11;
  }
  if (obj12) {
    /* Verify that the user supplied a valid pool */
    if (obj12 != Py_None && obj12 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj12);
      SWIG_arg_fail(svn_argnum_obj12);
      SWIG_fail;
    }
  }
  if (obj13) {
    /* Verify that the user supplied a valid pool */
    if (obj13 != Py_None && obj13 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj13);
      SWIG_arg_fail(svn_argnum_obj13);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_status_editor5((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,arg10,arg11,arg12,(apr_array_header_t const *)arg13,arg14,arg15,arg16,arg17,arg18,arg19);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_status_editor4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  void **arg3 = (void **) 0 ;
  svn_revnum_t *arg4 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg5 = (svn_wc_adm_access_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_array_header_t *arg10 = (apr_array_header_t *) 0 ;
  svn_wc_status_func3_t arg11 = (svn_wc_status_func3_t) 0 ;
  void *arg12 = (void *) 0 ;
  svn_cancel_func_t arg13 = (svn_cancel_func_t) 0 ;
  void *arg14 = (void *) 0 ;
  svn_wc_traversal_info_t *arg15 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg16 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg16 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOOO|O:svn_wc_get_status_editor4",&obj0,&arg6,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  {
    arg5 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_status_func3_t * tmp =
    svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj6);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg11 = *tmp;
  }
  {
    if (obj7 == Py_None) {
      arg12 = NULL;
    } else if (SWIG_ConvertPtr(obj7, (void **) &arg12, 0, 0) == -1) {
      arg12 = (void *) obj7;
      PyErr_Clear();
    }
  }
  {
    arg13 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg14 = obj8;
  }
  {
    arg15 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj9);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_status_editor4((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12,arg13,arg14,arg15,arg16);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_status_editor3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  void **arg3 = (void **) 0 ;
  svn_revnum_t *arg4 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg5 = (svn_wc_adm_access_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_array_header_t *arg10 = (apr_array_header_t *) 0 ;
  svn_wc_status_func2_t arg11 = (svn_wc_status_func2_t) 0 ;
  void *arg12 = (void *) 0 ;
  svn_cancel_func_t arg13 = (svn_cancel_func_t) 0 ;
  void *arg14 = (void *) 0 ;
  svn_wc_traversal_info_t *arg15 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg16 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg16 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOO|O:svn_wc_get_status_editor3",&obj0,&arg6,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg5 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_wc_status_func2_t) svn_swig_py_status_func2;
    arg12 = obj6;
  }
  {
    arg13 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg14 = obj7;
  }
  {
    arg15 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_status_editor3((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12,arg13,arg14,arg15,arg16);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_status_editor2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  void **arg3 = (void **) 0 ;
  svn_revnum_t *arg4 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg5 = (svn_wc_adm_access_t *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_wc_status_func2_t arg11 = (svn_wc_status_func2_t) 0 ;
  void *arg12 = (void *) 0 ;
  svn_cancel_func_t arg13 = (svn_cancel_func_t) 0 ;
  void *arg14 = (void *) 0 ;
  svn_wc_traversal_info_t *arg15 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg16 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg16 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOO|O:svn_wc_get_status_editor2",&obj0,&arg6,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg5 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg7 = svn_swig_py_struct_ptr_hash_from_dict(obj2,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_wc_status_func2_t) svn_swig_py_status_func2;
    arg12 = obj6;
  }
  {
    arg13 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg14 = obj7;
  }
  {
    arg15 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_status_editor2((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15,arg16);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_status_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  svn_revnum_t *arg3 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_wc_status_func_t arg10 = (svn_wc_status_func_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_cancel_func_t arg12 = (svn_cancel_func_t) 0 ;
  void *arg13 = (void *) 0 ;
  svn_wc_traversal_info_t *arg14 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg15 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOO|O:svn_wc_get_status_editor",&obj0,&arg5,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_struct_ptr_hash_from_dict(obj2,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_wc_status_func_t) svn_swig_py_status_func;
    arg11 = obj6;
  }
  {
    arg12 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg13 = obj7;
  }
  {
    arg14 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_status_editor((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_status_set_repos_locks(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void *arg1 = (void *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOs|O:svn_wc_status_set_repos_locks",&obj0,&obj1,&arg3,&obj3)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg1 = NULL;
    } else if (SWIG_ConvertPtr(obj0, (void **) &arg1, 0, 0) == -1) {
      arg1 = (void *) obj0;
      PyErr_Clear();
    }
  }
  {
    arg2 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_apr_hash_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_status_set_repos_locks(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_copy3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OssOOO|O:svn_wc_copy3",&obj0,&arg2,&arg3,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_copy3(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_copy2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_cancel_func_t arg4 = (svn_cancel_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_wc_notify_func2_t arg6 = (svn_wc_notify_func2_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOsOO|O:svn_wc_copy2",&arg1,&obj1,&arg3,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg5 = obj3;
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_copy2((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_copy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_cancel_func_t arg4 = (svn_cancel_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_wc_notify_func_t arg6 = (svn_wc_notify_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOsOO|O:svn_wc_copy",&arg1,&obj1,&arg3,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg5 = obj3;
  }
  {
    arg6 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_copy((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_move(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OssOOO|O:svn_wc_move",&obj0,&arg2,&arg3,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_move(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_delete4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOO|O:svn_wc_delete4",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_delete4(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_delete3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  svn_wc_notify_func2_t arg5 = (svn_wc_notify_func2_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_boolean_t arg7 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOO|O:svn_wc_delete3",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg4 = obj2;
  }
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg6 = obj3;
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_delete3((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_delete2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  svn_wc_notify_func2_t arg5 = (svn_wc_notify_func2_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOO|O:svn_wc_delete2",&arg1,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg4 = obj2;
  }
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg6 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_delete2((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_delete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  svn_wc_notify_func_t arg5 = (svn_wc_notify_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOO|O:svn_wc_delete",&arg1,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg4 = obj2;
  }
  {
    arg5 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg6 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_delete((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_from_disk3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_hash_t *arg3 = (apr_hash_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_wc_notify_func2_t arg5 = (svn_wc_notify_func2_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_add_from_disk3",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg3 = svn_swig_py_prophash_from_dict(obj2, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg6 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_from_disk3(arg1,(char const *)arg2,(apr_hash_t const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_from_disk2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_hash_t *arg3 = (apr_hash_t *) 0 ;
  svn_wc_notify_func2_t arg4 = (svn_wc_notify_func2_t) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_add_from_disk2",&obj0,&arg2,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg3 = svn_swig_py_prophash_from_dict(obj2, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg5 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_from_disk2(arg1,(char const *)arg2,(apr_hash_t const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_from_disk(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_notify_func2_t arg3 = (svn_wc_notify_func2_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsO|O:svn_wc_add_from_disk",&obj0,&arg2,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg4 = obj2;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_from_disk(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOzOOO|O:svn_wc_add4",&obj0,&arg2,&obj2,&arg4,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj5;
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add4(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_depth_t arg3 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOzOOO|O:svn_wc_add3",&arg1,&obj1,&obj2,&arg4,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj5;
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add3((char const *)arg1,arg2,arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOzOOO|O:svn_wc_add2",&arg1,&obj1,&arg3,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add2((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func_t arg7 = (svn_wc_notify_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOzOOO|O:svn_wc_add",&arg1,&obj1,&arg3,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_repos_file4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  svn_stream_t *arg4 = (svn_stream_t *) 0 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_revnum_t arg8 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOzOO|O:svn_wc_add_repos_file4",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&arg7,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_make_stream (obj2, _global_pool);
  }
  {
    arg4 = svn_swig_py_make_stream (obj3, _global_pool);
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_prophash_from_dict(obj4, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_prophash_from_dict(obj5, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj8;
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_repos_file4(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,(char const *)arg7,arg8,arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_repos_file3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  svn_stream_t *arg4 = (svn_stream_t *) 0 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_revnum_t arg8 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_wc_notify_func2_t arg11 = (svn_wc_notify_func2_t) 0 ;
  void *arg12 = (void *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg13 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOzOOO|O:svn_wc_add_repos_file3",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&arg7,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_make_stream (obj2, _global_pool);
  }
  {
    arg4 = svn_swig_py_make_stream (obj3, _global_pool);
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_prophash_from_dict(obj4, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_prophash_from_dict(obj5, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj8;
  }
  {
    arg11 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg12 = obj9;
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_repos_file3((char const *)arg1,arg2,arg3,arg4,arg5,arg6,(char const *)arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_repos_file2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_revnum_t arg8 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOszOOzO|O:svn_wc_add_repos_file2",&arg1,&obj1,&arg3,&arg4,&obj4,&obj5,&arg7,&obj7,&obj8)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_prophash_from_dict(obj4, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_prophash_from_dict(obj5, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_repos_file2((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,(char const *)arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_repos_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_revnum_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOzOzO|O:svn_wc_add_repos_file",&arg1,&obj1,&arg3,&obj3,&arg5,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg4 = svn_swig_py_prophash_from_dict(obj3, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_revnum_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_repos_file((char const *)arg1,arg2,(char const *)arg3,arg4,(char const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_remove_from_revision_control2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_remove_from_revision_control2",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_remove_from_revision_control2(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_remove_from_revision_control(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_remove_from_revision_control",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_remove_from_revision_control(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_resolved_conflict5(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  svn_wc_conflict_choice_t arg7 ;
  svn_cancel_func_t arg8 = (svn_cancel_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_wc_notify_func2_t arg10 = (svn_wc_notify_func2_t) 0 ;
  void *arg11 = (void *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOsOOOO|O:svn_wc_resolved_conflict5",&obj0,&arg2,&obj2,&obj3,&arg5,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_wc_conflict_choice_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg9 = obj7;
  }
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg11 = obj8;
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_resolved_conflict5(arg1,(char const *)arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_resolved_conflict4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_wc_conflict_choice_t arg7 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_cancel_func_t arg10 = (svn_cancel_func_t) 0 ;
  void *arg11 = (void *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOOOO|O:svn_wc_resolved_conflict4",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_wc_conflict_choice_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj7;
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg11 = obj8;
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_resolved_conflict4((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_resolved_conflict3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_depth_t arg5 ;
  svn_wc_conflict_choice_t arg6 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOOO|O:svn_wc_resolved_conflict3",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_wc_conflict_choice_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj6;
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj7;
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_resolved_conflict3((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_resolved_conflict2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func2_t arg6 = (svn_wc_notify_func2_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_cancel_func_t arg8 = (svn_cancel_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOO|O:svn_wc_resolved_conflict2",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg7 = obj5;
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg9 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_resolved_conflict2((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_resolved_conflict(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func_t arg6 = (svn_wc_notify_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOO|O:svn_wc_resolved_conflict",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg7 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_resolved_conflict((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_committed_queue_create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  svn_wc_committed_queue_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg1 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"|O:svn_wc_committed_queue_create",&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_wc_committed_queue_t *)svn_wc_committed_queue_create(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_committed_queue_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_queue_committed4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_committed_queue_t *arg1 = (svn_wc_committed_queue_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_array_header_t *arg6 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_checksum_t *arg9 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOsOOOOOO|O:svn_wc_queue_committed4",&obj0,&obj1,&arg3,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_committed_queue_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_committed_queue_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_proparray_from_dict(obj5, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_checksum_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_checksum_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_queue_committed4(arg1,arg2,(char const *)arg3,arg4,arg5,(apr_array_header_t const *)arg6,arg7,arg8,(struct svn_checksum_t const *)arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_queue_committed3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_committed_queue_t *arg1 = (svn_wc_committed_queue_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_checksum_t *arg8 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOsOOOOO|O:svn_wc_queue_committed3",&obj0,&obj1,&arg3,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg1 = (svn_wc_committed_queue_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_committed_queue_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_proparray_from_dict(obj4, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_checksum_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_checksum_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_queue_committed3(arg1,arg2,(char const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,(struct svn_checksum_t const *)arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_queue_committed2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_committed_queue_t *arg1 = (svn_wc_committed_queue_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_checksum_t *arg8 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOO|O:svn_wc_queue_committed2",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg1 = (svn_wc_committed_queue_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_committed_queue_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_proparray_from_dict(obj4, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_checksum_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_checksum_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_queue_committed2(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,(struct svn_checksum_t const *)arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_queue_committed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_committed_queue_t **arg1 = (svn_wc_committed_queue_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  unsigned char *arg8 = (unsigned char *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_committed_queue_t *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOO|O:svn_wc_queue_committed",&arg2,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_proparray_from_dict(obj3, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg8 = NULL;
    } else {
      arg8 = (unsigned char *) PyString_AsString(obj6);
      if (arg8 == NULL) SWIG_fail;
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_queue_committed(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,(unsigned char const *)arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_committed_queue_t,
        _global_py_pool, args))
    
    ;
  }
  
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_process_committed_queue2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_committed_queue_t *arg1 = (svn_wc_committed_queue_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  svn_revnum_t arg3 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOOzzO|O:svn_wc_process_committed_queue2",&obj0,&obj1,&obj2,&arg4,&arg5,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_committed_queue_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_committed_queue_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_process_committed_queue2(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_process_committed_queue(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_committed_queue_t *arg1 = (svn_wc_committed_queue_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_revnum_t arg3 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOOzz|O:svn_wc_process_committed_queue",&obj0,&obj1,&obj2,&arg4,&arg5,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_committed_queue_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_committed_queue_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_process_committed_queue(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_process_committed4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  unsigned char *arg10 = (unsigned char *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOzzOOOO|O:svn_wc_process_committed4",&arg1,&obj1,&obj2,&obj3,&arg5,&arg6,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg7 = svn_swig_py_proparray_from_dict(obj6, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    if (obj9 == Py_None) {
      arg10 = NULL;
    } else {
      arg10 = (unsigned char *) PyString_AsString(obj9);
      if (arg10 == NULL) SWIG_fail;
    }
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_process_committed4((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8,arg9,(unsigned char const *)arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_process_committed3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg8 ;
  unsigned char *arg9 = (unsigned char *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOzzOOO|O:svn_wc_process_committed3",&arg1,&obj1,&obj2,&obj3,&arg5,&arg6,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg7 = svn_swig_py_proparray_from_dict(obj6, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    if (obj8 == Py_None) {
      arg9 = NULL;
    } else {
      arg9 = (unsigned char *) PyString_AsString(obj8);
      if (arg9 == NULL) SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_process_committed3((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8,(unsigned char const *)arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_process_committed2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg8 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOzzOO|O:svn_wc_process_committed2",&arg1,&obj1,&obj2,&obj3,&arg5,&arg6,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg7 = svn_swig_py_proparray_from_dict(obj6, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_process_committed2((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_process_committed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOzzO|O:svn_wc_process_committed",&arg1,&obj1,&obj2,&obj3,&arg5,&arg6,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg7 = svn_swig_py_proparray_from_dict(obj6, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_process_committed((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_crawl_revisions5(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_ra_reporter3_t *arg3 = (svn_ra_reporter3_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_cancel_func_t arg10 = (svn_cancel_func_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_wc_notify_func2_t arg12 = (svn_wc_notify_func2_t) 0 ;
  void *arg13 = (void *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg14 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOOOO|O:svn_wc_crawl_revisions5",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg11 = obj9;
  }
  {
    arg12 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg13 = obj10;
  }
  if (obj11) {
    /* Verify that the user supplied a valid pool */
    if (obj11 != Py_None && obj11 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj11);
      SWIG_arg_fail(svn_argnum_obj11);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_crawl_revisions5(arg1,(char const *)arg2,(struct svn_ra_reporter3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_crawl_revisions4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_ra_reporter3_t *arg3 = (svn_ra_reporter3_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_wc_notify_func2_t arg10 = (svn_wc_notify_func2_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_wc_traversal_info_t *arg12 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg13 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOOOOOO|O:svn_wc_crawl_revisions4",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg11 = obj9;
  }
  {
    arg12 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj10, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj10);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj11) {
    /* Verify that the user supplied a valid pool */
    if (obj11 != Py_None && obj11 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj11);
      SWIG_arg_fail(svn_argnum_obj11);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_crawl_revisions4((char const *)arg1,arg2,(struct svn_ra_reporter3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_crawl_revisions3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_ra_reporter3_t *arg3 = (svn_ra_reporter3_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_wc_notify_func2_t arg9 = (svn_wc_notify_func2_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_wc_traversal_info_t *arg11 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOOOOO|O:svn_wc_crawl_revisions3",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg10 = obj8;
  }
  {
    arg11 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj9);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_crawl_revisions3((char const *)arg1,arg2,(struct svn_ra_reporter3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_crawl_revisions2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_ra_reporter2_t *arg3 = (svn_ra_reporter2_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_wc_traversal_info_t *arg10 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOOO|O:svn_wc_crawl_revisions2",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_ra_reporter2_t *) svn_swig_py_get_ra_reporter2();
    arg4 = obj2;
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj6;
  }
  {
    arg10 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_crawl_revisions2((char const *)arg1,arg2,(struct svn_ra_reporter2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_crawl_revisions(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_ra_reporter_t *arg3 = (svn_ra_reporter_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_notify_func_t arg8 = (svn_wc_notify_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_wc_traversal_info_t *arg10 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOOOO|O:svn_wc_crawl_revisions",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg9 = obj7;
  }
  {
    arg10 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_crawl_revisions((char const *)arg1,arg2,(struct svn_ra_reporter_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_check_root(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_node_kind_t *arg3 = (svn_node_kind_t *) 0 ;
  svn_wc_context_t *arg4 = (svn_wc_context_t *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  svn_node_kind_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_check_root",&obj0,&arg5,&obj2)) SWIG_fail;
  {
    arg4 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_check_root(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_svn_node_kind_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_is_wc_root2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_is_wc_root2",&obj0,&arg3,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_is_wc_root2(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_is_wc_root(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_is_wc_root",&arg2,&obj1,&obj2)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_is_wc_root(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_actual_target2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  char **arg2 = (char **) 0 ;
  svn_wc_context_t *arg3 = (svn_wc_context_t *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp1 ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"Os|OO:svn_wc_get_actual_target2",&obj0,&arg4,&obj2,&obj3)) SWIG_fail;
  {
    arg3 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_actual_target2((char const **)arg1,(char const **)arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg1);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_actual_target(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char **arg2 = (char **) 0 ;
  char **arg3 = (char **) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  char *temp3 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"s|O:svn_wc_get_actual_target",&arg1,&obj1)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_actual_target((char const *)arg1,(char const **)arg2,(char const **)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    PyObject *s;
    if (*arg3 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg3);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_update_editor4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  svn_revnum_t *arg3 = (svn_revnum_t *) 0 ;
  svn_wc_context_t *arg4 = (svn_wc_context_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_boolean_t arg7 ;
  svn_depth_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  svn_boolean_t arg13 ;
  char *arg14 = (char *) 0 ;
  apr_array_header_t *arg15 = (apr_array_header_t *) 0 ;
  svn_wc_dirents_func_t arg16 = (svn_wc_dirents_func_t) 0 ;
  void *arg17 = (void *) 0 ;
  svn_wc_conflict_resolver_func2_t arg18 = (svn_wc_conflict_resolver_func2_t) 0 ;
  void *arg19 = (void *) 0 ;
  svn_wc_external_update_t arg20 = (svn_wc_external_update_t) 0 ;
  void *arg21 = (void *) 0 ;
  svn_cancel_func_t arg22 = (svn_cancel_func_t) 0 ;
  void *arg23 = (void *) 0 ;
  svn_wc_notify_func2_t arg24 = (svn_wc_notify_func2_t) 0 ;
  void *arg25 = (void *) 0 ;
  apr_pool_t *arg26 = (apr_pool_t *) 0 ;
  apr_pool_t *arg27 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  PyObject * obj15 = 0 ;
  PyObject * obj16 = 0 ;
  PyObject * obj17 = 0 ;
  PyObject * obj18 = 0 ;
  PyObject * obj19 = 0 ;
  PyObject * obj20 = 0 ;
  PyObject * obj21 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg26 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg27 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOOOOOzOOOOOOOOO|OO:svn_wc_get_update_editor4",&obj0,&arg5,&arg6,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&arg14,&obj11,&obj12,&obj13,&obj14,&obj15,&obj16,&obj17,&obj18,&obj19,&obj20,&obj21)) SWIG_fail;
  {
    arg4 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg12 = (svn_boolean_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg13 = (svn_boolean_t)SWIG_As_long (obj9);
    if (SWIG_arg_fail(svn_argnum_obj9)) {
      SWIG_fail;
    }
  }
  {
    arg15 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj11,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    svn_wc_dirents_func_t * tmp =
    svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj12);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg16 = *tmp;
  }
  {
    if (obj13 == Py_None) {
      arg17 = NULL;
    } else if (SWIG_ConvertPtr(obj13, (void **) &arg17, 0, 0) == -1) {
      arg17 = (void *) obj13;
      PyErr_Clear();
    }
  }
  {
    svn_wc_conflict_resolver_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj14, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj14);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg18 = *tmp;
  }
  {
    if (obj15 == Py_None) {
      arg19 = NULL;
    } else if (SWIG_ConvertPtr(obj15, (void **) &arg19, 0, 0) == -1) {
      arg19 = (void *) obj15;
      PyErr_Clear();
    }
  }
  {
    svn_wc_external_update_t * tmp =
    svn_swig_py_must_get_ptr(obj16, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj16);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg20 = *tmp;
  }
  {
    if (obj17 == Py_None) {
      arg21 = NULL;
    } else if (SWIG_ConvertPtr(obj17, (void **) &arg21, 0, 0) == -1) {
      arg21 = (void *) obj17;
      PyErr_Clear();
    }
  }
  {
    arg22 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg23 = obj18;
  }
  {
    arg24 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg25 = obj19;
  }
  if (obj20) {
    /* Verify that the user supplied a valid pool */
    if (obj20 != Py_None && obj20 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj20);
      SWIG_arg_fail(svn_argnum_obj20);
      SWIG_fail;
    }
  }
  if (obj21) {
    /* Verify that the user supplied a valid pool */
    if (obj21 != Py_None && obj21 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj21);
      SWIG_arg_fail(svn_argnum_obj21);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_update_editor4((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,(char const *)arg14,(apr_array_header_t const *)arg15,arg16,arg17,arg18,arg19,arg20,arg21,arg22,arg23,arg24,arg25,arg26,arg27);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_update_editor3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_cancel_func_t arg10 = (svn_cancel_func_t) 0 ;
  void *arg11 = (void *) 0 ;
  svn_wc_conflict_resolver_func_t arg12 = (svn_wc_conflict_resolver_func_t) 0 ;
  void *arg13 = (void *) 0 ;
  svn_wc_get_file_t arg14 = (svn_wc_get_file_t) 0 ;
  void *arg15 = (void *) 0 ;
  char *arg16 = (char *) 0 ;
  apr_array_header_t *arg17 = (apr_array_header_t *) 0 ;
  svn_delta_editor_t **arg18 = (svn_delta_editor_t **) 0 ;
  void **arg19 = (void **) 0 ;
  svn_wc_traversal_info_t *arg20 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg21 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_delta_editor_t *temp18 ;
  void *temp19 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  PyObject * obj15 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg21 = _global_pool;
  arg1 = &temp1;
  arg18 = &temp18;
  arg19 = &temp19;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOOOOOzOO|O:svn_wc_get_update_editor3",&obj0,&arg3,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11,&arg16,&obj13,&obj14,&obj15)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj6;
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg11 = obj7;
  }
  {
    svn_wc_conflict_resolver_func_t * tmp =
    svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj8);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg12 = *tmp;
  }
  {
    if (obj9 == Py_None) {
      arg13 = NULL;
    } else if (SWIG_ConvertPtr(obj9, (void **) &arg13, 0, 0) == -1) {
      arg13 = (void *) obj9;
      PyErr_Clear();
    }
  }
  {
    svn_wc_get_file_t * tmp =
    svn_swig_py_must_get_ptr(obj10, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj10);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg14 = *tmp;
  }
  {
    if (obj11 == Py_None) {
      arg15 = NULL;
    } else if (SWIG_ConvertPtr(obj11, (void **) &arg15, 0, 0) == -1) {
      arg15 = (void *) obj11;
      PyErr_Clear();
    }
  }
  {
    arg17 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj13,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg20 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj14, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj14);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj15) {
    /* Verify that the user supplied a valid pool */
    if (obj15 != Py_None && obj15 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj15);
      SWIG_arg_fail(svn_argnum_obj15);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_update_editor3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15,(char const *)arg16,(apr_array_header_t const *)arg17,(struct svn_delta_editor_t const **)arg18,arg19,arg20,arg21);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg18, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg19, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_update_editor2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func2_t arg6 = (svn_wc_notify_func2_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_cancel_func_t arg8 = (svn_cancel_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  char *arg10 = (char *) 0 ;
  svn_delta_editor_t **arg11 = (svn_delta_editor_t **) 0 ;
  void **arg12 = (void **) 0 ;
  svn_wc_traversal_info_t *arg13 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg14 = _global_pool;
  arg1 = &temp1;
  arg11 = &temp11;
  arg12 = &temp12;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOzO|O:svn_wc_get_update_editor2",&obj0,&arg3,&obj2,&obj3,&obj4,&obj5,&arg10,&obj7,&obj8)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg7 = obj4;
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg9 = obj5;
  }
  {
    arg13 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_update_editor2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,(char const *)arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13,arg14);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg11, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_update_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func_t arg6 = (svn_wc_notify_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_cancel_func_t arg8 = (svn_cancel_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  char *arg10 = (char *) 0 ;
  svn_delta_editor_t **arg11 = (svn_delta_editor_t **) 0 ;
  void **arg12 = (void **) 0 ;
  svn_wc_traversal_info_t *arg13 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg14 = _global_pool;
  arg1 = &temp1;
  arg11 = &temp11;
  arg12 = &temp12;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOzO|O:svn_wc_get_update_editor",&obj0,&arg3,&obj2,&obj3,&obj4,&obj5,&arg10,&obj7,&obj8)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg7 = obj4;
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg9 = obj5;
  }
  {
    arg13 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_update_editor(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,(char const *)arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13,arg14);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg11, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_switch_editor4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  svn_revnum_t *arg3 = (svn_revnum_t *) 0 ;
  svn_wc_context_t *arg4 = (svn_wc_context_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_boolean_t arg8 ;
  svn_depth_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  char *arg13 = (char *) 0 ;
  apr_array_header_t *arg14 = (apr_array_header_t *) 0 ;
  svn_wc_dirents_func_t arg15 = (svn_wc_dirents_func_t) 0 ;
  void *arg16 = (void *) 0 ;
  svn_wc_conflict_resolver_func2_t arg17 = (svn_wc_conflict_resolver_func2_t) 0 ;
  void *arg18 = (void *) 0 ;
  svn_wc_external_update_t arg19 = (svn_wc_external_update_t) 0 ;
  void *arg20 = (void *) 0 ;
  svn_cancel_func_t arg21 = (svn_cancel_func_t) 0 ;
  void *arg22 = (void *) 0 ;
  svn_wc_notify_func2_t arg23 = (svn_wc_notify_func2_t) 0 ;
  void *arg24 = (void *) 0 ;
  apr_pool_t *arg25 = (apr_pool_t *) 0 ;
  apr_pool_t *arg26 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  PyObject * obj15 = 0 ;
  PyObject * obj16 = 0 ;
  PyObject * obj17 = 0 ;
  PyObject * obj18 = 0 ;
  PyObject * obj19 = 0 ;
  PyObject * obj20 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg25 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg26 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OsssOOOOOzOOOOOOOOO|OO:svn_wc_get_switch_editor4",&obj0,&arg5,&arg6,&arg7,&obj4,&obj5,&obj6,&obj7,&obj8,&arg13,&obj10,&obj11,&obj12,&obj13,&obj14,&obj15,&obj16,&obj17,&obj18,&obj19,&obj20)) SWIG_fail;
  {
    arg4 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg12 = (svn_boolean_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg14 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj10,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    svn_wc_dirents_func_t * tmp =
    svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj11);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg15 = *tmp;
  }
  {
    if (obj12 == Py_None) {
      arg16 = NULL;
    } else if (SWIG_ConvertPtr(obj12, (void **) &arg16, 0, 0) == -1) {
      arg16 = (void *) obj12;
      PyErr_Clear();
    }
  }
  {
    svn_wc_conflict_resolver_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj13, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj13);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg17 = *tmp;
  }
  {
    if (obj14 == Py_None) {
      arg18 = NULL;
    } else if (SWIG_ConvertPtr(obj14, (void **) &arg18, 0, 0) == -1) {
      arg18 = (void *) obj14;
      PyErr_Clear();
    }
  }
  {
    svn_wc_external_update_t * tmp =
    svn_swig_py_must_get_ptr(obj15, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj15);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg19 = *tmp;
  }
  {
    if (obj16 == Py_None) {
      arg20 = NULL;
    } else if (SWIG_ConvertPtr(obj16, (void **) &arg20, 0, 0) == -1) {
      arg20 = (void *) obj16;
      PyErr_Clear();
    }
  }
  {
    arg21 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg22 = obj17;
  }
  {
    arg23 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg24 = obj18;
  }
  if (obj19) {
    /* Verify that the user supplied a valid pool */
    if (obj19 != Py_None && obj19 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj19);
      SWIG_arg_fail(svn_argnum_obj19);
      SWIG_fail;
    }
  }
  if (obj20) {
    /* Verify that the user supplied a valid pool */
    if (obj20 != Py_None && obj20 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj20);
      SWIG_arg_fail(svn_argnum_obj20);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_switch_editor4((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,arg10,arg11,arg12,(char const *)arg13,(apr_array_header_t const *)arg14,arg15,arg16,arg17,arg18,arg19,arg20,arg21,arg22,arg23,arg24,arg25,arg26);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_switch_editor3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_wc_notify_func2_t arg9 = (svn_wc_notify_func2_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_cancel_func_t arg11 = (svn_cancel_func_t) 0 ;
  void *arg12 = (void *) 0 ;
  svn_wc_conflict_resolver_func_t arg13 = (svn_wc_conflict_resolver_func_t) 0 ;
  void *arg14 = (void *) 0 ;
  char *arg15 = (char *) 0 ;
  apr_array_header_t *arg16 = (apr_array_header_t *) 0 ;
  svn_delta_editor_t **arg17 = (svn_delta_editor_t **) 0 ;
  void **arg18 = (void **) 0 ;
  svn_wc_traversal_info_t *arg19 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg20 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_delta_editor_t *temp17 ;
  void *temp18 ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg20 = _global_pool;
  arg1 = &temp1;
  arg17 = &temp17;
  arg18 = &temp18;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOOOOOOzOO|O:svn_wc_get_switch_editor3",&obj0,&arg3,&arg4,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&arg15,&obj12,&obj13,&obj14)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg10 = obj7;
  }
  {
    arg11 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg12 = obj8;
  }
  {
    svn_wc_conflict_resolver_func_t * tmp =
    svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj9);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg13 = *tmp;
  }
  {
    if (obj10 == Py_None) {
      arg14 = NULL;
    } else if (SWIG_ConvertPtr(obj10, (void **) &arg14, 0, 0) == -1) {
      arg14 = (void *) obj10;
      PyErr_Clear();
    }
  }
  {
    arg16 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj12,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg19 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj13, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj13);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj14) {
    /* Verify that the user supplied a valid pool */
    if (obj14 != Py_None && obj14 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj14);
      SWIG_arg_fail(svn_argnum_obj14);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_switch_editor3(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,(char const *)arg15,(apr_array_header_t const *)arg16,(struct svn_delta_editor_t const **)arg17,arg18,arg19,arg20);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg17, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg18, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_switch_editor2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  char *arg11 = (char *) 0 ;
  svn_delta_editor_t **arg12 = (svn_delta_editor_t **) 0 ;
  void **arg13 = (void **) 0 ;
  svn_wc_traversal_info_t *arg14 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg15 = _global_pool;
  arg1 = &temp1;
  arg12 = &temp12;
  arg13 = &temp13;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOOzO|O:svn_wc_get_switch_editor2",&obj0,&arg3,&arg4,&obj3,&obj4,&obj5,&obj6,&arg11,&obj8,&obj9)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj5;
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj6;
  }
  {
    arg14 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_switch_editor2(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,(char const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14,arg15);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg13, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_switch_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_wc_notify_func_t arg7 = (svn_wc_notify_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  char *arg11 = (char *) 0 ;
  svn_delta_editor_t **arg12 = (svn_delta_editor_t **) 0 ;
  void **arg13 = (void **) 0 ;
  svn_wc_traversal_info_t *arg14 = (svn_wc_traversal_info_t *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg15 = _global_pool;
  arg1 = &temp1;
  arg12 = &temp12;
  arg13 = &temp13;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOOzO|O:svn_wc_get_switch_editor",&obj0,&arg3,&arg4,&obj3,&obj4,&obj5,&obj6,&arg11,&obj8,&obj9)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg8 = obj5;
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj6;
  }
  {
    arg14 = (svn_wc_traversal_info_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_traversal_info_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_switch_editor(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,(char const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14,arg15);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg13, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_list2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|OO:svn_wc_prop_list2",&obj0,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_list2(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg1));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_list(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_prop_list",&arg2,&obj1,&obj2)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_list(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg1));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_pristine_props(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|OO:svn_wc_get_pristine_props",&obj0,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_pristine_props(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg1));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_get2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Oss|OO:svn_wc_prop_get2",&obj0,&arg3,&arg4,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_get2((struct svn_string_t const **)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromStringAndSize((*arg1)->data, (*arg1)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t *temp1 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"ssO|O:svn_wc_prop_get",&arg2,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_get((struct svn_string_t const **)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromStringAndSize((*arg1)->data, (*arg1)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_set4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_cancel_func_t arg8 = (svn_cancel_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_wc_notify_func2_t arg10 = (svn_wc_notify_func2_t) 0 ;
  void *arg11 = (void *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOOOO|O:svn_wc_prop_set4",&obj0,&arg2,&arg3,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None)
    arg4 = NULL;
    else {
      if (!PyString_Check(obj3)) {
        PyErr_SetString(PyExc_TypeError, "not a string");
        SWIG_fail;
      }
      value4.data = PyString_AS_STRING(obj3);
      value4.len = PyString_GET_SIZE(obj3);
      arg4 = &value4;
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg9 = obj7;
  }
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg11 = obj8;
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_set4(arg1,(char const *)arg2,(char const *)arg3,(struct svn_string_t const *)arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_set3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func2_t arg6 = (svn_wc_notify_func2_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value2 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOsOOO|O:svn_wc_prop_set3",&arg1,&obj1,&arg3,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    if (obj1 == Py_None)
    arg2 = NULL;
    else {
      if (!PyString_Check(obj1)) {
        PyErr_SetString(PyExc_TypeError, "not a string");
        SWIG_fail;
      }
      value2.data = PyString_AS_STRING(obj1);
      value2.len = PyString_GET_SIZE(obj1);
      arg2 = &value2;
    }
  }
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg7 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_set3((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_set2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value2 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOsOO|O:svn_wc_prop_set2",&arg1,&obj1,&arg3,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    if (obj1 == Py_None)
    arg2 = NULL;
    else {
      if (!PyString_Check(obj1)) {
        PyErr_SetString(PyExc_TypeError, "not a string");
        SWIG_fail;
      }
      value2.data = PyString_AS_STRING(obj1);
      value2.len = PyString_GET_SIZE(obj1);
      arg2 = &value2;
    }
  }
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_set2((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_prop_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value2 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOsO|O:svn_wc_prop_set",&arg1,&obj1,&arg3,&obj3,&obj4)) SWIG_fail;
  {
    if (obj1 == Py_None)
    arg2 = NULL;
    else {
      if (!PyString_Check(obj1)) {
        PyErr_SetString(PyExc_TypeError, "not a string");
        SWIG_fail;
      }
      value2.data = PyString_AS_STRING(obj1);
      value2.len = PyString_GET_SIZE(obj1);
      arg2 = &value2;
    }
  }
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_prop_set((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_is_normal_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"s:svn_wc_is_normal_prop",&arg1)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_boolean_t)svn_wc_is_normal_prop((char const *)arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_is_wc_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"s:svn_wc_is_wc_prop",&arg1)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_boolean_t)svn_wc_is_wc_prop((char const *)arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_is_entry_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"s:svn_wc_is_entry_prop",&arg1)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_boolean_t)svn_wc_is_entry_prop((char const *)arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_canonicalize_svn_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_node_kind_t arg5 ;
  svn_boolean_t arg6 ;
  svn_wc_canonicalize_svn_prop_get_file_t arg7 = (svn_wc_canonicalize_svn_prop_get_file_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t *temp1 ;
  svn_string_t value3 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sOsOOOO|O:svn_wc_canonicalize_svn_prop",&arg2,&obj1,&arg4,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    if (obj1 == Py_None)
    arg3 = NULL;
    else {
      if (!PyString_Check(obj1)) {
        PyErr_SetString(PyExc_TypeError, "not a string");
        SWIG_fail;
      }
      value3.data = PyString_AS_STRING(obj1);
      value3.len = PyString_GET_SIZE(obj1);
      arg3 = &value3;
    }
  }
  {
    arg5 = (svn_node_kind_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    svn_wc_canonicalize_svn_prop_get_file_t * tmp =
    svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj5);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg7 = *tmp;
  }
  {
    if (obj6 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_canonicalize_svn_prop((struct svn_string_t const **)arg1,(char const *)arg2,(struct svn_string_t const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromStringAndSize((*arg1)->data, (*arg1)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_diff_editor6(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  svn_wc_context_t *arg3 = (svn_wc_context_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  apr_array_header_t *arg13 = (apr_array_header_t *) 0 ;
  svn_wc_diff_callbacks4_t *arg14 = (svn_wc_diff_callbacks4_t *) 0 ;
  void *arg15 = (void *) 0 ;
  svn_cancel_func_t arg16 = (svn_cancel_func_t) 0 ;
  void *arg17 = (void *) 0 ;
  apr_pool_t *arg18 = (apr_pool_t *) 0 ;
  apr_pool_t *arg19 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  PyObject * obj15 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg18 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg19 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOOOOOOOOO|OO:svn_wc_get_diff_editor6",&obj0,&arg4,&arg5,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11,&obj12,&obj13,&obj14,&obj15)) SWIG_fail;
  {
    arg3 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_boolean_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg12 = (svn_boolean_t)SWIG_As_long (obj9);
    if (SWIG_arg_fail(svn_argnum_obj9)) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj10, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj10);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg14 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj11);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj12 == Py_None) {
      arg15 = NULL;
    } else if (SWIG_ConvertPtr(obj12, (void **) &arg15, 0, 0) == -1) {
      arg15 = (void *) obj12;
      PyErr_Clear();
    }
  }
  {
    arg16 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg17 = obj13;
  }
  if (obj14) {
    /* Verify that the user supplied a valid pool */
    if (obj14 != Py_None && obj14 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj14);
      SWIG_arg_fail(svn_argnum_obj14);
      SWIG_fail;
    }
  }
  if (obj15) {
    /* Verify that the user supplied a valid pool */
    if (obj15 != Py_None && obj15 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj15);
      SWIG_arg_fail(svn_argnum_obj15);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_diff_editor6((struct svn_delta_editor_t const **)arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,(apr_array_header_t const *)arg13,(struct svn_wc_diff_callbacks4_t const *)arg14,arg15,arg16,arg17,arg18,arg19);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_diff_editor5(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks3_t *arg3 = (svn_wc_diff_callbacks3_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  svn_delta_editor_t **arg12 = (svn_delta_editor_t **) 0 ;
  void **arg13 = (void **) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg14 = _global_pool;
  arg12 = &temp12;
  arg13 = &temp13;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOOO|O:svn_wc_get_diff_editor5",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj8;
  }
  {
    arg11 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj9);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_diff_editor5(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg13, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_diff_editor4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks2_t *arg3 = (svn_wc_diff_callbacks2_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  svn_delta_editor_t **arg12 = (svn_delta_editor_t **) 0 ;
  void **arg13 = (void **) 0 ;
  apr_pool_t *arg14 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg14 = _global_pool;
  arg12 = &temp12;
  arg13 = &temp13;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOO|O:svn_wc_get_diff_editor4",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_py_setup_wc_diff_callbacks2(&arg4, obj2,
      _global_pool);
    arg4 = obj2;
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj7;
  }
  {
    arg11 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_diff_editor4(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg13, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_diff_editor3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks2_t *arg3 = (svn_wc_diff_callbacks2_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_delta_editor_t **arg11 = (svn_delta_editor_t **) 0 ;
  void **arg12 = (void **) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg13 = _global_pool;
  arg11 = &temp11;
  arg12 = &temp12;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOO|O:svn_wc_get_diff_editor3",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_py_setup_wc_diff_callbacks2(&arg4, obj2,
      _global_pool);
    arg4 = obj2;
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj7;
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_diff_editor3(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg11, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_diff_editor2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks_t *arg3 = (svn_wc_diff_callbacks_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_delta_editor_t **arg11 = (svn_delta_editor_t **) 0 ;
  void **arg12 = (void **) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg13 = _global_pool;
  arg11 = &temp11;
  arg12 = &temp12;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOO|O:svn_wc_get_diff_editor2",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg10 = obj8;
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_diff_editor2(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg11, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg12, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_diff_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks_t *arg3 = (svn_wc_diff_callbacks_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_cancel_func_t arg8 = (svn_cancel_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_delta_editor_t **arg10 = (svn_delta_editor_t **) 0 ;
  void **arg11 = (void **) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp10 ;
  void *temp11 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  arg10 = &temp10;
  arg11 = &temp11;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOO|O:svn_wc_get_diff_editor",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg9 = obj7;
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_diff_editor(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,(struct svn_delta_editor_t const **)arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg10, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg11, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff6(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks4_t *arg3 = (svn_wc_diff_callbacks4_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  apr_array_header_t *arg9 = (apr_array_header_t *) 0 ;
  svn_cancel_func_t arg10 = (svn_cancel_func_t) 0 ;
  void *arg11 = (void *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOOO|O:svn_wc_diff6",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg11 = obj9;
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff6(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks4_t const *)arg3,arg4,arg5,arg6,arg7,arg8,(apr_array_header_t const *)arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff5(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks3_t *arg3 = (svn_wc_diff_callbacks3_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOO|O:svn_wc_diff5",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff5(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks3_t const *)arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks2_t *arg3 = (svn_wc_diff_callbacks2_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOO|O:svn_wc_diff4",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_py_setup_wc_diff_callbacks2(&arg4, obj2,
      _global_pool);
    arg4 = obj2;
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff4(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks2_t *arg3 = (svn_wc_diff_callbacks2_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_diff3",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_py_setup_wc_diff_callbacks2(&arg4, obj2,
      _global_pool);
    arg4 = obj2;
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff3(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks_t *arg3 = (svn_wc_diff_callbacks_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOO|O:svn_wc_diff2",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff2(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_diff_callbacks_t *arg3 = (svn_wc_diff_callbacks_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_diff",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_prop_diffs2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_hash_t **arg2 = (apr_hash_t **) 0 ;
  svn_wc_context_t *arg3 = (svn_wc_context_t *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp1 ;
  apr_hash_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"Os|OO:svn_wc_get_prop_diffs2",&obj0,&arg4,&obj2,&obj3)) SWIG_fail;
  {
    arg3 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_prop_diffs2(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_proparray_to_dict(*arg1));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg2));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_prop_diffs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_hash_t **arg2 = (apr_hash_t **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp1 ;
  apr_hash_t *temp2 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_get_prop_diffs",&arg3,&obj1,&obj2)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_prop_diffs(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_proparray_to_dict(*arg1));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg2));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge5(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  enum svn_wc_merge_outcome_t *arg1 = (enum svn_wc_merge_outcome_t *) 0 ;
  enum svn_wc_notify_state_t *arg2 = (enum svn_wc_notify_state_t *) 0 ;
  svn_wc_context_t *arg3 = (svn_wc_context_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  svn_wc_conflict_version_t *arg10 = (svn_wc_conflict_version_t *) 0 ;
  svn_wc_conflict_version_t *arg11 = (svn_wc_conflict_version_t *) 0 ;
  svn_boolean_t arg12 ;
  char *arg13 = (char *) 0 ;
  apr_array_header_t *arg14 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg15 = (apr_hash_t *) 0 ;
  apr_array_header_t *arg16 = (apr_array_header_t *) 0 ;
  svn_wc_conflict_resolver_func2_t arg17 = (svn_wc_conflict_resolver_func2_t) 0 ;
  void *arg18 = (void *) 0 ;
  svn_cancel_func_t arg19 = (svn_cancel_func_t) 0 ;
  void *arg20 = (void *) 0 ;
  apr_pool_t *arg21 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  PyObject * obj15 = 0 ;
  PyObject * obj16 = 0 ;
  PyObject * obj17 = 0 ;
  PyObject * obj18 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg21 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OOssssssOOOzOOOOOO|O:svn_wc_merge5",&obj0,&obj1,&arg4,&arg5,&arg6,&arg7,&arg8,&arg9,&obj8,&obj9,&obj10,&arg13,&obj12,&obj13,&obj14,&obj15,&obj16,&obj17,&obj18)) SWIG_fail;
  {
    arg2 = (enum svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj9);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg12 = (svn_boolean_t)SWIG_As_long (obj10);
    if (SWIG_arg_fail(svn_argnum_obj10)) {
      SWIG_fail;
    }
  }
  {
    arg14 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj12);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg15 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj13, SWIGTYPE_p_apr_hash_t, svn_argnum_obj13);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg16 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj14, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj14);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_conflict_resolver_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj15, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj15);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg17 = *tmp;
  }
  {
    if (obj16 == Py_None) {
      arg18 = NULL;
    } else if (SWIG_ConvertPtr(obj16, (void **) &arg18, 0, 0) == -1) {
      arg18 = (void *) obj16;
      PyErr_Clear();
    }
  }
  {
    arg19 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg20 = obj17;
  }
  if (obj18) {
    /* Verify that the user supplied a valid pool */
    if (obj18 != Py_None && obj18 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj18);
      SWIG_arg_fail(svn_argnum_obj18);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge5(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,(struct svn_wc_conflict_version_t const *)arg10,(struct svn_wc_conflict_version_t const *)arg11,arg12,(char const *)arg13,(apr_array_header_t const *)arg14,arg15,(apr_array_header_t const *)arg16,arg17,arg18,arg19,arg20,arg21);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  enum svn_wc_merge_outcome_t *arg1 = (enum svn_wc_merge_outcome_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  svn_wc_conflict_version_t *arg9 = (svn_wc_conflict_version_t *) 0 ;
  svn_wc_conflict_version_t *arg10 = (svn_wc_conflict_version_t *) 0 ;
  svn_boolean_t arg11 ;
  char *arg12 = (char *) 0 ;
  apr_array_header_t *arg13 = (apr_array_header_t *) 0 ;
  apr_array_header_t *arg14 = (apr_array_header_t *) 0 ;
  svn_wc_conflict_resolver_func2_t arg15 = (svn_wc_conflict_resolver_func2_t) 0 ;
  void *arg16 = (void *) 0 ;
  svn_cancel_func_t arg17 = (svn_cancel_func_t) 0 ;
  void *arg18 = (void *) 0 ;
  apr_pool_t *arg19 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  PyObject * obj15 = 0 ;
  PyObject * obj16 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg19 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OssssssOOOzOOOOO|O:svn_wc_merge4",&obj0,&arg3,&arg4,&arg5,&arg6,&arg7,&arg8,&obj7,&obj8,&obj9,&arg12,&obj11,&obj12,&obj13,&obj14,&obj15,&obj16)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_boolean_t)SWIG_As_long (obj9);
    if (SWIG_arg_fail(svn_argnum_obj9)) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj11);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg14 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj12);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_conflict_resolver_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj13, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj13);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg15 = *tmp;
  }
  {
    if (obj14 == Py_None) {
      arg16 = NULL;
    } else if (SWIG_ConvertPtr(obj14, (void **) &arg16, 0, 0) == -1) {
      arg16 = (void *) obj14;
      PyErr_Clear();
    }
  }
  {
    arg17 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg18 = obj15;
  }
  if (obj16) {
    /* Verify that the user supplied a valid pool */
    if (obj16 != Py_None && obj16 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj16);
      SWIG_arg_fail(svn_argnum_obj16);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge4(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(struct svn_wc_conflict_version_t const *)arg9,(struct svn_wc_conflict_version_t const *)arg10,arg11,(char const *)arg12,(apr_array_header_t const *)arg13,(apr_array_header_t const *)arg14,arg15,arg16,arg17,arg18,arg19);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  enum svn_wc_merge_outcome_t *arg1 = (enum svn_wc_merge_outcome_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_wc_adm_access_t *arg5 = (svn_wc_adm_access_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  svn_boolean_t arg9 ;
  char *arg10 = (char *) 0 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  svn_wc_conflict_resolver_func_t arg13 = (svn_wc_conflict_resolver_func_t) 0 ;
  void *arg14 = (void *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj3 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg15 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sssOsssOzOOOO|O:svn_wc_merge3",&arg2,&arg3,&arg4,&obj3,&arg6,&arg7,&arg8,&obj7,&arg10,&obj9,&obj10,&obj11,&obj12,&obj13)) SWIG_fail;
  {
    arg5 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj9);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg12 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj10, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj10);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_conflict_resolver_func_t * tmp =
    svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj11);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg13 = *tmp;
  }
  {
    if (obj12 == Py_None) {
      arg14 = NULL;
    } else if (SWIG_ConvertPtr(obj12, (void **) &arg14, 0, 0) == -1) {
      arg14 = (void *) obj12;
      PyErr_Clear();
    }
  }
  if (obj13) {
    /* Verify that the user supplied a valid pool */
    if (obj13 != Py_None && obj13 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj13);
      SWIG_arg_fail(svn_argnum_obj13);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge3(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,(char const *)arg10,(apr_array_header_t const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14,arg15);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  enum svn_wc_merge_outcome_t *arg1 = (enum svn_wc_merge_outcome_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_wc_adm_access_t *arg5 = (svn_wc_adm_access_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  svn_boolean_t arg9 ;
  char *arg10 = (char *) 0 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj3 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sssOsssOzO|O:svn_wc_merge2",&arg2,&arg3,&arg4,&obj3,&arg6,&arg7,&arg8,&obj7,&arg10,&obj9,&obj10)) SWIG_fail;
  {
    arg5 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj9, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj9);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj10);
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge2(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,(char const *)arg10,(apr_array_header_t const *)arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_boolean_t arg8 ;
  enum svn_wc_merge_outcome_t *arg9 = (enum svn_wc_merge_outcome_t *) 0 ;
  char *arg10 = (char *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  enum svn_wc_merge_outcome_t temp9 ;
  int res9 = SWIG_TMPOBJ ;
  PyObject * obj3 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  arg9 = &temp9;
  if (!PyArg_ParseTuple(args,(char *)"sssOsssOz|O:svn_wc_merge",&arg1,&arg2,&arg3,&obj3,&arg5,&arg6,&arg7,&obj7,&arg10,&obj9)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res9)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg9)));
  } else {
    int new_flags = SWIG_IsNewObj(res9) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg9), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge_props3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_notify_state_t *arg1 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_conflict_version_t *arg4 = (svn_wc_conflict_version_t *) 0 ;
  svn_wc_conflict_version_t *arg5 = (svn_wc_conflict_version_t *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  apr_array_header_t *arg7 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg8 ;
  svn_wc_conflict_resolver_func2_t arg9 = (svn_wc_conflict_resolver_func2_t) 0 ;
  void *arg10 = (void *) 0 ;
  svn_cancel_func_t arg11 = (svn_cancel_func_t) 0 ;
  void *arg12 = (void *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg13 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOsOOOOOOOO|O:svn_wc_merge_props3",&obj0,&obj1,&arg3,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11)) SWIG_fail;
  {
    arg1 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_wc_conflict_version_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_wc_conflict_version_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_prophash_from_dict(obj5, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    svn_wc_conflict_resolver_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj8);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg9 = *tmp;
  }
  {
    if (obj9 == Py_None) {
      arg10 = NULL;
    } else if (SWIG_ConvertPtr(obj9, (void **) &arg10, 0, 0) == -1) {
      arg10 = (void *) obj9;
      PyErr_Clear();
    }
  }
  {
    arg11 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg12 = obj10;
  }
  if (obj11) {
    /* Verify that the user supplied a valid pool */
    if (obj11 != Py_None && obj11 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj11);
      SWIG_arg_fail(svn_argnum_obj11);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge_props3(arg1,arg2,(char const *)arg3,(struct svn_wc_conflict_version_t const *)arg4,(struct svn_wc_conflict_version_t const *)arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge_props2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_notify_state_t *arg1 = (svn_wc_notify_state_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_conflict_resolver_func_t arg8 = (svn_wc_conflict_resolver_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOO|O:svn_wc_merge_props2",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg4 = svn_swig_py_prophash_from_dict(obj3, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    svn_wc_conflict_resolver_func_t * tmp =
    svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj7);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg8 = *tmp;
  }
  {
    if (obj8 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj8;
      PyErr_Clear();
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge_props2(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge_props(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_notify_state_t *arg1 = (svn_wc_notify_state_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOO|O:svn_wc_merge_props",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg4 = svn_swig_py_prophash_from_dict(obj3, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge_props(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_merge_prop_diffs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_notify_state_t *arg1 = (svn_wc_notify_state_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_array_header_t *arg4 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOO|O:svn_wc_merge_prop_diffs",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_merge_prop_diffs(arg1,(char const *)arg2,arg3,(apr_array_header_t const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_pristine_contents2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_stream_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"Os|OO:svn_wc_get_pristine_contents2",&obj0,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_pristine_contents2(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_stream_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_pristine_contents(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_stream_t *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"s|OO:svn_wc_get_pristine_contents",&arg2,&obj1,&obj2)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_pristine_contents(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_stream_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_pristine_copy_path(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char **arg2 = (char **) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"s|O:svn_wc_get_pristine_copy_path",&arg1,&obj1)) SWIG_fail;
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_pristine_copy_path((char const *)arg1,(char const **)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_cleanup4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_cancel_func_t arg7 = (svn_cancel_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  svn_wc_notify_func2_t arg9 = (svn_wc_notify_func2_t) 0 ;
  void *arg10 = (void *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOO|O:svn_wc_cleanup4",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg8 = obj6;
  }
  {
    arg9 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg10 = obj7;
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj8);
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_cleanup4(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_cleanup3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsO|O:svn_wc_cleanup3",&obj0,&arg2,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg4 = obj2;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_cleanup3(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_cleanup2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"szO|O:svn_wc_cleanup2",&arg1,&arg2,&obj2,&obj3)) SWIG_fail;
  {
    arg3 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg4 = obj2;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_cleanup2((char const *)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_cleanup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_cancel_func_t arg4 = (svn_cancel_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOzO|O:svn_wc_cleanup",&arg1,&obj1,&arg3,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg5 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_cleanup((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_upgrade(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_upgrade_get_repos_info_t arg3 = (svn_wc_upgrade_get_repos_info_t) 0 ;
  void *arg4 = (void *) 0 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOO|O:svn_wc_upgrade",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_upgrade_get_repos_info_t * tmp =
    svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj2);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg3 = *tmp;
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_upgrade(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_relocate4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_wc_relocation_validator3_t arg5 = (svn_wc_relocation_validator3_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsssOO|O:svn_wc_relocate4",&obj0,&arg2,&arg3,&arg4,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_wc_relocation_validator3_t * tmp =
    svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj4);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg5 = *tmp;
  }
  {
    if (obj5 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_relocate4(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_relocate3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_wc_relocation_validator3_t arg6 = (svn_wc_relocation_validator3_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOssOOO|O:svn_wc_relocate3",&arg1,&obj1,&arg3,&arg4,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    svn_wc_relocation_validator3_t * tmp =
    svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj5);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg6 = *tmp;
  }
  {
    if (obj6 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_relocate3((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_relocate2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_wc_relocation_validator2_t arg6 = (svn_wc_relocation_validator2_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOssOOO|O:svn_wc_relocate2",&arg1,&obj1,&arg3,&arg4,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    svn_wc_relocation_validator2_t * tmp =
    svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj5);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg6 = *tmp;
  }
  {
    if (obj6 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_relocate2((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_relocate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_wc_relocation_validator_t arg6 = (svn_wc_relocation_validator_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOssOOO|O:svn_wc_relocate",&arg1,&obj1,&arg3,&arg4,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    svn_wc_relocation_validator_t * tmp =
    svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t, svn_argnum_obj5);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg6 = *tmp;
  }
  {
    if (obj6 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_relocate((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revert5(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_cancel_func_t arg8 = (svn_cancel_func_t) 0 ;
  void *arg9 = (void *) 0 ;
  svn_wc_notify_func2_t arg10 = (svn_wc_notify_func2_t) 0 ;
  void *arg11 = (void *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOOOO|O:svn_wc_revert5",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg9 = obj7;
  }
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg11 = obj8;
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_revert5(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revert4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOO|O:svn_wc_revert4",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj5;
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_revert4(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revert3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOOO|O:svn_wc_revert3",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj5;
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_revert3((char const *)arg1,arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revert2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func2_t arg7 = (svn_wc_notify_func2_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOO|O:svn_wc_revert2",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_revert2((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revert(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_wc_notify_func_t arg7 = (svn_wc_notify_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOOOOO|O:svn_wc_revert",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj4;
  }
  {
    arg7 = (svn_wc_notify_func_t) svn_swig_py_notify_func;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_revert((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_restore(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsO|O:svn_wc_restore",&obj0,&arg2,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_restore(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_create_tmp_file2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_io_file_del_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_file_t *temp1 ;
  char *temp2 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_create_tmp_file2",&arg3,&obj1,&obj2)) SWIG_fail;
  {
    arg4 = (svn_io_file_del_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_create_tmp_file2(arg1,(char const **)arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_apr_file_t,
        _global_py_pool, args))
    
    ;
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_create_tmp_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_file_t *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_create_tmp_file",&arg2,&obj1,&obj2)) SWIG_fail;
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_create_tmp_file(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_apr_file_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_translated_file2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  apr_uint32_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp1 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"ssOO|O:svn_wc_translated_file2",&arg2,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_uint32_t)SWIG_As_unsigned_SS_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_translated_file2((char const **)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg1);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_translated_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp1 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"sOO|O:svn_wc_translated_file",&arg2,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_translated_file((char const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg1);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_translated_stream(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  apr_uint32_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_stream_t *temp1 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"ssOO|O:svn_wc_translated_stream",&arg2,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_uint32_t)SWIG_As_unsigned_SS_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_translated_stream(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_stream_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_transmit_text_deltas3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_checksum_t **arg1 = (svn_checksum_t **) 0 ;
  svn_checksum_t **arg2 = (svn_checksum_t **) 0 ;
  svn_wc_context_t *arg3 = (svn_wc_context_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_delta_editor_t *arg6 = (svn_delta_editor_t *) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_checksum_t *temp1 ;
  svn_checksum_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|OO:svn_wc_transmit_text_deltas3",&obj0,&arg4,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg3 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj4;
      PyErr_Clear();
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_transmit_text_deltas3((struct svn_checksum_t const **)arg1,(struct svn_checksum_t const **)arg2,arg3,(char const *)arg4,arg5,(struct svn_delta_editor_t const *)arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_transmit_text_deltas3 arg 1 (svn_checksum_t const **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_transmit_text_deltas3 is not implemented yet");
    
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_transmit_text_deltas3 arg 2 (svn_checksum_t const **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_transmit_text_deltas3 is not implemented yet");
    
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_transmit_text_deltas2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  unsigned char *arg2 ;
  char *arg3 = (char *) 0 ;
  svn_wc_adm_access_t *arg4 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg5 ;
  svn_delta_editor_t *arg6 = (svn_delta_editor_t *) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp1 ;
  unsigned char temp2[APR_MD5_DIGESTSIZE] ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg1 = &temp1;
  arg2 = temp2;
  if (!PyArg_ParseTuple(args,(char *)"sOOOO|O:svn_wc_transmit_text_deltas2",&arg3,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg4 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj4;
      PyErr_Clear();
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_transmit_text_deltas2((char const **)arg1,arg2,(char const *)arg3,arg4,arg5,(struct svn_delta_editor_t const *)arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg1);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, PyString_FromStringAndSize((char *)arg2, APR_MD5_DIGESTSIZE));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_transmit_text_deltas(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_delta_editor_t *arg4 = (svn_delta_editor_t *) 0 ;
  void *arg5 = (void *) 0 ;
  char **arg6 = (char **) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp6 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg6 = &temp6;
  if (!PyArg_ParseTuple(args,(char *)"sOOOO|O:svn_wc_transmit_text_deltas",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj4;
      PyErr_Clear();
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_transmit_text_deltas((char const *)arg1,arg2,arg3,(struct svn_delta_editor_t const *)arg4,arg5,(char const **)arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg6 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg6);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_transmit_prop_deltas2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_delta_editor_t *arg3 = (svn_delta_editor_t *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_transmit_prop_deltas2",&obj0,&arg2,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_transmit_prop_deltas2(arg1,(char const *)arg2,(struct svn_delta_editor_t const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_transmit_prop_deltas(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_entry_t *arg3 = (svn_wc_entry_t *) 0 ;
  svn_delta_editor_t *arg4 = (svn_delta_editor_t *) 0 ;
  void *arg5 = (void *) 0 ;
  char **arg6 = (char **) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp6 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg6 = &temp6;
  if (!PyArg_ParseTuple(args,(char *)"sOOOO|O:svn_wc_transmit_prop_deltas",&arg1,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj4;
      PyErr_Clear();
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_transmit_prop_deltas((char const *)arg1,arg2,(struct svn_wc_entry_t const *)arg3,(struct svn_delta_editor_t const *)arg4,arg5,(char const **)arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg6 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg6);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_default_ignores(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"O|O:svn_wc_get_default_ignores",&obj0,&obj1)) SWIG_fail;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg2 = svn_swig_py_struct_ptr_hash_from_dict(obj0,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_default_ignores(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_array_to_list(*arg1));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_ignores2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_hash_t *arg4 = (apr_hash_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsO|OO:svn_wc_get_ignores2",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg4 = svn_swig_py_struct_ptr_hash_from_dict(obj2,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_ignores2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_array_to_list(*arg1));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_ignores(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OO|O:svn_wc_get_ignores",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg2 = svn_swig_py_struct_ptr_hash_from_dict(obj0,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_ignores(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_array_to_list(*arg1));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_match_ignore_list(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_boolean_t result;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_match_ignore_list",&arg1,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj1,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_boolean_t)svn_wc_match_ignore_list((char const *)arg1,(apr_array_header_t const *)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_From_long((long)(result));
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_lock2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_lock_t *arg3 = (svn_lock_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsO|O:svn_wc_add_lock2",&obj0,&arg2,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_lock_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_lock_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_lock2(arg1,(char const *)arg2,(struct svn_lock_t const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_add_lock(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_lock_t *arg2 = (svn_lock_t *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sOO|O:svn_wc_add_lock",&arg1,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg2 = (svn_lock_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_lock_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_add_lock((char const *)arg1,(struct svn_lock_t const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_remove_lock2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"Os|O:svn_wc_remove_lock2",&obj0,&arg2,&obj2)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_remove_lock2(arg1,(char const *)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_remove_lock(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"sO|O:svn_wc_remove_lock",&arg1,&obj1,&obj2)) SWIG_fail;
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_remove_lock((char const *)arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_min_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_revision_status_t_min_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->min_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_min_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_revision_status_t_min_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->min_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_max_rev_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  svn_revnum_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_revision_status_t_max_rev_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->max_rev = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_max_rev_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_revnum_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_revision_status_t_max_rev_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_revnum_t) ((arg1)->max_rev);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_switched_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_revision_status_t_switched_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->switched = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_switched_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_revision_status_t_switched_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->switched);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_modified_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_revision_status_t_modified_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->modified = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_modified_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_revision_status_t_modified_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->modified);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_sparse_checkout_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  svn_boolean_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:svn_wc_revision_status_t_sparse_checkout_set",&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_boolean_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->sparse_checkout = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status_t_sparse_checkout_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_boolean_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:svn_wc_revision_status_t_sparse_checkout_get",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_boolean_t) ((arg1)->sparse_checkout);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_wc_revision_status_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_svn_wc_revision_status_t")) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_wc_revision_status_t *)calloc(1, sizeof(struct svn_wc_revision_status_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_wc_revision_status_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_wc_revision_status_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_svn_wc_revision_status_t",&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_wc_revision_status_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_revision_status_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_wc_revision_status_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_revision_status_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_revision_status2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_revision_status_t **arg1 = (svn_wc_revision_status_t **) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_revision_status_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OszOO|OO:svn_wc_revision_status2",&obj0,&arg3,&arg4,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_revision_status2(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_revision_status_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_revision_status(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_revision_status_t **arg1 = (svn_wc_revision_status_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = (svn_cancel_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_revision_status_t *temp1 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"szOO|O:svn_wc_revision_status",&arg2,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg6 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_revision_status(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_wc_revision_status_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_set_changelist2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_depth_t arg4 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  svn_wc_notify_func2_t arg8 = (svn_wc_notify_func2_t) 0 ;
  void *arg9 = (void *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OssOOOO|O:svn_wc_set_changelist2",&obj0,&arg2,&arg3,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj5;
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg9 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_set_changelist2(arg1,(char const *)arg2,(char const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_set_changelist(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_adm_access_t *arg3 = (svn_wc_adm_access_t *) 0 ;
  svn_cancel_func_t arg4 = (svn_cancel_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_wc_notify_func2_t arg6 = (svn_wc_notify_func2_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"ssOOO|O:svn_wc_set_changelist",&arg1,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg3 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg5 = obj3;
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_set_changelist((char const *)arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_get_changelists(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  apr_array_header_t *arg4 = (apr_array_header_t *) 0 ;
  svn_changelist_receiver_t arg5 = (svn_changelist_receiver_t) 0 ;
  void *arg6 = (void *) 0 ;
  svn_cancel_func_t arg7 = (svn_cancel_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOOOO|O:svn_wc_get_changelists",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_changelist_receiver_t * tmp =
    svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj4);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg5 = *tmp;
  }
  {
    if (obj5 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj5;
      PyErr_Clear();
    }
  }
  {
    arg7 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg8 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_get_changelists(arg1,(char const *)arg2,arg3,(apr_array_header_t const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_crop_tree2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_cancel_func_t arg4 = (svn_cancel_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_wc_notify_func2_t arg6 = (svn_wc_notify_func2_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_crop_tree2",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg5 = obj3;
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_crop_tree2(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_crop_tree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_adm_access_t *arg1 = (svn_wc_adm_access_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_depth_t arg3 ;
  svn_wc_notify_func2_t arg4 = (svn_wc_notify_func2_t) 0 ;
  void *arg5 = (void *) 0 ;
  svn_cancel_func_t arg6 = (svn_cancel_func_t) 0 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOOO|O:svn_wc_crop_tree",&obj0,&arg2,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg5 = obj3;
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg7 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_crop_tree(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_exclude(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_context_t *arg1 = (svn_wc_context_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  svn_wc_notify_func2_t arg5 = (svn_wc_notify_func2_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_exclude",&obj0,&arg2,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg4 = obj2;
  }
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_py_notify_func2;
    arg6 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_exclude(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_read_kind2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_node_kind_t *arg1 = (svn_node_kind_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_node_kind_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_read_kind2",&obj0,&arg3,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_read_kind2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_node_kind_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_read_kind(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_node_kind_t *arg1 = (svn_node_kind_t *) 0 ;
  svn_wc_context_t *arg2 = (svn_wc_context_t *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_node_kind_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if (!PyArg_ParseTuple(args,(char *)"OsO|O:svn_wc_read_kind",&obj0,&arg3,&obj2,&obj3)) SWIG_fail;
  {
    arg2 = (svn_wc_context_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_context_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_read_kind(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_node_kind_t, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *svn_wc_context_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_context_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_adm_access_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_adm_access_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_traversal_info_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_traversal_info_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_committed_queue_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_wc_committed_queue_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_file_opened(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_diff_callbacks4_invoke_file_opened",&obj0,&arg4,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_opened(arg1,arg2,arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_file_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_wc_notify_state_t *arg2 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = (char *) 0 ;
  char *arg11 = (char *) 0 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg13 = (apr_hash_t *) 0 ;
  void *arg14 = (void *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg15 = _global_pool;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OOOsssOOssOOO|O:svn_wc_diff_callbacks4_invoke_file_changed",&obj0,&obj1,&obj2,&arg5,&arg6,&arg7,&obj6,&obj7,&arg10,&arg11,&obj10,&obj11,&obj12,&obj13)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg12 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj10, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj10);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_apr_hash_t, svn_argnum_obj11);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj12 == Py_None) {
      arg14 = NULL;
    } else if (SWIG_ConvertPtr(obj12, (void **) &arg14, 0, 0) == -1) {
      arg14 = (void *) obj12;
      PyErr_Clear();
    }
  }
  if (obj13) {
    /* Verify that the user supplied a valid pool */
    if (obj13 != Py_None && obj13 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj13);
      SWIG_arg_fail(svn_argnum_obj13);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_changed(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14,arg15);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_file_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_wc_notify_state_t *arg2 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = (char *) 0 ;
  char *arg11 = (char *) 0 ;
  char *arg12 = (char *) 0 ;
  svn_revnum_t arg13 ;
  apr_array_header_t *arg14 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg15 = (apr_hash_t *) 0 ;
  void *arg16 = (void *) 0 ;
  apr_pool_t *arg17 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  PyObject * obj14 = 0 ;
  PyObject * obj15 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg17 = _global_pool;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OOOsssOOsszOOOO|O:svn_wc_diff_callbacks4_invoke_file_added",&obj0,&obj1,&obj2,&arg5,&arg6,&arg7,&obj6,&obj7,&arg10,&arg11,&arg12,&obj11,&obj12,&obj13,&obj14,&obj15)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg13 = (svn_revnum_t)SWIG_As_long (obj11);
    if (SWIG_arg_fail(svn_argnum_obj11)) {
      SWIG_fail;
    }
  }
  {
    arg14 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj12);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg15 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj13, SWIGTYPE_p_apr_hash_t, svn_argnum_obj13);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj14 == Py_None) {
      arg16 = NULL;
    } else if (SWIG_ConvertPtr(obj14, (void **) &arg16, 0, 0) == -1) {
      arg16 = (void *) obj14;
      PyErr_Clear();
    }
  }
  if (obj15) {
    /* Verify that the user supplied a valid pool */
    if (obj15 != Py_None && obj15 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj15);
      SWIG_arg_fail(svn_argnum_obj15);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_added(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(char const *)arg12,arg13,(apr_array_header_t const *)arg14,arg15,arg16,arg17);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_file_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_wc_notify_state_t *arg2 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  apr_hash_t *arg9 = (apr_hash_t *) 0 ;
  void *arg10 = (void *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OOsssssOO|O:svn_wc_diff_callbacks4_invoke_file_deleted",&obj0,&obj1,&arg4,&arg5,&arg6,&arg7,&arg8,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg9 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_apr_hash_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj8 == Py_None) {
      arg10 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg10, 0, 0) == -1) {
      arg10 = (void *) obj8;
      PyErr_Clear();
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj9);
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_deleted(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_dir_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_wc_notify_state_t *arg2 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OOsO|O:svn_wc_diff_callbacks4_invoke_dir_deleted",&obj0,&obj1,&arg4,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_deleted(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_dir_opened(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_revnum_t arg6 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_diff_callbacks4_invoke_dir_opened",&obj0,&arg5,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_opened(arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_dir_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_wc_notify_state_t *arg2 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg5 = (svn_boolean_t *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_revnum_t arg7 ;
  char *arg8 = (char *) 0 ;
  svn_revnum_t arg9 ;
  void *arg10 = (void *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  svn_boolean_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  if (!PyArg_ParseTuple(args,(char *)"OOsOzOO|O:svn_wc_diff_callbacks4_invoke_dir_added",&obj0,&obj1,&arg6,&obj3,&arg8,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg10 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg10, 0, 0) == -1) {
      arg10 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_added(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,(char const *)arg8,arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_dir_props_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_wc_notify_state_t *arg2 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  apr_array_header_t *arg6 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OOsOOOO|O:svn_wc_diff_callbacks4_invoke_dir_props_changed",&obj0,&obj1,&arg4,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_hash_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_props_changed(arg1,arg2,arg3,(char const *)arg4,arg5,(apr_array_header_t const *)arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks4_invoke_dir_closed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks4_t *arg1 = (svn_wc_diff_callbacks4_t *) 0 ;
  svn_wc_notify_state_t *arg2 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  void *arg7 = (void *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OOOsOO|O:svn_wc_diff_callbacks4_invoke_dir_closed",&obj0,&obj1,&obj2,&arg5,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks4_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks4_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_closed(arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_file_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg4 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg5 = (svn_boolean_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  svn_revnum_t arg9 ;
  svn_revnum_t arg10 ;
  char *arg11 = (char *) 0 ;
  char *arg12 = (char *) 0 ;
  apr_array_header_t *arg13 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg14 = (apr_hash_t *) 0 ;
  void *arg15 = (void *) 0 ;
  svn_boolean_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  svn_error_t *result = 0 ;
  
  arg5 = &temp5;
  if (!PyArg_ParseTuple(args,(char *)"OOOOsssOOssOOO:svn_wc_diff_callbacks3_invoke_file_changed",&obj0,&obj1,&obj2,&obj3,&arg6,&arg7,&arg8,&obj7,&obj8,&arg11,&arg12,&obj11,&obj12,&obj13)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_revnum_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj11);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg14 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_apr_hash_t, svn_argnum_obj12);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj13 == Py_None) {
      arg15 = NULL;
    } else if (SWIG_ConvertPtr(obj13, (void **) &arg15, 0, 0) == -1) {
      arg15 = (void *) obj13;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_file_changed(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10,(char const *)arg11,(char const *)arg12,(apr_array_header_t const *)arg13,arg14,arg15);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_file_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg4 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg5 = (svn_boolean_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  svn_revnum_t arg9 ;
  svn_revnum_t arg10 ;
  char *arg11 = (char *) 0 ;
  char *arg12 = (char *) 0 ;
  apr_array_header_t *arg13 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg14 = (apr_hash_t *) 0 ;
  void *arg15 = (void *) 0 ;
  svn_boolean_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  svn_error_t *result = 0 ;
  
  arg5 = &temp5;
  if (!PyArg_ParseTuple(args,(char *)"OOOOsssOOssOOO:svn_wc_diff_callbacks3_invoke_file_added",&obj0,&obj1,&obj2,&obj3,&arg6,&arg7,&arg8,&obj7,&obj8,&arg11,&arg12,&obj11,&obj12,&obj13)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_revnum_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj11);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg14 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_apr_hash_t, svn_argnum_obj12);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj13 == Py_None) {
      arg15 = NULL;
    } else if (SWIG_ConvertPtr(obj13, (void **) &arg15, 0, 0) == -1) {
      arg15 = (void *) obj13;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_file_added(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10,(char const *)arg11,(char const *)arg12,(apr_array_header_t const *)arg13,arg14,arg15);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_file_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  apr_hash_t *arg10 = (apr_hash_t *) 0 ;
  void *arg11 = (void *) 0 ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OOOsssssOO:svn_wc_diff_callbacks3_invoke_file_deleted",&obj0,&obj1,&obj2,&arg5,&arg6,&arg7,&arg8,&arg9,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg10 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_apr_hash_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj9 == Py_None) {
      arg11 = NULL;
    } else if (SWIG_ConvertPtr(obj9, (void **) &arg11, 0, 0) == -1) {
      arg11 = (void *) obj9;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_file_deleted(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_dir_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_revnum_t arg6 ;
  void *arg7 = (void *) 0 ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OOOsOO:svn_wc_diff_callbacks3_invoke_dir_added",&obj0,&obj1,&obj2,&arg5,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj5;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_added(arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_dir_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  void *arg6 = (void *) 0 ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OOOsO:svn_wc_diff_callbacks3_invoke_dir_deleted",&obj0,&obj1,&obj2,&arg5,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj4;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_deleted(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_dir_props_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_array_header_t *arg6 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg7 = (apr_hash_t *) 0 ;
  void *arg8 = (void *) 0 ;
  svn_boolean_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  arg4 = &temp4;
  if (!PyArg_ParseTuple(args,(char *)"OOOsOOO:svn_wc_diff_callbacks3_invoke_dir_props_changed",&obj0,&obj1,&obj2,&arg5,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_hash_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj6;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_props_changed(arg1,arg2,arg3,arg4,(char const *)arg5,(apr_array_header_t const *)arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_dir_opened(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = (void *) 0 ;
  svn_boolean_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OOsOO:svn_wc_diff_callbacks3_invoke_dir_opened",&obj0,&obj1,&arg4,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj4;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_opened(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks3_invoke_dir_closed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks3_t *arg1 = (svn_wc_diff_callbacks3_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg4 = (svn_wc_notify_state_t *) 0 ;
  svn_boolean_t *arg5 = (svn_boolean_t *) 0 ;
  char *arg6 = (char *) 0 ;
  void *arg7 = (void *) 0 ;
  svn_boolean_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  arg5 = &temp5;
  if (!PyArg_ParseTuple(args,(char *)"OOOOsO:svn_wc_diff_callbacks3_invoke_dir_closed",&obj0,&obj1,&obj2,&obj3,&arg6,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj5;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_closed(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_invoke_file_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks2_t *arg1 = (svn_wc_diff_callbacks2_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg4 = (svn_wc_notify_state_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = (char *) 0 ;
  char *arg11 = (char *) 0 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg13 = (apr_hash_t *) 0 ;
  void *arg14 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOsssOOssOOO:svn_wc_diff_callbacks2_invoke_file_changed",&obj0,&obj1,&obj2,&obj3,&arg5,&arg6,&arg7,&obj7,&obj8,&arg10,&arg11,&obj11,&obj12,&obj13)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg12 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj11);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_apr_hash_t, svn_argnum_obj12);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj13 == Py_None) {
      arg14 = NULL;
    } else if (SWIG_ConvertPtr(obj13, (void **) &arg14, 0, 0) == -1) {
      arg14 = (void *) obj13;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_file_changed(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_invoke_file_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks2_t *arg1 = (svn_wc_diff_callbacks2_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  svn_wc_notify_state_t *arg4 = (svn_wc_notify_state_t *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = (char *) 0 ;
  char *arg11 = (char *) 0 ;
  apr_array_header_t *arg12 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg13 = (apr_hash_t *) 0 ;
  void *arg14 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOsssOOssOOO:svn_wc_diff_callbacks2_invoke_file_added",&obj0,&obj1,&obj2,&obj3,&arg5,&arg6,&arg7,&obj7,&obj8,&arg10,&arg11,&obj11,&obj12,&obj13)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    arg12 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj11, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj11);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg13 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj12, SWIGTYPE_p_apr_hash_t, svn_argnum_obj12);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj13 == Py_None) {
      arg14 = NULL;
    } else if (SWIG_ConvertPtr(obj13, (void **) &arg14, 0, 0) == -1) {
      arg14 = (void *) obj13;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_file_added(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_invoke_file_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks2_t *arg1 = (svn_wc_diff_callbacks2_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  apr_hash_t *arg9 = (apr_hash_t *) 0 ;
  void *arg10 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsssssOO:svn_wc_diff_callbacks2_invoke_file_deleted",&obj0,&obj1,&obj2,&arg4,&arg5,&arg6,&arg7,&arg8,&obj8,&obj9)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg9 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj8, SWIGTYPE_p_apr_hash_t, svn_argnum_obj8);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj9 == Py_None) {
      arg10 = NULL;
    } else if (SWIG_ConvertPtr(obj9, (void **) &arg10, 0, 0) == -1) {
      arg10 = (void *) obj9;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_file_deleted(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_invoke_dir_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks2_t *arg1 = (svn_wc_diff_callbacks2_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsOO:svn_wc_diff_callbacks2_invoke_dir_added",&obj0,&obj1,&obj2,&arg4,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj5;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_dir_added(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_invoke_dir_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks2_t *arg1 = (svn_wc_diff_callbacks2_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  void *arg5 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsO:svn_wc_diff_callbacks2_invoke_dir_deleted",&obj0,&obj1,&obj2,&arg4,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj4;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_dir_deleted(arg1,arg2,arg3,(char const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks2_invoke_dir_props_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks2_t *arg1 = (svn_wc_diff_callbacks2_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  void *arg7 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsOOO:svn_wc_diff_callbacks2_invoke_dir_props_changed",&obj0,&obj1,&obj2,&arg4,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_hash_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj6;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_dir_props_changed(arg1,arg2,arg3,(char const *)arg4,(apr_array_header_t const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_invoke_file_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks_t *arg1 = (svn_wc_diff_callbacks_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_revnum_t arg7 ;
  svn_revnum_t arg8 ;
  char *arg9 = (char *) 0 ;
  char *arg10 = (char *) 0 ;
  void *arg11 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsssOOssO:svn_wc_diff_callbacks_invoke_file_changed",&obj0,&obj1,&obj2,&arg4,&arg5,&arg6,&obj6,&obj7,&arg9,&arg10,&obj10)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_revnum_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    if (obj10 == Py_None) {
      arg11 = NULL;
    } else if (SWIG_ConvertPtr(obj10, (void **) &arg11, 0, 0) == -1) {
      arg11 = (void *) obj10;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_file_changed(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7,arg8,(char const *)arg9,(char const *)arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_invoke_file_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks_t *arg1 = (svn_wc_diff_callbacks_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_revnum_t arg7 ;
  svn_revnum_t arg8 ;
  char *arg9 = (char *) 0 ;
  char *arg10 = (char *) 0 ;
  void *arg11 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsssOOssO:svn_wc_diff_callbacks_invoke_file_added",&obj0,&obj1,&obj2,&arg4,&arg5,&arg6,&obj6,&obj7,&arg9,&arg10,&obj10)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_revnum_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_revnum_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    if (obj10 == Py_None) {
      arg11 = NULL;
    } else if (SWIG_ConvertPtr(obj10, (void **) &arg11, 0, 0) == -1) {
      arg11 = (void *) obj10;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_file_added(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7,arg8,(char const *)arg9,(char const *)arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_invoke_file_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks_t *arg1 = (svn_wc_diff_callbacks_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  void *arg9 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsssssO:svn_wc_diff_callbacks_invoke_file_deleted",&obj0,&obj1,&obj2,&arg4,&arg5,&arg6,&arg7,&arg8,&obj8)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj8 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj8;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_file_deleted(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_invoke_dir_added(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks_t *arg1 = (svn_wc_diff_callbacks_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsOO:svn_wc_diff_callbacks_invoke_dir_added",&obj0,&obj1,&obj2,&arg4,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj5;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_dir_added(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_invoke_dir_deleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks_t *arg1 = (svn_wc_diff_callbacks_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  void *arg5 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsO:svn_wc_diff_callbacks_invoke_dir_deleted",&obj0,&obj1,&obj2,&arg4,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj4;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_dir_deleted(arg1,arg2,arg3,(char const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_diff_callbacks_invoke_props_changed(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_diff_callbacks_t *arg1 = (svn_wc_diff_callbacks_t *) 0 ;
  svn_wc_adm_access_t *arg2 = (svn_wc_adm_access_t *) 0 ;
  svn_wc_notify_state_t *arg3 = (svn_wc_notify_state_t *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_array_header_t *arg5 = (apr_array_header_t *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  void *arg7 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOsOOO:svn_wc_diff_callbacks_invoke_props_changed",&obj0,&obj1,&obj2,&arg4,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_wc_diff_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_diff_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_wc_adm_access_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_adm_access_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_notify_state_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_state_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_hash_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj6;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_props_changed(arg1,arg2,arg3,(char const *)arg4,(apr_array_header_t const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks2_invoke_found_entry(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_entry_callbacks2_t *arg1 = (svn_wc_entry_callbacks2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_entry_t *arg3 = (svn_wc_entry_t *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_entry_callbacks2_invoke_found_entry",&obj0,&arg2,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_entry_callbacks2_invoke_found_entry(arg1,(char const *)arg2,(struct svn_wc_entry_t const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks2_invoke_handle_error(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_entry_callbacks2_t *arg1 = (svn_wc_entry_callbacks2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_error_t *arg3 = (svn_error_t *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_entry_callbacks2_invoke_handle_error",&obj0,&arg2,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_entry_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_error_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_error_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_entry_callbacks2_invoke_handle_error(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_entry_callbacks_invoke_found_entry(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_entry_callbacks_t *arg1 = (svn_wc_entry_callbacks_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_wc_entry_t *arg3 = (svn_wc_entry_t *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OsOO|O:svn_wc_entry_callbacks_invoke_found_entry",&obj0,&arg2,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_wc_entry_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_wc_entry_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_wc_entry_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_entry_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_entry_callbacks_invoke_found_entry(arg1,(char const *)arg2,(struct svn_wc_entry_t const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_external_update(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_external_update_t arg1 = (svn_wc_external_update_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  svn_string_t *arg5 = (svn_string_t *) 0 ;
  svn_depth_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value4 ;
  svn_string_t value5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOsOOO|O:svn_wc_invoke_external_update",&obj0,&obj1,&arg3,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    svn_wc_external_update_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    if (obj3 == Py_None)
    arg4 = NULL;
    else {
      if (!PyString_Check(obj3)) {
        PyErr_SetString(PyExc_TypeError, "not a string");
        SWIG_fail;
      }
      value4.data = PyString_AS_STRING(obj3);
      value4.len = PyString_GET_SIZE(obj3);
      arg4 = &value4;
    }
  }
  {
    if (obj4 == Py_None)
    arg5 = NULL;
    else {
      if (!PyString_Check(obj4)) {
        PyErr_SetString(PyExc_TypeError, "not a string");
        SWIG_fail;
      }
      value5.data = PyString_AS_STRING(obj4);
      value5.len = PyString_GET_SIZE(obj4);
      arg5 = &value5;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_external_update(arg1,arg2,(char const *)arg3,(struct svn_string_t const *)arg4,(struct svn_string_t const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_notify_func2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_notify_func2_t arg1 = (svn_wc_notify_func2_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_wc_notify_t *arg3 = (svn_wc_notify_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOO|O:svn_wc_invoke_notify_func2",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    svn_wc_notify_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_wc_notify_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_wc_notify_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_wc_invoke_notify_func2(arg1,arg2,(struct svn_wc_notify_t const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_notify_func(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_notify_func_t arg1 = (svn_wc_notify_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_notify_action_t arg4 ;
  svn_node_kind_t arg5 ;
  char *arg6 = (char *) 0 ;
  svn_wc_notify_state_t arg7 ;
  svn_wc_notify_state_t arg8 ;
  svn_revnum_t arg9 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOsOOsOOO:svn_wc_invoke_notify_func",&obj0,&obj1,&arg3,&obj3,&obj4,&arg6,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    svn_wc_notify_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_wc_notify_action_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_node_kind_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_wc_notify_state_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_wc_notify_state_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_revnum_t)SWIG_As_long (obj8);
    if (SWIG_arg_fail(svn_argnum_obj8)) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_wc_invoke_notify_func(arg1,arg2,(char const *)arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_conflict_resolver_func2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_conflict_resolver_func2_t arg1 = (svn_wc_conflict_resolver_func2_t) 0 ;
  svn_wc_conflict_result_t **arg2 = (svn_wc_conflict_result_t **) 0 ;
  svn_wc_conflict_description2_t *arg3 = (svn_wc_conflict_description2_t *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_conflict_result_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"OOO|OO:svn_wc_invoke_conflict_resolver_func2",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    svn_wc_conflict_resolver_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg3 = (svn_wc_conflict_description2_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_conflict_description2_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_conflict_resolver_func2(arg1,arg2,(struct svn_wc_conflict_description2_t const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_invoke_conflict_resolver_func2 arg 2 (svn_wc_conflict_result_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_invoke_conflict_resolver_func2 is not implemented yet");
    
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_conflict_resolver_func(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_conflict_resolver_func_t arg1 = (svn_wc_conflict_resolver_func_t) 0 ;
  svn_wc_conflict_result_t **arg2 = (svn_wc_conflict_result_t **) 0 ;
  svn_wc_conflict_description_t *arg3 = (svn_wc_conflict_description_t *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_wc_conflict_result_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"OOO|O:svn_wc_invoke_conflict_resolver_func",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    svn_wc_conflict_resolver_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg3 = (svn_wc_conflict_description_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_wc_conflict_description_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_conflict_resolver_func(arg1,arg2,(struct svn_wc_conflict_description_t const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_invoke_conflict_resolver_func arg 2 (svn_wc_conflict_result_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_invoke_conflict_resolver_func is not implemented yet");
    
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_status_func4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status_func4_t arg1 = (svn_wc_status_func4_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_status3_t *arg4 = (svn_wc_status3_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOsO|O:svn_wc_invoke_status_func4",&obj0,&obj1,&arg3,&obj3,&obj4)) SWIG_fail;
  {
    svn_wc_status_func4_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_wc_status3_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_status3_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_status_func4(arg1,arg2,(char const *)arg3,(struct svn_wc_status3_t const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_status_func3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status_func3_t arg1 = (svn_wc_status_func3_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_status2_t *arg4 = (svn_wc_status2_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOsO|O:svn_wc_invoke_status_func3",&obj0,&obj1,&arg3,&obj3,&obj4)) SWIG_fail;
  {
    svn_wc_status_func3_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_status_func3(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_status_func2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status_func2_t arg1 = (svn_wc_status_func2_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_status2_t *arg4 = (svn_wc_status2_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOsO:svn_wc_invoke_status_func2",&obj0,&obj1,&arg3,&obj3)) SWIG_fail;
  {
    svn_wc_status_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_wc_status2_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_status2_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_wc_invoke_status_func2(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_status_func(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_status_func_t arg1 = (svn_wc_status_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_wc_status_t *arg4 = (svn_wc_status_t *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOsO:svn_wc_invoke_status_func",&obj0,&obj1,&arg3,&obj3)) SWIG_fail;
  {
    svn_wc_status_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_wc_status_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_wc_status_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_wc_invoke_status_func(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_get_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_get_file_t arg1 = (svn_wc_get_file_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  svn_stream_t *arg5 = (svn_stream_t *) 0 ;
  svn_revnum_t *arg6 = (svn_revnum_t *) 0 ;
  apr_hash_t **arg7 = (apr_hash_t **) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  apr_hash_t *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg6 = &temp6;
  arg7 = &temp7;
  if (!PyArg_ParseTuple(args,(char *)"OOsOO|O:svn_wc_invoke_get_file",&obj0,&obj1,&arg3,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    svn_wc_get_file_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_make_stream (obj4, _global_pool);
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_get_file(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  if (SWIG_IsTmpObj(res6)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg6)));
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_long, new_flags));
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg7));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_dirents_func(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_dirents_func_t arg1 = (svn_wc_dirents_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  apr_hash_t **arg3 = (apr_hash_t **) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OOss|OO:svn_wc_invoke_dirents_func",&obj0,&obj1,&arg4,&arg5,&obj4,&obj5)) SWIG_fail;
  {
    svn_wc_dirents_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_dirents_func(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg3,
        SWIGTYPE_p_svn_dirent_t,
        _global_py_pool))
    
    
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_canonicalize_svn_prop_get_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_canonicalize_svn_prop_get_file_t arg1 = (svn_wc_canonicalize_svn_prop_get_file_t) 0 ;
  svn_string_t **arg2 = (svn_string_t **) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg2 = &temp2;
  if (!PyArg_ParseTuple(args,(char *)"OOO|O:svn_wc_invoke_canonicalize_svn_prop_get_file",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    svn_wc_canonicalize_svn_prop_get_file_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg3 = svn_swig_py_make_stream (obj1, _global_pool);
  }
  {
    if (obj2 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_canonicalize_svn_prop_get_file(arg1,(struct svn_string_t const **)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromStringAndSize((*arg2)->data, (*arg2)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_upgrade_get_repos_info(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_upgrade_get_repos_info_t arg1 = (svn_wc_upgrade_get_repos_info_t) 0 ;
  char **arg2 = (char **) 0 ;
  char **arg3 = (char **) 0 ;
  void *arg4 = (void *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  char *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  if (!PyArg_ParseTuple(args,(char *)"OOs|OO:svn_wc_invoke_upgrade_get_repos_info",&obj0,&obj1,&arg5,&obj3,&obj4)) SWIG_fail;
  {
    svn_wc_upgrade_get_repos_info_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_upgrade_get_repos_info(arg1,(char const **)arg2,(char const **)arg3,arg4,(char const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    PyObject *s;
    if (*arg3 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyString_FromString(*arg3);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_relocation_validator3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_relocation_validator3_t arg1 = (svn_wc_relocation_validator3_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOzss|O:svn_wc_invoke_relocation_validator3",&obj0,&obj1,&arg3,&arg4,&arg5,&obj5)) SWIG_fail;
  {
    svn_wc_relocation_validator3_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_relocation_validator3(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_relocation_validator2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_relocation_validator2_t arg1 = (svn_wc_relocation_validator2_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOzsO|O:svn_wc_invoke_relocation_validator2",&obj0,&obj1,&arg3,&arg4,&obj4,&obj5)) SWIG_fail;
  {
    svn_wc_relocation_validator2_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_relocation_validator2(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_wc_invoke_relocation_validator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_wc_relocation_validator_t arg1 = (svn_wc_relocation_validator_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOzs:svn_wc_invoke_relocation_validator",&obj0,&obj1,&arg3,&arg4)) SWIG_fail;
  {
    svn_wc_relocation_validator_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_invoke_relocation_validator(arg1,arg2,(char const *)arg3,(char const *)arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_changelist_invoke_receiver(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_changelist_receiver_t arg1 = (svn_changelist_receiver_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"OOss|O:svn_changelist_invoke_receiver",&obj0,&obj1,&arg3,&arg4,&obj4)) SWIG_fail;
  {
    svn_changelist_receiver_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_changelist_invoke_receiver(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *svn_wc_external_update_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_notify_func2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_notify_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_conflict_resolver_func2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_conflict_resolver_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_status_func4_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_status_func3_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_status_func2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_status_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_get_file_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_dirents_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_canonicalize_svn_prop_get_file_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_upgrade_get_repos_info_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_relocation_validator3_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_relocation_validator2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_wc_relocation_validator_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_changelist_receiver_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_wc_swig_init_asp_dot_net_hack(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg1 = _global_pool;
  if (!PyArg_ParseTuple(args,(char *)"|O:svn_wc_swig_init_asp_dot_net_hack",&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_wc_swig_init_asp_dot_net_hack(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_INCREF(Py_None);
    resultobj = Py_None;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


static PyMethodDef SwigMethods[] = {
	 { (char *)"SWIG_PyInstanceMethod_New", (PyCFunction)SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { (char *)"svn_wc_version", _wrap_svn_wc_version, METH_VARARGS, (char *)"svn_wc_version() -> svn_version_t const *"},
	 { (char *)"svn_wc_context_create", _wrap_svn_wc_context_create, METH_VARARGS, (char *)"svn_wc_context_create(svn_config_t const * config, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_context_destroy", _wrap_svn_wc_context_destroy, METH_VARARGS, (char *)"svn_wc_context_destroy(svn_wc_context_t * wc_ctx) -> svn_error_t"},
	 { (char *)"svn_wc_adm_open3", _wrap_svn_wc_adm_open3, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_open3(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, int levels_to_lock, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_open2", _wrap_svn_wc_adm_open2, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_open2(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, int levels_to_lock, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_open", _wrap_svn_wc_adm_open, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_open(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, svn_boolean_t tree_lock, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_probe_open3", _wrap_svn_wc_adm_probe_open3, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_probe_open3(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, int levels_to_lock, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_probe_open2", _wrap_svn_wc_adm_probe_open2, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_probe_open2(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, int levels_to_lock, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_probe_open", _wrap_svn_wc_adm_probe_open, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_probe_open(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, svn_boolean_t tree_lock, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_open_anchor", _wrap_svn_wc_adm_open_anchor, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_open_anchor(char const * path, svn_boolean_t write_lock, int levels_to_lock, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_retrieve", _wrap_svn_wc_adm_retrieve, METH_VARARGS, (char *)"svn_wc_adm_retrieve(svn_wc_adm_access_t * associated, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_adm_probe_retrieve", _wrap_svn_wc_adm_probe_retrieve, METH_VARARGS, (char *)"svn_wc_adm_probe_retrieve(svn_wc_adm_access_t * associated, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_adm_probe_try3", _wrap_svn_wc_adm_probe_try3, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_probe_try3(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, int levels_to_lock, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_probe_try2", _wrap_svn_wc_adm_probe_try2, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_probe_try2(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, int levels_to_lock, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_probe_try", _wrap_svn_wc_adm_probe_try, METH_VARARGS, (char *)"\n"
		"svn_wc_adm_probe_try(svn_wc_adm_access_t * associated, char const * path, svn_boolean_t write_lock, svn_boolean_t tree_lock, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_adm_close2", _wrap_svn_wc_adm_close2, METH_VARARGS, (char *)"svn_wc_adm_close2(svn_wc_adm_access_t * adm_access, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_adm_close", _wrap_svn_wc_adm_close, METH_VARARGS, (char *)"svn_wc_adm_close(svn_wc_adm_access_t * adm_access) -> svn_error_t"},
	 { (char *)"svn_wc_adm_access_path", _wrap_svn_wc_adm_access_path, METH_VARARGS, (char *)"svn_wc_adm_access_path(svn_wc_adm_access_t const * adm_access) -> char const *"},
	 { (char *)"svn_wc_adm_access_pool", _wrap_svn_wc_adm_access_pool, METH_VARARGS, (char *)"svn_wc_adm_access_pool(svn_wc_adm_access_t const * adm_access) -> apr_pool_t"},
	 { (char *)"svn_wc_adm_locked", _wrap_svn_wc_adm_locked, METH_VARARGS, (char *)"svn_wc_adm_locked(svn_wc_adm_access_t const * adm_access) -> svn_boolean_t"},
	 { (char *)"svn_wc_locked2", _wrap_svn_wc_locked2, METH_VARARGS, (char *)"svn_wc_locked2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_locked", _wrap_svn_wc_locked, METH_VARARGS, (char *)"svn_wc_locked(char const * path, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_is_adm_dir", _wrap_svn_wc_is_adm_dir, METH_VARARGS, (char *)"svn_wc_is_adm_dir(char const * name, apr_pool_t pool) -> svn_boolean_t"},
	 { (char *)"svn_wc_get_adm_dir", _wrap_svn_wc_get_adm_dir, METH_VARARGS, (char *)"svn_wc_get_adm_dir(apr_pool_t pool) -> char const *"},
	 { (char *)"svn_wc_set_adm_dir", _wrap_svn_wc_set_adm_dir, METH_VARARGS, (char *)"svn_wc_set_adm_dir(char const * name, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_init_traversal_info", _wrap_svn_wc_init_traversal_info, METH_VARARGS, (char *)"svn_wc_init_traversal_info(apr_pool_t pool) -> svn_wc_traversal_info_t *"},
	 { (char *)"svn_wc_edited_externals", _wrap_svn_wc_edited_externals, METH_VARARGS, (char *)"svn_wc_edited_externals(svn_wc_traversal_info_t * traversal_info)"},
	 { (char *)"svn_wc_traversed_depths", _wrap_svn_wc_traversed_depths, METH_VARARGS, (char *)"svn_wc_traversed_depths(svn_wc_traversal_info_t * traversal_info)"},
	 { (char *)"svn_wc_external_item2_t_target_dir_set", _wrap_svn_wc_external_item2_t_target_dir_set, METH_VARARGS, (char *)"svn_wc_external_item2_t_target_dir_set(svn_wc_external_item2_t self, char const * target_dir)"},
	 { (char *)"svn_wc_external_item2_t_target_dir_get", _wrap_svn_wc_external_item2_t_target_dir_get, METH_VARARGS, (char *)"svn_wc_external_item2_t_target_dir_get(svn_wc_external_item2_t self) -> char const *"},
	 { (char *)"svn_wc_external_item2_t_url_set", _wrap_svn_wc_external_item2_t_url_set, METH_VARARGS, (char *)"svn_wc_external_item2_t_url_set(svn_wc_external_item2_t self, char const * url)"},
	 { (char *)"svn_wc_external_item2_t_url_get", _wrap_svn_wc_external_item2_t_url_get, METH_VARARGS, (char *)"svn_wc_external_item2_t_url_get(svn_wc_external_item2_t self) -> char const *"},
	 { (char *)"svn_wc_external_item2_t_revision_set", _wrap_svn_wc_external_item2_t_revision_set, METH_VARARGS, (char *)"svn_wc_external_item2_t_revision_set(svn_wc_external_item2_t self, svn_opt_revision_t revision)"},
	 { (char *)"svn_wc_external_item2_t_revision_get", _wrap_svn_wc_external_item2_t_revision_get, METH_VARARGS, (char *)"svn_wc_external_item2_t_revision_get(svn_wc_external_item2_t self) -> svn_opt_revision_t"},
	 { (char *)"svn_wc_external_item2_t_peg_revision_set", _wrap_svn_wc_external_item2_t_peg_revision_set, METH_VARARGS, (char *)"svn_wc_external_item2_t_peg_revision_set(svn_wc_external_item2_t self, svn_opt_revision_t peg_revision)"},
	 { (char *)"svn_wc_external_item2_t_peg_revision_get", _wrap_svn_wc_external_item2_t_peg_revision_get, METH_VARARGS, (char *)"svn_wc_external_item2_t_peg_revision_get(svn_wc_external_item2_t self) -> svn_opt_revision_t"},
	 { (char *)"new_svn_wc_external_item2_t", _wrap_new_svn_wc_external_item2_t, METH_VARARGS, (char *)"new_svn_wc_external_item2_t() -> svn_wc_external_item2_t"},
	 { (char *)"delete_svn_wc_external_item2_t", _wrap_delete_svn_wc_external_item2_t, METH_VARARGS, (char *)"delete_svn_wc_external_item2_t(svn_wc_external_item2_t self)"},
	 { (char *)"svn_wc_external_item2_t_swigregister", svn_wc_external_item2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_external_item2_create", _wrap_svn_wc_external_item2_create, METH_VARARGS, (char *)"svn_wc_external_item2_create(apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_external_item_create", _wrap_svn_wc_external_item_create, METH_VARARGS, (char *)"svn_wc_external_item_create(apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_external_item2_dup", _wrap_svn_wc_external_item2_dup, METH_VARARGS, (char *)"svn_wc_external_item2_dup(svn_wc_external_item2_t item, apr_pool_t pool) -> svn_wc_external_item2_t"},
	 { (char *)"svn_wc_external_item_t_target_dir_set", _wrap_svn_wc_external_item_t_target_dir_set, METH_VARARGS, (char *)"svn_wc_external_item_t_target_dir_set(svn_wc_external_item_t self, char const * target_dir)"},
	 { (char *)"svn_wc_external_item_t_target_dir_get", _wrap_svn_wc_external_item_t_target_dir_get, METH_VARARGS, (char *)"svn_wc_external_item_t_target_dir_get(svn_wc_external_item_t self) -> char const *"},
	 { (char *)"svn_wc_external_item_t_url_set", _wrap_svn_wc_external_item_t_url_set, METH_VARARGS, (char *)"svn_wc_external_item_t_url_set(svn_wc_external_item_t self, char const * url)"},
	 { (char *)"svn_wc_external_item_t_url_get", _wrap_svn_wc_external_item_t_url_get, METH_VARARGS, (char *)"svn_wc_external_item_t_url_get(svn_wc_external_item_t self) -> char const *"},
	 { (char *)"svn_wc_external_item_t_revision_set", _wrap_svn_wc_external_item_t_revision_set, METH_VARARGS, (char *)"svn_wc_external_item_t_revision_set(svn_wc_external_item_t self, svn_opt_revision_t revision)"},
	 { (char *)"svn_wc_external_item_t_revision_get", _wrap_svn_wc_external_item_t_revision_get, METH_VARARGS, (char *)"svn_wc_external_item_t_revision_get(svn_wc_external_item_t self) -> svn_opt_revision_t"},
	 { (char *)"new_svn_wc_external_item_t", _wrap_new_svn_wc_external_item_t, METH_VARARGS, (char *)"new_svn_wc_external_item_t() -> svn_wc_external_item_t"},
	 { (char *)"delete_svn_wc_external_item_t", _wrap_delete_svn_wc_external_item_t, METH_VARARGS, (char *)"delete_svn_wc_external_item_t(svn_wc_external_item_t self)"},
	 { (char *)"svn_wc_external_item_t_swigregister", svn_wc_external_item_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_external_item_dup", _wrap_svn_wc_external_item_dup, METH_VARARGS, (char *)"svn_wc_external_item_dup(svn_wc_external_item_t item, apr_pool_t pool) -> svn_wc_external_item_t"},
	 { (char *)"svn_wc_parse_externals_description3", _wrap_svn_wc_parse_externals_description3, METH_VARARGS, (char *)"\n"
		"svn_wc_parse_externals_description3(char const * defining_directory, char const * desc, svn_boolean_t canonicalize_url, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_parse_externals_description2", _wrap_svn_wc_parse_externals_description2, METH_VARARGS, (char *)"svn_wc_parse_externals_description2(char const * parent_directory, char const * desc, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_parse_externals_description", _wrap_svn_wc_parse_externals_description, METH_VARARGS, (char *)"svn_wc_parse_externals_description(char const * parent_directory, char const * desc, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_notify_t_path_set", _wrap_svn_wc_notify_t_path_set, METH_VARARGS, (char *)"svn_wc_notify_t_path_set(svn_wc_notify_t self, char const * path)"},
	 { (char *)"svn_wc_notify_t_path_get", _wrap_svn_wc_notify_t_path_get, METH_VARARGS, (char *)"svn_wc_notify_t_path_get(svn_wc_notify_t self) -> char const *"},
	 { (char *)"svn_wc_notify_t_action_set", _wrap_svn_wc_notify_t_action_set, METH_VARARGS, (char *)"svn_wc_notify_t_action_set(svn_wc_notify_t self, svn_wc_notify_action_t action)"},
	 { (char *)"svn_wc_notify_t_action_get", _wrap_svn_wc_notify_t_action_get, METH_VARARGS, (char *)"svn_wc_notify_t_action_get(svn_wc_notify_t self) -> svn_wc_notify_action_t"},
	 { (char *)"svn_wc_notify_t_kind_set", _wrap_svn_wc_notify_t_kind_set, METH_VARARGS, (char *)"svn_wc_notify_t_kind_set(svn_wc_notify_t self, svn_node_kind_t kind)"},
	 { (char *)"svn_wc_notify_t_kind_get", _wrap_svn_wc_notify_t_kind_get, METH_VARARGS, (char *)"svn_wc_notify_t_kind_get(svn_wc_notify_t self) -> svn_node_kind_t"},
	 { (char *)"svn_wc_notify_t_mime_type_set", _wrap_svn_wc_notify_t_mime_type_set, METH_VARARGS, (char *)"svn_wc_notify_t_mime_type_set(svn_wc_notify_t self, char const * mime_type)"},
	 { (char *)"svn_wc_notify_t_mime_type_get", _wrap_svn_wc_notify_t_mime_type_get, METH_VARARGS, (char *)"svn_wc_notify_t_mime_type_get(svn_wc_notify_t self) -> char const *"},
	 { (char *)"svn_wc_notify_t_lock_set", _wrap_svn_wc_notify_t_lock_set, METH_VARARGS, (char *)"svn_wc_notify_t_lock_set(svn_wc_notify_t self, svn_lock_t lock)"},
	 { (char *)"svn_wc_notify_t_lock_get", _wrap_svn_wc_notify_t_lock_get, METH_VARARGS, (char *)"svn_wc_notify_t_lock_get(svn_wc_notify_t self) -> svn_lock_t"},
	 { (char *)"svn_wc_notify_t_err_set", _wrap_svn_wc_notify_t_err_set, METH_VARARGS, (char *)"svn_wc_notify_t_err_set(svn_wc_notify_t self, svn_error_t err)"},
	 { (char *)"svn_wc_notify_t_err_get", _wrap_svn_wc_notify_t_err_get, METH_VARARGS, (char *)"svn_wc_notify_t_err_get(svn_wc_notify_t self) -> svn_error_t"},
	 { (char *)"svn_wc_notify_t_content_state_set", _wrap_svn_wc_notify_t_content_state_set, METH_VARARGS, (char *)"svn_wc_notify_t_content_state_set(svn_wc_notify_t self, svn_wc_notify_state_t content_state)"},
	 { (char *)"svn_wc_notify_t_content_state_get", _wrap_svn_wc_notify_t_content_state_get, METH_VARARGS, (char *)"svn_wc_notify_t_content_state_get(svn_wc_notify_t self) -> svn_wc_notify_state_t"},
	 { (char *)"svn_wc_notify_t_prop_state_set", _wrap_svn_wc_notify_t_prop_state_set, METH_VARARGS, (char *)"svn_wc_notify_t_prop_state_set(svn_wc_notify_t self, svn_wc_notify_state_t prop_state)"},
	 { (char *)"svn_wc_notify_t_prop_state_get", _wrap_svn_wc_notify_t_prop_state_get, METH_VARARGS, (char *)"svn_wc_notify_t_prop_state_get(svn_wc_notify_t self) -> svn_wc_notify_state_t"},
	 { (char *)"svn_wc_notify_t_lock_state_set", _wrap_svn_wc_notify_t_lock_state_set, METH_VARARGS, (char *)"svn_wc_notify_t_lock_state_set(svn_wc_notify_t self, svn_wc_notify_lock_state_t lock_state)"},
	 { (char *)"svn_wc_notify_t_lock_state_get", _wrap_svn_wc_notify_t_lock_state_get, METH_VARARGS, (char *)"svn_wc_notify_t_lock_state_get(svn_wc_notify_t self) -> svn_wc_notify_lock_state_t"},
	 { (char *)"svn_wc_notify_t_revision_set", _wrap_svn_wc_notify_t_revision_set, METH_VARARGS, (char *)"svn_wc_notify_t_revision_set(svn_wc_notify_t self, svn_revnum_t revision)"},
	 { (char *)"svn_wc_notify_t_revision_get", _wrap_svn_wc_notify_t_revision_get, METH_VARARGS, (char *)"svn_wc_notify_t_revision_get(svn_wc_notify_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_notify_t_changelist_name_set", _wrap_svn_wc_notify_t_changelist_name_set, METH_VARARGS, (char *)"svn_wc_notify_t_changelist_name_set(svn_wc_notify_t self, char const * changelist_name)"},
	 { (char *)"svn_wc_notify_t_changelist_name_get", _wrap_svn_wc_notify_t_changelist_name_get, METH_VARARGS, (char *)"svn_wc_notify_t_changelist_name_get(svn_wc_notify_t self) -> char const *"},
	 { (char *)"svn_wc_notify_t_merge_range_set", _wrap_svn_wc_notify_t_merge_range_set, METH_VARARGS, (char *)"svn_wc_notify_t_merge_range_set(svn_wc_notify_t self, svn_merge_range_t merge_range)"},
	 { (char *)"svn_wc_notify_t_merge_range_get", _wrap_svn_wc_notify_t_merge_range_get, METH_VARARGS, (char *)"svn_wc_notify_t_merge_range_get(svn_wc_notify_t self) -> svn_merge_range_t"},
	 { (char *)"svn_wc_notify_t_url_set", _wrap_svn_wc_notify_t_url_set, METH_VARARGS, (char *)"svn_wc_notify_t_url_set(svn_wc_notify_t self, char const * url)"},
	 { (char *)"svn_wc_notify_t_url_get", _wrap_svn_wc_notify_t_url_get, METH_VARARGS, (char *)"svn_wc_notify_t_url_get(svn_wc_notify_t self) -> char const *"},
	 { (char *)"svn_wc_notify_t_path_prefix_set", _wrap_svn_wc_notify_t_path_prefix_set, METH_VARARGS, (char *)"svn_wc_notify_t_path_prefix_set(svn_wc_notify_t self, char const * path_prefix)"},
	 { (char *)"svn_wc_notify_t_path_prefix_get", _wrap_svn_wc_notify_t_path_prefix_get, METH_VARARGS, (char *)"svn_wc_notify_t_path_prefix_get(svn_wc_notify_t self) -> char const *"},
	 { (char *)"svn_wc_notify_t_prop_name_set", _wrap_svn_wc_notify_t_prop_name_set, METH_VARARGS, (char *)"svn_wc_notify_t_prop_name_set(svn_wc_notify_t self, char const * prop_name)"},
	 { (char *)"svn_wc_notify_t_prop_name_get", _wrap_svn_wc_notify_t_prop_name_get, METH_VARARGS, (char *)"svn_wc_notify_t_prop_name_get(svn_wc_notify_t self) -> char const *"},
	 { (char *)"svn_wc_notify_t_rev_props_set", _wrap_svn_wc_notify_t_rev_props_set, METH_VARARGS, (char *)"svn_wc_notify_t_rev_props_set(svn_wc_notify_t self, apr_hash_t rev_props)"},
	 { (char *)"svn_wc_notify_t_rev_props_get", _wrap_svn_wc_notify_t_rev_props_get, METH_VARARGS, (char *)"svn_wc_notify_t_rev_props_get(svn_wc_notify_t self) -> apr_hash_t"},
	 { (char *)"svn_wc_notify_t_old_revision_set", _wrap_svn_wc_notify_t_old_revision_set, METH_VARARGS, (char *)"svn_wc_notify_t_old_revision_set(svn_wc_notify_t self, svn_revnum_t old_revision)"},
	 { (char *)"svn_wc_notify_t_old_revision_get", _wrap_svn_wc_notify_t_old_revision_get, METH_VARARGS, (char *)"svn_wc_notify_t_old_revision_get(svn_wc_notify_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_notify_t_hunk_original_start_set", _wrap_svn_wc_notify_t_hunk_original_start_set, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_original_start_set(svn_wc_notify_t self, svn_linenum_t hunk_original_start)"},
	 { (char *)"svn_wc_notify_t_hunk_original_start_get", _wrap_svn_wc_notify_t_hunk_original_start_get, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_original_start_get(svn_wc_notify_t self) -> svn_linenum_t"},
	 { (char *)"svn_wc_notify_t_hunk_original_length_set", _wrap_svn_wc_notify_t_hunk_original_length_set, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_original_length_set(svn_wc_notify_t self, svn_linenum_t hunk_original_length)"},
	 { (char *)"svn_wc_notify_t_hunk_original_length_get", _wrap_svn_wc_notify_t_hunk_original_length_get, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_original_length_get(svn_wc_notify_t self) -> svn_linenum_t"},
	 { (char *)"svn_wc_notify_t_hunk_modified_start_set", _wrap_svn_wc_notify_t_hunk_modified_start_set, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_modified_start_set(svn_wc_notify_t self, svn_linenum_t hunk_modified_start)"},
	 { (char *)"svn_wc_notify_t_hunk_modified_start_get", _wrap_svn_wc_notify_t_hunk_modified_start_get, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_modified_start_get(svn_wc_notify_t self) -> svn_linenum_t"},
	 { (char *)"svn_wc_notify_t_hunk_modified_length_set", _wrap_svn_wc_notify_t_hunk_modified_length_set, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_modified_length_set(svn_wc_notify_t self, svn_linenum_t hunk_modified_length)"},
	 { (char *)"svn_wc_notify_t_hunk_modified_length_get", _wrap_svn_wc_notify_t_hunk_modified_length_get, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_modified_length_get(svn_wc_notify_t self) -> svn_linenum_t"},
	 { (char *)"svn_wc_notify_t_hunk_matched_line_set", _wrap_svn_wc_notify_t_hunk_matched_line_set, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_matched_line_set(svn_wc_notify_t self, svn_linenum_t hunk_matched_line)"},
	 { (char *)"svn_wc_notify_t_hunk_matched_line_get", _wrap_svn_wc_notify_t_hunk_matched_line_get, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_matched_line_get(svn_wc_notify_t self) -> svn_linenum_t"},
	 { (char *)"svn_wc_notify_t_hunk_fuzz_set", _wrap_svn_wc_notify_t_hunk_fuzz_set, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_fuzz_set(svn_wc_notify_t self, svn_linenum_t hunk_fuzz)"},
	 { (char *)"svn_wc_notify_t_hunk_fuzz_get", _wrap_svn_wc_notify_t_hunk_fuzz_get, METH_VARARGS, (char *)"svn_wc_notify_t_hunk_fuzz_get(svn_wc_notify_t self) -> svn_linenum_t"},
	 { (char *)"new_svn_wc_notify_t", _wrap_new_svn_wc_notify_t, METH_VARARGS, (char *)"new_svn_wc_notify_t() -> svn_wc_notify_t"},
	 { (char *)"delete_svn_wc_notify_t", _wrap_delete_svn_wc_notify_t, METH_VARARGS, (char *)"delete_svn_wc_notify_t(svn_wc_notify_t self)"},
	 { (char *)"svn_wc_notify_t_swigregister", svn_wc_notify_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_create_notify", _wrap_svn_wc_create_notify, METH_VARARGS, (char *)"svn_wc_create_notify(char const * path, svn_wc_notify_action_t action, apr_pool_t pool) -> svn_wc_notify_t"},
	 { (char *)"svn_wc_create_notify_url", _wrap_svn_wc_create_notify_url, METH_VARARGS, (char *)"svn_wc_create_notify_url(char const * url, svn_wc_notify_action_t action, apr_pool_t pool) -> svn_wc_notify_t"},
	 { (char *)"svn_wc_dup_notify", _wrap_svn_wc_dup_notify, METH_VARARGS, (char *)"svn_wc_dup_notify(svn_wc_notify_t notify, apr_pool_t pool) -> svn_wc_notify_t"},
	 { (char *)"svn_wc_conflict_version_t_repos_url_set", _wrap_svn_wc_conflict_version_t_repos_url_set, METH_VARARGS, (char *)"svn_wc_conflict_version_t_repos_url_set(svn_wc_conflict_version_t self, char const * repos_url)"},
	 { (char *)"svn_wc_conflict_version_t_repos_url_get", _wrap_svn_wc_conflict_version_t_repos_url_get, METH_VARARGS, (char *)"svn_wc_conflict_version_t_repos_url_get(svn_wc_conflict_version_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_version_t_peg_rev_set", _wrap_svn_wc_conflict_version_t_peg_rev_set, METH_VARARGS, (char *)"svn_wc_conflict_version_t_peg_rev_set(svn_wc_conflict_version_t self, svn_revnum_t peg_rev)"},
	 { (char *)"svn_wc_conflict_version_t_peg_rev_get", _wrap_svn_wc_conflict_version_t_peg_rev_get, METH_VARARGS, (char *)"svn_wc_conflict_version_t_peg_rev_get(svn_wc_conflict_version_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_conflict_version_t_path_in_repos_set", _wrap_svn_wc_conflict_version_t_path_in_repos_set, METH_VARARGS, (char *)"svn_wc_conflict_version_t_path_in_repos_set(svn_wc_conflict_version_t self, char const * path_in_repos)"},
	 { (char *)"svn_wc_conflict_version_t_path_in_repos_get", _wrap_svn_wc_conflict_version_t_path_in_repos_get, METH_VARARGS, (char *)"svn_wc_conflict_version_t_path_in_repos_get(svn_wc_conflict_version_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_version_t_node_kind_set", _wrap_svn_wc_conflict_version_t_node_kind_set, METH_VARARGS, (char *)"svn_wc_conflict_version_t_node_kind_set(svn_wc_conflict_version_t self, svn_node_kind_t node_kind)"},
	 { (char *)"svn_wc_conflict_version_t_node_kind_get", _wrap_svn_wc_conflict_version_t_node_kind_get, METH_VARARGS, (char *)"svn_wc_conflict_version_t_node_kind_get(svn_wc_conflict_version_t self) -> svn_node_kind_t"},
	 { (char *)"svn_wc_conflict_version_t_repos_uuid_set", _wrap_svn_wc_conflict_version_t_repos_uuid_set, METH_VARARGS, (char *)"svn_wc_conflict_version_t_repos_uuid_set(svn_wc_conflict_version_t self, char const * repos_uuid)"},
	 { (char *)"svn_wc_conflict_version_t_repos_uuid_get", _wrap_svn_wc_conflict_version_t_repos_uuid_get, METH_VARARGS, (char *)"svn_wc_conflict_version_t_repos_uuid_get(svn_wc_conflict_version_t self) -> char const *"},
	 { (char *)"new_svn_wc_conflict_version_t", _wrap_new_svn_wc_conflict_version_t, METH_VARARGS, (char *)"new_svn_wc_conflict_version_t() -> svn_wc_conflict_version_t"},
	 { (char *)"delete_svn_wc_conflict_version_t", _wrap_delete_svn_wc_conflict_version_t, METH_VARARGS, (char *)"delete_svn_wc_conflict_version_t(svn_wc_conflict_version_t self)"},
	 { (char *)"svn_wc_conflict_version_t_swigregister", svn_wc_conflict_version_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_conflict_version_create2", _wrap_svn_wc_conflict_version_create2, METH_VARARGS, (char *)"\n"
		"svn_wc_conflict_version_create2(char const * repos_root_url, char const * repos_uuid, char const * repos_relpath, \n"
		"    svn_revnum_t revision, svn_node_kind_t kind, apr_pool_t result_pool) -> svn_wc_conflict_version_t\n"
		""},
	 { (char *)"svn_wc_conflict_version_create", _wrap_svn_wc_conflict_version_create, METH_VARARGS, (char *)"\n"
		"svn_wc_conflict_version_create(char const * repos_url, char const * path_in_repos, svn_revnum_t peg_rev, svn_node_kind_t node_kind, \n"
		"    apr_pool_t pool) -> svn_wc_conflict_version_t\n"
		""},
	 { (char *)"svn_wc_conflict_version_dup", _wrap_svn_wc_conflict_version_dup, METH_VARARGS, (char *)"svn_wc_conflict_version_dup(svn_wc_conflict_version_t version, apr_pool_t pool) -> svn_wc_conflict_version_t"},
	 { (char *)"svn_wc_conflict_description_t_path_set", _wrap_svn_wc_conflict_description_t_path_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_path_set(svn_wc_conflict_description_t self, char const * path)"},
	 { (char *)"svn_wc_conflict_description_t_path_get", _wrap_svn_wc_conflict_description_t_path_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_path_get(svn_wc_conflict_description_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_description_t_node_kind_set", _wrap_svn_wc_conflict_description_t_node_kind_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_node_kind_set(svn_wc_conflict_description_t self, svn_node_kind_t node_kind)"},
	 { (char *)"svn_wc_conflict_description_t_node_kind_get", _wrap_svn_wc_conflict_description_t_node_kind_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_node_kind_get(svn_wc_conflict_description_t self) -> svn_node_kind_t"},
	 { (char *)"svn_wc_conflict_description_t_kind_set", _wrap_svn_wc_conflict_description_t_kind_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_kind_set(svn_wc_conflict_description_t self, svn_wc_conflict_kind_t kind)"},
	 { (char *)"svn_wc_conflict_description_t_kind_get", _wrap_svn_wc_conflict_description_t_kind_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_kind_get(svn_wc_conflict_description_t self) -> svn_wc_conflict_kind_t"},
	 { (char *)"svn_wc_conflict_description_t_property_name_set", _wrap_svn_wc_conflict_description_t_property_name_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_property_name_set(svn_wc_conflict_description_t self, char const * property_name)"},
	 { (char *)"svn_wc_conflict_description_t_property_name_get", _wrap_svn_wc_conflict_description_t_property_name_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_property_name_get(svn_wc_conflict_description_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_description_t_is_binary_set", _wrap_svn_wc_conflict_description_t_is_binary_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_is_binary_set(svn_wc_conflict_description_t self, svn_boolean_t is_binary)"},
	 { (char *)"svn_wc_conflict_description_t_is_binary_get", _wrap_svn_wc_conflict_description_t_is_binary_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_is_binary_get(svn_wc_conflict_description_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_conflict_description_t_mime_type_set", _wrap_svn_wc_conflict_description_t_mime_type_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_mime_type_set(svn_wc_conflict_description_t self, char const * mime_type)"},
	 { (char *)"svn_wc_conflict_description_t_mime_type_get", _wrap_svn_wc_conflict_description_t_mime_type_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_mime_type_get(svn_wc_conflict_description_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_description_t_access_set", _wrap_svn_wc_conflict_description_t_access_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_access_set(svn_wc_conflict_description_t self, svn_wc_adm_access_t * access)"},
	 { (char *)"svn_wc_conflict_description_t_access_get", _wrap_svn_wc_conflict_description_t_access_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_access_get(svn_wc_conflict_description_t self) -> svn_wc_adm_access_t *"},
	 { (char *)"svn_wc_conflict_description_t_action_set", _wrap_svn_wc_conflict_description_t_action_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_action_set(svn_wc_conflict_description_t self, svn_wc_conflict_action_t action)"},
	 { (char *)"svn_wc_conflict_description_t_action_get", _wrap_svn_wc_conflict_description_t_action_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_action_get(svn_wc_conflict_description_t self) -> svn_wc_conflict_action_t"},
	 { (char *)"svn_wc_conflict_description_t_reason_set", _wrap_svn_wc_conflict_description_t_reason_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_reason_set(svn_wc_conflict_description_t self, svn_wc_conflict_reason_t reason)"},
	 { (char *)"svn_wc_conflict_description_t_reason_get", _wrap_svn_wc_conflict_description_t_reason_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_reason_get(svn_wc_conflict_description_t self) -> svn_wc_conflict_reason_t"},
	 { (char *)"svn_wc_conflict_description_t_base_file_set", _wrap_svn_wc_conflict_description_t_base_file_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_base_file_set(svn_wc_conflict_description_t self, char const * base_file)"},
	 { (char *)"svn_wc_conflict_description_t_base_file_get", _wrap_svn_wc_conflict_description_t_base_file_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_base_file_get(svn_wc_conflict_description_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_description_t_their_file_set", _wrap_svn_wc_conflict_description_t_their_file_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_their_file_set(svn_wc_conflict_description_t self, char const * their_file)"},
	 { (char *)"svn_wc_conflict_description_t_their_file_get", _wrap_svn_wc_conflict_description_t_their_file_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_their_file_get(svn_wc_conflict_description_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_description_t_my_file_set", _wrap_svn_wc_conflict_description_t_my_file_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_my_file_set(svn_wc_conflict_description_t self, char const * my_file)"},
	 { (char *)"svn_wc_conflict_description_t_my_file_get", _wrap_svn_wc_conflict_description_t_my_file_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_my_file_get(svn_wc_conflict_description_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_description_t_merged_file_set", _wrap_svn_wc_conflict_description_t_merged_file_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_merged_file_set(svn_wc_conflict_description_t self, char const * merged_file)"},
	 { (char *)"svn_wc_conflict_description_t_merged_file_get", _wrap_svn_wc_conflict_description_t_merged_file_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_merged_file_get(svn_wc_conflict_description_t self) -> char const *"},
	 { (char *)"svn_wc_conflict_description_t_operation_set", _wrap_svn_wc_conflict_description_t_operation_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_operation_set(svn_wc_conflict_description_t self, svn_wc_operation_t operation)"},
	 { (char *)"svn_wc_conflict_description_t_operation_get", _wrap_svn_wc_conflict_description_t_operation_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_operation_get(svn_wc_conflict_description_t self) -> svn_wc_operation_t"},
	 { (char *)"svn_wc_conflict_description_t_src_left_version_set", _wrap_svn_wc_conflict_description_t_src_left_version_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_src_left_version_set(svn_wc_conflict_description_t self, svn_wc_conflict_version_t src_left_version)"},
	 { (char *)"svn_wc_conflict_description_t_src_left_version_get", _wrap_svn_wc_conflict_description_t_src_left_version_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_src_left_version_get(svn_wc_conflict_description_t self) -> svn_wc_conflict_version_t"},
	 { (char *)"svn_wc_conflict_description_t_src_right_version_set", _wrap_svn_wc_conflict_description_t_src_right_version_set, METH_VARARGS, (char *)"svn_wc_conflict_description_t_src_right_version_set(svn_wc_conflict_description_t self, svn_wc_conflict_version_t src_right_version)"},
	 { (char *)"svn_wc_conflict_description_t_src_right_version_get", _wrap_svn_wc_conflict_description_t_src_right_version_get, METH_VARARGS, (char *)"svn_wc_conflict_description_t_src_right_version_get(svn_wc_conflict_description_t self) -> svn_wc_conflict_version_t"},
	 { (char *)"new_svn_wc_conflict_description_t", _wrap_new_svn_wc_conflict_description_t, METH_VARARGS, (char *)"new_svn_wc_conflict_description_t() -> svn_wc_conflict_description_t"},
	 { (char *)"delete_svn_wc_conflict_description_t", _wrap_delete_svn_wc_conflict_description_t, METH_VARARGS, (char *)"delete_svn_wc_conflict_description_t(svn_wc_conflict_description_t self)"},
	 { (char *)"svn_wc_conflict_description_t_swigregister", svn_wc_conflict_description_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_conflict_description_create_text2", _wrap_svn_wc_conflict_description_create_text2, METH_VARARGS, (char *)"svn_wc_conflict_description_create_text2(char const * local_abspath, apr_pool_t result_pool) -> svn_wc_conflict_description2_t *"},
	 { (char *)"svn_wc_conflict_description_create_text", _wrap_svn_wc_conflict_description_create_text, METH_VARARGS, (char *)"svn_wc_conflict_description_create_text(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_wc_conflict_description_t"},
	 { (char *)"svn_wc_conflict_description_create_prop2", _wrap_svn_wc_conflict_description_create_prop2, METH_VARARGS, (char *)"\n"
		"svn_wc_conflict_description_create_prop2(char const * local_abspath, svn_node_kind_t node_kind, char const * property_name, \n"
		"    apr_pool_t result_pool) -> svn_wc_conflict_description2_t *\n"
		""},
	 { (char *)"svn_wc_conflict_description_create_prop", _wrap_svn_wc_conflict_description_create_prop, METH_VARARGS, (char *)"\n"
		"svn_wc_conflict_description_create_prop(char const * path, svn_wc_adm_access_t * adm_access, svn_node_kind_t node_kind, char const * property_name, \n"
		"    apr_pool_t pool) -> svn_wc_conflict_description_t\n"
		""},
	 { (char *)"svn_wc_conflict_description_create_tree2", _wrap_svn_wc_conflict_description_create_tree2, METH_VARARGS, (char *)"\n"
		"svn_wc_conflict_description_create_tree2(char const * local_abspath, svn_node_kind_t node_kind, svn_wc_operation_t operation, \n"
		"    svn_wc_conflict_version_t src_left_version, svn_wc_conflict_version_t src_right_version, \n"
		"    apr_pool_t result_pool) -> svn_wc_conflict_description2_t *\n"
		""},
	 { (char *)"svn_wc_conflict_description_create_tree", _wrap_svn_wc_conflict_description_create_tree, METH_VARARGS, (char *)"\n"
		"svn_wc_conflict_description_create_tree(char const * path, svn_wc_adm_access_t * adm_access, svn_node_kind_t node_kind, svn_wc_operation_t operation, \n"
		"    svn_wc_conflict_version_t src_left_version, svn_wc_conflict_version_t src_right_version, \n"
		"    apr_pool_t pool) -> svn_wc_conflict_description_t\n"
		""},
	 { (char *)"svn_wc_conflict_description2_dup", _wrap_svn_wc_conflict_description2_dup, METH_VARARGS, (char *)"svn_wc_conflict_description2_dup(svn_wc_conflict_description2_t const * conflict, apr_pool_t result_pool) -> svn_wc_conflict_description2_t *"},
	 { (char *)"svn_wc__conflict_description2_dup", _wrap_svn_wc__conflict_description2_dup, METH_VARARGS, (char *)"svn_wc__conflict_description2_dup(svn_wc_conflict_description2_t const * conflict, apr_pool_t result_pool) -> svn_wc_conflict_description2_t *"},
	 { (char *)"svn_wc_create_conflict_result", _wrap_svn_wc_create_conflict_result, METH_VARARGS, (char *)"svn_wc_create_conflict_result(svn_wc_conflict_choice_t choice, char const * merged_file, apr_pool_t pool) -> svn_wc_conflict_result_t *"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_opened_set", _wrap_svn_wc_diff_callbacks4_t_file_opened_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_opened_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) file_opened)"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_opened_get", _wrap_svn_wc_diff_callbacks4_t_file_opened_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_opened_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_changed_set", _wrap_svn_wc_diff_callbacks4_t_file_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_changed_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) file_changed)"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_changed_get", _wrap_svn_wc_diff_callbacks4_t_file_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_changed_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_added_set", _wrap_svn_wc_diff_callbacks4_t_file_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_added_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) file_added)"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_added_get", _wrap_svn_wc_diff_callbacks4_t_file_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_added_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_deleted_set", _wrap_svn_wc_diff_callbacks4_t_file_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_deleted_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *) file_deleted)"},
	 { (char *)"svn_wc_diff_callbacks4_t_file_deleted_get", _wrap_svn_wc_diff_callbacks4_t_file_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_file_deleted_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_deleted_set", _wrap_svn_wc_diff_callbacks4_t_dir_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_deleted_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *) dir_deleted)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_deleted_get", _wrap_svn_wc_diff_callbacks4_t_dir_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_deleted_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_opened_set", _wrap_svn_wc_diff_callbacks4_t_dir_opened_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_opened_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) dir_opened)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_opened_get", _wrap_svn_wc_diff_callbacks4_t_dir_opened_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_opened_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_added_set", _wrap_svn_wc_diff_callbacks4_t_dir_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_added_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *) dir_added)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_added_get", _wrap_svn_wc_diff_callbacks4_t_dir_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_added_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_props_changed_set", _wrap_svn_wc_diff_callbacks4_t_dir_props_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_props_changed_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) dir_props_changed)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_props_changed_get", _wrap_svn_wc_diff_callbacks4_t_dir_props_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_props_changed_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_closed_set", _wrap_svn_wc_diff_callbacks4_t_dir_closed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_closed_set(svn_wc_diff_callbacks4_t self, svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *) dir_closed)"},
	 { (char *)"svn_wc_diff_callbacks4_t_dir_closed_get", _wrap_svn_wc_diff_callbacks4_t_dir_closed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks4_t_dir_closed_get(svn_wc_diff_callbacks4_t self) -> svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)"},
	 { (char *)"new_svn_wc_diff_callbacks4_t", _wrap_new_svn_wc_diff_callbacks4_t, METH_VARARGS, (char *)"new_svn_wc_diff_callbacks4_t() -> svn_wc_diff_callbacks4_t"},
	 { (char *)"delete_svn_wc_diff_callbacks4_t", _wrap_delete_svn_wc_diff_callbacks4_t, METH_VARARGS, (char *)"delete_svn_wc_diff_callbacks4_t(svn_wc_diff_callbacks4_t self)"},
	 { (char *)"svn_wc_diff_callbacks4_t_swigregister", svn_wc_diff_callbacks4_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_diff_callbacks3_t_file_changed_set", _wrap_svn_wc_diff_callbacks3_t_file_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_file_changed_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) file_changed)"},
	 { (char *)"svn_wc_diff_callbacks3_t_file_changed_get", _wrap_svn_wc_diff_callbacks3_t_file_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_file_changed_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)"},
	 { (char *)"svn_wc_diff_callbacks3_t_file_added_set", _wrap_svn_wc_diff_callbacks3_t_file_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_file_added_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) file_added)"},
	 { (char *)"svn_wc_diff_callbacks3_t_file_added_get", _wrap_svn_wc_diff_callbacks3_t_file_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_file_added_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)"},
	 { (char *)"svn_wc_diff_callbacks3_t_file_deleted_set", _wrap_svn_wc_diff_callbacks3_t_file_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_file_deleted_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) file_deleted)"},
	 { (char *)"svn_wc_diff_callbacks3_t_file_deleted_get", _wrap_svn_wc_diff_callbacks3_t_file_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_file_deleted_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_added_set", _wrap_svn_wc_diff_callbacks3_t_dir_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_added_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) dir_added)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_added_get", _wrap_svn_wc_diff_callbacks3_t_dir_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_added_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_deleted_set", _wrap_svn_wc_diff_callbacks3_t_dir_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_deleted_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) dir_deleted)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_deleted_get", _wrap_svn_wc_diff_callbacks3_t_dir_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_deleted_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_props_changed_set", _wrap_svn_wc_diff_callbacks3_t_dir_props_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_props_changed_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) dir_props_changed)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_props_changed_get", _wrap_svn_wc_diff_callbacks3_t_dir_props_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_props_changed_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_opened_set", _wrap_svn_wc_diff_callbacks3_t_dir_opened_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_opened_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) dir_opened)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_opened_get", _wrap_svn_wc_diff_callbacks3_t_dir_opened_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_opened_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_closed_set", _wrap_svn_wc_diff_callbacks3_t_dir_closed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_closed_set(svn_wc_diff_callbacks3_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) dir_closed)"},
	 { (char *)"svn_wc_diff_callbacks3_t_dir_closed_get", _wrap_svn_wc_diff_callbacks3_t_dir_closed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks3_t_dir_closed_get(svn_wc_diff_callbacks3_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)"},
	 { (char *)"new_svn_wc_diff_callbacks3_t", _wrap_new_svn_wc_diff_callbacks3_t, METH_VARARGS, (char *)"new_svn_wc_diff_callbacks3_t() -> svn_wc_diff_callbacks3_t"},
	 { (char *)"delete_svn_wc_diff_callbacks3_t", _wrap_delete_svn_wc_diff_callbacks3_t, METH_VARARGS, (char *)"delete_svn_wc_diff_callbacks3_t(svn_wc_diff_callbacks3_t self)"},
	 { (char *)"svn_wc_diff_callbacks3_t_swigregister", svn_wc_diff_callbacks3_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_diff_callbacks2_t_file_changed_set", _wrap_svn_wc_diff_callbacks2_t_file_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_file_changed_set(svn_wc_diff_callbacks2_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) file_changed)"},
	 { (char *)"svn_wc_diff_callbacks2_t_file_changed_get", _wrap_svn_wc_diff_callbacks2_t_file_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_file_changed_get(svn_wc_diff_callbacks2_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)"},
	 { (char *)"svn_wc_diff_callbacks2_t_file_added_set", _wrap_svn_wc_diff_callbacks2_t_file_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_file_added_set(svn_wc_diff_callbacks2_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) file_added)"},
	 { (char *)"svn_wc_diff_callbacks2_t_file_added_get", _wrap_svn_wc_diff_callbacks2_t_file_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_file_added_get(svn_wc_diff_callbacks2_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)"},
	 { (char *)"svn_wc_diff_callbacks2_t_file_deleted_set", _wrap_svn_wc_diff_callbacks2_t_file_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_file_deleted_set(svn_wc_diff_callbacks2_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) file_deleted)"},
	 { (char *)"svn_wc_diff_callbacks2_t_file_deleted_get", _wrap_svn_wc_diff_callbacks2_t_file_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_file_deleted_get(svn_wc_diff_callbacks2_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)"},
	 { (char *)"svn_wc_diff_callbacks2_t_dir_added_set", _wrap_svn_wc_diff_callbacks2_t_dir_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_dir_added_set(svn_wc_diff_callbacks2_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) dir_added)"},
	 { (char *)"svn_wc_diff_callbacks2_t_dir_added_get", _wrap_svn_wc_diff_callbacks2_t_dir_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_dir_added_get(svn_wc_diff_callbacks2_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)"},
	 { (char *)"svn_wc_diff_callbacks2_t_dir_deleted_set", _wrap_svn_wc_diff_callbacks2_t_dir_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_dir_deleted_set(svn_wc_diff_callbacks2_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) dir_deleted)"},
	 { (char *)"svn_wc_diff_callbacks2_t_dir_deleted_get", _wrap_svn_wc_diff_callbacks2_t_dir_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_dir_deleted_get(svn_wc_diff_callbacks2_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)"},
	 { (char *)"svn_wc_diff_callbacks2_t_dir_props_changed_set", _wrap_svn_wc_diff_callbacks2_t_dir_props_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_dir_props_changed_set(svn_wc_diff_callbacks2_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) dir_props_changed)"},
	 { (char *)"svn_wc_diff_callbacks2_t_dir_props_changed_get", _wrap_svn_wc_diff_callbacks2_t_dir_props_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks2_t_dir_props_changed_get(svn_wc_diff_callbacks2_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)"},
	 { (char *)"new_svn_wc_diff_callbacks2_t", _wrap_new_svn_wc_diff_callbacks2_t, METH_VARARGS, (char *)"new_svn_wc_diff_callbacks2_t() -> svn_wc_diff_callbacks2_t"},
	 { (char *)"delete_svn_wc_diff_callbacks2_t", _wrap_delete_svn_wc_diff_callbacks2_t, METH_VARARGS, (char *)"delete_svn_wc_diff_callbacks2_t(svn_wc_diff_callbacks2_t self)"},
	 { (char *)"svn_wc_diff_callbacks2_t_swigregister", svn_wc_diff_callbacks2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_diff_callbacks_t_file_changed_set", _wrap_svn_wc_diff_callbacks_t_file_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_file_changed_set(svn_wc_diff_callbacks_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) file_changed)"},
	 { (char *)"svn_wc_diff_callbacks_t_file_changed_get", _wrap_svn_wc_diff_callbacks_t_file_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_file_changed_get(svn_wc_diff_callbacks_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)"},
	 { (char *)"svn_wc_diff_callbacks_t_file_added_set", _wrap_svn_wc_diff_callbacks_t_file_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_file_added_set(svn_wc_diff_callbacks_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) file_added)"},
	 { (char *)"svn_wc_diff_callbacks_t_file_added_get", _wrap_svn_wc_diff_callbacks_t_file_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_file_added_get(svn_wc_diff_callbacks_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)"},
	 { (char *)"svn_wc_diff_callbacks_t_file_deleted_set", _wrap_svn_wc_diff_callbacks_t_file_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_file_deleted_set(svn_wc_diff_callbacks_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *) file_deleted)"},
	 { (char *)"svn_wc_diff_callbacks_t_file_deleted_get", _wrap_svn_wc_diff_callbacks_t_file_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_file_deleted_get(svn_wc_diff_callbacks_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)"},
	 { (char *)"svn_wc_diff_callbacks_t_dir_added_set", _wrap_svn_wc_diff_callbacks_t_dir_added_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_dir_added_set(svn_wc_diff_callbacks_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) dir_added)"},
	 { (char *)"svn_wc_diff_callbacks_t_dir_added_get", _wrap_svn_wc_diff_callbacks_t_dir_added_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_dir_added_get(svn_wc_diff_callbacks_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)"},
	 { (char *)"svn_wc_diff_callbacks_t_dir_deleted_set", _wrap_svn_wc_diff_callbacks_t_dir_deleted_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_dir_deleted_set(svn_wc_diff_callbacks_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) dir_deleted)"},
	 { (char *)"svn_wc_diff_callbacks_t_dir_deleted_get", _wrap_svn_wc_diff_callbacks_t_dir_deleted_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_dir_deleted_get(svn_wc_diff_callbacks_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)"},
	 { (char *)"svn_wc_diff_callbacks_t_props_changed_set", _wrap_svn_wc_diff_callbacks_t_props_changed_set, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_props_changed_set(svn_wc_diff_callbacks_t self, svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) props_changed)"},
	 { (char *)"svn_wc_diff_callbacks_t_props_changed_get", _wrap_svn_wc_diff_callbacks_t_props_changed_get, METH_VARARGS, (char *)"svn_wc_diff_callbacks_t_props_changed_get(svn_wc_diff_callbacks_t self) -> svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)"},
	 { (char *)"new_svn_wc_diff_callbacks_t", _wrap_new_svn_wc_diff_callbacks_t, METH_VARARGS, (char *)"new_svn_wc_diff_callbacks_t() -> svn_wc_diff_callbacks_t"},
	 { (char *)"delete_svn_wc_diff_callbacks_t", _wrap_delete_svn_wc_diff_callbacks_t, METH_VARARGS, (char *)"delete_svn_wc_diff_callbacks_t(svn_wc_diff_callbacks_t self)"},
	 { (char *)"svn_wc_diff_callbacks_t_swigregister", svn_wc_diff_callbacks_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_check_wc2", _wrap_svn_wc_check_wc2, METH_VARARGS, (char *)"svn_wc_check_wc2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_check_wc", _wrap_svn_wc_check_wc, METH_VARARGS, (char *)"svn_wc_check_wc(char const * path, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_has_binary_prop", _wrap_svn_wc_has_binary_prop, METH_VARARGS, (char *)"svn_wc_has_binary_prop(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_text_modified_p2", _wrap_svn_wc_text_modified_p2, METH_VARARGS, (char *)"svn_wc_text_modified_p2(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_boolean_t unused, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_text_modified_p", _wrap_svn_wc_text_modified_p, METH_VARARGS, (char *)"\n"
		"svn_wc_text_modified_p(char const * filename, svn_boolean_t force_comparison, svn_wc_adm_access_t * adm_access, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_props_modified_p2", _wrap_svn_wc_props_modified_p2, METH_VARARGS, (char *)"svn_wc_props_modified_p2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_props_modified_p", _wrap_svn_wc_props_modified_p, METH_VARARGS, (char *)"svn_wc_props_modified_p(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_entry_t_name_set", _wrap_svn_wc_entry_t_name_set, METH_VARARGS, (char *)"svn_wc_entry_t_name_set(svn_wc_entry_t self, char const * name)"},
	 { (char *)"svn_wc_entry_t_name_get", _wrap_svn_wc_entry_t_name_get, METH_VARARGS, (char *)"svn_wc_entry_t_name_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_revision_set", _wrap_svn_wc_entry_t_revision_set, METH_VARARGS, (char *)"svn_wc_entry_t_revision_set(svn_wc_entry_t self, svn_revnum_t revision)"},
	 { (char *)"svn_wc_entry_t_revision_get", _wrap_svn_wc_entry_t_revision_get, METH_VARARGS, (char *)"svn_wc_entry_t_revision_get(svn_wc_entry_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_entry_t_url_set", _wrap_svn_wc_entry_t_url_set, METH_VARARGS, (char *)"svn_wc_entry_t_url_set(svn_wc_entry_t self, char const * url)"},
	 { (char *)"svn_wc_entry_t_url_get", _wrap_svn_wc_entry_t_url_get, METH_VARARGS, (char *)"svn_wc_entry_t_url_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_repos_set", _wrap_svn_wc_entry_t_repos_set, METH_VARARGS, (char *)"svn_wc_entry_t_repos_set(svn_wc_entry_t self, char const * repos)"},
	 { (char *)"svn_wc_entry_t_repos_get", _wrap_svn_wc_entry_t_repos_get, METH_VARARGS, (char *)"svn_wc_entry_t_repos_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_uuid_set", _wrap_svn_wc_entry_t_uuid_set, METH_VARARGS, (char *)"svn_wc_entry_t_uuid_set(svn_wc_entry_t self, char const * uuid)"},
	 { (char *)"svn_wc_entry_t_uuid_get", _wrap_svn_wc_entry_t_uuid_get, METH_VARARGS, (char *)"svn_wc_entry_t_uuid_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_kind_set", _wrap_svn_wc_entry_t_kind_set, METH_VARARGS, (char *)"svn_wc_entry_t_kind_set(svn_wc_entry_t self, svn_node_kind_t kind)"},
	 { (char *)"svn_wc_entry_t_kind_get", _wrap_svn_wc_entry_t_kind_get, METH_VARARGS, (char *)"svn_wc_entry_t_kind_get(svn_wc_entry_t self) -> svn_node_kind_t"},
	 { (char *)"svn_wc_entry_t_schedule_set", _wrap_svn_wc_entry_t_schedule_set, METH_VARARGS, (char *)"svn_wc_entry_t_schedule_set(svn_wc_entry_t self, svn_wc_schedule_t schedule)"},
	 { (char *)"svn_wc_entry_t_schedule_get", _wrap_svn_wc_entry_t_schedule_get, METH_VARARGS, (char *)"svn_wc_entry_t_schedule_get(svn_wc_entry_t self) -> svn_wc_schedule_t"},
	 { (char *)"svn_wc_entry_t_copied_set", _wrap_svn_wc_entry_t_copied_set, METH_VARARGS, (char *)"svn_wc_entry_t_copied_set(svn_wc_entry_t self, svn_boolean_t copied)"},
	 { (char *)"svn_wc_entry_t_copied_get", _wrap_svn_wc_entry_t_copied_get, METH_VARARGS, (char *)"svn_wc_entry_t_copied_get(svn_wc_entry_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_entry_t_deleted_set", _wrap_svn_wc_entry_t_deleted_set, METH_VARARGS, (char *)"svn_wc_entry_t_deleted_set(svn_wc_entry_t self, svn_boolean_t deleted)"},
	 { (char *)"svn_wc_entry_t_deleted_get", _wrap_svn_wc_entry_t_deleted_get, METH_VARARGS, (char *)"svn_wc_entry_t_deleted_get(svn_wc_entry_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_entry_t_absent_set", _wrap_svn_wc_entry_t_absent_set, METH_VARARGS, (char *)"svn_wc_entry_t_absent_set(svn_wc_entry_t self, svn_boolean_t absent)"},
	 { (char *)"svn_wc_entry_t_absent_get", _wrap_svn_wc_entry_t_absent_get, METH_VARARGS, (char *)"svn_wc_entry_t_absent_get(svn_wc_entry_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_entry_t_incomplete_set", _wrap_svn_wc_entry_t_incomplete_set, METH_VARARGS, (char *)"svn_wc_entry_t_incomplete_set(svn_wc_entry_t self, svn_boolean_t incomplete)"},
	 { (char *)"svn_wc_entry_t_incomplete_get", _wrap_svn_wc_entry_t_incomplete_get, METH_VARARGS, (char *)"svn_wc_entry_t_incomplete_get(svn_wc_entry_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_entry_t_copyfrom_url_set", _wrap_svn_wc_entry_t_copyfrom_url_set, METH_VARARGS, (char *)"svn_wc_entry_t_copyfrom_url_set(svn_wc_entry_t self, char const * copyfrom_url)"},
	 { (char *)"svn_wc_entry_t_copyfrom_url_get", _wrap_svn_wc_entry_t_copyfrom_url_get, METH_VARARGS, (char *)"svn_wc_entry_t_copyfrom_url_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_copyfrom_rev_set", _wrap_svn_wc_entry_t_copyfrom_rev_set, METH_VARARGS, (char *)"svn_wc_entry_t_copyfrom_rev_set(svn_wc_entry_t self, svn_revnum_t copyfrom_rev)"},
	 { (char *)"svn_wc_entry_t_copyfrom_rev_get", _wrap_svn_wc_entry_t_copyfrom_rev_get, METH_VARARGS, (char *)"svn_wc_entry_t_copyfrom_rev_get(svn_wc_entry_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_entry_t_conflict_old_set", _wrap_svn_wc_entry_t_conflict_old_set, METH_VARARGS, (char *)"svn_wc_entry_t_conflict_old_set(svn_wc_entry_t self, char const * conflict_old)"},
	 { (char *)"svn_wc_entry_t_conflict_old_get", _wrap_svn_wc_entry_t_conflict_old_get, METH_VARARGS, (char *)"svn_wc_entry_t_conflict_old_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_conflict_new_set", _wrap_svn_wc_entry_t_conflict_new_set, METH_VARARGS, (char *)"svn_wc_entry_t_conflict_new_set(svn_wc_entry_t self, char const * conflict_new)"},
	 { (char *)"svn_wc_entry_t_conflict_new_get", _wrap_svn_wc_entry_t_conflict_new_get, METH_VARARGS, (char *)"svn_wc_entry_t_conflict_new_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_conflict_wrk_set", _wrap_svn_wc_entry_t_conflict_wrk_set, METH_VARARGS, (char *)"svn_wc_entry_t_conflict_wrk_set(svn_wc_entry_t self, char const * conflict_wrk)"},
	 { (char *)"svn_wc_entry_t_conflict_wrk_get", _wrap_svn_wc_entry_t_conflict_wrk_get, METH_VARARGS, (char *)"svn_wc_entry_t_conflict_wrk_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_prejfile_set", _wrap_svn_wc_entry_t_prejfile_set, METH_VARARGS, (char *)"svn_wc_entry_t_prejfile_set(svn_wc_entry_t self, char const * prejfile)"},
	 { (char *)"svn_wc_entry_t_prejfile_get", _wrap_svn_wc_entry_t_prejfile_get, METH_VARARGS, (char *)"svn_wc_entry_t_prejfile_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_text_time_set", _wrap_svn_wc_entry_t_text_time_set, METH_VARARGS, (char *)"svn_wc_entry_t_text_time_set(svn_wc_entry_t self, apr_time_t text_time)"},
	 { (char *)"svn_wc_entry_t_text_time_get", _wrap_svn_wc_entry_t_text_time_get, METH_VARARGS, (char *)"svn_wc_entry_t_text_time_get(svn_wc_entry_t self) -> apr_time_t"},
	 { (char *)"svn_wc_entry_t_prop_time_set", _wrap_svn_wc_entry_t_prop_time_set, METH_VARARGS, (char *)"svn_wc_entry_t_prop_time_set(svn_wc_entry_t self, apr_time_t prop_time)"},
	 { (char *)"svn_wc_entry_t_prop_time_get", _wrap_svn_wc_entry_t_prop_time_get, METH_VARARGS, (char *)"svn_wc_entry_t_prop_time_get(svn_wc_entry_t self) -> apr_time_t"},
	 { (char *)"svn_wc_entry_t_checksum_set", _wrap_svn_wc_entry_t_checksum_set, METH_VARARGS, (char *)"svn_wc_entry_t_checksum_set(svn_wc_entry_t self, char const * checksum)"},
	 { (char *)"svn_wc_entry_t_checksum_get", _wrap_svn_wc_entry_t_checksum_get, METH_VARARGS, (char *)"svn_wc_entry_t_checksum_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_cmt_rev_set", _wrap_svn_wc_entry_t_cmt_rev_set, METH_VARARGS, (char *)"svn_wc_entry_t_cmt_rev_set(svn_wc_entry_t self, svn_revnum_t cmt_rev)"},
	 { (char *)"svn_wc_entry_t_cmt_rev_get", _wrap_svn_wc_entry_t_cmt_rev_get, METH_VARARGS, (char *)"svn_wc_entry_t_cmt_rev_get(svn_wc_entry_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_entry_t_cmt_date_set", _wrap_svn_wc_entry_t_cmt_date_set, METH_VARARGS, (char *)"svn_wc_entry_t_cmt_date_set(svn_wc_entry_t self, apr_time_t cmt_date)"},
	 { (char *)"svn_wc_entry_t_cmt_date_get", _wrap_svn_wc_entry_t_cmt_date_get, METH_VARARGS, (char *)"svn_wc_entry_t_cmt_date_get(svn_wc_entry_t self) -> apr_time_t"},
	 { (char *)"svn_wc_entry_t_cmt_author_set", _wrap_svn_wc_entry_t_cmt_author_set, METH_VARARGS, (char *)"svn_wc_entry_t_cmt_author_set(svn_wc_entry_t self, char const * cmt_author)"},
	 { (char *)"svn_wc_entry_t_cmt_author_get", _wrap_svn_wc_entry_t_cmt_author_get, METH_VARARGS, (char *)"svn_wc_entry_t_cmt_author_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_lock_token_set", _wrap_svn_wc_entry_t_lock_token_set, METH_VARARGS, (char *)"svn_wc_entry_t_lock_token_set(svn_wc_entry_t self, char const * lock_token)"},
	 { (char *)"svn_wc_entry_t_lock_token_get", _wrap_svn_wc_entry_t_lock_token_get, METH_VARARGS, (char *)"svn_wc_entry_t_lock_token_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_lock_owner_set", _wrap_svn_wc_entry_t_lock_owner_set, METH_VARARGS, (char *)"svn_wc_entry_t_lock_owner_set(svn_wc_entry_t self, char const * lock_owner)"},
	 { (char *)"svn_wc_entry_t_lock_owner_get", _wrap_svn_wc_entry_t_lock_owner_get, METH_VARARGS, (char *)"svn_wc_entry_t_lock_owner_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_lock_comment_set", _wrap_svn_wc_entry_t_lock_comment_set, METH_VARARGS, (char *)"svn_wc_entry_t_lock_comment_set(svn_wc_entry_t self, char const * lock_comment)"},
	 { (char *)"svn_wc_entry_t_lock_comment_get", _wrap_svn_wc_entry_t_lock_comment_get, METH_VARARGS, (char *)"svn_wc_entry_t_lock_comment_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_lock_creation_date_set", _wrap_svn_wc_entry_t_lock_creation_date_set, METH_VARARGS, (char *)"svn_wc_entry_t_lock_creation_date_set(svn_wc_entry_t self, apr_time_t lock_creation_date)"},
	 { (char *)"svn_wc_entry_t_lock_creation_date_get", _wrap_svn_wc_entry_t_lock_creation_date_get, METH_VARARGS, (char *)"svn_wc_entry_t_lock_creation_date_get(svn_wc_entry_t self) -> apr_time_t"},
	 { (char *)"svn_wc_entry_t_has_props_set", _wrap_svn_wc_entry_t_has_props_set, METH_VARARGS, (char *)"svn_wc_entry_t_has_props_set(svn_wc_entry_t self, svn_boolean_t has_props)"},
	 { (char *)"svn_wc_entry_t_has_props_get", _wrap_svn_wc_entry_t_has_props_get, METH_VARARGS, (char *)"svn_wc_entry_t_has_props_get(svn_wc_entry_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_entry_t_has_prop_mods_set", _wrap_svn_wc_entry_t_has_prop_mods_set, METH_VARARGS, (char *)"svn_wc_entry_t_has_prop_mods_set(svn_wc_entry_t self, svn_boolean_t has_prop_mods)"},
	 { (char *)"svn_wc_entry_t_has_prop_mods_get", _wrap_svn_wc_entry_t_has_prop_mods_get, METH_VARARGS, (char *)"svn_wc_entry_t_has_prop_mods_get(svn_wc_entry_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_entry_t_cachable_props_set", _wrap_svn_wc_entry_t_cachable_props_set, METH_VARARGS, (char *)"svn_wc_entry_t_cachable_props_set(svn_wc_entry_t self, char const * cachable_props)"},
	 { (char *)"svn_wc_entry_t_cachable_props_get", _wrap_svn_wc_entry_t_cachable_props_get, METH_VARARGS, (char *)"svn_wc_entry_t_cachable_props_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_present_props_set", _wrap_svn_wc_entry_t_present_props_set, METH_VARARGS, (char *)"svn_wc_entry_t_present_props_set(svn_wc_entry_t self, char const * present_props)"},
	 { (char *)"svn_wc_entry_t_present_props_get", _wrap_svn_wc_entry_t_present_props_get, METH_VARARGS, (char *)"svn_wc_entry_t_present_props_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_changelist_set", _wrap_svn_wc_entry_t_changelist_set, METH_VARARGS, (char *)"svn_wc_entry_t_changelist_set(svn_wc_entry_t self, char const * changelist)"},
	 { (char *)"svn_wc_entry_t_changelist_get", _wrap_svn_wc_entry_t_changelist_get, METH_VARARGS, (char *)"svn_wc_entry_t_changelist_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_working_size_set", _wrap_svn_wc_entry_t_working_size_set, METH_VARARGS, (char *)"svn_wc_entry_t_working_size_set(svn_wc_entry_t self, apr_off_t working_size)"},
	 { (char *)"svn_wc_entry_t_working_size_get", _wrap_svn_wc_entry_t_working_size_get, METH_VARARGS, (char *)"svn_wc_entry_t_working_size_get(svn_wc_entry_t self) -> apr_off_t"},
	 { (char *)"svn_wc_entry_t_keep_local_set", _wrap_svn_wc_entry_t_keep_local_set, METH_VARARGS, (char *)"svn_wc_entry_t_keep_local_set(svn_wc_entry_t self, svn_boolean_t keep_local)"},
	 { (char *)"svn_wc_entry_t_keep_local_get", _wrap_svn_wc_entry_t_keep_local_get, METH_VARARGS, (char *)"svn_wc_entry_t_keep_local_get(svn_wc_entry_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_entry_t_depth_set", _wrap_svn_wc_entry_t_depth_set, METH_VARARGS, (char *)"svn_wc_entry_t_depth_set(svn_wc_entry_t self, svn_depth_t depth)"},
	 { (char *)"svn_wc_entry_t_depth_get", _wrap_svn_wc_entry_t_depth_get, METH_VARARGS, (char *)"svn_wc_entry_t_depth_get(svn_wc_entry_t self) -> svn_depth_t"},
	 { (char *)"svn_wc_entry_t_tree_conflict_data_set", _wrap_svn_wc_entry_t_tree_conflict_data_set, METH_VARARGS, (char *)"svn_wc_entry_t_tree_conflict_data_set(svn_wc_entry_t self, char const * tree_conflict_data)"},
	 { (char *)"svn_wc_entry_t_tree_conflict_data_get", _wrap_svn_wc_entry_t_tree_conflict_data_get, METH_VARARGS, (char *)"svn_wc_entry_t_tree_conflict_data_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_file_external_path_set", _wrap_svn_wc_entry_t_file_external_path_set, METH_VARARGS, (char *)"svn_wc_entry_t_file_external_path_set(svn_wc_entry_t self, char const * file_external_path)"},
	 { (char *)"svn_wc_entry_t_file_external_path_get", _wrap_svn_wc_entry_t_file_external_path_get, METH_VARARGS, (char *)"svn_wc_entry_t_file_external_path_get(svn_wc_entry_t self) -> char const *"},
	 { (char *)"svn_wc_entry_t_file_external_peg_rev_set", _wrap_svn_wc_entry_t_file_external_peg_rev_set, METH_VARARGS, (char *)"svn_wc_entry_t_file_external_peg_rev_set(svn_wc_entry_t self, svn_opt_revision_t file_external_peg_rev)"},
	 { (char *)"svn_wc_entry_t_file_external_peg_rev_get", _wrap_svn_wc_entry_t_file_external_peg_rev_get, METH_VARARGS, (char *)"svn_wc_entry_t_file_external_peg_rev_get(svn_wc_entry_t self) -> svn_opt_revision_t"},
	 { (char *)"svn_wc_entry_t_file_external_rev_set", _wrap_svn_wc_entry_t_file_external_rev_set, METH_VARARGS, (char *)"svn_wc_entry_t_file_external_rev_set(svn_wc_entry_t self, svn_opt_revision_t file_external_rev)"},
	 { (char *)"svn_wc_entry_t_file_external_rev_get", _wrap_svn_wc_entry_t_file_external_rev_get, METH_VARARGS, (char *)"svn_wc_entry_t_file_external_rev_get(svn_wc_entry_t self) -> svn_opt_revision_t"},
	 { (char *)"new_svn_wc_entry_t", _wrap_new_svn_wc_entry_t, METH_VARARGS, (char *)"new_svn_wc_entry_t() -> svn_wc_entry_t"},
	 { (char *)"delete_svn_wc_entry_t", _wrap_delete_svn_wc_entry_t, METH_VARARGS, (char *)"delete_svn_wc_entry_t(svn_wc_entry_t self)"},
	 { (char *)"svn_wc_entry_t_swigregister", svn_wc_entry_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_entry", _wrap_svn_wc_entry, METH_VARARGS, (char *)"svn_wc_entry(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t show_hidden, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_entries_read", _wrap_svn_wc_entries_read, METH_VARARGS, (char *)"svn_wc_entries_read(svn_wc_adm_access_t * adm_access, svn_boolean_t show_hidden, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_entry_dup", _wrap_svn_wc_entry_dup, METH_VARARGS, (char *)"svn_wc_entry_dup(svn_wc_entry_t entry, apr_pool_t pool) -> svn_wc_entry_t"},
	 { (char *)"svn_wc_info_t_schedule_set", _wrap_svn_wc_info_t_schedule_set, METH_VARARGS, (char *)"svn_wc_info_t_schedule_set(svn_wc_info_t self, svn_wc_schedule_t schedule)"},
	 { (char *)"svn_wc_info_t_schedule_get", _wrap_svn_wc_info_t_schedule_get, METH_VARARGS, (char *)"svn_wc_info_t_schedule_get(svn_wc_info_t self) -> svn_wc_schedule_t"},
	 { (char *)"svn_wc_info_t_copyfrom_url_set", _wrap_svn_wc_info_t_copyfrom_url_set, METH_VARARGS, (char *)"svn_wc_info_t_copyfrom_url_set(svn_wc_info_t self, char const * copyfrom_url)"},
	 { (char *)"svn_wc_info_t_copyfrom_url_get", _wrap_svn_wc_info_t_copyfrom_url_get, METH_VARARGS, (char *)"svn_wc_info_t_copyfrom_url_get(svn_wc_info_t self) -> char const *"},
	 { (char *)"svn_wc_info_t_copyfrom_rev_set", _wrap_svn_wc_info_t_copyfrom_rev_set, METH_VARARGS, (char *)"svn_wc_info_t_copyfrom_rev_set(svn_wc_info_t self, svn_revnum_t copyfrom_rev)"},
	 { (char *)"svn_wc_info_t_copyfrom_rev_get", _wrap_svn_wc_info_t_copyfrom_rev_get, METH_VARARGS, (char *)"svn_wc_info_t_copyfrom_rev_get(svn_wc_info_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_info_t_checksum_set", _wrap_svn_wc_info_t_checksum_set, METH_VARARGS, (char *)"svn_wc_info_t_checksum_set(svn_wc_info_t self, svn_checksum_t checksum)"},
	 { (char *)"svn_wc_info_t_checksum_get", _wrap_svn_wc_info_t_checksum_get, METH_VARARGS, (char *)"svn_wc_info_t_checksum_get(svn_wc_info_t self) -> svn_checksum_t"},
	 { (char *)"svn_wc_info_t_changelist_set", _wrap_svn_wc_info_t_changelist_set, METH_VARARGS, (char *)"svn_wc_info_t_changelist_set(svn_wc_info_t self, char const * changelist)"},
	 { (char *)"svn_wc_info_t_changelist_get", _wrap_svn_wc_info_t_changelist_get, METH_VARARGS, (char *)"svn_wc_info_t_changelist_get(svn_wc_info_t self) -> char const *"},
	 { (char *)"svn_wc_info_t_depth_set", _wrap_svn_wc_info_t_depth_set, METH_VARARGS, (char *)"svn_wc_info_t_depth_set(svn_wc_info_t self, svn_depth_t depth)"},
	 { (char *)"svn_wc_info_t_depth_get", _wrap_svn_wc_info_t_depth_get, METH_VARARGS, (char *)"svn_wc_info_t_depth_get(svn_wc_info_t self) -> svn_depth_t"},
	 { (char *)"svn_wc_info_t_recorded_size_set", _wrap_svn_wc_info_t_recorded_size_set, METH_VARARGS, (char *)"svn_wc_info_t_recorded_size_set(svn_wc_info_t self, svn_filesize_t recorded_size)"},
	 { (char *)"svn_wc_info_t_recorded_size_get", _wrap_svn_wc_info_t_recorded_size_get, METH_VARARGS, (char *)"svn_wc_info_t_recorded_size_get(svn_wc_info_t self) -> svn_filesize_t"},
	 { (char *)"svn_wc_info_t_recorded_time_set", _wrap_svn_wc_info_t_recorded_time_set, METH_VARARGS, (char *)"svn_wc_info_t_recorded_time_set(svn_wc_info_t self, apr_time_t recorded_time)"},
	 { (char *)"svn_wc_info_t_recorded_time_get", _wrap_svn_wc_info_t_recorded_time_get, METH_VARARGS, (char *)"svn_wc_info_t_recorded_time_get(svn_wc_info_t self) -> apr_time_t"},
	 { (char *)"svn_wc_info_t_conflicts_set", _wrap_svn_wc_info_t_conflicts_set, METH_VARARGS, (char *)"svn_wc_info_t_conflicts_set(svn_wc_info_t self, apr_array_header_t conflicts)"},
	 { (char *)"svn_wc_info_t_conflicts_get", _wrap_svn_wc_info_t_conflicts_get, METH_VARARGS, (char *)"svn_wc_info_t_conflicts_get(svn_wc_info_t self) -> apr_array_header_t"},
	 { (char *)"svn_wc_info_t_wcroot_abspath_set", _wrap_svn_wc_info_t_wcroot_abspath_set, METH_VARARGS, (char *)"svn_wc_info_t_wcroot_abspath_set(svn_wc_info_t self, char const * wcroot_abspath)"},
	 { (char *)"svn_wc_info_t_wcroot_abspath_get", _wrap_svn_wc_info_t_wcroot_abspath_get, METH_VARARGS, (char *)"svn_wc_info_t_wcroot_abspath_get(svn_wc_info_t self) -> char const *"},
	 { (char *)"svn_wc_info_t_moved_from_abspath_set", _wrap_svn_wc_info_t_moved_from_abspath_set, METH_VARARGS, (char *)"svn_wc_info_t_moved_from_abspath_set(svn_wc_info_t self, char const * moved_from_abspath)"},
	 { (char *)"svn_wc_info_t_moved_from_abspath_get", _wrap_svn_wc_info_t_moved_from_abspath_get, METH_VARARGS, (char *)"svn_wc_info_t_moved_from_abspath_get(svn_wc_info_t self) -> char const *"},
	 { (char *)"svn_wc_info_t_moved_to_abspath_set", _wrap_svn_wc_info_t_moved_to_abspath_set, METH_VARARGS, (char *)"svn_wc_info_t_moved_to_abspath_set(svn_wc_info_t self, char const * moved_to_abspath)"},
	 { (char *)"svn_wc_info_t_moved_to_abspath_get", _wrap_svn_wc_info_t_moved_to_abspath_get, METH_VARARGS, (char *)"svn_wc_info_t_moved_to_abspath_get(svn_wc_info_t self) -> char const *"},
	 { (char *)"new_svn_wc_info_t", _wrap_new_svn_wc_info_t, METH_VARARGS, (char *)"new_svn_wc_info_t() -> svn_wc_info_t"},
	 { (char *)"delete_svn_wc_info_t", _wrap_delete_svn_wc_info_t, METH_VARARGS, (char *)"delete_svn_wc_info_t(svn_wc_info_t self)"},
	 { (char *)"svn_wc_info_t_swigregister", svn_wc_info_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_info_dup", _wrap_svn_wc_info_dup, METH_VARARGS, (char *)"svn_wc_info_dup(svn_wc_info_t info, apr_pool_t pool) -> svn_wc_info_t"},
	 { (char *)"svn_wc_conflicted_p3", _wrap_svn_wc_conflicted_p3, METH_VARARGS, (char *)"svn_wc_conflicted_p3(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_conflicted_p2", _wrap_svn_wc_conflicted_p2, METH_VARARGS, (char *)"svn_wc_conflicted_p2(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_conflicted_p", _wrap_svn_wc_conflicted_p, METH_VARARGS, (char *)"svn_wc_conflicted_p(char const * dir_path, svn_wc_entry_t entry, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_ancestry", _wrap_svn_wc_get_ancestry, METH_VARARGS, (char *)"svn_wc_get_ancestry(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_entry_callbacks2_t_found_entry_set", _wrap_svn_wc_entry_callbacks2_t_found_entry_set, METH_VARARGS, (char *)"svn_wc_entry_callbacks2_t_found_entry_set(svn_wc_entry_callbacks2_t self, svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) found_entry)"},
	 { (char *)"svn_wc_entry_callbacks2_t_found_entry_get", _wrap_svn_wc_entry_callbacks2_t_found_entry_get, METH_VARARGS, (char *)"svn_wc_entry_callbacks2_t_found_entry_get(svn_wc_entry_callbacks2_t self) -> svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)"},
	 { (char *)"svn_wc_entry_callbacks2_t_handle_error_set", _wrap_svn_wc_entry_callbacks2_t_handle_error_set, METH_VARARGS, (char *)"svn_wc_entry_callbacks2_t_handle_error_set(svn_wc_entry_callbacks2_t self, svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *) handle_error)"},
	 { (char *)"svn_wc_entry_callbacks2_t_handle_error_get", _wrap_svn_wc_entry_callbacks2_t_handle_error_get, METH_VARARGS, (char *)"svn_wc_entry_callbacks2_t_handle_error_get(svn_wc_entry_callbacks2_t self) -> svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)"},
	 { (char *)"new_svn_wc_entry_callbacks2_t", _wrap_new_svn_wc_entry_callbacks2_t, METH_VARARGS, (char *)"new_svn_wc_entry_callbacks2_t() -> svn_wc_entry_callbacks2_t"},
	 { (char *)"delete_svn_wc_entry_callbacks2_t", _wrap_delete_svn_wc_entry_callbacks2_t, METH_VARARGS, (char *)"delete_svn_wc_entry_callbacks2_t(svn_wc_entry_callbacks2_t self)"},
	 { (char *)"svn_wc_entry_callbacks2_t_swigregister", svn_wc_entry_callbacks2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_entry_callbacks_t_found_entry_set", _wrap_svn_wc_entry_callbacks_t_found_entry_set, METH_VARARGS, (char *)"svn_wc_entry_callbacks_t_found_entry_set(svn_wc_entry_callbacks_t self, svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) found_entry)"},
	 { (char *)"svn_wc_entry_callbacks_t_found_entry_get", _wrap_svn_wc_entry_callbacks_t_found_entry_get, METH_VARARGS, (char *)"svn_wc_entry_callbacks_t_found_entry_get(svn_wc_entry_callbacks_t self) -> svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)"},
	 { (char *)"new_svn_wc_entry_callbacks_t", _wrap_new_svn_wc_entry_callbacks_t, METH_VARARGS, (char *)"new_svn_wc_entry_callbacks_t() -> svn_wc_entry_callbacks_t"},
	 { (char *)"delete_svn_wc_entry_callbacks_t", _wrap_delete_svn_wc_entry_callbacks_t, METH_VARARGS, (char *)"delete_svn_wc_entry_callbacks_t(svn_wc_entry_callbacks_t self)"},
	 { (char *)"svn_wc_entry_callbacks_t_swigregister", svn_wc_entry_callbacks_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_walk_entries3", _wrap_svn_wc_walk_entries3, METH_VARARGS, (char *)"\n"
		"svn_wc_walk_entries3(char const * path, svn_wc_adm_access_t * adm_access, svn_wc_entry_callbacks2_t walk_callbacks, \n"
		"    void * walk_baton, svn_depth_t depth, svn_boolean_t show_hidden, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_walk_entries2", _wrap_svn_wc_walk_entries2, METH_VARARGS, (char *)"\n"
		"svn_wc_walk_entries2(char const * path, svn_wc_adm_access_t * adm_access, svn_wc_entry_callbacks_t walk_callbacks, \n"
		"    void * walk_baton, svn_boolean_t show_hidden, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_walk_entries", _wrap_svn_wc_walk_entries, METH_VARARGS, (char *)"\n"
		"svn_wc_walk_entries(char const * path, svn_wc_adm_access_t * adm_access, svn_wc_entry_callbacks_t walk_callbacks, \n"
		"    void * walk_baton, svn_boolean_t show_hidden, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_mark_missing_deleted", _wrap_svn_wc_mark_missing_deleted, METH_VARARGS, (char *)"svn_wc_mark_missing_deleted(char const * path, svn_wc_adm_access_t * parent, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_ensure_adm4", _wrap_svn_wc_ensure_adm4, METH_VARARGS, (char *)"\n"
		"svn_wc_ensure_adm4(svn_wc_context_t * wc_ctx, char const * local_abspath, char const * url, char const * repos_root_url, \n"
		"    char const * repos_uuid, svn_revnum_t revision, svn_depth_t depth, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_ensure_adm3", _wrap_svn_wc_ensure_adm3, METH_VARARGS, (char *)"\n"
		"svn_wc_ensure_adm3(char const * path, char const * uuid, char const * url, char const * repos, svn_revnum_t revision, \n"
		"    svn_depth_t depth, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_ensure_adm2", _wrap_svn_wc_ensure_adm2, METH_VARARGS, (char *)"\n"
		"svn_wc_ensure_adm2(char const * path, char const * uuid, char const * url, char const * repos, svn_revnum_t revision, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_ensure_adm", _wrap_svn_wc_ensure_adm, METH_VARARGS, (char *)"svn_wc_ensure_adm(char const * path, char const * uuid, char const * url, svn_revnum_t revision, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_maybe_set_repos_root", _wrap_svn_wc_maybe_set_repos_root, METH_VARARGS, (char *)"svn_wc_maybe_set_repos_root(svn_wc_adm_access_t * adm_access, char const * path, char const * repos, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_status3_t_kind_set", _wrap_svn_wc_status3_t_kind_set, METH_VARARGS, (char *)"svn_wc_status3_t_kind_set(svn_wc_status3_t self, svn_node_kind_t kind)"},
	 { (char *)"svn_wc_status3_t_kind_get", _wrap_svn_wc_status3_t_kind_get, METH_VARARGS, (char *)"svn_wc_status3_t_kind_get(svn_wc_status3_t self) -> svn_node_kind_t"},
	 { (char *)"svn_wc_status3_t_depth_set", _wrap_svn_wc_status3_t_depth_set, METH_VARARGS, (char *)"svn_wc_status3_t_depth_set(svn_wc_status3_t self, svn_depth_t depth)"},
	 { (char *)"svn_wc_status3_t_depth_get", _wrap_svn_wc_status3_t_depth_get, METH_VARARGS, (char *)"svn_wc_status3_t_depth_get(svn_wc_status3_t self) -> svn_depth_t"},
	 { (char *)"svn_wc_status3_t_filesize_set", _wrap_svn_wc_status3_t_filesize_set, METH_VARARGS, (char *)"svn_wc_status3_t_filesize_set(svn_wc_status3_t self, svn_filesize_t filesize)"},
	 { (char *)"svn_wc_status3_t_filesize_get", _wrap_svn_wc_status3_t_filesize_get, METH_VARARGS, (char *)"svn_wc_status3_t_filesize_get(svn_wc_status3_t self) -> svn_filesize_t"},
	 { (char *)"svn_wc_status3_t_versioned_set", _wrap_svn_wc_status3_t_versioned_set, METH_VARARGS, (char *)"svn_wc_status3_t_versioned_set(svn_wc_status3_t self, svn_boolean_t versioned)"},
	 { (char *)"svn_wc_status3_t_versioned_get", _wrap_svn_wc_status3_t_versioned_get, METH_VARARGS, (char *)"svn_wc_status3_t_versioned_get(svn_wc_status3_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status3_t_conflicted_set", _wrap_svn_wc_status3_t_conflicted_set, METH_VARARGS, (char *)"svn_wc_status3_t_conflicted_set(svn_wc_status3_t self, svn_boolean_t conflicted)"},
	 { (char *)"svn_wc_status3_t_conflicted_get", _wrap_svn_wc_status3_t_conflicted_get, METH_VARARGS, (char *)"svn_wc_status3_t_conflicted_get(svn_wc_status3_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status3_t_node_status_set", _wrap_svn_wc_status3_t_node_status_set, METH_VARARGS, (char *)"svn_wc_status3_t_node_status_set(svn_wc_status3_t self, enum svn_wc_status_kind node_status)"},
	 { (char *)"svn_wc_status3_t_node_status_get", _wrap_svn_wc_status3_t_node_status_get, METH_VARARGS, (char *)"svn_wc_status3_t_node_status_get(svn_wc_status3_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status3_t_text_status_set", _wrap_svn_wc_status3_t_text_status_set, METH_VARARGS, (char *)"svn_wc_status3_t_text_status_set(svn_wc_status3_t self, enum svn_wc_status_kind text_status)"},
	 { (char *)"svn_wc_status3_t_text_status_get", _wrap_svn_wc_status3_t_text_status_get, METH_VARARGS, (char *)"svn_wc_status3_t_text_status_get(svn_wc_status3_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status3_t_prop_status_set", _wrap_svn_wc_status3_t_prop_status_set, METH_VARARGS, (char *)"svn_wc_status3_t_prop_status_set(svn_wc_status3_t self, enum svn_wc_status_kind prop_status)"},
	 { (char *)"svn_wc_status3_t_prop_status_get", _wrap_svn_wc_status3_t_prop_status_get, METH_VARARGS, (char *)"svn_wc_status3_t_prop_status_get(svn_wc_status3_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status3_t_copied_set", _wrap_svn_wc_status3_t_copied_set, METH_VARARGS, (char *)"svn_wc_status3_t_copied_set(svn_wc_status3_t self, svn_boolean_t copied)"},
	 { (char *)"svn_wc_status3_t_copied_get", _wrap_svn_wc_status3_t_copied_get, METH_VARARGS, (char *)"svn_wc_status3_t_copied_get(svn_wc_status3_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status3_t_revision_set", _wrap_svn_wc_status3_t_revision_set, METH_VARARGS, (char *)"svn_wc_status3_t_revision_set(svn_wc_status3_t self, svn_revnum_t revision)"},
	 { (char *)"svn_wc_status3_t_revision_get", _wrap_svn_wc_status3_t_revision_get, METH_VARARGS, (char *)"svn_wc_status3_t_revision_get(svn_wc_status3_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_status3_t_changed_rev_set", _wrap_svn_wc_status3_t_changed_rev_set, METH_VARARGS, (char *)"svn_wc_status3_t_changed_rev_set(svn_wc_status3_t self, svn_revnum_t changed_rev)"},
	 { (char *)"svn_wc_status3_t_changed_rev_get", _wrap_svn_wc_status3_t_changed_rev_get, METH_VARARGS, (char *)"svn_wc_status3_t_changed_rev_get(svn_wc_status3_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_status3_t_changed_date_set", _wrap_svn_wc_status3_t_changed_date_set, METH_VARARGS, (char *)"svn_wc_status3_t_changed_date_set(svn_wc_status3_t self, apr_time_t changed_date)"},
	 { (char *)"svn_wc_status3_t_changed_date_get", _wrap_svn_wc_status3_t_changed_date_get, METH_VARARGS, (char *)"svn_wc_status3_t_changed_date_get(svn_wc_status3_t self) -> apr_time_t"},
	 { (char *)"svn_wc_status3_t_changed_author_set", _wrap_svn_wc_status3_t_changed_author_set, METH_VARARGS, (char *)"svn_wc_status3_t_changed_author_set(svn_wc_status3_t self, char const * changed_author)"},
	 { (char *)"svn_wc_status3_t_changed_author_get", _wrap_svn_wc_status3_t_changed_author_get, METH_VARARGS, (char *)"svn_wc_status3_t_changed_author_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_repos_root_url_set", _wrap_svn_wc_status3_t_repos_root_url_set, METH_VARARGS, (char *)"svn_wc_status3_t_repos_root_url_set(svn_wc_status3_t self, char const * repos_root_url)"},
	 { (char *)"svn_wc_status3_t_repos_root_url_get", _wrap_svn_wc_status3_t_repos_root_url_get, METH_VARARGS, (char *)"svn_wc_status3_t_repos_root_url_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_repos_uuid_set", _wrap_svn_wc_status3_t_repos_uuid_set, METH_VARARGS, (char *)"svn_wc_status3_t_repos_uuid_set(svn_wc_status3_t self, char const * repos_uuid)"},
	 { (char *)"svn_wc_status3_t_repos_uuid_get", _wrap_svn_wc_status3_t_repos_uuid_get, METH_VARARGS, (char *)"svn_wc_status3_t_repos_uuid_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_repos_relpath_set", _wrap_svn_wc_status3_t_repos_relpath_set, METH_VARARGS, (char *)"svn_wc_status3_t_repos_relpath_set(svn_wc_status3_t self, char const * repos_relpath)"},
	 { (char *)"svn_wc_status3_t_repos_relpath_get", _wrap_svn_wc_status3_t_repos_relpath_get, METH_VARARGS, (char *)"svn_wc_status3_t_repos_relpath_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_switched_set", _wrap_svn_wc_status3_t_switched_set, METH_VARARGS, (char *)"svn_wc_status3_t_switched_set(svn_wc_status3_t self, svn_boolean_t switched)"},
	 { (char *)"svn_wc_status3_t_switched_get", _wrap_svn_wc_status3_t_switched_get, METH_VARARGS, (char *)"svn_wc_status3_t_switched_get(svn_wc_status3_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status3_t_locked_set", _wrap_svn_wc_status3_t_locked_set, METH_VARARGS, (char *)"svn_wc_status3_t_locked_set(svn_wc_status3_t self, svn_boolean_t locked)"},
	 { (char *)"svn_wc_status3_t_locked_get", _wrap_svn_wc_status3_t_locked_get, METH_VARARGS, (char *)"svn_wc_status3_t_locked_get(svn_wc_status3_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status3_t_lock_set", _wrap_svn_wc_status3_t_lock_set, METH_VARARGS, (char *)"svn_wc_status3_t_lock_set(svn_wc_status3_t self, svn_lock_t lock)"},
	 { (char *)"svn_wc_status3_t_lock_get", _wrap_svn_wc_status3_t_lock_get, METH_VARARGS, (char *)"svn_wc_status3_t_lock_get(svn_wc_status3_t self) -> svn_lock_t"},
	 { (char *)"svn_wc_status3_t_changelist_set", _wrap_svn_wc_status3_t_changelist_set, METH_VARARGS, (char *)"svn_wc_status3_t_changelist_set(svn_wc_status3_t self, char const * changelist)"},
	 { (char *)"svn_wc_status3_t_changelist_get", _wrap_svn_wc_status3_t_changelist_get, METH_VARARGS, (char *)"svn_wc_status3_t_changelist_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_ood_kind_set", _wrap_svn_wc_status3_t_ood_kind_set, METH_VARARGS, (char *)"svn_wc_status3_t_ood_kind_set(svn_wc_status3_t self, svn_node_kind_t ood_kind)"},
	 { (char *)"svn_wc_status3_t_ood_kind_get", _wrap_svn_wc_status3_t_ood_kind_get, METH_VARARGS, (char *)"svn_wc_status3_t_ood_kind_get(svn_wc_status3_t self) -> svn_node_kind_t"},
	 { (char *)"svn_wc_status3_t_repos_node_status_set", _wrap_svn_wc_status3_t_repos_node_status_set, METH_VARARGS, (char *)"svn_wc_status3_t_repos_node_status_set(svn_wc_status3_t self, enum svn_wc_status_kind repos_node_status)"},
	 { (char *)"svn_wc_status3_t_repos_node_status_get", _wrap_svn_wc_status3_t_repos_node_status_get, METH_VARARGS, (char *)"svn_wc_status3_t_repos_node_status_get(svn_wc_status3_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status3_t_repos_text_status_set", _wrap_svn_wc_status3_t_repos_text_status_set, METH_VARARGS, (char *)"svn_wc_status3_t_repos_text_status_set(svn_wc_status3_t self, enum svn_wc_status_kind repos_text_status)"},
	 { (char *)"svn_wc_status3_t_repos_text_status_get", _wrap_svn_wc_status3_t_repos_text_status_get, METH_VARARGS, (char *)"svn_wc_status3_t_repos_text_status_get(svn_wc_status3_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status3_t_repos_prop_status_set", _wrap_svn_wc_status3_t_repos_prop_status_set, METH_VARARGS, (char *)"svn_wc_status3_t_repos_prop_status_set(svn_wc_status3_t self, enum svn_wc_status_kind repos_prop_status)"},
	 { (char *)"svn_wc_status3_t_repos_prop_status_get", _wrap_svn_wc_status3_t_repos_prop_status_get, METH_VARARGS, (char *)"svn_wc_status3_t_repos_prop_status_get(svn_wc_status3_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status3_t_repos_lock_set", _wrap_svn_wc_status3_t_repos_lock_set, METH_VARARGS, (char *)"svn_wc_status3_t_repos_lock_set(svn_wc_status3_t self, svn_lock_t repos_lock)"},
	 { (char *)"svn_wc_status3_t_repos_lock_get", _wrap_svn_wc_status3_t_repos_lock_get, METH_VARARGS, (char *)"svn_wc_status3_t_repos_lock_get(svn_wc_status3_t self) -> svn_lock_t"},
	 { (char *)"svn_wc_status3_t_ood_changed_rev_set", _wrap_svn_wc_status3_t_ood_changed_rev_set, METH_VARARGS, (char *)"svn_wc_status3_t_ood_changed_rev_set(svn_wc_status3_t self, svn_revnum_t ood_changed_rev)"},
	 { (char *)"svn_wc_status3_t_ood_changed_rev_get", _wrap_svn_wc_status3_t_ood_changed_rev_get, METH_VARARGS, (char *)"svn_wc_status3_t_ood_changed_rev_get(svn_wc_status3_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_status3_t_ood_changed_date_set", _wrap_svn_wc_status3_t_ood_changed_date_set, METH_VARARGS, (char *)"svn_wc_status3_t_ood_changed_date_set(svn_wc_status3_t self, apr_time_t ood_changed_date)"},
	 { (char *)"svn_wc_status3_t_ood_changed_date_get", _wrap_svn_wc_status3_t_ood_changed_date_get, METH_VARARGS, (char *)"svn_wc_status3_t_ood_changed_date_get(svn_wc_status3_t self) -> apr_time_t"},
	 { (char *)"svn_wc_status3_t_ood_changed_author_set", _wrap_svn_wc_status3_t_ood_changed_author_set, METH_VARARGS, (char *)"svn_wc_status3_t_ood_changed_author_set(svn_wc_status3_t self, char const * ood_changed_author)"},
	 { (char *)"svn_wc_status3_t_ood_changed_author_get", _wrap_svn_wc_status3_t_ood_changed_author_get, METH_VARARGS, (char *)"svn_wc_status3_t_ood_changed_author_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_moved_from_abspath_set", _wrap_svn_wc_status3_t_moved_from_abspath_set, METH_VARARGS, (char *)"svn_wc_status3_t_moved_from_abspath_set(svn_wc_status3_t self, char const * moved_from_abspath)"},
	 { (char *)"svn_wc_status3_t_moved_from_abspath_get", _wrap_svn_wc_status3_t_moved_from_abspath_get, METH_VARARGS, (char *)"svn_wc_status3_t_moved_from_abspath_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_moved_to_abspath_set", _wrap_svn_wc_status3_t_moved_to_abspath_set, METH_VARARGS, (char *)"svn_wc_status3_t_moved_to_abspath_set(svn_wc_status3_t self, char const * moved_to_abspath)"},
	 { (char *)"svn_wc_status3_t_moved_to_abspath_get", _wrap_svn_wc_status3_t_moved_to_abspath_get, METH_VARARGS, (char *)"svn_wc_status3_t_moved_to_abspath_get(svn_wc_status3_t self) -> char const *"},
	 { (char *)"svn_wc_status3_t_file_external_set", _wrap_svn_wc_status3_t_file_external_set, METH_VARARGS, (char *)"svn_wc_status3_t_file_external_set(svn_wc_status3_t self, svn_boolean_t file_external)"},
	 { (char *)"svn_wc_status3_t_file_external_get", _wrap_svn_wc_status3_t_file_external_get, METH_VARARGS, (char *)"svn_wc_status3_t_file_external_get(svn_wc_status3_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status3_t_actual_kind_set", _wrap_svn_wc_status3_t_actual_kind_set, METH_VARARGS, (char *)"svn_wc_status3_t_actual_kind_set(svn_wc_status3_t self, svn_node_kind_t actual_kind)"},
	 { (char *)"svn_wc_status3_t_actual_kind_get", _wrap_svn_wc_status3_t_actual_kind_get, METH_VARARGS, (char *)"svn_wc_status3_t_actual_kind_get(svn_wc_status3_t self) -> svn_node_kind_t"},
	 { (char *)"new_svn_wc_status3_t", _wrap_new_svn_wc_status3_t, METH_VARARGS, (char *)"new_svn_wc_status3_t() -> svn_wc_status3_t"},
	 { (char *)"delete_svn_wc_status3_t", _wrap_delete_svn_wc_status3_t, METH_VARARGS, (char *)"delete_svn_wc_status3_t(svn_wc_status3_t self)"},
	 { (char *)"svn_wc_status3_t_swigregister", svn_wc_status3_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_status2_t_entry_set", _wrap_svn_wc_status2_t_entry_set, METH_VARARGS, (char *)"svn_wc_status2_t_entry_set(svn_wc_status2_t self, svn_wc_entry_t entry)"},
	 { (char *)"svn_wc_status2_t_entry_get", _wrap_svn_wc_status2_t_entry_get, METH_VARARGS, (char *)"svn_wc_status2_t_entry_get(svn_wc_status2_t self) -> svn_wc_entry_t"},
	 { (char *)"svn_wc_status2_t_text_status_set", _wrap_svn_wc_status2_t_text_status_set, METH_VARARGS, (char *)"svn_wc_status2_t_text_status_set(svn_wc_status2_t self, enum svn_wc_status_kind text_status)"},
	 { (char *)"svn_wc_status2_t_text_status_get", _wrap_svn_wc_status2_t_text_status_get, METH_VARARGS, (char *)"svn_wc_status2_t_text_status_get(svn_wc_status2_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status2_t_prop_status_set", _wrap_svn_wc_status2_t_prop_status_set, METH_VARARGS, (char *)"svn_wc_status2_t_prop_status_set(svn_wc_status2_t self, enum svn_wc_status_kind prop_status)"},
	 { (char *)"svn_wc_status2_t_prop_status_get", _wrap_svn_wc_status2_t_prop_status_get, METH_VARARGS, (char *)"svn_wc_status2_t_prop_status_get(svn_wc_status2_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status2_t_locked_set", _wrap_svn_wc_status2_t_locked_set, METH_VARARGS, (char *)"svn_wc_status2_t_locked_set(svn_wc_status2_t self, svn_boolean_t locked)"},
	 { (char *)"svn_wc_status2_t_locked_get", _wrap_svn_wc_status2_t_locked_get, METH_VARARGS, (char *)"svn_wc_status2_t_locked_get(svn_wc_status2_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status2_t_copied_set", _wrap_svn_wc_status2_t_copied_set, METH_VARARGS, (char *)"svn_wc_status2_t_copied_set(svn_wc_status2_t self, svn_boolean_t copied)"},
	 { (char *)"svn_wc_status2_t_copied_get", _wrap_svn_wc_status2_t_copied_get, METH_VARARGS, (char *)"svn_wc_status2_t_copied_get(svn_wc_status2_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status2_t_switched_set", _wrap_svn_wc_status2_t_switched_set, METH_VARARGS, (char *)"svn_wc_status2_t_switched_set(svn_wc_status2_t self, svn_boolean_t switched)"},
	 { (char *)"svn_wc_status2_t_switched_get", _wrap_svn_wc_status2_t_switched_get, METH_VARARGS, (char *)"svn_wc_status2_t_switched_get(svn_wc_status2_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status2_t_repos_text_status_set", _wrap_svn_wc_status2_t_repos_text_status_set, METH_VARARGS, (char *)"svn_wc_status2_t_repos_text_status_set(svn_wc_status2_t self, enum svn_wc_status_kind repos_text_status)"},
	 { (char *)"svn_wc_status2_t_repos_text_status_get", _wrap_svn_wc_status2_t_repos_text_status_get, METH_VARARGS, (char *)"svn_wc_status2_t_repos_text_status_get(svn_wc_status2_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status2_t_repos_prop_status_set", _wrap_svn_wc_status2_t_repos_prop_status_set, METH_VARARGS, (char *)"svn_wc_status2_t_repos_prop_status_set(svn_wc_status2_t self, enum svn_wc_status_kind repos_prop_status)"},
	 { (char *)"svn_wc_status2_t_repos_prop_status_get", _wrap_svn_wc_status2_t_repos_prop_status_get, METH_VARARGS, (char *)"svn_wc_status2_t_repos_prop_status_get(svn_wc_status2_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status2_t_repos_lock_set", _wrap_svn_wc_status2_t_repos_lock_set, METH_VARARGS, (char *)"svn_wc_status2_t_repos_lock_set(svn_wc_status2_t self, svn_lock_t repos_lock)"},
	 { (char *)"svn_wc_status2_t_repos_lock_get", _wrap_svn_wc_status2_t_repos_lock_get, METH_VARARGS, (char *)"svn_wc_status2_t_repos_lock_get(svn_wc_status2_t self) -> svn_lock_t"},
	 { (char *)"svn_wc_status2_t_url_set", _wrap_svn_wc_status2_t_url_set, METH_VARARGS, (char *)"svn_wc_status2_t_url_set(svn_wc_status2_t self, char const * url)"},
	 { (char *)"svn_wc_status2_t_url_get", _wrap_svn_wc_status2_t_url_get, METH_VARARGS, (char *)"svn_wc_status2_t_url_get(svn_wc_status2_t self) -> char const *"},
	 { (char *)"svn_wc_status2_t_ood_last_cmt_rev_set", _wrap_svn_wc_status2_t_ood_last_cmt_rev_set, METH_VARARGS, (char *)"svn_wc_status2_t_ood_last_cmt_rev_set(svn_wc_status2_t self, svn_revnum_t ood_last_cmt_rev)"},
	 { (char *)"svn_wc_status2_t_ood_last_cmt_rev_get", _wrap_svn_wc_status2_t_ood_last_cmt_rev_get, METH_VARARGS, (char *)"svn_wc_status2_t_ood_last_cmt_rev_get(svn_wc_status2_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_status2_t_ood_last_cmt_date_set", _wrap_svn_wc_status2_t_ood_last_cmt_date_set, METH_VARARGS, (char *)"svn_wc_status2_t_ood_last_cmt_date_set(svn_wc_status2_t self, apr_time_t ood_last_cmt_date)"},
	 { (char *)"svn_wc_status2_t_ood_last_cmt_date_get", _wrap_svn_wc_status2_t_ood_last_cmt_date_get, METH_VARARGS, (char *)"svn_wc_status2_t_ood_last_cmt_date_get(svn_wc_status2_t self) -> apr_time_t"},
	 { (char *)"svn_wc_status2_t_ood_kind_set", _wrap_svn_wc_status2_t_ood_kind_set, METH_VARARGS, (char *)"svn_wc_status2_t_ood_kind_set(svn_wc_status2_t self, svn_node_kind_t ood_kind)"},
	 { (char *)"svn_wc_status2_t_ood_kind_get", _wrap_svn_wc_status2_t_ood_kind_get, METH_VARARGS, (char *)"svn_wc_status2_t_ood_kind_get(svn_wc_status2_t self) -> svn_node_kind_t"},
	 { (char *)"svn_wc_status2_t_ood_last_cmt_author_set", _wrap_svn_wc_status2_t_ood_last_cmt_author_set, METH_VARARGS, (char *)"svn_wc_status2_t_ood_last_cmt_author_set(svn_wc_status2_t self, char const * ood_last_cmt_author)"},
	 { (char *)"svn_wc_status2_t_ood_last_cmt_author_get", _wrap_svn_wc_status2_t_ood_last_cmt_author_get, METH_VARARGS, (char *)"svn_wc_status2_t_ood_last_cmt_author_get(svn_wc_status2_t self) -> char const *"},
	 { (char *)"svn_wc_status2_t_tree_conflict_set", _wrap_svn_wc_status2_t_tree_conflict_set, METH_VARARGS, (char *)"svn_wc_status2_t_tree_conflict_set(svn_wc_status2_t self, svn_wc_conflict_description_t tree_conflict)"},
	 { (char *)"svn_wc_status2_t_tree_conflict_get", _wrap_svn_wc_status2_t_tree_conflict_get, METH_VARARGS, (char *)"svn_wc_status2_t_tree_conflict_get(svn_wc_status2_t self) -> svn_wc_conflict_description_t"},
	 { (char *)"svn_wc_status2_t_file_external_set", _wrap_svn_wc_status2_t_file_external_set, METH_VARARGS, (char *)"svn_wc_status2_t_file_external_set(svn_wc_status2_t self, svn_boolean_t file_external)"},
	 { (char *)"svn_wc_status2_t_file_external_get", _wrap_svn_wc_status2_t_file_external_get, METH_VARARGS, (char *)"svn_wc_status2_t_file_external_get(svn_wc_status2_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status2_t_pristine_text_status_set", _wrap_svn_wc_status2_t_pristine_text_status_set, METH_VARARGS, (char *)"svn_wc_status2_t_pristine_text_status_set(svn_wc_status2_t self, enum svn_wc_status_kind pristine_text_status)"},
	 { (char *)"svn_wc_status2_t_pristine_text_status_get", _wrap_svn_wc_status2_t_pristine_text_status_get, METH_VARARGS, (char *)"svn_wc_status2_t_pristine_text_status_get(svn_wc_status2_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status2_t_pristine_prop_status_set", _wrap_svn_wc_status2_t_pristine_prop_status_set, METH_VARARGS, (char *)"svn_wc_status2_t_pristine_prop_status_set(svn_wc_status2_t self, enum svn_wc_status_kind pristine_prop_status)"},
	 { (char *)"svn_wc_status2_t_pristine_prop_status_get", _wrap_svn_wc_status2_t_pristine_prop_status_get, METH_VARARGS, (char *)"svn_wc_status2_t_pristine_prop_status_get(svn_wc_status2_t self) -> enum svn_wc_status_kind"},
	 { (char *)"new_svn_wc_status2_t", _wrap_new_svn_wc_status2_t, METH_VARARGS, (char *)"new_svn_wc_status2_t() -> svn_wc_status2_t"},
	 { (char *)"delete_svn_wc_status2_t", _wrap_delete_svn_wc_status2_t, METH_VARARGS, (char *)"delete_svn_wc_status2_t(svn_wc_status2_t self)"},
	 { (char *)"svn_wc_status2_t_swigregister", svn_wc_status2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_status_t_entry_set", _wrap_svn_wc_status_t_entry_set, METH_VARARGS, (char *)"svn_wc_status_t_entry_set(svn_wc_status_t self, svn_wc_entry_t entry)"},
	 { (char *)"svn_wc_status_t_entry_get", _wrap_svn_wc_status_t_entry_get, METH_VARARGS, (char *)"svn_wc_status_t_entry_get(svn_wc_status_t self) -> svn_wc_entry_t"},
	 { (char *)"svn_wc_status_t_text_status_set", _wrap_svn_wc_status_t_text_status_set, METH_VARARGS, (char *)"svn_wc_status_t_text_status_set(svn_wc_status_t self, enum svn_wc_status_kind text_status)"},
	 { (char *)"svn_wc_status_t_text_status_get", _wrap_svn_wc_status_t_text_status_get, METH_VARARGS, (char *)"svn_wc_status_t_text_status_get(svn_wc_status_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status_t_prop_status_set", _wrap_svn_wc_status_t_prop_status_set, METH_VARARGS, (char *)"svn_wc_status_t_prop_status_set(svn_wc_status_t self, enum svn_wc_status_kind prop_status)"},
	 { (char *)"svn_wc_status_t_prop_status_get", _wrap_svn_wc_status_t_prop_status_get, METH_VARARGS, (char *)"svn_wc_status_t_prop_status_get(svn_wc_status_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status_t_locked_set", _wrap_svn_wc_status_t_locked_set, METH_VARARGS, (char *)"svn_wc_status_t_locked_set(svn_wc_status_t self, svn_boolean_t locked)"},
	 { (char *)"svn_wc_status_t_locked_get", _wrap_svn_wc_status_t_locked_get, METH_VARARGS, (char *)"svn_wc_status_t_locked_get(svn_wc_status_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status_t_copied_set", _wrap_svn_wc_status_t_copied_set, METH_VARARGS, (char *)"svn_wc_status_t_copied_set(svn_wc_status_t self, svn_boolean_t copied)"},
	 { (char *)"svn_wc_status_t_copied_get", _wrap_svn_wc_status_t_copied_get, METH_VARARGS, (char *)"svn_wc_status_t_copied_get(svn_wc_status_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status_t_switched_set", _wrap_svn_wc_status_t_switched_set, METH_VARARGS, (char *)"svn_wc_status_t_switched_set(svn_wc_status_t self, svn_boolean_t switched)"},
	 { (char *)"svn_wc_status_t_switched_get", _wrap_svn_wc_status_t_switched_get, METH_VARARGS, (char *)"svn_wc_status_t_switched_get(svn_wc_status_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_status_t_repos_text_status_set", _wrap_svn_wc_status_t_repos_text_status_set, METH_VARARGS, (char *)"svn_wc_status_t_repos_text_status_set(svn_wc_status_t self, enum svn_wc_status_kind repos_text_status)"},
	 { (char *)"svn_wc_status_t_repos_text_status_get", _wrap_svn_wc_status_t_repos_text_status_get, METH_VARARGS, (char *)"svn_wc_status_t_repos_text_status_get(svn_wc_status_t self) -> enum svn_wc_status_kind"},
	 { (char *)"svn_wc_status_t_repos_prop_status_set", _wrap_svn_wc_status_t_repos_prop_status_set, METH_VARARGS, (char *)"svn_wc_status_t_repos_prop_status_set(svn_wc_status_t self, enum svn_wc_status_kind repos_prop_status)"},
	 { (char *)"svn_wc_status_t_repos_prop_status_get", _wrap_svn_wc_status_t_repos_prop_status_get, METH_VARARGS, (char *)"svn_wc_status_t_repos_prop_status_get(svn_wc_status_t self) -> enum svn_wc_status_kind"},
	 { (char *)"new_svn_wc_status_t", _wrap_new_svn_wc_status_t, METH_VARARGS, (char *)"new_svn_wc_status_t() -> svn_wc_status_t"},
	 { (char *)"delete_svn_wc_status_t", _wrap_delete_svn_wc_status_t, METH_VARARGS, (char *)"delete_svn_wc_status_t(svn_wc_status_t self)"},
	 { (char *)"svn_wc_status_t_swigregister", svn_wc_status_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_dup_status3", _wrap_svn_wc_dup_status3, METH_VARARGS, (char *)"svn_wc_dup_status3(svn_wc_status3_t orig_stat, apr_pool_t pool) -> svn_wc_status3_t"},
	 { (char *)"svn_wc_dup_status2", _wrap_svn_wc_dup_status2, METH_VARARGS, (char *)"svn_wc_dup_status2(svn_wc_status2_t orig_stat, apr_pool_t pool) -> svn_wc_status2_t"},
	 { (char *)"svn_wc_dup_status", _wrap_svn_wc_dup_status, METH_VARARGS, (char *)"svn_wc_dup_status(svn_wc_status_t orig_stat, apr_pool_t pool) -> svn_wc_status_t"},
	 { (char *)"svn_wc_status3", _wrap_svn_wc_status3, METH_VARARGS, (char *)"svn_wc_status3(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_status2", _wrap_svn_wc_status2, METH_VARARGS, (char *)"svn_wc_status2(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_status", _wrap_svn_wc_status, METH_VARARGS, (char *)"svn_wc_status(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_walk_status", _wrap_svn_wc_walk_status, METH_VARARGS, (char *)"\n"
		"svn_wc_walk_status(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_depth_t depth, svn_boolean_t get_all, \n"
		"    svn_boolean_t no_ignore, svn_boolean_t ignore_text_mods, apr_array_header_t ignore_patterns, \n"
		"    svn_wc_status_func4_t status_func, void * status_baton, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_status_editor5", _wrap_svn_wc_get_status_editor5, METH_VARARGS, (char *)"\n"
		"svn_wc_get_status_editor5(svn_wc_context_t * wc_ctx, char const * anchor_abspath, char const * target_basename, \n"
		"    svn_depth_t depth, svn_boolean_t get_all, svn_boolean_t no_ignore, svn_boolean_t depth_as_sticky, \n"
		"    svn_boolean_t server_performs_filtering, apr_array_header_t ignore_patterns, \n"
		"    svn_wc_status_func4_t status_func, void * status_baton, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_status_editor4", _wrap_svn_wc_get_status_editor4, METH_VARARGS, (char *)"\n"
		"svn_wc_get_status_editor4(svn_wc_adm_access_t * anchor, char const * target, svn_depth_t depth, svn_boolean_t get_all, \n"
		"    svn_boolean_t no_ignore, apr_array_header_t ignore_patterns, svn_wc_status_func3_t status_func, \n"
		"    void * status_baton, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_traversal_info_t * traversal_info, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_status_editor3", _wrap_svn_wc_get_status_editor3, METH_VARARGS, (char *)"\n"
		"svn_wc_get_status_editor3(svn_wc_adm_access_t * anchor, char const * target, svn_depth_t depth, svn_boolean_t get_all, \n"
		"    svn_boolean_t no_ignore, apr_array_header_t ignore_patterns, svn_wc_status_func2_t status_func, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_traversal_info_t * traversal_info, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_status_editor2", _wrap_svn_wc_get_status_editor2, METH_VARARGS, (char *)"\n"
		"svn_wc_get_status_editor2(svn_wc_adm_access_t * anchor, char const * target, apr_hash_t config, svn_boolean_t recurse, \n"
		"    svn_boolean_t get_all, svn_boolean_t no_ignore, svn_wc_status_func2_t status_func, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_traversal_info_t * traversal_info, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_status_editor", _wrap_svn_wc_get_status_editor, METH_VARARGS, (char *)"\n"
		"svn_wc_get_status_editor(svn_wc_adm_access_t * anchor, char const * target, apr_hash_t config, svn_boolean_t recurse, \n"
		"    svn_boolean_t get_all, svn_boolean_t no_ignore, svn_wc_status_func_t status_func, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_traversal_info_t * traversal_info, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_status_set_repos_locks", _wrap_svn_wc_status_set_repos_locks, METH_VARARGS, (char *)"svn_wc_status_set_repos_locks(void * set_locks_baton, apr_hash_t locks, char const * repos_root, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_copy3", _wrap_svn_wc_copy3, METH_VARARGS, (char *)"\n"
		"svn_wc_copy3(svn_wc_context_t * wc_ctx, char const * src_abspath, char const * dst_abspath, svn_boolean_t metadata_only, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_copy2", _wrap_svn_wc_copy2, METH_VARARGS, (char *)"\n"
		"svn_wc_copy2(char const * src, svn_wc_adm_access_t * dst_parent, char const * dst_basename, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_copy", _wrap_svn_wc_copy, METH_VARARGS, (char *)"\n"
		"svn_wc_copy(char const * src, svn_wc_adm_access_t * dst_parent, char const * dst_basename, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func_t notify_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_move", _wrap_svn_wc_move, METH_VARARGS, (char *)"\n"
		"svn_wc_move(svn_wc_context_t * wc_ctx, char const * src_abspath, char const * dst_abspath, svn_boolean_t metadata_only, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_delete4", _wrap_svn_wc_delete4, METH_VARARGS, (char *)"\n"
		"svn_wc_delete4(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_boolean_t keep_local, \n"
		"    svn_boolean_t delete_unversioned_target, svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_delete3", _wrap_svn_wc_delete3, METH_VARARGS, (char *)"\n"
		"svn_wc_delete3(char const * path, svn_wc_adm_access_t * adm_access, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, svn_boolean_t keep_local, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_delete2", _wrap_svn_wc_delete2, METH_VARARGS, (char *)"\n"
		"svn_wc_delete2(char const * path, svn_wc_adm_access_t * adm_access, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_delete", _wrap_svn_wc_delete, METH_VARARGS, (char *)"\n"
		"svn_wc_delete(char const * path, svn_wc_adm_access_t * adm_access, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func_t notify_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add_from_disk3", _wrap_svn_wc_add_from_disk3, METH_VARARGS, (char *)"\n"
		"svn_wc_add_from_disk3(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_hash_t props, svn_boolean_t skip_checks, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add_from_disk2", _wrap_svn_wc_add_from_disk2, METH_VARARGS, (char *)"\n"
		"svn_wc_add_from_disk2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_hash_t props, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add_from_disk", _wrap_svn_wc_add_from_disk, METH_VARARGS, (char *)"\n"
		"svn_wc_add_from_disk(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add4", _wrap_svn_wc_add4, METH_VARARGS, (char *)"\n"
		"svn_wc_add4(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_depth_t depth, char const * copyfrom_url, \n"
		"    svn_revnum_t copyfrom_rev, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add3", _wrap_svn_wc_add3, METH_VARARGS, (char *)"\n"
		"svn_wc_add3(char const * path, svn_wc_adm_access_t * parent_access, svn_depth_t depth, char const * copyfrom_url, \n"
		"    svn_revnum_t copyfrom_rev, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add2", _wrap_svn_wc_add2, METH_VARARGS, (char *)"\n"
		"svn_wc_add2(char const * path, svn_wc_adm_access_t * parent_access, char const * copyfrom_url, \n"
		"    svn_revnum_t copyfrom_rev, svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add", _wrap_svn_wc_add, METH_VARARGS, (char *)"\n"
		"svn_wc_add(char const * path, svn_wc_adm_access_t * parent_access, char const * copyfrom_url, \n"
		"    svn_revnum_t copyfrom_rev, svn_cancel_func_t cancel_func, svn_wc_notify_func_t notify_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add_repos_file4", _wrap_svn_wc_add_repos_file4, METH_VARARGS, (char *)"\n"
		"svn_wc_add_repos_file4(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_stream_t * new_base_contents, \n"
		"    svn_stream_t * new_contents, apr_hash_t new_base_props, apr_hash_t new_props, \n"
		"    char const * copyfrom_url, svn_revnum_t copyfrom_rev, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add_repos_file3", _wrap_svn_wc_add_repos_file3, METH_VARARGS, (char *)"\n"
		"svn_wc_add_repos_file3(char const * dst_path, svn_wc_adm_access_t * adm_access, svn_stream_t * new_base_contents, \n"
		"    svn_stream_t * new_contents, apr_hash_t new_base_props, apr_hash_t new_props, \n"
		"    char const * copyfrom_url, svn_revnum_t copyfrom_rev, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add_repos_file2", _wrap_svn_wc_add_repos_file2, METH_VARARGS, (char *)"\n"
		"svn_wc_add_repos_file2(char const * dst_path, svn_wc_adm_access_t * adm_access, char const * new_text_base_path, \n"
		"    char const * new_text_path, apr_hash_t new_base_props, apr_hash_t new_props, \n"
		"    char const * copyfrom_url, svn_revnum_t copyfrom_rev, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_add_repos_file", _wrap_svn_wc_add_repos_file, METH_VARARGS, (char *)"\n"
		"svn_wc_add_repos_file(char const * dst_path, svn_wc_adm_access_t * adm_access, char const * new_text_path, \n"
		"    apr_hash_t new_props, char const * copyfrom_url, svn_revnum_t copyfrom_rev, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_remove_from_revision_control2", _wrap_svn_wc_remove_from_revision_control2, METH_VARARGS, (char *)"\n"
		"svn_wc_remove_from_revision_control2(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_boolean_t destroy_wf, \n"
		"    svn_boolean_t instant_error, svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_remove_from_revision_control", _wrap_svn_wc_remove_from_revision_control, METH_VARARGS, (char *)"\n"
		"svn_wc_remove_from_revision_control(svn_wc_adm_access_t * adm_access, char const * name, svn_boolean_t destroy_wf, svn_boolean_t instant_error, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_resolved_conflict5", _wrap_svn_wc_resolved_conflict5, METH_VARARGS, (char *)"\n"
		"svn_wc_resolved_conflict5(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_depth_t depth, svn_boolean_t resolve_text, \n"
		"    char const * resolve_prop, svn_boolean_t resolve_tree, \n"
		"    svn_wc_conflict_choice_t conflict_choice, svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_resolved_conflict4", _wrap_svn_wc_resolved_conflict4, METH_VARARGS, (char *)"\n"
		"svn_wc_resolved_conflict4(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t resolve_text, \n"
		"    svn_boolean_t resolve_props, svn_boolean_t resolve_tree, svn_depth_t depth, \n"
		"    svn_wc_conflict_choice_t conflict_choice, svn_wc_notify_func2_t notify_func, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_resolved_conflict3", _wrap_svn_wc_resolved_conflict3, METH_VARARGS, (char *)"\n"
		"svn_wc_resolved_conflict3(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t resolve_text, \n"
		"    svn_boolean_t resolve_props, svn_depth_t depth, svn_wc_conflict_choice_t conflict_choice, \n"
		"    svn_wc_notify_func2_t notify_func, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_resolved_conflict2", _wrap_svn_wc_resolved_conflict2, METH_VARARGS, (char *)"\n"
		"svn_wc_resolved_conflict2(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t resolve_text, \n"
		"    svn_boolean_t resolve_props, svn_boolean_t recurse, svn_wc_notify_func2_t notify_func, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_resolved_conflict", _wrap_svn_wc_resolved_conflict, METH_VARARGS, (char *)"\n"
		"svn_wc_resolved_conflict(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t resolve_text, \n"
		"    svn_boolean_t resolve_props, svn_boolean_t recurse, svn_wc_notify_func_t notify_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_committed_queue_create", _wrap_svn_wc_committed_queue_create, METH_VARARGS, (char *)"svn_wc_committed_queue_create(apr_pool_t pool) -> svn_wc_committed_queue_t *"},
	 { (char *)"svn_wc_queue_committed4", _wrap_svn_wc_queue_committed4, METH_VARARGS, (char *)"\n"
		"svn_wc_queue_committed4(svn_wc_committed_queue_t * queue, svn_wc_context_t * wc_ctx, char const * local_abspath, \n"
		"    svn_boolean_t recurse, svn_boolean_t is_committed, apr_array_header_t wcprop_changes, \n"
		"    svn_boolean_t remove_lock, svn_boolean_t remove_changelist, \n"
		"    svn_checksum_t sha1_checksum, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_queue_committed3", _wrap_svn_wc_queue_committed3, METH_VARARGS, (char *)"\n"
		"svn_wc_queue_committed3(svn_wc_committed_queue_t * queue, svn_wc_context_t * wc_ctx, char const * local_abspath, \n"
		"    svn_boolean_t recurse, apr_array_header_t wcprop_changes, svn_boolean_t remove_lock, \n"
		"    svn_boolean_t remove_changelist, svn_checksum_t sha1_checksum, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_queue_committed2", _wrap_svn_wc_queue_committed2, METH_VARARGS, (char *)"\n"
		"svn_wc_queue_committed2(svn_wc_committed_queue_t * queue, char const * path, svn_wc_adm_access_t * adm_access, \n"
		"    svn_boolean_t recurse, apr_array_header_t wcprop_changes, svn_boolean_t remove_lock, \n"
		"    svn_boolean_t remove_changelist, svn_checksum_t md5_checksum, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_queue_committed", _wrap_svn_wc_queue_committed, METH_VARARGS, (char *)"\n"
		"svn_wc_queue_committed(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t recurse, apr_array_header_t wcprop_changes, \n"
		"    svn_boolean_t remove_lock, svn_boolean_t remove_changelist, \n"
		"    unsigned char const * digest, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_process_committed_queue2", _wrap_svn_wc_process_committed_queue2, METH_VARARGS, (char *)"\n"
		"svn_wc_process_committed_queue2(svn_wc_committed_queue_t * queue, svn_wc_context_t * wc_ctx, svn_revnum_t new_revnum, \n"
		"    char const * rev_date, char const * rev_author, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_process_committed_queue", _wrap_svn_wc_process_committed_queue, METH_VARARGS, (char *)"\n"
		"svn_wc_process_committed_queue(svn_wc_committed_queue_t * queue, svn_wc_adm_access_t * adm_access, svn_revnum_t new_revnum, \n"
		"    char const * rev_date, char const * rev_author, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_process_committed4", _wrap_svn_wc_process_committed4, METH_VARARGS, (char *)"\n"
		"svn_wc_process_committed4(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t recurse, svn_revnum_t new_revnum, \n"
		"    char const * rev_date, char const * rev_author, apr_array_header_t wcprop_changes, \n"
		"    svn_boolean_t remove_lock, svn_boolean_t remove_changelist, \n"
		"    unsigned char const * digest, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_process_committed3", _wrap_svn_wc_process_committed3, METH_VARARGS, (char *)"\n"
		"svn_wc_process_committed3(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t recurse, svn_revnum_t new_revnum, \n"
		"    char const * rev_date, char const * rev_author, apr_array_header_t wcprop_changes, \n"
		"    svn_boolean_t remove_lock, unsigned char const * digest, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_process_committed2", _wrap_svn_wc_process_committed2, METH_VARARGS, (char *)"\n"
		"svn_wc_process_committed2(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t recurse, svn_revnum_t new_revnum, \n"
		"    char const * rev_date, char const * rev_author, apr_array_header_t wcprop_changes, \n"
		"    svn_boolean_t remove_lock, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_process_committed", _wrap_svn_wc_process_committed, METH_VARARGS, (char *)"\n"
		"svn_wc_process_committed(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t recurse, svn_revnum_t new_revnum, \n"
		"    char const * rev_date, char const * rev_author, apr_array_header_t wcprop_changes, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_crawl_revisions5", _wrap_svn_wc_crawl_revisions5, METH_VARARGS, (char *)"\n"
		"svn_wc_crawl_revisions5(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_ra_reporter3_t reporter, \n"
		"    void * report_baton, svn_boolean_t restore_files, svn_depth_t depth, svn_boolean_t honor_depth_exclude, \n"
		"    svn_boolean_t depth_compatibility_trick, svn_boolean_t use_commit_times, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_crawl_revisions4", _wrap_svn_wc_crawl_revisions4, METH_VARARGS, (char *)"\n"
		"svn_wc_crawl_revisions4(char const * path, svn_wc_adm_access_t * adm_access, svn_ra_reporter3_t reporter, \n"
		"    void * report_baton, svn_boolean_t restore_files, svn_depth_t depth, svn_boolean_t honor_depth_exclude, \n"
		"    svn_boolean_t depth_compatibility_trick, svn_boolean_t use_commit_times, \n"
		"    svn_wc_notify_func2_t notify_func, svn_wc_traversal_info_t * traversal_info, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_crawl_revisions3", _wrap_svn_wc_crawl_revisions3, METH_VARARGS, (char *)"\n"
		"svn_wc_crawl_revisions3(char const * path, svn_wc_adm_access_t * adm_access, svn_ra_reporter3_t reporter, \n"
		"    void * report_baton, svn_boolean_t restore_files, svn_depth_t depth, svn_boolean_t depth_compatibility_trick, \n"
		"    svn_boolean_t use_commit_times, svn_wc_notify_func2_t notify_func, \n"
		"    svn_wc_traversal_info_t * traversal_info, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_crawl_revisions2", _wrap_svn_wc_crawl_revisions2, METH_VARARGS, (char *)"\n"
		"svn_wc_crawl_revisions2(char const * path, svn_wc_adm_access_t * adm_access, svn_ra_reporter2_t reporter, \n"
		"    svn_boolean_t restore_files, svn_boolean_t recurse, svn_boolean_t use_commit_times, \n"
		"    svn_wc_notify_func2_t notify_func, svn_wc_traversal_info_t * traversal_info, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_crawl_revisions", _wrap_svn_wc_crawl_revisions, METH_VARARGS, (char *)"\n"
		"svn_wc_crawl_revisions(char const * path, svn_wc_adm_access_t * adm_access, svn_ra_reporter_t reporter, \n"
		"    void * report_baton, svn_boolean_t restore_files, svn_boolean_t recurse, svn_boolean_t use_commit_times, \n"
		"    svn_wc_notify_func_t notify_func, svn_wc_traversal_info_t * traversal_info, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_check_root", _wrap_svn_wc_check_root, METH_VARARGS, (char *)"svn_wc_check_root(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_is_wc_root2", _wrap_svn_wc_is_wc_root2, METH_VARARGS, (char *)"svn_wc_is_wc_root2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_is_wc_root", _wrap_svn_wc_is_wc_root, METH_VARARGS, (char *)"svn_wc_is_wc_root(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_actual_target2", _wrap_svn_wc_get_actual_target2, METH_VARARGS, (char *)"svn_wc_get_actual_target2(svn_wc_context_t * wc_ctx, char const * path, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_actual_target", _wrap_svn_wc_get_actual_target, METH_VARARGS, (char *)"svn_wc_get_actual_target(char const * path, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_update_editor4", _wrap_svn_wc_get_update_editor4, METH_VARARGS, (char *)"\n"
		"svn_wc_get_update_editor4(svn_wc_context_t * wc_ctx, char const * anchor_abspath, char const * target_basename, \n"
		"    svn_boolean_t use_commit_times, svn_depth_t depth, svn_boolean_t depth_is_sticky, \n"
		"    svn_boolean_t allow_unver_obstructions, svn_boolean_t adds_as_modification, \n"
		"    svn_boolean_t server_performs_filtering, svn_boolean_t clean_checkout, \n"
		"    char const * diff3_cmd, apr_array_header_t preserved_exts, svn_wc_dirents_func_t fetch_dirents_func, \n"
		"    void * fetch_dirents_baton, svn_wc_conflict_resolver_func2_t conflict_func, \n"
		"    void * conflict_baton, svn_wc_external_update_t external_func, \n"
		"    void * external_baton, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_update_editor3", _wrap_svn_wc_get_update_editor3, METH_VARARGS, (char *)"\n"
		"svn_wc_get_update_editor3(svn_wc_adm_access_t * anchor, char const * target, svn_boolean_t use_commit_times, \n"
		"    svn_depth_t depth, svn_boolean_t depth_is_sticky, svn_boolean_t allow_unver_obstructions, \n"
		"    svn_wc_notify_func2_t notify_func, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_conflict_resolver_func_t conflict_func, void * conflict_baton, \n"
		"    svn_wc_get_file_t fetch_func, void * fetch_baton, char const * diff3_cmd, \n"
		"    apr_array_header_t preserved_exts, svn_wc_traversal_info_t * ti, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_update_editor2", _wrap_svn_wc_get_update_editor2, METH_VARARGS, (char *)"\n"
		"svn_wc_get_update_editor2(svn_wc_adm_access_t * anchor, char const * target, svn_boolean_t use_commit_times, \n"
		"    svn_boolean_t recurse, svn_wc_notify_func2_t notify_func, svn_cancel_func_t cancel_func, \n"
		"    char const * diff3_cmd, svn_wc_traversal_info_t * ti, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_update_editor", _wrap_svn_wc_get_update_editor, METH_VARARGS, (char *)"\n"
		"svn_wc_get_update_editor(svn_wc_adm_access_t * anchor, char const * target, svn_boolean_t use_commit_times, \n"
		"    svn_boolean_t recurse, svn_wc_notify_func_t notify_func, svn_cancel_func_t cancel_func, \n"
		"    char const * diff3_cmd, svn_wc_traversal_info_t * ti, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_switch_editor4", _wrap_svn_wc_get_switch_editor4, METH_VARARGS, (char *)"\n"
		"svn_wc_get_switch_editor4(svn_wc_context_t * wc_ctx, char const * anchor_abspath, char const * target_basename, \n"
		"    char const * switch_url, svn_boolean_t use_commit_times, svn_depth_t depth, \n"
		"    svn_boolean_t depth_is_sticky, svn_boolean_t allow_unver_obstructions, \n"
		"    svn_boolean_t server_performs_filtering, char const * diff3_cmd, apr_array_header_t preserved_exts, \n"
		"    svn_wc_dirents_func_t fetch_dirents_func, void * fetch_dirents_baton, \n"
		"    svn_wc_conflict_resolver_func2_t conflict_func, void * conflict_baton, \n"
		"    svn_wc_external_update_t external_func, void * external_baton, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, apr_pool_t result_pool, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_switch_editor3", _wrap_svn_wc_get_switch_editor3, METH_VARARGS, (char *)"\n"
		"svn_wc_get_switch_editor3(svn_wc_adm_access_t * anchor, char const * target, char const * switch_url, svn_boolean_t use_commit_times, \n"
		"    svn_depth_t depth, svn_boolean_t depth_is_sticky, \n"
		"    svn_boolean_t allow_unver_obstructions, svn_wc_notify_func2_t notify_func, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_conflict_resolver_func_t conflict_func, \n"
		"    void * conflict_baton, char const * diff3_cmd, apr_array_header_t preserved_exts, \n"
		"    svn_wc_traversal_info_t * ti, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_switch_editor2", _wrap_svn_wc_get_switch_editor2, METH_VARARGS, (char *)"\n"
		"svn_wc_get_switch_editor2(svn_wc_adm_access_t * anchor, char const * target, char const * switch_url, svn_boolean_t use_commit_times, \n"
		"    svn_boolean_t recurse, svn_wc_notify_func2_t notify_func, \n"
		"    svn_cancel_func_t cancel_func, char const * diff3_cmd, svn_wc_traversal_info_t * ti, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_switch_editor", _wrap_svn_wc_get_switch_editor, METH_VARARGS, (char *)"\n"
		"svn_wc_get_switch_editor(svn_wc_adm_access_t * anchor, char const * target, char const * switch_url, svn_boolean_t use_commit_times, \n"
		"    svn_boolean_t recurse, svn_wc_notify_func_t notify_func, \n"
		"    svn_cancel_func_t cancel_func, char const * diff3_cmd, svn_wc_traversal_info_t * ti, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_prop_list2", _wrap_svn_wc_prop_list2, METH_VARARGS, (char *)"svn_wc_prop_list2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_prop_list", _wrap_svn_wc_prop_list, METH_VARARGS, (char *)"svn_wc_prop_list(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_pristine_props", _wrap_svn_wc_get_pristine_props, METH_VARARGS, (char *)"svn_wc_get_pristine_props(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_prop_get2", _wrap_svn_wc_prop_get2, METH_VARARGS, (char *)"\n"
		"svn_wc_prop_get2(svn_wc_context_t * wc_ctx, char const * local_abspath, char const * name, apr_pool_t result_pool, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_prop_get", _wrap_svn_wc_prop_get, METH_VARARGS, (char *)"svn_wc_prop_get(char const * name, char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_prop_set4", _wrap_svn_wc_prop_set4, METH_VARARGS, (char *)"\n"
		"svn_wc_prop_set4(svn_wc_context_t * wc_ctx, char const * local_abspath, char const * name, svn_string_t const * value, \n"
		"    svn_depth_t depth, svn_boolean_t skip_checks, apr_array_header_t changelist_filter, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_prop_set3", _wrap_svn_wc_prop_set3, METH_VARARGS, (char *)"\n"
		"svn_wc_prop_set3(char const * name, svn_string_t const * value, char const * path, svn_wc_adm_access_t * adm_access, \n"
		"    svn_boolean_t skip_checks, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_prop_set2", _wrap_svn_wc_prop_set2, METH_VARARGS, (char *)"\n"
		"svn_wc_prop_set2(char const * name, svn_string_t const * value, char const * path, svn_wc_adm_access_t * adm_access, \n"
		"    svn_boolean_t skip_checks, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_prop_set", _wrap_svn_wc_prop_set, METH_VARARGS, (char *)"\n"
		"svn_wc_prop_set(char const * name, svn_string_t const * value, char const * path, svn_wc_adm_access_t * adm_access, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_is_normal_prop", _wrap_svn_wc_is_normal_prop, METH_VARARGS, (char *)"svn_wc_is_normal_prop(char const * name) -> svn_boolean_t"},
	 { (char *)"svn_wc_is_wc_prop", _wrap_svn_wc_is_wc_prop, METH_VARARGS, (char *)"svn_wc_is_wc_prop(char const * name) -> svn_boolean_t"},
	 { (char *)"svn_wc_is_entry_prop", _wrap_svn_wc_is_entry_prop, METH_VARARGS, (char *)"svn_wc_is_entry_prop(char const * name) -> svn_boolean_t"},
	 { (char *)"svn_wc_canonicalize_svn_prop", _wrap_svn_wc_canonicalize_svn_prop, METH_VARARGS, (char *)"\n"
		"svn_wc_canonicalize_svn_prop(char const * propname, svn_string_t const * propval, char const * path, svn_node_kind_t kind, \n"
		"    svn_boolean_t skip_some_checks, svn_wc_canonicalize_svn_prop_get_file_t prop_getter, \n"
		"    void * getter_baton, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_diff_editor6", _wrap_svn_wc_get_diff_editor6, METH_VARARGS, (char *)"\n"
		"svn_wc_get_diff_editor6(svn_wc_context_t * wc_ctx, char const * anchor_abspath, char const * target, svn_depth_t depth, \n"
		"    svn_boolean_t ignore_ancestry, svn_boolean_t show_copies_as_adds, \n"
		"    svn_boolean_t use_git_diff_format, svn_boolean_t use_text_base, svn_boolean_t reverse_order, \n"
		"    svn_boolean_t server_performs_filtering, apr_array_header_t changelist_filter, \n"
		"    svn_wc_diff_callbacks4_t callbacks, void * callback_baton, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_diff_editor5", _wrap_svn_wc_get_diff_editor5, METH_VARARGS, (char *)"\n"
		"svn_wc_get_diff_editor5(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks3_t callbacks, \n"
		"    void * callback_baton, svn_depth_t depth, svn_boolean_t ignore_ancestry, \n"
		"    svn_boolean_t use_text_base, svn_boolean_t reverse_order, svn_cancel_func_t cancel_func, \n"
		"    apr_array_header_t changelist_filter, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_diff_editor4", _wrap_svn_wc_get_diff_editor4, METH_VARARGS, (char *)"\n"
		"svn_wc_get_diff_editor4(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks2_t callbacks, \n"
		"    svn_depth_t depth, svn_boolean_t ignore_ancestry, svn_boolean_t use_text_base, \n"
		"    svn_boolean_t reverse_order, svn_cancel_func_t cancel_func, apr_array_header_t changelist_filter, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_diff_editor3", _wrap_svn_wc_get_diff_editor3, METH_VARARGS, (char *)"\n"
		"svn_wc_get_diff_editor3(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks2_t callbacks, \n"
		"    svn_boolean_t recurse, svn_boolean_t ignore_ancestry, svn_boolean_t use_text_base, \n"
		"    svn_boolean_t reverse_order, svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_diff_editor2", _wrap_svn_wc_get_diff_editor2, METH_VARARGS, (char *)"\n"
		"svn_wc_get_diff_editor2(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks_t callbacks, \n"
		"    void * callback_baton, svn_boolean_t recurse, svn_boolean_t ignore_ancestry, \n"
		"    svn_boolean_t use_text_base, svn_boolean_t reverse_order, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_diff_editor", _wrap_svn_wc_get_diff_editor, METH_VARARGS, (char *)"\n"
		"svn_wc_get_diff_editor(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks_t callbacks, \n"
		"    void * callback_baton, svn_boolean_t recurse, svn_boolean_t use_text_base, \n"
		"    svn_boolean_t reverse_order, svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff6", _wrap_svn_wc_diff6, METH_VARARGS, (char *)"\n"
		"svn_wc_diff6(svn_wc_context_t * wc_ctx, char const * target_abspath, svn_wc_diff_callbacks4_t callbacks, \n"
		"    void * callback_baton, svn_depth_t depth, svn_boolean_t ignore_ancestry, \n"
		"    svn_boolean_t show_copies_as_adds, svn_boolean_t use_git_diff_format, \n"
		"    apr_array_header_t changelist_filter, svn_cancel_func_t cancel_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff5", _wrap_svn_wc_diff5, METH_VARARGS, (char *)"\n"
		"svn_wc_diff5(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks3_t callbacks, \n"
		"    void * callback_baton, svn_depth_t depth, svn_boolean_t ignore_ancestry, \n"
		"    apr_array_header_t changelist_filter, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff4", _wrap_svn_wc_diff4, METH_VARARGS, (char *)"\n"
		"svn_wc_diff4(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks2_t callbacks, \n"
		"    svn_depth_t depth, svn_boolean_t ignore_ancestry, apr_array_header_t changelist_filter, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff3", _wrap_svn_wc_diff3, METH_VARARGS, (char *)"\n"
		"svn_wc_diff3(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks2_t callbacks, \n"
		"    svn_boolean_t recurse, svn_boolean_t ignore_ancestry, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff2", _wrap_svn_wc_diff2, METH_VARARGS, (char *)"\n"
		"svn_wc_diff2(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks_t callbacks, \n"
		"    void * callback_baton, svn_boolean_t recurse, svn_boolean_t ignore_ancestry, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff", _wrap_svn_wc_diff, METH_VARARGS, (char *)"\n"
		"svn_wc_diff(svn_wc_adm_access_t * anchor, char const * target, svn_wc_diff_callbacks_t callbacks, \n"
		"    void * callback_baton, svn_boolean_t recurse, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_prop_diffs2", _wrap_svn_wc_get_prop_diffs2, METH_VARARGS, (char *)"svn_wc_get_prop_diffs2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_prop_diffs", _wrap_svn_wc_get_prop_diffs, METH_VARARGS, (char *)"svn_wc_get_prop_diffs(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_merge5", _wrap_svn_wc_merge5, METH_VARARGS, (char *)"\n"
		"svn_wc_merge5(enum svn_wc_notify_state_t * merge_props_state, svn_wc_context_t * wc_ctx, char const * left_abspath, \n"
		"    char const * right_abspath, char const * target_abspath, \n"
		"    char const * left_label, char const * right_label, char const * target_label, \n"
		"    svn_wc_conflict_version_t left_version, svn_wc_conflict_version_t right_version, \n"
		"    svn_boolean_t dry_run, char const * diff3_cmd, apr_array_header_t merge_options, \n"
		"    apr_hash_t original_props, apr_array_header_t prop_diff, \n"
		"    svn_wc_conflict_resolver_func2_t conflict_func, void * conflict_baton, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge4", _wrap_svn_wc_merge4, METH_VARARGS, (char *)"\n"
		"svn_wc_merge4(svn_wc_context_t * wc_ctx, char const * left_abspath, char const * right_abspath, \n"
		"    char const * target_abspath, char const * left_label, char const * right_label, \n"
		"    char const * target_label, svn_wc_conflict_version_t left_version, svn_wc_conflict_version_t right_version, \n"
		"    svn_boolean_t dry_run, char const * diff3_cmd, \n"
		"    apr_array_header_t merge_options, apr_array_header_t prop_diff, \n"
		"    svn_wc_conflict_resolver_func2_t conflict_func, void * conflict_baton, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge3", _wrap_svn_wc_merge3, METH_VARARGS, (char *)"\n"
		"svn_wc_merge3(char const * left, char const * right, char const * merge_target, svn_wc_adm_access_t * adm_access, \n"
		"    char const * left_label, char const * right_label, char const * target_label, \n"
		"    svn_boolean_t dry_run, char const * diff3_cmd, apr_array_header_t merge_options, \n"
		"    apr_array_header_t prop_diff, svn_wc_conflict_resolver_func_t conflict_func, \n"
		"    void * conflict_baton, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge2", _wrap_svn_wc_merge2, METH_VARARGS, (char *)"\n"
		"svn_wc_merge2(char const * left, char const * right, char const * merge_target, svn_wc_adm_access_t * adm_access, \n"
		"    char const * left_label, char const * right_label, char const * target_label, \n"
		"    svn_boolean_t dry_run, char const * diff3_cmd, apr_array_header_t merge_options, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge", _wrap_svn_wc_merge, METH_VARARGS, (char *)"\n"
		"svn_wc_merge(char const * left, char const * right, char const * merge_target, svn_wc_adm_access_t * adm_access, \n"
		"    char const * left_label, char const * right_label, char const * target_label, \n"
		"    svn_boolean_t dry_run, char const * diff3_cmd, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge_props3", _wrap_svn_wc_merge_props3, METH_VARARGS, (char *)"\n"
		"svn_wc_merge_props3(svn_wc_notify_state_t * state, svn_wc_context_t * wc_ctx, char const * local_abspath, \n"
		"    svn_wc_conflict_version_t left_version, svn_wc_conflict_version_t right_version, \n"
		"    apr_hash_t baseprops, apr_array_header_t propchanges, svn_boolean_t dry_run, \n"
		"    svn_wc_conflict_resolver_func2_t conflict_func, void * conflict_baton, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge_props2", _wrap_svn_wc_merge_props2, METH_VARARGS, (char *)"\n"
		"svn_wc_merge_props2(svn_wc_notify_state_t * state, char const * path, svn_wc_adm_access_t * adm_access, \n"
		"    apr_hash_t baseprops, apr_array_header_t propchanges, svn_boolean_t base_merge, \n"
		"    svn_boolean_t dry_run, svn_wc_conflict_resolver_func_t conflict_func, \n"
		"    void * conflict_baton, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge_props", _wrap_svn_wc_merge_props, METH_VARARGS, (char *)"\n"
		"svn_wc_merge_props(svn_wc_notify_state_t * state, char const * path, svn_wc_adm_access_t * adm_access, \n"
		"    apr_hash_t baseprops, apr_array_header_t propchanges, svn_boolean_t base_merge, \n"
		"    svn_boolean_t dry_run, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_merge_prop_diffs", _wrap_svn_wc_merge_prop_diffs, METH_VARARGS, (char *)"\n"
		"svn_wc_merge_prop_diffs(svn_wc_notify_state_t * state, char const * path, svn_wc_adm_access_t * adm_access, \n"
		"    apr_array_header_t propchanges, svn_boolean_t base_merge, svn_boolean_t dry_run, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_pristine_contents2", _wrap_svn_wc_get_pristine_contents2, METH_VARARGS, (char *)"svn_wc_get_pristine_contents2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_pristine_contents", _wrap_svn_wc_get_pristine_contents, METH_VARARGS, (char *)"svn_wc_get_pristine_contents(char const * path, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_pristine_copy_path", _wrap_svn_wc_get_pristine_copy_path, METH_VARARGS, (char *)"svn_wc_get_pristine_copy_path(char const * path, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_cleanup4", _wrap_svn_wc_cleanup4, METH_VARARGS, (char *)"\n"
		"svn_wc_cleanup4(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_boolean_t break_locks, \n"
		"    svn_boolean_t fix_recorded_timestamps, svn_boolean_t clear_dav_cache, svn_boolean_t vacuum_pristines, \n"
		"    svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_cleanup3", _wrap_svn_wc_cleanup3, METH_VARARGS, (char *)"\n"
		"svn_wc_cleanup3(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_cleanup2", _wrap_svn_wc_cleanup2, METH_VARARGS, (char *)"svn_wc_cleanup2(char const * path, char const * diff3_cmd, svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_cleanup", _wrap_svn_wc_cleanup, METH_VARARGS, (char *)"\n"
		"svn_wc_cleanup(char const * path, svn_wc_adm_access_t * optional_adm_access, char const * diff3_cmd, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_upgrade", _wrap_svn_wc_upgrade, METH_VARARGS, (char *)"\n"
		"svn_wc_upgrade(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_wc_upgrade_get_repos_info_t repos_info_func, \n"
		"    void * repos_info_baton, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_relocate4", _wrap_svn_wc_relocate4, METH_VARARGS, (char *)"\n"
		"svn_wc_relocate4(svn_wc_context_t * wc_ctx, char const * wcroot_abspath, char const * _from, char const * to, \n"
		"    svn_wc_relocation_validator3_t validator, void * validator_baton, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_relocate3", _wrap_svn_wc_relocate3, METH_VARARGS, (char *)"\n"
		"svn_wc_relocate3(char const * path, svn_wc_adm_access_t * adm_access, char const * _from, char const * to, \n"
		"    svn_boolean_t recurse, svn_wc_relocation_validator3_t validator, void * validator_baton, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_relocate2", _wrap_svn_wc_relocate2, METH_VARARGS, (char *)"\n"
		"svn_wc_relocate2(char const * path, svn_wc_adm_access_t * adm_access, char const * _from, char const * to, \n"
		"    svn_boolean_t recurse, svn_wc_relocation_validator2_t validator, void * validator_baton, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_relocate", _wrap_svn_wc_relocate, METH_VARARGS, (char *)"\n"
		"svn_wc_relocate(char const * path, svn_wc_adm_access_t * adm_access, char const * _from, char const * to, \n"
		"    svn_boolean_t recurse, svn_wc_relocation_validator_t validator, void * validator_baton, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_revert5", _wrap_svn_wc_revert5, METH_VARARGS, (char *)"\n"
		"svn_wc_revert5(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_depth_t depth, svn_boolean_t use_commit_times, \n"
		"    apr_array_header_t changelist_filter, svn_boolean_t clear_changelists, \n"
		"    svn_boolean_t metadata_only, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_revert4", _wrap_svn_wc_revert4, METH_VARARGS, (char *)"\n"
		"svn_wc_revert4(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_depth_t depth, svn_boolean_t use_commit_times, \n"
		"    apr_array_header_t changelist_filter, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_revert3", _wrap_svn_wc_revert3, METH_VARARGS, (char *)"\n"
		"svn_wc_revert3(char const * path, svn_wc_adm_access_t * parent_access, svn_depth_t depth, svn_boolean_t use_commit_times, \n"
		"    apr_array_header_t changelist_filter, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_revert2", _wrap_svn_wc_revert2, METH_VARARGS, (char *)"\n"
		"svn_wc_revert2(char const * path, svn_wc_adm_access_t * parent_access, svn_boolean_t recursive, \n"
		"    svn_boolean_t use_commit_times, svn_cancel_func_t cancel_func, svn_wc_notify_func2_t notify_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_revert", _wrap_svn_wc_revert, METH_VARARGS, (char *)"\n"
		"svn_wc_revert(char const * path, svn_wc_adm_access_t * parent_access, svn_boolean_t recursive, \n"
		"    svn_boolean_t use_commit_times, svn_cancel_func_t cancel_func, svn_wc_notify_func_t notify_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_restore", _wrap_svn_wc_restore, METH_VARARGS, (char *)"\n"
		"svn_wc_restore(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_boolean_t use_commit_times, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_create_tmp_file2", _wrap_svn_wc_create_tmp_file2, METH_VARARGS, (char *)"svn_wc_create_tmp_file2(char const * path, svn_io_file_del_t delete_when, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_create_tmp_file", _wrap_svn_wc_create_tmp_file, METH_VARARGS, (char *)"svn_wc_create_tmp_file(char const * path, svn_boolean_t delete_on_close, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_translated_file2", _wrap_svn_wc_translated_file2, METH_VARARGS, (char *)"\n"
		"svn_wc_translated_file2(char const * src, char const * versioned_file, svn_wc_adm_access_t * adm_access, \n"
		"    apr_uint32_t flags, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_translated_file", _wrap_svn_wc_translated_file, METH_VARARGS, (char *)"\n"
		"svn_wc_translated_file(char const * vfile, svn_wc_adm_access_t * adm_access, svn_boolean_t force_repair, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_translated_stream", _wrap_svn_wc_translated_stream, METH_VARARGS, (char *)"\n"
		"svn_wc_translated_stream(char const * path, char const * versioned_file, svn_wc_adm_access_t * adm_access, \n"
		"    apr_uint32_t flags, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_transmit_text_deltas3", _wrap_svn_wc_transmit_text_deltas3, METH_VARARGS, (char *)"\n"
		"svn_wc_transmit_text_deltas3(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_boolean_t fulltext, svn_delta_editor_t editor, \n"
		"    void * file_baton, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_transmit_text_deltas2", _wrap_svn_wc_transmit_text_deltas2, METH_VARARGS, (char *)"\n"
		"svn_wc_transmit_text_deltas2(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t fulltext, svn_delta_editor_t editor, \n"
		"    void * file_baton, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_transmit_text_deltas", _wrap_svn_wc_transmit_text_deltas, METH_VARARGS, (char *)"\n"
		"svn_wc_transmit_text_deltas(char const * path, svn_wc_adm_access_t * adm_access, svn_boolean_t fulltext, svn_delta_editor_t editor, \n"
		"    void * file_baton, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_transmit_prop_deltas2", _wrap_svn_wc_transmit_prop_deltas2, METH_VARARGS, (char *)"\n"
		"svn_wc_transmit_prop_deltas2(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_delta_editor_t editor, \n"
		"    void * baton, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_transmit_prop_deltas", _wrap_svn_wc_transmit_prop_deltas, METH_VARARGS, (char *)"\n"
		"svn_wc_transmit_prop_deltas(char const * path, svn_wc_adm_access_t * adm_access, svn_wc_entry_t entry, svn_delta_editor_t editor, \n"
		"    void * baton, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_default_ignores", _wrap_svn_wc_get_default_ignores, METH_VARARGS, (char *)"svn_wc_get_default_ignores(apr_hash_t config, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_get_ignores2", _wrap_svn_wc_get_ignores2, METH_VARARGS, (char *)"\n"
		"svn_wc_get_ignores2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_hash_t config, apr_pool_t result_pool, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_ignores", _wrap_svn_wc_get_ignores, METH_VARARGS, (char *)"svn_wc_get_ignores(apr_hash_t config, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_match_ignore_list", _wrap_svn_wc_match_ignore_list, METH_VARARGS, (char *)"svn_wc_match_ignore_list(char const * str, apr_array_header_t list, apr_pool_t pool) -> svn_boolean_t"},
	 { (char *)"svn_wc_add_lock2", _wrap_svn_wc_add_lock2, METH_VARARGS, (char *)"svn_wc_add_lock2(svn_wc_context_t * wc_ctx, char const * abspath, svn_lock_t lock, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_add_lock", _wrap_svn_wc_add_lock, METH_VARARGS, (char *)"svn_wc_add_lock(char const * path, svn_lock_t lock, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_remove_lock2", _wrap_svn_wc_remove_lock2, METH_VARARGS, (char *)"svn_wc_remove_lock2(svn_wc_context_t * wc_ctx, char const * local_abspath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_remove_lock", _wrap_svn_wc_remove_lock, METH_VARARGS, (char *)"svn_wc_remove_lock(char const * path, svn_wc_adm_access_t * adm_access, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_wc_revision_status_t_min_rev_set", _wrap_svn_wc_revision_status_t_min_rev_set, METH_VARARGS, (char *)"svn_wc_revision_status_t_min_rev_set(svn_wc_revision_status_t self, svn_revnum_t min_rev)"},
	 { (char *)"svn_wc_revision_status_t_min_rev_get", _wrap_svn_wc_revision_status_t_min_rev_get, METH_VARARGS, (char *)"svn_wc_revision_status_t_min_rev_get(svn_wc_revision_status_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_revision_status_t_max_rev_set", _wrap_svn_wc_revision_status_t_max_rev_set, METH_VARARGS, (char *)"svn_wc_revision_status_t_max_rev_set(svn_wc_revision_status_t self, svn_revnum_t max_rev)"},
	 { (char *)"svn_wc_revision_status_t_max_rev_get", _wrap_svn_wc_revision_status_t_max_rev_get, METH_VARARGS, (char *)"svn_wc_revision_status_t_max_rev_get(svn_wc_revision_status_t self) -> svn_revnum_t"},
	 { (char *)"svn_wc_revision_status_t_switched_set", _wrap_svn_wc_revision_status_t_switched_set, METH_VARARGS, (char *)"svn_wc_revision_status_t_switched_set(svn_wc_revision_status_t self, svn_boolean_t switched)"},
	 { (char *)"svn_wc_revision_status_t_switched_get", _wrap_svn_wc_revision_status_t_switched_get, METH_VARARGS, (char *)"svn_wc_revision_status_t_switched_get(svn_wc_revision_status_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_revision_status_t_modified_set", _wrap_svn_wc_revision_status_t_modified_set, METH_VARARGS, (char *)"svn_wc_revision_status_t_modified_set(svn_wc_revision_status_t self, svn_boolean_t modified)"},
	 { (char *)"svn_wc_revision_status_t_modified_get", _wrap_svn_wc_revision_status_t_modified_get, METH_VARARGS, (char *)"svn_wc_revision_status_t_modified_get(svn_wc_revision_status_t self) -> svn_boolean_t"},
	 { (char *)"svn_wc_revision_status_t_sparse_checkout_set", _wrap_svn_wc_revision_status_t_sparse_checkout_set, METH_VARARGS, (char *)"svn_wc_revision_status_t_sparse_checkout_set(svn_wc_revision_status_t self, svn_boolean_t sparse_checkout)"},
	 { (char *)"svn_wc_revision_status_t_sparse_checkout_get", _wrap_svn_wc_revision_status_t_sparse_checkout_get, METH_VARARGS, (char *)"svn_wc_revision_status_t_sparse_checkout_get(svn_wc_revision_status_t self) -> svn_boolean_t"},
	 { (char *)"new_svn_wc_revision_status_t", _wrap_new_svn_wc_revision_status_t, METH_VARARGS, (char *)"new_svn_wc_revision_status_t() -> svn_wc_revision_status_t"},
	 { (char *)"delete_svn_wc_revision_status_t", _wrap_delete_svn_wc_revision_status_t, METH_VARARGS, (char *)"delete_svn_wc_revision_status_t(svn_wc_revision_status_t self)"},
	 { (char *)"svn_wc_revision_status_t_swigregister", svn_wc_revision_status_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_revision_status2", _wrap_svn_wc_revision_status2, METH_VARARGS, (char *)"\n"
		"svn_wc_revision_status2(svn_wc_context_t * wc_ctx, char const * local_abspath, char const * trail_url, svn_boolean_t committed, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t result_pool, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_revision_status", _wrap_svn_wc_revision_status, METH_VARARGS, (char *)"\n"
		"svn_wc_revision_status(char const * wc_path, char const * trail_url, svn_boolean_t committed, svn_cancel_func_t cancel_func, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_set_changelist2", _wrap_svn_wc_set_changelist2, METH_VARARGS, (char *)"\n"
		"svn_wc_set_changelist2(svn_wc_context_t * wc_ctx, char const * local_abspath, char const * changelist, svn_depth_t depth, \n"
		"    apr_array_header_t changelist_filter, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_set_changelist", _wrap_svn_wc_set_changelist, METH_VARARGS, (char *)"\n"
		"svn_wc_set_changelist(char const * path, char const * changelist, svn_wc_adm_access_t * adm_access, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_get_changelists", _wrap_svn_wc_get_changelists, METH_VARARGS, (char *)"\n"
		"svn_wc_get_changelists(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_depth_t depth, apr_array_header_t changelist_filter, \n"
		"    svn_changelist_receiver_t callback_func, void * callback_baton, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_crop_tree2", _wrap_svn_wc_crop_tree2, METH_VARARGS, (char *)"\n"
		"svn_wc_crop_tree2(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_depth_t depth, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_crop_tree", _wrap_svn_wc_crop_tree, METH_VARARGS, (char *)"\n"
		"svn_wc_crop_tree(svn_wc_adm_access_t * anchor, char const * target, svn_depth_t depth, svn_wc_notify_func2_t notify_func, \n"
		"    svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_exclude", _wrap_svn_wc_exclude, METH_VARARGS, (char *)"\n"
		"svn_wc_exclude(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_cancel_func_t cancel_func, \n"
		"    svn_wc_notify_func2_t notify_func, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_read_kind2", _wrap_svn_wc_read_kind2, METH_VARARGS, (char *)"\n"
		"svn_wc_read_kind2(svn_wc_context_t * wc_ctx, char const * local_abspath, svn_boolean_t show_deleted, \n"
		"    svn_boolean_t show_hidden, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_read_kind", _wrap_svn_wc_read_kind, METH_VARARGS, (char *)"svn_wc_read_kind(svn_wc_context_t * wc_ctx, char const * abspath, svn_boolean_t show_hidden, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_wc_context_t_swigregister", svn_wc_context_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_adm_access_t_swigregister", svn_wc_adm_access_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_traversal_info_t_swigregister", svn_wc_traversal_info_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_committed_queue_t_swigregister", svn_wc_committed_queue_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_diff_callbacks4_invoke_file_opened", _wrap_svn_wc_diff_callbacks4_invoke_file_opened, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_file_opened(svn_wc_diff_callbacks4_t _obj, char const * path, svn_revnum_t rev, void * diff_baton, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_file_changed", _wrap_svn_wc_diff_callbacks4_invoke_file_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_file_changed(svn_wc_diff_callbacks4_t _obj, svn_wc_notify_state_t * contentstate, svn_wc_notify_state_t * propstate, \n"
		"    char const * path, char const * tmpfile1, char const * tmpfile2, \n"
		"    svn_revnum_t rev1, svn_revnum_t rev2, char const * mimetype1, \n"
		"    char const * mimetype2, apr_array_header_t propchanges, apr_hash_t originalprops, \n"
		"    void * diff_baton, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_file_added", _wrap_svn_wc_diff_callbacks4_invoke_file_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_file_added(svn_wc_diff_callbacks4_t _obj, svn_wc_notify_state_t * contentstate, svn_wc_notify_state_t * propstate, \n"
		"    char const * path, char const * tmpfile1, char const * tmpfile2, \n"
		"    svn_revnum_t rev1, svn_revnum_t rev2, char const * mimetype1, \n"
		"    char const * mimetype2, char const * copyfrom_path, svn_revnum_t copyfrom_revision, \n"
		"    apr_array_header_t propchanges, apr_hash_t originalprops, void * diff_baton, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_file_deleted", _wrap_svn_wc_diff_callbacks4_invoke_file_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_file_deleted(svn_wc_diff_callbacks4_t _obj, svn_wc_notify_state_t * state, char const * path, \n"
		"    char const * tmpfile1, char const * tmpfile2, char const * mimetype1, char const * mimetype2, \n"
		"    apr_hash_t originalprops, void * diff_baton, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks4_invoke_dir_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_dir_deleted(svn_wc_diff_callbacks4_t _obj, svn_wc_notify_state_t * state, char const * path, \n"
		"    void * diff_baton, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_dir_opened", _wrap_svn_wc_diff_callbacks4_invoke_dir_opened, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_dir_opened(svn_wc_diff_callbacks4_t _obj, char const * path, svn_revnum_t rev, void * diff_baton, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_dir_added", _wrap_svn_wc_diff_callbacks4_invoke_dir_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_dir_added(svn_wc_diff_callbacks4_t _obj, svn_wc_notify_state_t * state, char const * path, \n"
		"    svn_revnum_t rev, char const * copyfrom_path, svn_revnum_t copyfrom_revision, \n"
		"    void * diff_baton, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_dir_props_changed", _wrap_svn_wc_diff_callbacks4_invoke_dir_props_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_dir_props_changed(svn_wc_diff_callbacks4_t _obj, svn_wc_notify_state_t * propstate, char const * path, \n"
		"    svn_boolean_t dir_was_added, apr_array_header_t propchanges, apr_hash_t original_props, \n"
		"    void * diff_baton, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks4_invoke_dir_closed", _wrap_svn_wc_diff_callbacks4_invoke_dir_closed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks4_invoke_dir_closed(svn_wc_diff_callbacks4_t _obj, svn_wc_notify_state_t * contentstate, svn_wc_notify_state_t * propstate, \n"
		"    char const * path, svn_boolean_t dir_was_added, void * diff_baton, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_file_changed", _wrap_svn_wc_diff_callbacks3_invoke_file_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_file_changed(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * contentstate, \n"
		"    svn_wc_notify_state_t * propstate, char const * path, \n"
		"    char const * tmpfile1, char const * tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, \n"
		"    char const * mimetype1, char const * mimetype2, apr_array_header_t propchanges, \n"
		"    apr_hash_t originalprops, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_file_added", _wrap_svn_wc_diff_callbacks3_invoke_file_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_file_added(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * contentstate, \n"
		"    svn_wc_notify_state_t * propstate, char const * path, \n"
		"    char const * tmpfile1, char const * tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, \n"
		"    char const * mimetype1, char const * mimetype2, apr_array_header_t propchanges, \n"
		"    apr_hash_t originalprops, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_file_deleted", _wrap_svn_wc_diff_callbacks3_invoke_file_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_file_deleted(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, char const * tmpfile1, char const * tmpfile2, \n"
		"    char const * mimetype1, char const * mimetype2, apr_hash_t originalprops, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_dir_added", _wrap_svn_wc_diff_callbacks3_invoke_dir_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_dir_added(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, svn_revnum_t rev, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks3_invoke_dir_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_dir_deleted(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_dir_props_changed", _wrap_svn_wc_diff_callbacks3_invoke_dir_props_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_dir_props_changed(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * propstate, \n"
		"    char const * path, apr_array_header_t propchanges, apr_hash_t original_props, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_dir_opened", _wrap_svn_wc_diff_callbacks3_invoke_dir_opened, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_dir_opened(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, char const * path, \n"
		"    svn_revnum_t rev, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks3_invoke_dir_closed", _wrap_svn_wc_diff_callbacks3_invoke_dir_closed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks3_invoke_dir_closed(svn_wc_diff_callbacks3_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * contentstate, \n"
		"    svn_wc_notify_state_t * propstate, char const * path, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks2_invoke_file_changed", _wrap_svn_wc_diff_callbacks2_invoke_file_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks2_invoke_file_changed(svn_wc_diff_callbacks2_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * contentstate, \n"
		"    svn_wc_notify_state_t * propstate, char const * path, \n"
		"    char const * tmpfile1, char const * tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, \n"
		"    char const * mimetype1, char const * mimetype2, apr_array_header_t propchanges, \n"
		"    apr_hash_t originalprops, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks2_invoke_file_added", _wrap_svn_wc_diff_callbacks2_invoke_file_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks2_invoke_file_added(svn_wc_diff_callbacks2_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * contentstate, \n"
		"    svn_wc_notify_state_t * propstate, char const * path, \n"
		"    char const * tmpfile1, char const * tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, \n"
		"    char const * mimetype1, char const * mimetype2, apr_array_header_t propchanges, \n"
		"    apr_hash_t originalprops, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks2_invoke_file_deleted", _wrap_svn_wc_diff_callbacks2_invoke_file_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks2_invoke_file_deleted(svn_wc_diff_callbacks2_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, char const * tmpfile1, char const * tmpfile2, \n"
		"    char const * mimetype1, char const * mimetype2, apr_hash_t originalprops, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks2_invoke_dir_added", _wrap_svn_wc_diff_callbacks2_invoke_dir_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks2_invoke_dir_added(svn_wc_diff_callbacks2_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, svn_revnum_t rev, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks2_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks2_invoke_dir_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks2_invoke_dir_deleted(svn_wc_diff_callbacks2_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks2_invoke_dir_props_changed", _wrap_svn_wc_diff_callbacks2_invoke_dir_props_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks2_invoke_dir_props_changed(svn_wc_diff_callbacks2_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, apr_array_header_t propchanges, apr_hash_t original_props, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks_invoke_file_changed", _wrap_svn_wc_diff_callbacks_invoke_file_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks_invoke_file_changed(svn_wc_diff_callbacks_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, char const * tmpfile1, char const * tmpfile2, \n"
		"    svn_revnum_t rev1, svn_revnum_t rev2, char const * mimetype1, char const * mimetype2, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks_invoke_file_added", _wrap_svn_wc_diff_callbacks_invoke_file_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks_invoke_file_added(svn_wc_diff_callbacks_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, char const * tmpfile1, char const * tmpfile2, \n"
		"    svn_revnum_t rev1, svn_revnum_t rev2, char const * mimetype1, char const * mimetype2, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks_invoke_file_deleted", _wrap_svn_wc_diff_callbacks_invoke_file_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks_invoke_file_deleted(svn_wc_diff_callbacks_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, char const * tmpfile1, char const * tmpfile2, \n"
		"    char const * mimetype1, char const * mimetype2, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks_invoke_dir_added", _wrap_svn_wc_diff_callbacks_invoke_dir_added, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks_invoke_dir_added(svn_wc_diff_callbacks_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, svn_revnum_t rev, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks_invoke_dir_deleted, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks_invoke_dir_deleted(svn_wc_diff_callbacks_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_diff_callbacks_invoke_props_changed", _wrap_svn_wc_diff_callbacks_invoke_props_changed, METH_VARARGS, (char *)"\n"
		"svn_wc_diff_callbacks_invoke_props_changed(svn_wc_diff_callbacks_t _obj, svn_wc_adm_access_t * adm_access, svn_wc_notify_state_t * state, \n"
		"    char const * path, apr_array_header_t propchanges, apr_hash_t original_props, \n"
		"    void * diff_baton) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_entry_callbacks2_invoke_found_entry", _wrap_svn_wc_entry_callbacks2_invoke_found_entry, METH_VARARGS, (char *)"\n"
		"svn_wc_entry_callbacks2_invoke_found_entry(svn_wc_entry_callbacks2_t _obj, char const * path, svn_wc_entry_t entry, void * walk_baton, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_entry_callbacks2_invoke_handle_error", _wrap_svn_wc_entry_callbacks2_invoke_handle_error, METH_VARARGS, (char *)"\n"
		"svn_wc_entry_callbacks2_invoke_handle_error(svn_wc_entry_callbacks2_t _obj, char const * path, svn_error_t err, void * walk_baton, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_entry_callbacks_invoke_found_entry", _wrap_svn_wc_entry_callbacks_invoke_found_entry, METH_VARARGS, (char *)"\n"
		"svn_wc_entry_callbacks_invoke_found_entry(svn_wc_entry_callbacks_t _obj, char const * path, svn_wc_entry_t entry, void * walk_baton, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_external_update", _wrap_svn_wc_invoke_external_update, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_external_update(svn_wc_external_update_t _obj, void * baton, char const * local_abspath, svn_string_t const * old_val, \n"
		"    svn_string_t const * new_val, svn_depth_t depth, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_notify_func2", _wrap_svn_wc_invoke_notify_func2, METH_VARARGS, (char *)"svn_wc_invoke_notify_func2(svn_wc_notify_func2_t _obj, void * baton, svn_wc_notify_t notify, apr_pool_t pool)"},
	 { (char *)"svn_wc_invoke_notify_func", _wrap_svn_wc_invoke_notify_func, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_notify_func(svn_wc_notify_func_t _obj, void * baton, char const * path, svn_wc_notify_action_t action, \n"
		"    svn_node_kind_t kind, char const * mime_type, svn_wc_notify_state_t content_state, \n"
		"    svn_wc_notify_state_t prop_state, svn_revnum_t revision)\n"
		""},
	 { (char *)"svn_wc_invoke_conflict_resolver_func2", _wrap_svn_wc_invoke_conflict_resolver_func2, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_conflict_resolver_func2(svn_wc_conflict_resolver_func2_t _obj, svn_wc_conflict_description2_t const * description, \n"
		"    void * baton, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_conflict_resolver_func", _wrap_svn_wc_invoke_conflict_resolver_func, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_conflict_resolver_func(svn_wc_conflict_resolver_func_t _obj, svn_wc_conflict_description_t description, \n"
		"    void * baton, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_status_func4", _wrap_svn_wc_invoke_status_func4, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_status_func4(svn_wc_status_func4_t _obj, void * baton, char const * local_abspath, svn_wc_status3_t status, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_status_func3", _wrap_svn_wc_invoke_status_func3, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_status_func3(svn_wc_status_func3_t _obj, void * baton, char const * path, svn_wc_status2_t status, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_status_func2", _wrap_svn_wc_invoke_status_func2, METH_VARARGS, (char *)"svn_wc_invoke_status_func2(svn_wc_status_func2_t _obj, void * baton, char const * path, svn_wc_status2_t status)"},
	 { (char *)"svn_wc_invoke_status_func", _wrap_svn_wc_invoke_status_func, METH_VARARGS, (char *)"svn_wc_invoke_status_func(svn_wc_status_func_t _obj, void * baton, char const * path, svn_wc_status_t status)"},
	 { (char *)"svn_wc_invoke_get_file", _wrap_svn_wc_invoke_get_file, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_get_file(svn_wc_get_file_t _obj, void * baton, char const * path, svn_revnum_t revision, svn_stream_t * stream, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_dirents_func", _wrap_svn_wc_invoke_dirents_func, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_dirents_func(svn_wc_dirents_func_t _obj, void * baton, char const * repos_root_url, char const * repos_relpath, \n"
		"    apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_canonicalize_svn_prop_get_file", _wrap_svn_wc_invoke_canonicalize_svn_prop_get_file, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_canonicalize_svn_prop_get_file(svn_wc_canonicalize_svn_prop_get_file_t _obj, svn_stream_t * stream, void * baton, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_upgrade_get_repos_info", _wrap_svn_wc_invoke_upgrade_get_repos_info, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_upgrade_get_repos_info(svn_wc_upgrade_get_repos_info_t _obj, void * baton, char const * url, apr_pool_t result_pool, \n"
		"    apr_pool_t scratch_pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_relocation_validator3", _wrap_svn_wc_invoke_relocation_validator3, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_relocation_validator3(svn_wc_relocation_validator3_t _obj, void * baton, char const * uuid, char const * url, \n"
		"    char const * root_url, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_relocation_validator2", _wrap_svn_wc_invoke_relocation_validator2, METH_VARARGS, (char *)"\n"
		"svn_wc_invoke_relocation_validator2(svn_wc_relocation_validator2_t _obj, void * baton, char const * uuid, char const * url, \n"
		"    svn_boolean_t root, apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_invoke_relocation_validator", _wrap_svn_wc_invoke_relocation_validator, METH_VARARGS, (char *)"svn_wc_invoke_relocation_validator(svn_wc_relocation_validator_t _obj, void * baton, char const * uuid, char const * url) -> svn_error_t"},
	 { (char *)"svn_changelist_invoke_receiver", _wrap_svn_changelist_invoke_receiver, METH_VARARGS, (char *)"\n"
		"svn_changelist_invoke_receiver(svn_changelist_receiver_t _obj, void * baton, char const * path, char const * changelist, \n"
		"    apr_pool_t pool) -> svn_error_t\n"
		""},
	 { (char *)"svn_wc_external_update_t_swigregister", svn_wc_external_update_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_notify_func2_t_swigregister", svn_wc_notify_func2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_notify_func_t_swigregister", svn_wc_notify_func_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_conflict_resolver_func2_t_swigregister", svn_wc_conflict_resolver_func2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_conflict_resolver_func_t_swigregister", svn_wc_conflict_resolver_func_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_status_func4_t_swigregister", svn_wc_status_func4_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_status_func3_t_swigregister", svn_wc_status_func3_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_status_func2_t_swigregister", svn_wc_status_func2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_status_func_t_swigregister", svn_wc_status_func_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_get_file_t_swigregister", svn_wc_get_file_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_dirents_func_t_swigregister", svn_wc_dirents_func_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_canonicalize_svn_prop_get_file_t_swigregister", svn_wc_canonicalize_svn_prop_get_file_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_upgrade_get_repos_info_t_swigregister", svn_wc_upgrade_get_repos_info_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_relocation_validator3_t_swigregister", svn_wc_relocation_validator3_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_relocation_validator2_t_swigregister", svn_wc_relocation_validator2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_relocation_validator_t_swigregister", svn_wc_relocation_validator_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_changelist_receiver_t_swigregister", svn_changelist_receiver_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_wc_swig_init_asp_dot_net_hack", _wrap_svn_wc_swig_init_asp_dot_net_hack, METH_VARARGS, (char *)"svn_wc_swig_init_asp_dot_net_hack(apr_pool_t pool) -> svn_error_t"},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_apr_array_header_t = {"_p_apr_array_header_t", "apr_array_header_t *|svn_rangelist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_file_t = {"_p_apr_file_t", "apr_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_hash_t = {"_p_apr_hash_t", "apr_hash_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int32_t = {"_p_apr_int32_t", "apr_int32_t *|time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int64_t = {"_p_apr_int64_t", "apr_int64_t *|svn_filesize_t *|apr_time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_off_t = {"_p_apr_off_t", "apr_off_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_pool_t = {"_p_apr_pool_t", "apr_pool_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_opt_subcommand_t *|struct svn_error_t *(*)(apr_getopt_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(char const **,char const **,void *,char const *,apr_pool_t *,apr_pool_t *)|svn_wc_upgrade_get_repos_info_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_string_t const **,svn_stream_t *,void *,apr_pool_t *)|svn_wc_canonicalize_svn_prop_get_file_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_wc_conflict_resolver_func2_t|struct svn_error_t *(*)(svn_wc_conflict_result_t **,svn_wc_conflict_description2_t const *,void *,apr_pool_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_wc_conflict_resolver_func_t|struct svn_error_t *(*)(svn_wc_conflict_result_t **,svn_wc_conflict_description_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)|svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)|svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)|struct svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)|svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)|struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)|svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)|svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)|svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)|svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)|svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)|svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)|svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)|svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__p_svn_error_t = {"_p_f_p_void__p_svn_error_t", "svn_cancel_func_t|struct svn_error_t *(*)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,apr_hash_t **,char const *,char const *,apr_pool_t *,apr_pool_t *)|svn_wc_dirents_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void = {"_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void", "void (*)(void *,char const *,enum svn_wc_notify_action_t,enum svn_node_kind_t,char const *,enum svn_wc_notify_state_t,enum svn_wc_notify_state_t,long)|svn_wc_notify_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t", "svn_wc_relocation_validator_t|struct svn_error_t *(*)(void *,char const *,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,char const *,apr_pool_t *)|svn_changelist_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,char const *,char const *,apr_pool_t *)|svn_wc_relocation_validator3_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,char const *,svn_boolean_t,apr_pool_t *)|svn_wc_relocation_validator2_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t", "svn_wc_external_update_t|struct svn_error_t *(*)(void *,char const *,svn_string_t const *,svn_string_t const *,svn_depth_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,svn_wc_status3_t const *,apr_pool_t *)|svn_wc_status_func4_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void = {"_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void", "void (*)(void *,char const *,struct svn_wc_status2_t *)|svn_wc_status_func2_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void = {"_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void", "void (*)(void *,char const *,struct svn_wc_status_t *)|svn_wc_status_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,svn_wc_status2_t *,apr_pool_t *)|svn_wc_status_func3_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *)|svn_wc_get_file_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void = {"_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void", "svn_wc_notify_func2_t|void (*)(void *,struct svn_wc_notify_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "int *|svn_boolean_t *|apr_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "long *|svn_revnum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_array_header_t = {"_p_p_apr_array_header_t", "apr_array_header_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_file_t = {"_p_p_apr_file_t", "apr_file_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_hash_t = {"_p_p_apr_hash_t", "apr_hash_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_wc_upgrade_get_repos_info_t *|struct svn_error_t *(**)(char const **,char const **,void *,char const *,apr_pool_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_wc_canonicalize_svn_prop_get_file_t *|struct svn_error_t *(**)(svn_string_t const **,svn_stream_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(**)(svn_wc_conflict_result_t **,svn_wc_conflict_description2_t const *,void *,apr_pool_t *,apr_pool_t *)|svn_wc_conflict_resolver_func2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(**)(svn_wc_conflict_result_t **,svn_wc_conflict_description_t const *,void *,apr_pool_t *)|svn_wc_conflict_resolver_func_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_wc_dirents_func_t *|struct svn_error_t *(**)(void *,apr_hash_t **,char const *,char const *,apr_pool_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void = {"_p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void", "void (**)(void *,char const *,enum svn_wc_notify_action_t,enum svn_node_kind_t,char const *,enum svn_wc_notify_state_t,enum svn_wc_notify_state_t,long)|svn_wc_notify_func_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t", "struct svn_error_t *(**)(void *,char const *,char const *)|svn_wc_relocation_validator_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(**)(void *,char const *,char const *,apr_pool_t *)|svn_changelist_receiver_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_wc_relocation_validator3_t *|struct svn_error_t *(**)(void *,char const *,char const *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(**)(void *,char const *,char const *,svn_boolean_t,apr_pool_t *)|svn_wc_relocation_validator2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(**)(void *,char const *,svn_string_t const *,svn_string_t const *,svn_depth_t,apr_pool_t *)|svn_wc_external_update_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t", "svn_wc_status_func4_t *|struct svn_error_t *(**)(void *,char const *,svn_wc_status3_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void = {"_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void", "void (**)(void *,char const *,struct svn_wc_status2_t *)|svn_wc_status_func2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void = {"_p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void", "void (**)(void *,char const *,struct svn_wc_status_t *)|svn_wc_status_func_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t", "svn_wc_status_func3_t *|struct svn_error_t *(**)(void *,char const *,svn_wc_status2_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(**)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *)|svn_wc_get_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void = {"_p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void", "void (**)(void *,struct svn_wc_notify_t const *,apr_pool_t *)|svn_wc_notify_func2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_checksum_t = {"_p_p_svn_checksum_t", "struct svn_checksum_t **|svn_checksum_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_delta_editor_t = {"_p_p_svn_delta_editor_t", "struct svn_delta_editor_t **|svn_delta_editor_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_stream_t = {"_p_p_svn_stream_t", "struct svn_stream_t **|svn_stream_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_string_t = {"_p_p_svn_string_t", "struct svn_string_t **|svn_string_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_adm_access_t = {"_p_p_svn_wc_adm_access_t", "struct svn_wc_adm_access_t **|svn_wc_adm_access_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_committed_queue_t = {"_p_p_svn_wc_committed_queue_t", "struct svn_wc_committed_queue_t **|svn_wc_committed_queue_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_conflict_result_t = {"_p_p_svn_wc_conflict_result_t", "struct svn_wc_conflict_result_t **|svn_wc_conflict_result_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_context_t = {"_p_p_svn_wc_context_t", "struct svn_wc_context_t **|svn_wc_context_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_entry_t = {"_p_p_svn_wc_entry_t", "struct svn_wc_entry_t **|svn_wc_entry_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_external_item2_t = {"_p_p_svn_wc_external_item2_t", "struct svn_wc_external_item2_t **|svn_wc_external_item2_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_revision_status_t = {"_p_p_svn_wc_revision_status_t", "struct svn_wc_revision_status_t **|svn_wc_revision_status_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_status2_t = {"_p_p_svn_wc_status2_t", "struct svn_wc_status2_t **|svn_wc_status2_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_status3_t = {"_p_p_svn_wc_status3_t", "struct svn_wc_status3_t **|svn_wc_status3_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_status_t = {"_p_p_svn_wc_status_t", "struct svn_wc_status_t **|svn_wc_status_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_baton_t = {"_p_svn_auth_baton_t", "struct svn_auth_baton_t *|svn_auth_baton_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_simple_t = {"_p_svn_auth_cred_simple_t", "struct svn_auth_cred_simple_t *|svn_auth_cred_simple_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_pw_t = {"_p_svn_auth_cred_ssl_client_cert_pw_t", "struct svn_auth_cred_ssl_client_cert_pw_t *|svn_auth_cred_ssl_client_cert_pw_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_t = {"_p_svn_auth_cred_ssl_client_cert_t", "struct svn_auth_cred_ssl_client_cert_t *|svn_auth_cred_ssl_client_cert_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_server_trust_t = {"_p_svn_auth_cred_ssl_server_trust_t", "struct svn_auth_cred_ssl_server_trust_t *|svn_auth_cred_ssl_server_trust_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_username_t = {"_p_svn_auth_cred_username_t", "struct svn_auth_cred_username_t *|svn_auth_cred_username_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_iterstate_t = {"_p_svn_auth_iterstate_t", "struct svn_auth_iterstate_t *|svn_auth_iterstate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_object_t = {"_p_svn_auth_provider_object_t", "struct svn_auth_provider_object_t *|svn_auth_provider_object_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_t = {"_p_svn_auth_provider_t", "struct svn_auth_provider_t *|svn_auth_provider_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_ssl_server_cert_info_t = {"_p_svn_auth_ssl_server_cert_info_t", "struct svn_auth_ssl_server_cert_info_t *|svn_auth_ssl_server_cert_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_ctx_t = {"_p_svn_checksum_ctx_t", "struct svn_checksum_ctx_t *|svn_checksum_ctx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_kind_t = {"_p_svn_checksum_kind_t", "enum svn_checksum_kind_t *|svn_checksum_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_t = {"_p_svn_checksum_t", "struct svn_checksum_t *|svn_checksum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_commit_info_t = {"_p_svn_commit_info_t", "struct svn_commit_info_t *|svn_commit_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_config_t = {"_p_svn_config_t", "struct svn_config_t *|svn_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_editor_t = {"_p_svn_delta_editor_t", "struct svn_delta_editor_t *|svn_delta_editor_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_depth_t = {"_p_svn_depth_t", "enum svn_depth_t *|svn_depth_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_hunk_t = {"_p_svn_diff_hunk_t", "svn_diff_hunk_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_dirent_t = {"_p_svn_dirent_t", "struct svn_dirent_t *|svn_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_errno_t = {"_p_svn_errno_t", "enum svn_errno_t *|svn_errno_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_error_t = {"_p_svn_error_t", "struct svn_error_t *|svn_error_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_fs_lock_target_t = {"_p_svn_fs_lock_target_t", "svn_fs_lock_target_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent2_t = {"_p_svn_io_dirent2_t", "struct svn_io_dirent2_t *|svn_io_dirent2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent_t = {"_p_svn_io_dirent_t", "struct svn_io_dirent_t *|svn_io_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_file_del_t = {"_p_svn_io_file_del_t", "enum svn_io_file_del_t *|svn_io_file_del_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_location_segment_t = {"_p_svn_location_segment_t", "struct svn_location_segment_t *|svn_location_segment_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_lock_t = {"_p_svn_lock_t", "struct svn_lock_t *|svn_lock_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path2_t = {"_p_svn_log_changed_path2_t", "struct svn_log_changed_path2_t *|svn_log_changed_path2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path_t = {"_p_svn_log_changed_path_t", "struct svn_log_changed_path_t *|svn_log_changed_path_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_entry_t = {"_p_svn_log_entry_t", "struct svn_log_entry_t *|svn_log_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_merge_range_t = {"_p_svn_merge_range_t", "struct svn_merge_range_t *|svn_merge_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_mergeinfo_inheritance_t = {"_p_svn_mergeinfo_inheritance_t", "enum svn_mergeinfo_inheritance_t *|svn_mergeinfo_inheritance_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_node_kind_t = {"_p_svn_node_kind_t", "enum svn_node_kind_t *|svn_node_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_range_t = {"_p_svn_opt_revision_range_t", "struct svn_opt_revision_range_t *|svn_opt_revision_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_t = {"_p_svn_opt_revision_t", "struct svn_opt_revision_t *|svn_opt_revision_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_value_t = {"_p_svn_opt_revision_value_t", "union svn_opt_revision_value_t *|svn_opt_revision_value_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc2_t = {"_p_svn_opt_subcommand_desc2_t", "struct svn_opt_subcommand_desc2_t *|svn_opt_subcommand_desc2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc_t = {"_p_svn_opt_subcommand_desc_t", "struct svn_opt_subcommand_desc_t *|svn_opt_subcommand_desc_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_inherited_item_t = {"_p_svn_prop_inherited_item_t", "struct svn_prop_inherited_item_t *|svn_prop_inherited_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_kind = {"_p_svn_prop_kind", "svn_prop_kind_t *|enum svn_prop_kind *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks2_t = {"_p_svn_ra_callbacks2_t", "struct svn_ra_callbacks2_t *|svn_ra_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks_t = {"_p_svn_ra_callbacks_t", "struct svn_ra_callbacks_t *|svn_ra_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_plugin_t = {"_p_svn_ra_plugin_t", "struct svn_ra_plugin_t *|svn_ra_plugin_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter2_t = {"_p_svn_ra_reporter2_t", "struct svn_ra_reporter2_t *|svn_ra_reporter2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter3_t = {"_p_svn_ra_reporter3_t", "struct svn_ra_reporter3_t *|svn_ra_reporter3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter_t = {"_p_svn_ra_reporter_t", "struct svn_ra_reporter_t *|svn_ra_reporter_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_session_t = {"_p_svn_ra_session_t", "struct svn_ra_session_t *|svn_ra_session_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_mark_t = {"_p_svn_stream_mark_t", "struct svn_stream_mark_t *|svn_stream_mark_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_t = {"_p_svn_stream_t", "struct svn_stream_t *|svn_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_string_t = {"_p_svn_string_t", "struct svn_string_t *|svn_string_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stringbuf_t = {"_p_svn_stringbuf_t", "struct svn_stringbuf_t *|svn_stringbuf_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_tristate_t = {"_p_svn_tristate_t", "enum svn_tristate_t *|svn_tristate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_op_t = {"_p_svn_txdelta_op_t", "struct svn_txdelta_op_t *|svn_txdelta_op_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_stream_t = {"_p_svn_txdelta_stream_t", "struct svn_txdelta_stream_t *|svn_txdelta_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_window_t = {"_p_svn_txdelta_window_t", "struct svn_txdelta_window_t *|svn_txdelta_window_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_checklist_t = {"_p_svn_version_checklist_t", "struct svn_version_checklist_t *|svn_version_checklist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_linked_lib_t = {"_p_svn_version_ext_linked_lib_t", "struct svn_version_ext_linked_lib_t *|svn_version_ext_linked_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_loaded_lib_t = {"_p_svn_version_ext_loaded_lib_t", "struct svn_version_ext_loaded_lib_t *|svn_version_ext_loaded_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_extended_t = {"_p_svn_version_extended_t", "struct svn_version_extended_t *|svn_version_extended_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_t = {"_p_svn_version_t", "struct svn_version_t *|svn_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_adm_access_t = {"_p_svn_wc_adm_access_t", "struct svn_wc_adm_access_t *|svn_wc_adm_access_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_committed_queue_t = {"_p_svn_wc_committed_queue_t", "struct svn_wc_committed_queue_t *|svn_wc_committed_queue_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_action_t = {"_p_svn_wc_conflict_action_t", "enum svn_wc_conflict_action_t *|svn_wc_conflict_action_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_choice_t = {"_p_svn_wc_conflict_choice_t", "enum svn_wc_conflict_choice_t *|svn_wc_conflict_choice_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_description2_t = {"_p_svn_wc_conflict_description2_t", "struct svn_wc_conflict_description2_t *|svn_wc_conflict_description2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_description_t = {"_p_svn_wc_conflict_description_t", "struct svn_wc_conflict_description_t *|svn_wc_conflict_description_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_kind_t = {"_p_svn_wc_conflict_kind_t", "enum svn_wc_conflict_kind_t *|svn_wc_conflict_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_reason_t = {"_p_svn_wc_conflict_reason_t", "enum svn_wc_conflict_reason_t *|svn_wc_conflict_reason_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_result_t = {"_p_svn_wc_conflict_result_t", "struct svn_wc_conflict_result_t *|svn_wc_conflict_result_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_version_t = {"_p_svn_wc_conflict_version_t", "struct svn_wc_conflict_version_t *|svn_wc_conflict_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_context_t = {"_p_svn_wc_context_t", "struct svn_wc_context_t *|svn_wc_context_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks2_t = {"_p_svn_wc_diff_callbacks2_t", "struct svn_wc_diff_callbacks2_t *|svn_wc_diff_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks3_t = {"_p_svn_wc_diff_callbacks3_t", "struct svn_wc_diff_callbacks3_t *|svn_wc_diff_callbacks3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks4_t = {"_p_svn_wc_diff_callbacks4_t", "struct svn_wc_diff_callbacks4_t *|svn_wc_diff_callbacks4_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks_t = {"_p_svn_wc_diff_callbacks_t", "struct svn_wc_diff_callbacks_t *|svn_wc_diff_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_callbacks2_t = {"_p_svn_wc_entry_callbacks2_t", "struct svn_wc_entry_callbacks2_t *|svn_wc_entry_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_callbacks_t = {"_p_svn_wc_entry_callbacks_t", "struct svn_wc_entry_callbacks_t *|svn_wc_entry_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_t = {"_p_svn_wc_entry_t", "struct svn_wc_entry_t *|svn_wc_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item2_t = {"_p_svn_wc_external_item2_t", "struct svn_wc_external_item2_t *|svn_wc_external_item2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item_t = {"_p_svn_wc_external_item_t", "struct svn_wc_external_item_t *|svn_wc_external_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_info_t = {"_p_svn_wc_info_t", "struct svn_wc_info_t *|svn_wc_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_merge_outcome_t = {"_p_svn_wc_merge_outcome_t", "enum svn_wc_merge_outcome_t *|svn_wc_merge_outcome_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_action_t = {"_p_svn_wc_notify_action_t", "enum svn_wc_notify_action_t *|svn_wc_notify_action_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_lock_state_t = {"_p_svn_wc_notify_lock_state_t", "enum svn_wc_notify_lock_state_t *|svn_wc_notify_lock_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_state_t = {"_p_svn_wc_notify_state_t", "enum svn_wc_notify_state_t *|svn_wc_notify_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_t = {"_p_svn_wc_notify_t", "struct svn_wc_notify_t *|svn_wc_notify_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_operation_t = {"_p_svn_wc_operation_t", "enum svn_wc_operation_t *|svn_wc_operation_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_revision_status_t = {"_p_svn_wc_revision_status_t", "struct svn_wc_revision_status_t *|svn_wc_revision_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_schedule_t = {"_p_svn_wc_schedule_t", "enum svn_wc_schedule_t *|svn_wc_schedule_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status2_t = {"_p_svn_wc_status2_t", "struct svn_wc_status2_t *|svn_wc_status2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status3_t = {"_p_svn_wc_status3_t", "struct svn_wc_status3_t *|svn_wc_status3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status_t = {"_p_svn_wc_status_t", "struct svn_wc_status_t *|svn_wc_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_traversal_info_t = {"_p_svn_wc_traversal_info_t", "struct svn_wc_traversal_info_t *|svn_wc_traversal_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "unsigned long *|svn_linenum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_apr_array_header_t,
  &_swigt__p_apr_file_t,
  &_swigt__p_apr_hash_t,
  &_swigt__p_apr_int32_t,
  &_swigt__p_apr_int64_t,
  &_swigt__p_apr_off_t,
  &_swigt__p_apr_pool_t,
  &_swigt__p_char,
  &_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  &_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  &_swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_p_apr_array_header_t,
  &_swigt__p_p_apr_file_t,
  &_swigt__p_p_apr_hash_t,
  &_swigt__p_p_char,
  &_swigt__p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  &_swigt__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  &_swigt__p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  &_swigt__p_p_svn_checksum_t,
  &_swigt__p_p_svn_delta_editor_t,
  &_swigt__p_p_svn_stream_t,
  &_swigt__p_p_svn_string_t,
  &_swigt__p_p_svn_wc_adm_access_t,
  &_swigt__p_p_svn_wc_committed_queue_t,
  &_swigt__p_p_svn_wc_conflict_result_t,
  &_swigt__p_p_svn_wc_context_t,
  &_swigt__p_p_svn_wc_entry_t,
  &_swigt__p_p_svn_wc_external_item2_t,
  &_swigt__p_p_svn_wc_revision_status_t,
  &_swigt__p_p_svn_wc_status2_t,
  &_swigt__p_p_svn_wc_status3_t,
  &_swigt__p_p_svn_wc_status_t,
  &_swigt__p_p_void,
  &_swigt__p_svn_auth_baton_t,
  &_swigt__p_svn_auth_cred_simple_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_pw_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_t,
  &_swigt__p_svn_auth_cred_ssl_server_trust_t,
  &_swigt__p_svn_auth_cred_username_t,
  &_swigt__p_svn_auth_iterstate_t,
  &_swigt__p_svn_auth_provider_object_t,
  &_swigt__p_svn_auth_provider_t,
  &_swigt__p_svn_auth_ssl_server_cert_info_t,
  &_swigt__p_svn_checksum_ctx_t,
  &_swigt__p_svn_checksum_kind_t,
  &_swigt__p_svn_checksum_t,
  &_swigt__p_svn_commit_info_t,
  &_swigt__p_svn_config_t,
  &_swigt__p_svn_delta_editor_t,
  &_swigt__p_svn_depth_t,
  &_swigt__p_svn_diff_hunk_t,
  &_swigt__p_svn_dirent_t,
  &_swigt__p_svn_errno_t,
  &_swigt__p_svn_error_t,
  &_swigt__p_svn_fs_lock_target_t,
  &_swigt__p_svn_io_dirent2_t,
  &_swigt__p_svn_io_dirent_t,
  &_swigt__p_svn_io_file_del_t,
  &_swigt__p_svn_location_segment_t,
  &_swigt__p_svn_lock_t,
  &_swigt__p_svn_log_changed_path2_t,
  &_swigt__p_svn_log_changed_path_t,
  &_swigt__p_svn_log_entry_t,
  &_swigt__p_svn_merge_range_t,
  &_swigt__p_svn_mergeinfo_inheritance_t,
  &_swigt__p_svn_node_kind_t,
  &_swigt__p_svn_opt_revision_range_t,
  &_swigt__p_svn_opt_revision_t,
  &_swigt__p_svn_opt_revision_value_t,
  &_swigt__p_svn_opt_subcommand_desc2_t,
  &_swigt__p_svn_opt_subcommand_desc_t,
  &_swigt__p_svn_prop_inherited_item_t,
  &_swigt__p_svn_prop_kind,
  &_swigt__p_svn_ra_callbacks2_t,
  &_swigt__p_svn_ra_callbacks_t,
  &_swigt__p_svn_ra_plugin_t,
  &_swigt__p_svn_ra_reporter2_t,
  &_swigt__p_svn_ra_reporter3_t,
  &_swigt__p_svn_ra_reporter_t,
  &_swigt__p_svn_ra_session_t,
  &_swigt__p_svn_stream_mark_t,
  &_swigt__p_svn_stream_t,
  &_swigt__p_svn_string_t,
  &_swigt__p_svn_stringbuf_t,
  &_swigt__p_svn_tristate_t,
  &_swigt__p_svn_txdelta_op_t,
  &_swigt__p_svn_txdelta_stream_t,
  &_swigt__p_svn_txdelta_window_t,
  &_swigt__p_svn_version_checklist_t,
  &_swigt__p_svn_version_ext_linked_lib_t,
  &_swigt__p_svn_version_ext_loaded_lib_t,
  &_swigt__p_svn_version_extended_t,
  &_swigt__p_svn_version_t,
  &_swigt__p_svn_wc_adm_access_t,
  &_swigt__p_svn_wc_committed_queue_t,
  &_swigt__p_svn_wc_conflict_action_t,
  &_swigt__p_svn_wc_conflict_choice_t,
  &_swigt__p_svn_wc_conflict_description2_t,
  &_swigt__p_svn_wc_conflict_description_t,
  &_swigt__p_svn_wc_conflict_kind_t,
  &_swigt__p_svn_wc_conflict_reason_t,
  &_swigt__p_svn_wc_conflict_result_t,
  &_swigt__p_svn_wc_conflict_version_t,
  &_swigt__p_svn_wc_context_t,
  &_swigt__p_svn_wc_diff_callbacks2_t,
  &_swigt__p_svn_wc_diff_callbacks3_t,
  &_swigt__p_svn_wc_diff_callbacks4_t,
  &_swigt__p_svn_wc_diff_callbacks_t,
  &_swigt__p_svn_wc_entry_callbacks2_t,
  &_swigt__p_svn_wc_entry_callbacks_t,
  &_swigt__p_svn_wc_entry_t,
  &_swigt__p_svn_wc_external_item2_t,
  &_swigt__p_svn_wc_external_item_t,
  &_swigt__p_svn_wc_info_t,
  &_swigt__p_svn_wc_merge_outcome_t,
  &_swigt__p_svn_wc_notify_action_t,
  &_swigt__p_svn_wc_notify_lock_state_t,
  &_swigt__p_svn_wc_notify_state_t,
  &_swigt__p_svn_wc_notify_t,
  &_swigt__p_svn_wc_operation_t,
  &_swigt__p_svn_wc_revision_status_t,
  &_swigt__p_svn_wc_schedule_t,
  &_swigt__p_svn_wc_status2_t,
  &_swigt__p_svn_wc_status3_t,
  &_swigt__p_svn_wc_status_t,
  &_swigt__p_svn_wc_traversal_info_t,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_long,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_apr_array_header_t[] = {  {&_swigt__p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_file_t[] = {  {&_swigt__p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_hash_t[] = {  {&_swigt__p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int32_t[] = {  {&_swigt__p_apr_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int64_t[] = {  {&_swigt__p_apr_int64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_off_t[] = {  {&_swigt__p_apr_off_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_pool_t[] = {  {&_swigt__p_apr_pool_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_array_header_t[] = {  {&_swigt__p_p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_file_t[] = {  {&_swigt__p_p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_hash_t[] = {  {&_swigt__p_p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void[] = {  {&_swigt__p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_checksum_t[] = {  {&_swigt__p_p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_delta_editor_t[] = {  {&_swigt__p_p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_stream_t[] = {  {&_swigt__p_p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_string_t[] = {  {&_swigt__p_p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_adm_access_t[] = {  {&_swigt__p_p_svn_wc_adm_access_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_committed_queue_t[] = {  {&_swigt__p_p_svn_wc_committed_queue_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_conflict_result_t[] = {  {&_swigt__p_p_svn_wc_conflict_result_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_context_t[] = {  {&_swigt__p_p_svn_wc_context_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_entry_t[] = {  {&_swigt__p_p_svn_wc_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_external_item2_t[] = {  {&_swigt__p_p_svn_wc_external_item2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_revision_status_t[] = {  {&_swigt__p_p_svn_wc_revision_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_status2_t[] = {  {&_swigt__p_p_svn_wc_status2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_status3_t[] = {  {&_swigt__p_p_svn_wc_status3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_status_t[] = {  {&_swigt__p_p_svn_wc_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_baton_t[] = {  {&_swigt__p_svn_auth_baton_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_simple_t[] = {  {&_swigt__p_svn_auth_cred_simple_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_pw_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_pw_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_server_trust_t[] = {  {&_swigt__p_svn_auth_cred_ssl_server_trust_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_username_t[] = {  {&_swigt__p_svn_auth_cred_username_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_iterstate_t[] = {  {&_swigt__p_svn_auth_iterstate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_object_t[] = {  {&_swigt__p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_t[] = {  {&_swigt__p_svn_auth_provider_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_ssl_server_cert_info_t[] = {  {&_swigt__p_svn_auth_ssl_server_cert_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_ctx_t[] = {  {&_swigt__p_svn_checksum_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_kind_t[] = {  {&_swigt__p_svn_checksum_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_t[] = {  {&_swigt__p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_commit_info_t[] = {  {&_swigt__p_svn_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_config_t[] = {  {&_swigt__p_svn_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_editor_t[] = {  {&_swigt__p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_depth_t[] = {  {&_swigt__p_svn_depth_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_hunk_t[] = {  {&_swigt__p_svn_diff_hunk_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_dirent_t[] = {  {&_swigt__p_svn_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_errno_t[] = {  {&_swigt__p_svn_errno_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_error_t[] = {  {&_swigt__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_fs_lock_target_t[] = {  {&_swigt__p_svn_fs_lock_target_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent2_t[] = {  {&_swigt__p_svn_io_dirent2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent_t[] = {  {&_swigt__p_svn_io_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_file_del_t[] = {  {&_swigt__p_svn_io_file_del_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_location_segment_t[] = {  {&_swigt__p_svn_location_segment_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_lock_t[] = {  {&_swigt__p_svn_lock_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path2_t[] = {  {&_swigt__p_svn_log_changed_path2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path_t[] = {  {&_swigt__p_svn_log_changed_path_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_entry_t[] = {  {&_swigt__p_svn_log_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_merge_range_t[] = {  {&_swigt__p_svn_merge_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_mergeinfo_inheritance_t[] = {  {&_swigt__p_svn_mergeinfo_inheritance_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_node_kind_t[] = {  {&_swigt__p_svn_node_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_range_t[] = {  {&_swigt__p_svn_opt_revision_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_t[] = {  {&_swigt__p_svn_opt_revision_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_value_t[] = {  {&_swigt__p_svn_opt_revision_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc2_t[] = {  {&_swigt__p_svn_opt_subcommand_desc2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc_t[] = {  {&_swigt__p_svn_opt_subcommand_desc_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_inherited_item_t[] = {  {&_swigt__p_svn_prop_inherited_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_kind[] = {  {&_swigt__p_svn_prop_kind, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks2_t[] = {  {&_swigt__p_svn_ra_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks_t[] = {  {&_swigt__p_svn_ra_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_plugin_t[] = {  {&_swigt__p_svn_ra_plugin_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter2_t[] = {  {&_swigt__p_svn_ra_reporter2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter3_t[] = {  {&_swigt__p_svn_ra_reporter3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter_t[] = {  {&_swigt__p_svn_ra_reporter_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_session_t[] = {  {&_swigt__p_svn_ra_session_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_mark_t[] = {  {&_swigt__p_svn_stream_mark_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_t[] = {  {&_swigt__p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_string_t[] = {  {&_swigt__p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stringbuf_t[] = {  {&_swigt__p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_tristate_t[] = {  {&_swigt__p_svn_tristate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_op_t[] = {  {&_swigt__p_svn_txdelta_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_stream_t[] = {  {&_swigt__p_svn_txdelta_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_window_t[] = {  {&_swigt__p_svn_txdelta_window_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_checklist_t[] = {  {&_swigt__p_svn_version_checklist_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_linked_lib_t[] = {  {&_swigt__p_svn_version_ext_linked_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_loaded_lib_t[] = {  {&_swigt__p_svn_version_ext_loaded_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_extended_t[] = {  {&_swigt__p_svn_version_extended_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_t[] = {  {&_swigt__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_adm_access_t[] = {  {&_swigt__p_svn_wc_adm_access_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_committed_queue_t[] = {  {&_swigt__p_svn_wc_committed_queue_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_action_t[] = {  {&_swigt__p_svn_wc_conflict_action_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_choice_t[] = {  {&_swigt__p_svn_wc_conflict_choice_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_description2_t[] = {  {&_swigt__p_svn_wc_conflict_description2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_description_t[] = {  {&_swigt__p_svn_wc_conflict_description_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_kind_t[] = {  {&_swigt__p_svn_wc_conflict_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_reason_t[] = {  {&_swigt__p_svn_wc_conflict_reason_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_result_t[] = {  {&_swigt__p_svn_wc_conflict_result_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_version_t[] = {  {&_swigt__p_svn_wc_conflict_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_context_t[] = {  {&_swigt__p_svn_wc_context_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks2_t[] = {  {&_swigt__p_svn_wc_diff_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks3_t[] = {  {&_swigt__p_svn_wc_diff_callbacks3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks4_t[] = {  {&_swigt__p_svn_wc_diff_callbacks4_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks_t[] = {  {&_swigt__p_svn_wc_diff_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_callbacks2_t[] = {  {&_swigt__p_svn_wc_entry_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_callbacks_t[] = {  {&_swigt__p_svn_wc_entry_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_t[] = {  {&_swigt__p_svn_wc_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item2_t[] = {  {&_swigt__p_svn_wc_external_item2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item_t[] = {  {&_swigt__p_svn_wc_external_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_info_t[] = {  {&_swigt__p_svn_wc_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_merge_outcome_t[] = {  {&_swigt__p_svn_wc_merge_outcome_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_action_t[] = {  {&_swigt__p_svn_wc_notify_action_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_lock_state_t[] = {  {&_swigt__p_svn_wc_notify_lock_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_state_t[] = {  {&_swigt__p_svn_wc_notify_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_t[] = {  {&_swigt__p_svn_wc_notify_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_operation_t[] = {  {&_swigt__p_svn_wc_operation_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_revision_status_t[] = {  {&_swigt__p_svn_wc_revision_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_schedule_t[] = {  {&_swigt__p_svn_wc_schedule_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status2_t[] = {  {&_swigt__p_svn_wc_status2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status3_t[] = {  {&_swigt__p_svn_wc_status3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status_t[] = {  {&_swigt__p_svn_wc_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_traversal_info_t[] = {  {&_swigt__p_svn_wc_traversal_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_apr_array_header_t,
  _swigc__p_apr_file_t,
  _swigc__p_apr_hash_t,
  _swigc__p_apr_int32_t,
  _swigc__p_apr_int64_t,
  _swigc__p_apr_off_t,
  _swigc__p_apr_pool_t,
  _swigc__p_char,
  _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void__p_svn_error_t,
  _swigc__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  _swigc__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_p_apr_array_header_t,
  _swigc__p_p_apr_file_t,
  _swigc__p_p_apr_hash_t,
  _swigc__p_p_char,
  _swigc__p_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  _swigc__p_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  _swigc__p_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  _swigc__p_p_svn_checksum_t,
  _swigc__p_p_svn_delta_editor_t,
  _swigc__p_p_svn_stream_t,
  _swigc__p_p_svn_string_t,
  _swigc__p_p_svn_wc_adm_access_t,
  _swigc__p_p_svn_wc_committed_queue_t,
  _swigc__p_p_svn_wc_conflict_result_t,
  _swigc__p_p_svn_wc_context_t,
  _swigc__p_p_svn_wc_entry_t,
  _swigc__p_p_svn_wc_external_item2_t,
  _swigc__p_p_svn_wc_revision_status_t,
  _swigc__p_p_svn_wc_status2_t,
  _swigc__p_p_svn_wc_status3_t,
  _swigc__p_p_svn_wc_status_t,
  _swigc__p_p_void,
  _swigc__p_svn_auth_baton_t,
  _swigc__p_svn_auth_cred_simple_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_pw_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_t,
  _swigc__p_svn_auth_cred_ssl_server_trust_t,
  _swigc__p_svn_auth_cred_username_t,
  _swigc__p_svn_auth_iterstate_t,
  _swigc__p_svn_auth_provider_object_t,
  _swigc__p_svn_auth_provider_t,
  _swigc__p_svn_auth_ssl_server_cert_info_t,
  _swigc__p_svn_checksum_ctx_t,
  _swigc__p_svn_checksum_kind_t,
  _swigc__p_svn_checksum_t,
  _swigc__p_svn_commit_info_t,
  _swigc__p_svn_config_t,
  _swigc__p_svn_delta_editor_t,
  _swigc__p_svn_depth_t,
  _swigc__p_svn_diff_hunk_t,
  _swigc__p_svn_dirent_t,
  _swigc__p_svn_errno_t,
  _swigc__p_svn_error_t,
  _swigc__p_svn_fs_lock_target_t,
  _swigc__p_svn_io_dirent2_t,
  _swigc__p_svn_io_dirent_t,
  _swigc__p_svn_io_file_del_t,
  _swigc__p_svn_location_segment_t,
  _swigc__p_svn_lock_t,
  _swigc__p_svn_log_changed_path2_t,
  _swigc__p_svn_log_changed_path_t,
  _swigc__p_svn_log_entry_t,
  _swigc__p_svn_merge_range_t,
  _swigc__p_svn_mergeinfo_inheritance_t,
  _swigc__p_svn_node_kind_t,
  _swigc__p_svn_opt_revision_range_t,
  _swigc__p_svn_opt_revision_t,
  _swigc__p_svn_opt_revision_value_t,
  _swigc__p_svn_opt_subcommand_desc2_t,
  _swigc__p_svn_opt_subcommand_desc_t,
  _swigc__p_svn_prop_inherited_item_t,
  _swigc__p_svn_prop_kind,
  _swigc__p_svn_ra_callbacks2_t,
  _swigc__p_svn_ra_callbacks_t,
  _swigc__p_svn_ra_plugin_t,
  _swigc__p_svn_ra_reporter2_t,
  _swigc__p_svn_ra_reporter3_t,
  _swigc__p_svn_ra_reporter_t,
  _swigc__p_svn_ra_session_t,
  _swigc__p_svn_stream_mark_t,
  _swigc__p_svn_stream_t,
  _swigc__p_svn_string_t,
  _swigc__p_svn_stringbuf_t,
  _swigc__p_svn_tristate_t,
  _swigc__p_svn_txdelta_op_t,
  _swigc__p_svn_txdelta_stream_t,
  _swigc__p_svn_txdelta_window_t,
  _swigc__p_svn_version_checklist_t,
  _swigc__p_svn_version_ext_linked_lib_t,
  _swigc__p_svn_version_ext_loaded_lib_t,
  _swigc__p_svn_version_extended_t,
  _swigc__p_svn_version_t,
  _swigc__p_svn_wc_adm_access_t,
  _swigc__p_svn_wc_committed_queue_t,
  _swigc__p_svn_wc_conflict_action_t,
  _swigc__p_svn_wc_conflict_choice_t,
  _swigc__p_svn_wc_conflict_description2_t,
  _swigc__p_svn_wc_conflict_description_t,
  _swigc__p_svn_wc_conflict_kind_t,
  _swigc__p_svn_wc_conflict_reason_t,
  _swigc__p_svn_wc_conflict_result_t,
  _swigc__p_svn_wc_conflict_version_t,
  _swigc__p_svn_wc_context_t,
  _swigc__p_svn_wc_diff_callbacks2_t,
  _swigc__p_svn_wc_diff_callbacks3_t,
  _swigc__p_svn_wc_diff_callbacks4_t,
  _swigc__p_svn_wc_diff_callbacks_t,
  _swigc__p_svn_wc_entry_callbacks2_t,
  _swigc__p_svn_wc_entry_callbacks_t,
  _swigc__p_svn_wc_entry_t,
  _swigc__p_svn_wc_external_item2_t,
  _swigc__p_svn_wc_external_item_t,
  _swigc__p_svn_wc_info_t,
  _swigc__p_svn_wc_merge_outcome_t,
  _swigc__p_svn_wc_notify_action_t,
  _swigc__p_svn_wc_notify_lock_state_t,
  _swigc__p_svn_wc_notify_state_t,
  _swigc__p_svn_wc_notify_t,
  _swigc__p_svn_wc_operation_t,
  _swigc__p_svn_wc_revision_status_t,
  _swigc__p_svn_wc_schedule_t,
  _swigc__p_svn_wc_status2_t,
  _swigc__p_svn_wc_status3_t,
  _swigc__p_svn_wc_status_t,
  _swigc__p_svn_wc_traversal_info_t,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_long,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned staticly to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int found, init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
    module_head = &swig_module;
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    found=0;
    iter=module_head;
    do {
      if (iter==&swig_module) {
        found=1;
        break;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* if the is found in the list, then all is done and we may leave */
    if (found) return;
    /* otherwise we must add out module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %d\n", swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
  
  /* -----------------------------------------------------------------------------
   * global variable support code.
   * ----------------------------------------------------------------------------- */
  
  typedef struct swig_globalvar {
    char       *name;                  /* Name of global variable */
    PyObject *(*get_attr)(void);       /* Return the current value */
    int       (*set_attr)(PyObject *); /* Set the value */
    struct swig_globalvar *next;
  } swig_globalvar;
  
  typedef struct swig_varlinkobject {
    PyObject_HEAD
    swig_globalvar *vars;
  } swig_varlinkobject;
  
  SWIGINTERN PyObject *
  swig_varlink_repr(swig_varlinkobject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
    return PyUnicode_InternFromString("<Swig global variables>");
#else
    return PyString_FromString("<Swig global variables>");
#endif
  }
  
  SWIGINTERN PyObject *
  swig_varlink_str(swig_varlinkobject *v) {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *str = PyUnicode_InternFromString("(");
    PyObject *tail;
    PyObject *joined;
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      tail = PyUnicode_FromString(var->name);
      joined = PyUnicode_Concat(str, tail);
      Py_DecRef(str);
      Py_DecRef(tail);
      str = joined;
      if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        Py_DecRef(str);
        Py_DecRef(tail);
        str = joined;
      }
    }
    tail = PyUnicode_InternFromString(")");
    joined = PyUnicode_Concat(str, tail);
    Py_DecRef(str);
    Py_DecRef(tail);
    str = joined;
#else
    PyObject *str = PyString_FromString("(");
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      PyString_ConcatAndDel(&str,PyString_FromString(var->name));
      if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
    }
    PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
    return str;
  }
  
  SWIGINTERN int
  swig_varlink_print(swig_varlinkobject *v, FILE *fp, int SWIGUNUSEDPARM(flags)) {
    char *tmp;
    PyObject *str = swig_varlink_str(v);
    fprintf(fp,"Swig global variables ");
    fprintf(fp,"%s\n", tmp = SWIG_Python_str_AsChar(str));
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(str);
    return 0;
  }
  
  SWIGINTERN void
  swig_varlink_dealloc(swig_varlinkobject *v) {
    swig_globalvar *var = v->vars;
    while (var) {
      swig_globalvar *n = var->next;
      free(var->name);
      free(var);
      var = n;
    }
  }
  
  SWIGINTERN PyObject *
  swig_varlink_getattr(swig_varlinkobject *v, char *n) {
    PyObject *res = NULL;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->get_attr)();
        break;
      }
      var = var->next;
    }
    if (res == NULL && !PyErr_Occurred()) {
      PyErr_SetString(PyExc_NameError,"Unknown C global variable");
    }
    return res;
  }
  
  SWIGINTERN int
  swig_varlink_setattr(swig_varlinkobject *v, char *n, PyObject *p) {
    int res = 1;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->set_attr)(p);
        break;
      }
      var = var->next;
    }
    if (res == 1 && !PyErr_Occurred()) {
      PyErr_SetString(PyExc_NameError,"Unknown C global variable");
    }
    return res;
  }
  
  SWIGINTERN PyTypeObject*
  swig_varlink_type(void) {
    static char varlink__doc__[] = "Swig var link object";
    static PyTypeObject varlink_type;
    static int type_init = 0;
    if (!type_init) {
      const PyTypeObject tmp = {
        /* PyObject header changed in Python 3 */
#if PY_VERSION_HEX >= 0x03000000
        PyVarObject_HEAD_INIT(NULL, 0)
#else
        PyObject_HEAD_INIT(NULL)
        0,                                  /* ob_size */
#endif
        (char *)"swigvarlink",              /* tp_name */
        sizeof(swig_varlinkobject),         /* tp_basicsize */
        0,                                  /* tp_itemsize */
        (destructor) swig_varlink_dealloc,  /* tp_dealloc */
        (printfunc) swig_varlink_print,     /* tp_print */
        (getattrfunc) swig_varlink_getattr, /* tp_getattr */
        (setattrfunc) swig_varlink_setattr, /* tp_setattr */
        0,                                  /* tp_compare */
        (reprfunc) swig_varlink_repr,       /* tp_repr */
        0,                                  /* tp_as_number */
        0,                                  /* tp_as_sequence */
        0,                                  /* tp_as_mapping */
        0,                                  /* tp_hash */
        0,                                  /* tp_call */
        (reprfunc) swig_varlink_str,        /* tp_str */
        0,                                  /* tp_getattro */
        0,                                  /* tp_setattro */
        0,                                  /* tp_as_buffer */
        0,                                  /* tp_flags */
        varlink__doc__,                     /* tp_doc */
        0,                                  /* tp_traverse */
        0,                                  /* tp_clear */
        0,                                  /* tp_richcompare */
        0,                                  /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
        0,                                  /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
        0,                                  /* tp_version */
#endif
#ifdef COUNT_ALLOCS
        0,0,0,0                             /* tp_alloc -> tp_next */
#endif
      };
      varlink_type = tmp;
      type_init = 1;
#if PY_VERSION_HEX < 0x02020000
      varlink_type.ob_type = &PyType_Type;
#else
      if (PyType_Ready(&varlink_type) < 0)
      return NULL;
#endif
    }
    return &varlink_type;
  }
  
  /* Create a variable linking object for use later */
  SWIGINTERN PyObject *
  SWIG_Python_newvarlink(void) {
    swig_varlinkobject *result = PyObject_NEW(swig_varlinkobject, swig_varlink_type());
    if (result) {
      result->vars = 0;
    }
    return ((PyObject*) result);
  }
  
  SWIGINTERN void 
  SWIG_Python_addvarlink(PyObject *p, char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
    swig_varlinkobject *v = (swig_varlinkobject *) p;
    swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
    if (gv) {
      size_t size = strlen(name)+1;
      gv->name = (char *)malloc(size);
      if (gv->name) {
        strncpy(gv->name,name,size);
        gv->get_attr = get_attr;
        gv->set_attr = set_attr;
        gv->next = v->vars;
      }
    }
    v->vars = gv;
  }
  
  SWIGINTERN PyObject *
  SWIG_globals(void) {
    static PyObject *_SWIG_globals = 0; 
    if (!_SWIG_globals) _SWIG_globals = SWIG_newvarlink();  
    return _SWIG_globals;
  }
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------*/
  /* Fix SwigMethods to carry the callback ptrs when needed */
  /* -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods,
    swig_const_info *const_table,
    swig_type_info **types,
    swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (c && (c = strstr(c, "swig_ptr: "))) {
        int j;
        swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              strncpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              strncpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
  PyObject *m, *d, *md;
#if PY_VERSION_HEX >= 0x03000000
  static struct PyModuleDef SWIG_module = {
# if PY_VERSION_HEX >= 0x03020000
    PyModuleDef_HEAD_INIT,
# else
    {
      PyObject_HEAD_INIT(NULL)
      NULL, /* m_init */
      0,    /* m_index */
      NULL, /* m_copy */
    },
# endif
    (char *) SWIG_name,
    NULL,
    -1,
    SwigMethods,
    NULL,
    NULL,
    NULL,
    NULL
  };
#endif
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    (PyCFunction) SwigPyObject_own,
    (PyCFunction) SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyObject *metatype_args;
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  
  /* metatype is used to implement static member variables. */
  metatype_args = Py_BuildValue("(s(O){})", "SwigPyObjectType", &PyType_Type);
  assert(metatype_args);
  metatype = (PyTypeObject *) PyType_Type.tp_call((PyObject *) &PyType_Type, metatype_args, NULL);
  assert(metatype);
  Py_DECREF(metatype_args);
  metatype->tp_setattro = (setattrofunc) &SwigPyObjectType_setattro;
  assert(PyType_Ready(metatype) >= 0);
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#if PY_VERSION_HEX >= 0x03000000
  m = PyModule_Create(&SWIG_module);
#else
  m = Py_InitModule((char *) SWIG_name, SwigMethods);
#endif
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = SwigPyObject_TypeOnce();
  } else if (SwigPyObject_TypeOnce()->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
# if PY_VERSION_HEX >= 0x03000000
    return NULL;
# else
    return;
# endif
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  SWIG_Python_SetConstant(d, "SVN_WC_TRANSLATE_FROM_NF",SWIG_From_long((long)(0x00000000)));
  SWIG_Python_SetConstant(d, "SVN_WC_TRANSLATE_TO_NF",SWIG_From_long((long)(0x00000001)));
  SWIG_Python_SetConstant(d, "SVN_WC_TRANSLATE_FORCE_EOL_REPAIR",SWIG_From_long((long)(0x00000002)));
  SWIG_Python_SetConstant(d, "SVN_WC_TRANSLATE_NO_OUTPUT_CLEANUP",SWIG_From_long((long)(0x00000004)));
  SWIG_Python_SetConstant(d, "SVN_WC_TRANSLATE_FORCE_COPY",SWIG_From_long((long)(0x00000008)));
  SWIG_Python_SetConstant(d, "SVN_WC_TRANSLATE_USE_GLOBAL_TMP",SWIG_From_long((long)(0x00000010)));
  SWIG_Python_SetConstant(d, "SVN_WC_ADM_DIR_NAME",SWIG_FromCharPtr(".svn"));
  SWIG_Python_SetConstant(d, "svn_wc_notify_add",SWIG_From_long((long)(svn_wc_notify_add)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_copy",SWIG_From_long((long)(svn_wc_notify_copy)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_delete",SWIG_From_long((long)(svn_wc_notify_delete)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_restore",SWIG_From_long((long)(svn_wc_notify_restore)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_revert",SWIG_From_long((long)(svn_wc_notify_revert)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_revert",SWIG_From_long((long)(svn_wc_notify_failed_revert)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_resolved",SWIG_From_long((long)(svn_wc_notify_resolved)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_skip",SWIG_From_long((long)(svn_wc_notify_skip)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_delete",SWIG_From_long((long)(svn_wc_notify_update_delete)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_add",SWIG_From_long((long)(svn_wc_notify_update_add)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_update",SWIG_From_long((long)(svn_wc_notify_update_update)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_completed",SWIG_From_long((long)(svn_wc_notify_update_completed)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_external",SWIG_From_long((long)(svn_wc_notify_update_external)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_status_completed",SWIG_From_long((long)(svn_wc_notify_status_completed)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_status_external",SWIG_From_long((long)(svn_wc_notify_status_external)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_modified",SWIG_From_long((long)(svn_wc_notify_commit_modified)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_added",SWIG_From_long((long)(svn_wc_notify_commit_added)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_deleted",SWIG_From_long((long)(svn_wc_notify_commit_deleted)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_replaced",SWIG_From_long((long)(svn_wc_notify_commit_replaced)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_postfix_txdelta",SWIG_From_long((long)(svn_wc_notify_commit_postfix_txdelta)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_blame_revision",SWIG_From_long((long)(svn_wc_notify_blame_revision)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_locked",SWIG_From_long((long)(svn_wc_notify_locked)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_unlocked",SWIG_From_long((long)(svn_wc_notify_unlocked)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_lock",SWIG_From_long((long)(svn_wc_notify_failed_lock)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_unlock",SWIG_From_long((long)(svn_wc_notify_failed_unlock)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_exists",SWIG_From_long((long)(svn_wc_notify_exists)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_changelist_set",SWIG_From_long((long)(svn_wc_notify_changelist_set)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_changelist_clear",SWIG_From_long((long)(svn_wc_notify_changelist_clear)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_changelist_moved",SWIG_From_long((long)(svn_wc_notify_changelist_moved)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_merge_begin",SWIG_From_long((long)(svn_wc_notify_merge_begin)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_foreign_merge_begin",SWIG_From_long((long)(svn_wc_notify_foreign_merge_begin)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_replace",SWIG_From_long((long)(svn_wc_notify_update_replace)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_property_added",SWIG_From_long((long)(svn_wc_notify_property_added)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_property_modified",SWIG_From_long((long)(svn_wc_notify_property_modified)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_property_deleted",SWIG_From_long((long)(svn_wc_notify_property_deleted)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_property_deleted_nonexistent",SWIG_From_long((long)(svn_wc_notify_property_deleted_nonexistent)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_revprop_set",SWIG_From_long((long)(svn_wc_notify_revprop_set)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_revprop_deleted",SWIG_From_long((long)(svn_wc_notify_revprop_deleted)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_merge_completed",SWIG_From_long((long)(svn_wc_notify_merge_completed)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_tree_conflict",SWIG_From_long((long)(svn_wc_notify_tree_conflict)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_external",SWIG_From_long((long)(svn_wc_notify_failed_external)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_started",SWIG_From_long((long)(svn_wc_notify_update_started)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_skip_obstruction",SWIG_From_long((long)(svn_wc_notify_update_skip_obstruction)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_skip_working_only",SWIG_From_long((long)(svn_wc_notify_update_skip_working_only)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_skip_access_denied",SWIG_From_long((long)(svn_wc_notify_update_skip_access_denied)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_external_removed",SWIG_From_long((long)(svn_wc_notify_update_external_removed)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_shadowed_add",SWIG_From_long((long)(svn_wc_notify_update_shadowed_add)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_shadowed_update",SWIG_From_long((long)(svn_wc_notify_update_shadowed_update)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_shadowed_delete",SWIG_From_long((long)(svn_wc_notify_update_shadowed_delete)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_merge_record_info",SWIG_From_long((long)(svn_wc_notify_merge_record_info)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_upgraded_path",SWIG_From_long((long)(svn_wc_notify_upgraded_path)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_merge_record_info_begin",SWIG_From_long((long)(svn_wc_notify_merge_record_info_begin)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_merge_elide_info",SWIG_From_long((long)(svn_wc_notify_merge_elide_info)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_patch",SWIG_From_long((long)(svn_wc_notify_patch)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_patch_applied_hunk",SWIG_From_long((long)(svn_wc_notify_patch_applied_hunk)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_patch_rejected_hunk",SWIG_From_long((long)(svn_wc_notify_patch_rejected_hunk)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_patch_hunk_already_applied",SWIG_From_long((long)(svn_wc_notify_patch_hunk_already_applied)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_copied",SWIG_From_long((long)(svn_wc_notify_commit_copied)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_copied_replaced",SWIG_From_long((long)(svn_wc_notify_commit_copied_replaced)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_url_redirect",SWIG_From_long((long)(svn_wc_notify_url_redirect)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_path_nonexistent",SWIG_From_long((long)(svn_wc_notify_path_nonexistent)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_exclude",SWIG_From_long((long)(svn_wc_notify_exclude)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_conflict",SWIG_From_long((long)(svn_wc_notify_failed_conflict)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_missing",SWIG_From_long((long)(svn_wc_notify_failed_missing)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_out_of_date",SWIG_From_long((long)(svn_wc_notify_failed_out_of_date)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_no_parent",SWIG_From_long((long)(svn_wc_notify_failed_no_parent)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_locked",SWIG_From_long((long)(svn_wc_notify_failed_locked)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_forbidden_by_server",SWIG_From_long((long)(svn_wc_notify_failed_forbidden_by_server)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_skip_conflicted",SWIG_From_long((long)(svn_wc_notify_skip_conflicted)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_update_broken_lock",SWIG_From_long((long)(svn_wc_notify_update_broken_lock)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_obstruction",SWIG_From_long((long)(svn_wc_notify_failed_obstruction)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_conflict_resolver_starting",SWIG_From_long((long)(svn_wc_notify_conflict_resolver_starting)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_conflict_resolver_done",SWIG_From_long((long)(svn_wc_notify_conflict_resolver_done)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_left_local_modifications",SWIG_From_long((long)(svn_wc_notify_left_local_modifications)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_foreign_copy_begin",SWIG_From_long((long)(svn_wc_notify_foreign_copy_begin)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_move_broken",SWIG_From_long((long)(svn_wc_notify_move_broken)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_cleanup_external",SWIG_From_long((long)(svn_wc_notify_cleanup_external)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_failed_requires_target",SWIG_From_long((long)(svn_wc_notify_failed_requires_target)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_info_external",SWIG_From_long((long)(svn_wc_notify_info_external)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_commit_finalizing",SWIG_From_long((long)(svn_wc_notify_commit_finalizing)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_inapplicable",SWIG_From_long((long)(svn_wc_notify_state_inapplicable)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_unknown",SWIG_From_long((long)(svn_wc_notify_state_unknown)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_unchanged",SWIG_From_long((long)(svn_wc_notify_state_unchanged)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_missing",SWIG_From_long((long)(svn_wc_notify_state_missing)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_obstructed",SWIG_From_long((long)(svn_wc_notify_state_obstructed)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_changed",SWIG_From_long((long)(svn_wc_notify_state_changed)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_merged",SWIG_From_long((long)(svn_wc_notify_state_merged)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_conflicted",SWIG_From_long((long)(svn_wc_notify_state_conflicted)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_state_source_missing",SWIG_From_long((long)(svn_wc_notify_state_source_missing)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_lock_state_inapplicable",SWIG_From_long((long)(svn_wc_notify_lock_state_inapplicable)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_lock_state_unknown",SWIG_From_long((long)(svn_wc_notify_lock_state_unknown)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_lock_state_unchanged",SWIG_From_long((long)(svn_wc_notify_lock_state_unchanged)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_lock_state_locked",SWIG_From_long((long)(svn_wc_notify_lock_state_locked)));
  SWIG_Python_SetConstant(d, "svn_wc_notify_lock_state_unlocked",SWIG_From_long((long)(svn_wc_notify_lock_state_unlocked)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_action_edit",SWIG_From_long((long)(svn_wc_conflict_action_edit)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_action_add",SWIG_From_long((long)(svn_wc_conflict_action_add)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_action_delete",SWIG_From_long((long)(svn_wc_conflict_action_delete)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_action_replace",SWIG_From_long((long)(svn_wc_conflict_action_replace)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_edited",SWIG_From_long((long)(svn_wc_conflict_reason_edited)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_obstructed",SWIG_From_long((long)(svn_wc_conflict_reason_obstructed)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_deleted",SWIG_From_long((long)(svn_wc_conflict_reason_deleted)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_missing",SWIG_From_long((long)(svn_wc_conflict_reason_missing)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_unversioned",SWIG_From_long((long)(svn_wc_conflict_reason_unversioned)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_added",SWIG_From_long((long)(svn_wc_conflict_reason_added)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_replaced",SWIG_From_long((long)(svn_wc_conflict_reason_replaced)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_moved_away",SWIG_From_long((long)(svn_wc_conflict_reason_moved_away)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_reason_moved_here",SWIG_From_long((long)(svn_wc_conflict_reason_moved_here)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_kind_text",SWIG_From_long((long)(svn_wc_conflict_kind_text)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_kind_property",SWIG_From_long((long)(svn_wc_conflict_kind_property)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_kind_tree",SWIG_From_long((long)(svn_wc_conflict_kind_tree)));
  SWIG_Python_SetConstant(d, "svn_wc_operation_none",SWIG_From_long((long)(svn_wc_operation_none)));
  SWIG_Python_SetConstant(d, "svn_wc_operation_update",SWIG_From_long((long)(svn_wc_operation_update)));
  SWIG_Python_SetConstant(d, "svn_wc_operation_switch",SWIG_From_long((long)(svn_wc_operation_switch)));
  SWIG_Python_SetConstant(d, "svn_wc_operation_merge",SWIG_From_long((long)(svn_wc_operation_merge)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_undefined",SWIG_From_long((long)(svn_wc_conflict_choose_undefined)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_postpone",SWIG_From_long((long)(svn_wc_conflict_choose_postpone)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_base",SWIG_From_long((long)(svn_wc_conflict_choose_base)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_theirs_full",SWIG_From_long((long)(svn_wc_conflict_choose_theirs_full)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_mine_full",SWIG_From_long((long)(svn_wc_conflict_choose_mine_full)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_theirs_conflict",SWIG_From_long((long)(svn_wc_conflict_choose_theirs_conflict)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_mine_conflict",SWIG_From_long((long)(svn_wc_conflict_choose_mine_conflict)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_merged",SWIG_From_long((long)(svn_wc_conflict_choose_merged)));
  SWIG_Python_SetConstant(d, "svn_wc_conflict_choose_unspecified",SWIG_From_long((long)(svn_wc_conflict_choose_unspecified)));
  SWIG_Python_SetConstant(d, "svn_wc_schedule_normal",SWIG_From_long((long)(svn_wc_schedule_normal)));
  SWIG_Python_SetConstant(d, "svn_wc_schedule_add",SWIG_From_long((long)(svn_wc_schedule_add)));
  SWIG_Python_SetConstant(d, "svn_wc_schedule_delete",SWIG_From_long((long)(svn_wc_schedule_delete)));
  SWIG_Python_SetConstant(d, "svn_wc_schedule_replace",SWIG_From_long((long)(svn_wc_schedule_replace)));
  SWIG_Python_SetConstant(d, "SVN_WC_ENTRY_WORKING_SIZE_UNKNOWN",SWIG_From_long((long)((-1))));
  SWIG_Python_SetConstant(d, "SVN_WC_ENTRY_THIS_DIR",SWIG_FromCharPtr(""));
  SWIG_Python_SetConstant(d, "svn_wc_status_none",SWIG_From_long((long)(svn_wc_status_none)));
  SWIG_Python_SetConstant(d, "svn_wc_status_unversioned",SWIG_From_long((long)(svn_wc_status_unversioned)));
  SWIG_Python_SetConstant(d, "svn_wc_status_normal",SWIG_From_long((long)(svn_wc_status_normal)));
  SWIG_Python_SetConstant(d, "svn_wc_status_added",SWIG_From_long((long)(svn_wc_status_added)));
  SWIG_Python_SetConstant(d, "svn_wc_status_missing",SWIG_From_long((long)(svn_wc_status_missing)));
  SWIG_Python_SetConstant(d, "svn_wc_status_deleted",SWIG_From_long((long)(svn_wc_status_deleted)));
  SWIG_Python_SetConstant(d, "svn_wc_status_replaced",SWIG_From_long((long)(svn_wc_status_replaced)));
  SWIG_Python_SetConstant(d, "svn_wc_status_modified",SWIG_From_long((long)(svn_wc_status_modified)));
  SWIG_Python_SetConstant(d, "svn_wc_status_merged",SWIG_From_long((long)(svn_wc_status_merged)));
  SWIG_Python_SetConstant(d, "svn_wc_status_conflicted",SWIG_From_long((long)(svn_wc_status_conflicted)));
  SWIG_Python_SetConstant(d, "svn_wc_status_ignored",SWIG_From_long((long)(svn_wc_status_ignored)));
  SWIG_Python_SetConstant(d, "svn_wc_status_obstructed",SWIG_From_long((long)(svn_wc_status_obstructed)));
  SWIG_Python_SetConstant(d, "svn_wc_status_external",SWIG_From_long((long)(svn_wc_status_external)));
  SWIG_Python_SetConstant(d, "svn_wc_status_incomplete",SWIG_From_long((long)(svn_wc_status_incomplete)));
  SWIG_Python_SetConstant(d, "svn_wc_merge_unchanged",SWIG_From_long((long)(svn_wc_merge_unchanged)));
  SWIG_Python_SetConstant(d, "svn_wc_merge_merged",SWIG_From_long((long)(svn_wc_merge_merged)));
  SWIG_Python_SetConstant(d, "svn_wc_merge_conflict",SWIG_From_long((long)(svn_wc_merge_conflict)));
  SWIG_Python_SetConstant(d, "svn_wc_merge_no_merge",SWIG_From_long((long)(svn_wc_merge_no_merge)));
#if PY_VERSION_HEX >= 0x03000000
  return m;
#else
  return;
#endif
}

